-- Functions for calling the Barvinok-binaries (see www.freshmeat.net or http://193.190.253.81/~skimo/barvinok/).
-- Is tested and used in VolIntCount.hs.
--
-- Author: Malte Rosenthal

module Barvinok (barvinok, barvinokStdOut) where

-- imports ----------------------------------------------------------------------------------------------------------------------------
import Debug.Trace(trace)
import System.Cmd(system)
import GHC.IOBase(ExitCode(..))

-- Calls the Barvinok-binary.
-- outputFile   	is the file where the output of the algorithm is printed. If it is the empty String "" it is printed to stdout.
-- binPath      	is the path to the Barvinok-binary (see www.freshmeat.net or http://193.190.253.81/~skimo/barvinok/).
-- polylibIneqSyst	are the inequality-systems in polylib-format (see PolylibInterf.showIneqSystemForPolylib)
barvinok :: String -> String -> String -> IO ()
barvinok outputFile binPath polylibIneqSyst =
	-- command that will be executed
    let cmd = if outputFile == "" then "echo \"" ++ polylibIneqSyst ++ "\" | " ++ binPath
    				  else "echo \"" ++ polylibIneqSyst ++ "\" | " ++ binPath ++ " >> " ++ outputFile
    in do
	exitCode <- do trace ("running Barvinok: " ++ (show cmd)) $ system cmd		-- execute command and remind exit-code
        let (ok, errorCode) = convertExitCode exitCode
        if ok then do putStrLn ("\nBarvinok completed.")
              else do putStrLn ("\nBarvinok exited with error code: " ++ (show errorCode))

-- Calls barvinok with "" as outputFile, i.e. all output will be printed on stdout.
barvinokStdOut :: String -> String -> IO ()
barvinokStdOut binPath polylibIneqSyst = barvinok "" binPath polylibIneqSyst



-- private functions -------------------------------------------------------------------------------------------------------------------

convertExitCode :: ExitCode -> (Bool, Int)
convertExitCode ExitSuccess	= (True, 0)
convertExitCode (ExitFailure x)	= (False, x)

