module CfFADA.GraphToList
    ( traverseCfGraph,
      -- listFromIO, listFromIOPiter, listFromIOTrivial, listFromIOExtreme
    )
where 

import Debug.Trace
import CfFADA.FilledCfGraph
import LP
import LoopoDS
import CfFADA.IterationVector

data Dependency = Input| TrueDep | Anti | Output deriving (Show, Eq) 






-- **************************** What the hack do this module? **********************************
-- >> This module is for the implementation of CfFada. You give them the Graph, and it will return the 
--    List in reversed order on which you have to evaluate the dependencies.
-- **************************** How can I call the function? ***********************************
-- >> You will find the function named listFromIO on the end of this module.
-- >> All Parameters are explained there.








-- Checks if all elements of a list (check) are elements of the list (list).
areAllElem :: (Eq a) => [a] -> [a] -> Bool
areAllElem check list = [y | y <- check, y `notElem` list] == [] 

-- Returns the elements of a list (check), which are elements of the list
-- (list).
areElem :: (Eq a) => [a] -> [a] -> [a]
areElem check list = [y | y <- check, y `elem` list] 

-- Returns the elements of a list (check), which are not elements of the list
-- (list).
areNotElem :: (Eq a) => [a] -> [a] -> [a]
areNotElem check list = [y | y <- check, y `notElem` list] 

-- An edge-type is relevant if it is StructuredBackward or UnstructuredBackward
isRelevantEdgeType :: CfgEdgeType -> Bool
isRelevantEdgeType (StructuredBackward _) = True
isRelevantEdgeType UnstructuredBackward   = True
isRelevantEdgeType _                      = False


-- Checks if a node (check) is the startpoint of a loop.
isLoopStart :: CfGraph -> NodeID -> Bool
isLoopStart graph check = let nodeType = getNodeType graph check
                          in (nodeType == LoopHead) || (nodeType == ForLoopHead) || (nodeType == DoWhileLoopHead)

-- Checks if a node (check) is the endpoint of a loop.
isLoopEnd :: CfGraph -> NodeID -> Bool
isLoopEnd graph check = [y| y <- (succIDs graph check), isRelevantEdgeType (edge y)] /= []
                           where edge x = getEdgeType graph (getEdgeBetween graph check x)


-- Returns the endpoint of a loop, which is specified by its startpoint.
getLoopEnd :: CfGraph -> NodeID -> NodeID
getLoopEnd graph start = head [y| y <- (predIDs graph start), isLoopEnd graph y]


-- Returns the startpoint of a loop, which is specified by its startpoint.
getNextLoopBegin :: CfGraph -> NodeID -> NodeID
getNextLoopBegin graph start = head [y| y <- (realPredIDs graph start), isLoopStart graph y]




-- Returns real successors - without following backward edges.
realSuccIDs :: CfGraph -> NodeID -> [NodeID]
realSuccIDs graph nodID = filter (not . test) (succIDs graph nodID)
    where  edgeType a = getEdgeType graph (edge a) 
	   edge a = getEdgeBetween graph nodID a 
	   test a = isRelevantEdgeType (edgeType a) || isRelevantEdgeType (edgeType a)

-- Returns real successors - without following backward edges.
realPredIDs :: CfGraph -> NodeID -> [NodeID]
realPredIDs graph nodID = filter (not . test) (predIDs graph nodID)
    where  edgeType a = getEdgeType graph (edge a) 
	   edge a = getEdgeBetween graph a nodID 
	   test a = isRelevantEdgeType (edgeType a) || isRelevantEdgeType (edgeType a)

-- Checks if an element of the list (list) exists, which is direct or indirect
-- successor of the node (nodeId).
hasSuccIn :: CfGraph -> [NodeID] -> NodeID -> Bool
hasSuccIn graph [] nodId  = False
hasSuccIn graph list nodId  
    | nodId `elem` list = True
    | otherwise = foldr ((||).(hasSuccIn graph list)) False (realSuccIDs graph nodId)

-- Checks if succ is a direct or indirect successor of prev
isSuccOf :: CfGraph -> NodeID -> NodeID -> Bool
isSuccOf graph succ prev = hasSuccIn graph [succ] prev

-- Returns only the nodes from a list
getNodeList :: [GraphID b] -> [NodeID]
getNodeList [] = []
getNodeList ((GNodeID a):xs) = a:getNodeList(xs)
getNodeList ((GEdge a):xs) = getNodeList(xs)

-- Remove all direct and indirect successors of head out of the list
removeSuccs :: CfGraph -> [NodeID] -> NodeID -> [NodeID]
removeSuccs graph list head = [i | i <- list, not (isSuccOf graph i head)]


-- get the Fork ID of an Join ID
getForkOfJoin :: CfGraph -> NodeID -> NodeID
getForkOfJoin graph id = getForkOfJoinHelp id 0 
	where
	getForkOfJoinHelp :: NodeID -> Integer -> NodeID
	getForkOfJoinHelp id x
		| getNodeType graph (head list) == Join 
			= getForkOfJoinHelp (head list) (x + 1)
		| getNodeType graph (head list) == Fork
		        && x /= 0 = getForkOfJoinHelp (head list) (x - 1)
		| getNodeType graph (head list) == Fork
		        && x == 0 = head list
		| otherwise = getForkOfJoinHelp (head list) x
		where 	list = realPredIDs graph id


-- get the Join ID of an Fork ID
getJoinOfFork :: CfGraph -> NodeID -> NodeID
getJoinOfFork graph id = getJoinOfForkHelp id 0
        where
	getJoinOfForkHelp :: NodeID -> Integer -> NodeID
	getJoinOfForkHelp id x
		 | getNodeType graph (head list) == Fork
                        = getJoinOfForkHelp (head list) (x + 1)
                 | getNodeType graph (head list) == Join
                        && x /= 0 = getJoinOfForkHelp (head list) (x - 1)
                 | getNodeType graph (head list) == Join
                        && x == 0 = head list
                 | otherwise = getJoinOfForkHelp (head list) x
                 where   list = (realSuccIDs graph id)

-- Evaluates if there is time to go a backwared edge.  
isTimeToGoBackward :: CfGraph -> NodeID -> IterationVector -> Bool
isTimeToGoBackward graph nodeID (IV iv) 
	| (iv)!!(fromIntegral(getIntOfLoopID(getLoopIDFromLoopHead graph nodeID) -1)) 
		== Star && getNodeType graph nodeID == ForLoopHead 	= True
        | isTimeToShift (IV iv) (getLoopIDFromLoopHead graph nodeID)	= True
	| otherwise							= False




-- traverses the graph to a list
-- done: There are all the loopheads stored which are already gone through
-- graph: the graph to transverse.          !!!!! Graph should be initialized with (getAdvancedCfGraph graph) !!!!!
-- currNodeID: There starts the list
-- list: is to store the current list. Start it with []
-- iv: is the IterationVector of an Statment
graphToList :: CfGraph -> NodeID -> [NodeID] -> [GraphID CfgEdgeData] -> IterationVector -> [GraphID CfgEdgeData]
graphToList graph currNodeID done list iv
	-- There is an IF. So return the list.
	| (getNodeType graph currNodeID) == Fork = (GNodeID currNodeID):list
	-- There is the Join of an IF. So get list s of the two different ways an concat them.
	| (getNodeType graph currNodeID) == Join 
		= let
			fork = (getForkOfJoin graph currNodeID)
			x = (realPredIDs graph currNodeID)!!0
			y = (realPredIDs graph currNodeID)!!1
			predOfFork = head (realPredIDs graph fork)

		in (graphToList graph predOfFork done [(GEdge (getEdgeInfoBetween graph predOfFork fork))] iv) 
						++ (graphToList graph y done [(GEdge (getEdgeInfoBetween graph y currNodeID))] iv) 
						++ (graphToList graph x done [(GEdge (getEdgeInfoBetween graph x currNodeID))] iv) 
						++ (GNodeID currNodeID):list
	-- Evaluates if there is time to go Backwared and run through 
	| (isLoopStart graph currNodeID) && (currNodeID `notElem` done) && (isTimeToGoBackward graph currNodeID iv)
		= let
			loopEnd = getLoopEnd graph currNodeID
			isTime =  isTimeToShift iv (getLoopIDFromLoopHead graph currNodeID)
			newList =  if isTime then (GIterVector ivNew):(GNodeID currNodeID):list else (GNodeID currNodeID):list
                	ivNew = if isTime then shift iv else iv 	
			doneNew = if isTime then [] else done
            	in graphToList graph loopEnd (currNodeID:doneNew) ((GEdge (getEdgeInfoBetween graph loopEnd currNodeID)):newList) ivNew
	-- go to the next Node on the top of you
	| realPredIDs graph currNodeID /= []
		= let 
			nextID = (head (realPredIDs graph currNodeID))
		in graphToList graph nextID done ((GEdge (getEdgeInfoBetween graph nextID currNodeID)):(GNodeID currNodeID):list) iv
	-- o. I am the toppest Node. So i have to finish this thing. 
	| otherwise
		= ((GNodeID currNodeID):list)

hasStatement :: CfgEdgeData -> Bool
hasStatement b = ((getMaybeOfEdgeData b) /= Nothing)

-- Fills the list of nodes and edges with Accesses and Statements, discarding the NodeIDs
cfGraphToFilledCfGraph :: ([Statement], [Access])
                       -> Access
                       -> AccessType
                       -> [GraphID CfgEdgeData]
                       -> [GraphID FilledCfgEdgeData]
cfGraphToFilledCfGraph (stmts,accs) destAcc srcAccType cedList =
    fillCfGraph cedList
    where
    fillCfGraph [] = []
    fillCfGraph (GNodeID _ : xs) = fillCfGraph xs
    fillCfGraph (GIterVector x : xs) = GIterVector x : fillCfGraph xs
    fillCfGraph (GEdge x : xs)
      | null srcAccs =
            (GEdge $
             EdgeInfo (edgeID x)
                      (FilledCfgEdgeData (getEdgeTypeOfEdge (edgeData x))
                                         Nothing
                      )
                      (edgeSrc x)
                      (edgeDst x)
            ) : fillCfGraph xs
      | otherwise =
           -- is it correct to simply emit an edge several times
           -- if the statement in the edge has several matching
           -- source accesses in it, or should we change the
           -- data type to contain a list of matching accesses
           -- (instead of one access only) ?
	   map mkEdge srcAccs ++ fillCfGraph xs
      where
      srcAccs = case getMaybeOfEdgeData $ edgeData x of
                   Just stmtID -> [a | a <- accs, accStmtID a == stmtID,
                                       accType a == srcAccType,
                                       accArrayID a == accArrayID destAcc
                                  ]
                   Nothing  ->  []
      mkEdge acc =
         GEdge $
         EdgeInfo (edgeID x)
                  (FilledCfgEdgeData (getEdgeTypeOfEdge (edgeData x))
	                             (Just (accStmtID acc
                                           ,stmt
                                           ,Just acc
                                           )
                                     )
                  )
                  (edgeSrc x)
                  (edgeDst x)
        where
        stmt = head [s | s@(stID,_) <- stmts, stID == accStmtID acc]


-- listFromIO is the main function in this module. See on the top of the module to get to know what this function is doing.
-- Examples for calling this function are below.
-- 1st  Parameter: It is the controll flow graph of the code which you want to evaluate the dependencies
-- 2end Parameter: This are the sourrounding loops of the Statement which you want to evaluate.
-- 3ed  Parameter: The statement id of the statement which you want to evaluate
-- 4ed  Parameter: Which kind of dependency do you want to evaluate?
-- 5ed  Parameter: The Location of the whole project. There have to be the *.cfg, *.slp files and so on.
listFromIO :: IO (CfGraph) -> IO (LPResult SurroundingLoops) -> Access -> AccessType -> String -> IO [GraphID FilledCfgEdgeData] 
listFromIO iogres ioslres destAcc srcAccType path = do
                      graph <- iogres
		      slRes <- ioslres
		      stAndAcc <- loadStatementsAndAccesses path
		      case slRes of
		         ParseOk x -> do
	                    let iv = (convBoolArrayToIV (lookupSurroundingLoops x (accStmtID destAcc)))
			    	liste = (graphToList (getAdvancedCfGraph graph) (getSrcOfEdgeFromStatementID graph (accStmtID destAcc)) [] [GIterVector iv] iv)
			    	result = cfGraphToFilledCfGraph stAndAcc destAcc srcAccType liste
			    print liste
			    return result
		         ParseFailed x -> return undefined


-- |
-- Traverse a control-flow graph in an order suitable for CfFADA analysis.
-- Starts the traversal at a given access and visits all the accesses of
-- the given type as possible sources
--
traverseCfGraph :: CfGraph
                -> [Statement]
                -> [Access]
                -> SurroundingLoops
                -> Access            -- ^ destination access of the dependence (i.e. the traversal starts there)
                -> AccessType        -- ^ type of the source accesses of the dependence
                -> [GraphID FilledCfgEdgeData]
traverseCfGraph cfg stmts accs slp destAcc srcAccType =
    cfGraphToFilledCfGraph (stmts,accs) destAcc srcAccType list
    where
    iv   = convBoolArrayToIV (lookupSurroundingLoops slp (accStmtID destAcc))
    list = graphToList (getAdvancedCfGraph cfg)
                       (getSrcOfEdgeFromStatementID cfg (accStmtID destAcc))
                       []
                       [GIterVector iv]
                       iv

-------------------------------------------------
-- here are some testcases 
-------------------------------------------------

undefAcc = Access { accStmtID=undefined,
                    accArrayID = undefined,
                    accID = undefined,
                    accIndexFct = undefined,
                    accDomain = undefined,
                    accType = undefined
                  }

listFromIOPiter = let file =  "./CfFADA/testdata/piter/piter.src.out"
		      iogres = loadCfGraph file
		      ioslres = readSLP (file ++ ".slp")
		      id = undefAcc { accStmtID = StatementID 4, accArrayID = integerToID 1 }
		      dep = Write
		  in  listFromIO iogres ioslres id dep file
resultPiter = "[N1,E12,N13,E13,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,E15,N14,E16,IV [Lower,Star],N13,E13,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,IV [Equal,Lower],N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,IV [Equal,Equal]]"

listFromIOTrivial = let file =  "./CfFADA/testdata/trivial/trivial.src.out"
                        iogres = loadCfGraph file
		        ioslres = readSLP (file ++ ".slp")
		        id = undefAcc { accStmtID = StatementID 5, accArrayID = integerToID 1 }
		        dep = Write
		    in  listFromIO iogres ioslres id dep file
resultTrivial = "[N1,E1,N2,E6,N6,E5,N7,E7,N2,E3,N3,E2,N4,E4,N5,IV []]"


listFromIOExtreme stNr = let file =  "./CfFADA/testdata/extreme/extreme.src.out"
                             iogres = loadCfGraph file
		             ioslres = readSLP (file ++ ".slp")
		             id = undefAcc { accStmtID = StatementID stNr, accArrayID = integerToID 1 }
		             dep = Write
		         in  listFromIO iogres ioslres id dep file
resultExtremeSt5 = "[N1,E19,N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,E12,N13,E14,N17,E15,N15,E13,N16,E17,N18,E18,N17,E16,N14,E22,N20,E23,IV [Lower,Star,Star],N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,IV [Equal,Star,Star]]"
resultExtremeSt4 = "[N1,E19,N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,E12,N13,E14,N17,E15,N15,E13,N16,E17,N18,E18,N17,E16,N14,E22,N20,E23,IV [Lower,Star,Star],N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,IV [Equal,Lower,Star],N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,IV [Equal,Equal,Star]]"
resultExtremeSt6 = "[N1,E19,N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,E12,N13,E14,N17,E15,N15,E13,N16,E17,N18,E18,N17,E16,N14,E22,N20,E23,IV [Lower,Star,Star],N19,E20,N3,E7,N11,E8,N5,E1,N6,E5,N6,E3,N7,E2,N8,E4,N9,E6,N10,E10,N12,E11,N11,E9,N4,E12,N13,E14,N17,E15,N15,E13,N16,E17,N18,E18,IV [Equal,Star,Lower],N17,E15,N15,IV [Equal,Star,Equal]]"

