-- |
-- Defines the ordering on the iteration vectors
-- in a CfFADA backward walk through the control flow graph.
--
module CfFADA.IterationVector
    ( Condition(..), IterationVector(..),
      convBoolArrayToIV, shift, isTimeToShift
    )
where

import LoopoDS.SurroundingLoops (LoopID(..))

-- The meaning of "Star" is that you know nothing about the iteration
-- The meaning of "Equal" is that you are in the same iteration
-- The meaning of "Lower" is that there is maybe in the past an dependence
data Condition = Star | Equal | Lower deriving (Show, Eq)

-- stores the condition for an statement in loops
data IterationVector = IV [Condition] deriving (Show, Eq)


-- konverts a Bool[] to IV
convBoolArrayToIV  :: [Bool] -> IterationVector
convBoolArrayToIV array = IV (cBA array) where
    cBA []                   = []
    cBA (x:xs) | x == True   = Equal : (cBA xs)
    cBA (x:xs) | x == False  = Star : (cBA xs)


-- shifts the condition vector to the left. Like (===_<**) -> (===<***)
shift :: IterationVector -> IterationVector
shift (IV list) = IV (shiftList list) where
    shiftList []                     = []
    shiftList [x]                    = [Lower]
    shiftList (x:y:xs) | y == Star && ((getLastEq xs 0 (-1)) == (-1))   
    				     = Lower:Star:(map (const Star) xs)
    shiftList (x:y:xs) | y == Star && ((getLastEq xs 0 (-1)) /= (-1))
                                         = x : y : (shiftList xs)
    shiftList (x:y:xs) | y == Lower  = Lower:Star:(map (const Star) xs)
    shiftList (x:xs)                 = x : shiftList xs
    getLastEq [] currNr lastEqNr = lastEqNr
    getLastEq (x:xs) currNr lastEqNr 	| x == Equal = getLastEq xs (currNr + 1) currNr
    					| otherwise  = getLastEq xs (currNr + 1) lastEqNr 


-- looks in the IterationVector if it is time to shift. So you have to go 
-- backward this Loop if isTimeToShift evaluates True.
isTimeToShift :: IterationVector -> LoopID -> Bool
isTimeToShift (IV xs) (LoopID pos) = (xs!!(fromIntegral pos - 1)) == Equal



