---
-- @author Martin Griebl
-- @version $Id: CodeGen.hs 4184 2009-08-13 15:59:24Z groessli $
--
module CodeGen(loopNest,genIfCascade,loopNestPretty,genIfCascadePretty,
	       getBoundText -- MC 2003-04-06: fuer CommCode.hs
	      )

where 

import Data.Map(Map, (!))
import qualified Data.Map as Map
import Data.Ratio
import LMath


myShow :: Rational -> String
myShow r = (if r>=0 then "+ " else "- " ) ++
	    if abs (denominator r) == 1 then show (abs (numerator r))
	    else show (abs r)


defaultNames :: Integer -> Integer -> Map Integer String 
defaultNames nrv nrp = Map.fromList (vars ++ params)
    where vars = [(i, 'i' : show i) |  i <- [0 .. nrv-1]]
	  params =  [(i, 'i' : show i) |  i <- [nrv .. nrv + nrp -1]]


testMap =  Map.fromList [(0,"i"), (1,"j"), (2,"k"), (3,"ii"), (4,"jj"), (5,"kk")]


factorToText :: Map Integer String -> [Rational] -> String -> Int -> String
factorToText names cs _ i = if cs!!i == 0 then "" 
		          else if cs!!i == 1 then "+ " ++ names!ii -- "_"
		          else myShow (cs!!i) ++ "* " ++ names!ii -- "_"
    where ii = fromIntegral i


test_factorToText_0 = factorToText testMap [1%1,0,0,-2,0,1] "x" 0
test_factorToText_1 = factorToText testMap [1%1,0,0,-2,0,1] "x" 1
test_factorToText_2 = factorToText testMap [1%1,0,0,-2,0,1] "x" 2
test_factorToText_3 = factorToText testMap [1%1,0,0,-2,0,1] "x" 3





exprToTextList :: Map Integer String -> IneqSystem -> ([String], Integer)
exprToTextList names is = if non_empty_result == [] then (["0"],1)
		        else (non_empty_result, denom)
    where row = head (rows' (getMatrix (toAffineExp is)))
	  nrv = fromIntegral (nVars is)
	  nrp = fromIntegral (nParams is)
	  denom = abs( foldr lcm 1 (map denominator row) )
          numerators = map (* (toRational denom)) row
	  non_empty_result =  filter (/= "") raw_result
	  raw_result = [ factorToText names numerators "egal" i | i <- [0 .. nrv-1] ] ++
		  [ factorToText names numerators "egal" i | i <- [nrv .. nrv + nrp -1] ] ++
		  ( if numerators!!(nrv+nrp) == 0 then []  
		    else [myShow (numerators!!(nrv+nrp))] )
	  
test_exprToTextList_1 = 
    exprToTextList testMap (ineqSystem ((4,1),matrix [[1%1,0,0,-2,5,1]] ))
test_exprToTextList_0 = 
    exprToTextList testMap (ineqSystem ((4,1),matrix [[0%1,0,0,0,0,0]] ))





-- if elim is a prefix of src, then drop it (e.g., for leading +es)
elimLeadingString :: String -> String -> String
elimLeadingString elim src = 
    if length elim > length src then src -- caller is dumb
    else if prefix == elim then rest else src
    where prefix = [ src!!i | i <- [0..length elim -1]]
	  rest = [ src!!i | i <- [length elim ..length src -1]]




exprToText :: Map Integer String -> BoundaryType -> IneqSystem -> String
exprToText names bnd is =  rational_result denom (elimLeadingString " + " int_result)
    where ettl = exprToTextList names is
	  denom = snd ettl
          int_result = foldr (appendWithSeparator " ")   ""   (fst ettl)
	  rational_result :: Integer -> String -> String
	  rational_result den s = 
		if denom==1 then s
                else if bnd == Upper then " (int)floor((" ++ s ++ ") /" ++ (show den) ++ ".0) "	
		     else " (int)ceil((" ++ s ++ ") /" ++ (show den) ++ ".0) "



test_exprToText = exprToText testMap Upper (ineqSystem ((4,1),matrix [[1%1,0,0,-2,5,1]] ))
test_exprToText_0 = exprToText testMap Lower (ineqSystem ((4,1),matrix [[0%1,0,0,0,0,0]] ))




appendWithSeparator :: String -> String -> String -> String
appendWithSeparator s1 sep s2 = s1 ++ sep ++ s2

chop2l :: String -> String
chop2l s = if length s >= 2 then reverse (tail (tail (reverse s))) else s

chop2r :: String -> String
chop2r s = if length s >= 2 then tail (tail s) else s




exprListToText :: Map Integer String -> BoundaryType -> IneqSystem -> Integer -> String
exprListToText names bnd is dim = chop2r raw_result 
    where raw_result = foldr (appendWithSeparator ", ") "" (map (exprToText names bnd) bounds_list)
          bounds_list = map toIneqSystem (rows (getMatrix (toAffineExp is)))
	  toIneqSystem :: Matrix -> IneqSystem
	  toIneqSystem m = ineqSystem ((nVars is, nParams is), m)

test_exprListToText_1 = 
    exprListToText testMap Upper (ineqSystem ((4,1),matrix [[1%1,0,0,-2,5,1]] )) 0
test_exprListToText_2 = 
    exprListToText testMap Lower (ineqSystem ((4,1),matrix [[1%1,0,0,-2,5,1],[0,-1,2,0,0,-1]] )) 0




eliminateBoundVariable :: IneqSystem -> Integer -> IneqSystem 
eliminateBoundVariable is dim = 
    ineqSystem ((nVars is, nParams is), reducedMat )
	where origMat =  getMatrix (toAffineExp is)
	      reducedMat = origMat + colOneMat 
	      colOneMat = colOneMatrix (nRows origMat) (nColumns origMat) dim



-- we assume that is contains only the ineqs needed for the actual bound
fullBoundToText  :: Map Integer String -> BoundaryType -> IneqSystem -> Integer -> String
fullBoundToText names bnd is dim = 
    if nRows is > 1 then 
       red_fct ++ "(" ++ show (nRows is) ++ " , " ++
	       exprListToText names bnd reduced_is dim ++ ")"
    else if nRows is == 1 then exprListToText names bnd reduced_is dim
    else " unknown "
    where red_fct = if bnd == Upper then "varMin" else "varMax"
	  reduced_is = eliminateBoundVariable is dim


--genOneLoop :: IneqSystem -> Integer -> String
--genOneLoop is dim = genOneLoopPretty is  dim
{---
genOneLoop :: IneqSystem -> Integer -> String
genOneLoop is dim = "for (i" ++ (show dim) ++ " = " ++ lower ++ 
			"; i" ++ (show dim) ++ " <= " ++ upper ++ 
			"; i" ++ (show dim) ++ "++)\n"
    where 
    lower = fullBoundToText Lower (smul (boundsOfDim Lower is dim) (-1)) dim
    upper = fullBoundToText Upper (boundsOfDim Upper is dim) dim
---}

genOneLoop :: Map Integer String -> IneqSystem -> Integer -> String
genOneLoop names is dim = "for (" ++ names!dim ++ " = " ++ lower ++ 
			        "; " ++ names!dim  ++ " <= " ++ upper ++ 
			        "; " ++ names!dim ++ "++)\n"
    where 
    lower = fullBoundToText names Lower (smul (boundsOfDim Lower is dim) (-1)) dim
    upper = fullBoundToText names Upper (boundsOfDim Upper is dim) dim

-- returns the text for (lower-, upper-bound) as a string
getBoundText :: Map Integer String -> IneqSystem -> Integer -> (String, String)
getBoundText names is dim = (lower, upper)	  
    where 
    lower = fullBoundToText names Lower (smul (boundsOfDim Lower is dim) (-1)) dim
    upper = fullBoundToText names Upper (boundsOfDim Upper is dim) dim


rawLoopNest :: Map Integer String -> IneqSystem -> String
rawLoopNest names is = concat  all_loop_texts
	       where all_loop_texts = [ genOneLoop names is d | d <- [0 .. nVars is-1]]


---
-- loopNest assumes that all projection and sorting is already done.
-- It just prints the Ineqsystem as a loop nest, thereby ignoring the
-- context.
loopNest  :: IneqSystem -> String
loopNest is = loopNestPretty (defaultNames (nVars is) (nParams is) )  is


---
-- loopNestPretty assumes that all projection and sorting is already done.
-- It just prints the Ineqsystem as a loop nest, thereby ignoring the
-- context. The names identify the column numbers of the matrix of the
-- IneqSystem is with the name of the corresponding variables.
loopNestPretty  :: Map Integer String -> IneqSystem -> String
loopNestPretty names is = (rawLoopNest names) $ simplifyTrivially $ makeDivisorfree $ dropContextOnlyInfo is


genIfCascade :: IneqSystem -> String
genIfCascade is = genIfCascadePretty (defaultNames (nVars is) (nParams is)) is

genIfCascadePretty  :: Map Integer String -> IneqSystem -> String
genIfCascadePretty names is = "if ( " ++ all_constraints ++ " )"
   where 
   all_constraints = elimLeadingString " && " (concat constr)
   lower dim = fullBoundToText names Lower (smul (boundsOfDim Lower is dim) (-1)) dim
   upper dim = fullBoundToText names Upper (boundsOfDim Upper is dim) dim
   lowText dim = if lower dim == " unknown " then "" 
                else " && " ++ names!dim ++ " >= " ++ lower dim
   upText dim = if upper dim == " unknown " then "" 
                else " && " ++ names!dim ++ " <= " ++ upper dim
   constr = [lowText dim  |  dim <- [0 .. nVars is-1]] 
           ++ 
           [upText dim |     dim <- [0 .. nVars is-1]]
