module CompSpread.CloogStatements
where

import LMath.Types
import LoopoDS.Statement
import LoopoDS.TrDep
import LoopoDS.Qual
import CompSpread.Data
import CompSpread.DomainDescriptions

                            

data CloogStatement = ComputeStatement { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       , clStmtTrStmtID :: TStatementID
                                       }
                    | WriteStatement   { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       , clStmtTrDepID  :: TrDependenceID
                                       }
                    | UnpackStatement  { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       , clStmtTrDepID  :: TrDependenceID
                                       }
                    | SendStatement    { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       }
                    | RecvStatement    { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       }
                    | SpreadStatement  { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       }
                    | RPMapStatement   { clStmtDomain   :: [IneqSystem]
                                       , clStmtDomDesc  :: StmtDomDescription
                                       }
                      deriving (Show)

data CloogStatements = CloogStatements { clCompStmts   :: [CloogStatement]
                                       , clWriteStmts  :: [CloogStatement]
                                       , clUnpackStmts :: [CloogStatement]
                                       , clSendStmts   :: [CloogStatement]
                                       , clRecvStmts   :: [CloogStatement]
                                       , clSpreadStmts :: [CloogStatement]
                                       }

data RPMapStatements = RPMapStatements { rpmStmts :: [CloogStatement] }

getStmtTypeNr :: CloogStatement -> Integer
getStmtTypeNr SpreadStatement {}   = 0
getStmtTypeNr WriteStatement {}    = 1
getStmtTypeNr SendStatement {}     = 2
getStmtTypeNr RecvStatement {}     = 3
getStmtTypeNr UnpackStatement {}   = 4
getStmtTypeNr ComputeStatement {}  = 5

isWriteStmt WriteStatement {} = True
isWriteStmt _                 = False

isUnpackStmt UnpackStatement {} = True
isUnpackStmt _                  = False

isSendStmt SendStatement {} = True
isSendStmt _                = False

isRecvStmt RecvStatement {} = True
isRecvStmt _                = False
