module CompSpread.Defines 
where

-- import System.Cmd(system)
import Debug.Trace(trace)
-- import GHC.IOBase(ExitCode(..))
import List
import Maybe

import LoopoDS
import LMath.Types
import CompSpread.Data
import CompSpread.CloogStatements
import CompSpread.DomainDescriptions
import CompSpread.Names
import CompSpread.CloogInput
import CompSpread.Helpers


defineHeader :: CompSpreadDS -> Integer -> CloogStatement -> String
defineHeader csDS cloogStmtNr cloogStmt =
    (define_t ++ " " ++ cloogStmt_t ++ (show cloogStmtNr) ++ "(" ++
     iteratorNames ++ ") ")
    where
    iteratorNames = concat $ intersperse ", " $ getUsedIteratorNames csDS cloogStmt


genCompStmtsDefines :: CompSpreadDS -> [(StatementID, String)] -> 
		       [CloogStatement] -> String
genCompStmtsDefines csDS origStatementsCode compStmts = 
    verbose "compStmtsDefines" $
    unlines compStmtsDefines
    where
    allTrStmtIDs = getAllTStatementIDs csDS
    compStmtIndexMapping = map (genCompStmtIndexMappingCode csDS) allTrStmtIDs
    compStmtsDefines = zipWith oneCompStmtDefine [1..] compStmts

    oneCompStmtDefine :: Integer -> CloogStatement -> String
    oneCompStmtDefine cloogStmtNr cloogStmt = 
	let trStmtID = clStmtTrStmtID cloogStmt
        in concat 
	       [ defineHeader csDS cloogStmtNr cloogStmt
	       , genCompStmtIndexMappingCode csDS trStmtID (getIteratorNames csDS)
	       , fromJust $ 
		 lookup (getStmtID csDS trStmtID) origStatementsCode
	       ]

genCompStmtIndexMappingCode :: CompSpreadDS -> TStatementID -> [String] -> String
genCompStmtIndexMappingCode csDS trStmtID iteratorNames = 
    concat $
    zipWith (\l c -> l ++ " = " ++ c ++ "; ") 
	    loopoIndices cloogIndices
    where
    origStmtID = getStmtID csDS trStmtID
    tilingDim = getTilingDim csDS
    ispcDim = getISPCdim csDS trStmtID
    (SLP slps) = csSLPs csDS
    slpVector = (fromJust $ lookup origStmtID slps)::[Bool]
    slp = (elemIndices True slpVector) :: [Int]
    loopoIndices = 
	[ genericIndex (csIndexNames csDS) lp | lp <- slp ]
    cloogIndices = 
	genericTake ispcDim $ genericDrop (tilingDim) iteratorNames


genWriteStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genWriteStmtDefines csDS cloogStmtNr 
                    cloogStmt@WriteStatement   { clStmtTrDepID = trDepNr } = 
    (defineHeader csDS cloogStmtNr cloogStmt ++
     writeBuffer_t)
    where
    arrayName = 
	genericIndex (csArrayNames csDS) (integerFromID arrayID - 1)
    arrAccNames = inner_t
        
    srcAcc = getSrcAcc csDS trDepNr
    accDim = targetDimIH $ accIndexFct srcAcc
    arrayID = accArrayID $ srcAcc

    -- Achtung: hier Annahme, dass nur 1-dimensionale Prozessor Tiles vorkommen!
    -- (evtl. doch moeglich: einfach entsprechende realprozessor koordinate verwenden...?)

    writeBuffer_t =
        -- e.g.: sendVect[destRP].push_back(y[coords1][coords2]);
        (sendVector_t ++ "[" ++ destProc_t ++ "1].push_back(" ++ arrayName ++
	 concat ["[" ++ arrAccNames ++ (show i) ++ "]" | i <- [1..accDim]] ++
	 ");")


genUnpackStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genUnpackStmtDefines csDS cloogStmtNr 
                    cloogStmt@UnpackStatement   { clStmtTrDepID = trDepNr } = 
    (defineHeader csDS cloogStmtNr cloogStmt ++
     unpackBuffer_t)
    where
    arrayName = 
	genericIndex (csArrayNames csDS) (integerFromID arrayID - 1)
    arrAccNames = inner_t

    -- Achtung: hier Annahme, dass nur 1-dimensionale Prozessor Tiles vorkommen!
    -- (evtl. doch moeglich: einfach entsprechende realprozessor koordinate verwenden...?)
        
    srcAcc = getSrcAcc csDS trDepNr
    accDim = targetDimIH $ accIndexFct srcAcc
    arrayID = accArrayID $ srcAcc
    unpackBuffer_t =
        (arrayName ++ 
         concat ["[" ++ arrAccNames ++ (show i) ++ "]" | i <- [1..accDim]] ++
	 " = " ++
         recvVector_t ++ "["++ srcProc_t ++ "1" ++ "]" ++ 
         "[" ++ rdispls_t ++ "[" ++ srcProc_t ++ "1" ++ "]++];")


genSendStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genSendStmtDefines csDS cloogStmtNr cloogStmt@SendStatement {} =    
    (defineHeader csDS cloogStmtNr cloogStmt ++ send_t)
    where
    send_t = 
        (mpiSend ++ "(&" ++ sendVector_t ++ "[" ++ destProc_t ++ "1" ++ "][0], " ++
         sendVector_t ++ "[" ++ destProc_t ++ "1" ++ "].size(), " ++
         "MPI_DOUBLE, " ++ 
         destRP_t ++ ", " ++
         srcProc_t ++ "1" ++ ", " ++ -- tag=srcProcTile
         mpiCommNodes ++ "); ")
    {--
      MPI_Bsend(&loopo___sendVect[destProc][0], loopo___sendVect[destProc].size(), MPI_DOUBLE, 
      destProc, destProc, MPI_COMM_WORLD);
      --}

genRecvStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genRecvStmtDefines csDS cloogStmtNr cloogStmt@RecvStatement {} =    
    (defineHeader csDS cloogStmtNr cloogStmt ++ recv_t)
    where
    recv_t = 
        (mpiProbe ++ "(" ++ srcRP_t ++ ", " ++ srcProc_t ++ "1" ++ ", "++
         mpiCommNodes ++ ", &" ++ mpiStatus_t ++ "); " ++
         mpiGetCount ++ "(&" ++ mpiStatus_t ++ ", " ++ "MPI_INT, " ++
         "&" ++ mpiCount_t ++ "); " ++
         recvVector_t ++ "[" ++ srcProc_t ++ "1" ++ "].resize(" ++ mpiCount_t ++ ");" ++
         mpiRecv ++ "(&" ++ recvVector_t ++ "[" ++ srcProc_t ++ "1" ++ "][0], " ++
         mpiCount_t ++ ", MPI_DOUBLE, " ++
         srcRP_t ++ ", " ++ 
         srcProc_t ++ "1" ++ ", " ++ 
         mpiCommNodes ++ 
         ", &" ++ mpiStatus_t ++ "); " ++
         rdispls_t ++ "[" ++ srcProc_t ++ "1" ++ "]" ++ " = 0;"
        )
         
        
    {--
      MPI_recv recvBuffer[srcProc] von lookupRP(srcProc, srcTime)

      srcRP = lookupRP(srcProc, srcTime)
      MPI_Probe(srcRP, MPI_ANY_TAG, MPI_COMM_WORLD, &status); //probe
      MPI_Get_count(&status, MPI_INT, &recvCount); //get receive length
      recvBuffer[srcProc].resize(recvCount);
      MPI_Recv(&recvBuffer[srcProc][0], recvCount, MPI_INT, srcRP, 
      LBP_MESSAGE_TAG, MPI_COMM_WORLD, &status);

      --}

mpiSend  = "MPI_Bsend"
mpiRecv  = "MPI_Recv"
mpiProbe = "MPI_Probe"
mpiGetCount = "MPI_Get_count"
mpiCommNodes  = "MPI_COMM_NODES"


genSpreadStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genSpreadStmtDefines csDS cloogStmtNr cloogStmt@SpreadStatement {} = spreadDefines
    where
    spreadDefines = 
        ((defineHeader csDS cloogStmtNr cloogStmt) ++ 
          (spreadCode csDS))


spreadCode :: CompSpreadDS -> String 
spreadCode csDS =
    let tempVar_t = "spread_tmp1" 
    in concat $
    ([
      "if(" ++ determineLoad_t ++ ") {"
     , "  " ++ determineLoad_t ++ " = false;"
     , "  " ++ loadTime_t ++ " = " ++ getUserLoadTime_t ++ "() - " ++ loadTime_t ++ ";"
     , "  " ++ loadInfo_t ++ ".spentTime = " ++ loadTime_t ++ ";"
     , "  " ++ loadInfo_t ++ ".globalTimeStep = " ++ glT_t ++ "1;"
     , "  MPI_Isend(&" ++ loadInfo_t ++ ", 1, MPI_LoadInfo, " ++ 
      lbpServerID_t ++ "," ++ lbpMessageTag_t ++ ", MPI_COMM_WORLD, &" ++ 
      request_t ++ ");"
     , "} else {"
     , "  " ++ determineLoad_t ++ " = true;"
     , "  MPI_Wait(&" ++ request_t ++ ", &" ++ mpiStatus_t ++ ");"
     , "  " ++ mpiProbe ++ "(" ++ lbpServerID_t ++ ", " ++ lbpMessageTag_t ++ 
      ", MPI_COMM_WORLD, &" ++ mpiStatus_t ++ ");"
     , "  " ++ mpiGetCount ++ "(&" ++ mpiStatus_t ++ ", MPI_INT, &" ++ mpiCount_t ++ ");"
     , "  " ++ recvBuffer_t ++ ".resize(" ++ mpiCount_t ++ ");"
     , "  " ++ mpiRecv ++ "(&" ++ recvBuffer_t ++ "[0], " ++ mpiCount_t ++ 
      ", MPI_INT, " ++ lbpServerID_t ++ ", " ++ lbpMessageTag_t ++ ", MPI_COMM_WORLD, &" ++ 
      mpiStatus_t ++ ");"
     , "  for(int " ++ tempVar_t ++ " = 0; " ++ tempVar_t ++ " < " ++ mpiCount_t ++ 
      " / sizeof(RedistributionInformation); " ++ tempVar_t ++ "++) {"
     , "    " ++ recvBufferOffset_t ++ " = " ++ tempVar_t ++ " * sizeof(RedistributionInformation);"
     , "    " ++ redistInfo_t ++ ".VPinterval.minimumProcessor = " ++ recvBuffer_t ++ "[" ++ 
      recvBufferOffset_t ++ "];"
     , "    " ++ redistInfo_t ++ ".VPinterval.maximumProcessor = " ++ recvBuffer_t ++ "[" ++ recvBufferOffset_t ++ "+1];"
     , "    " ++ redistInfo_t ++ ".RPnum = " ++ recvBuffer_t ++ "[" ++ recvBufferOffset_t ++ "+2];"
      --        , "    " ++ redistInfo_t ++ ".time = " ++ recvBuffer_t ++ "[" ++ recvBufferOffset_t ++ "+3];"
     , "    assert(" ++ redistInfo_t ++ ".VPinterval.minimumProcessor >= 0"
     , "       && " ++ redistInfo_t ++ ".VPinterval.maximumProcessor < " ++ nVP_t ++ ");"
     , "    assert(" ++ redistInfo_t ++ ".RPnum >= 0 && " ++ redistInfo_t ++ ".RPnum <= " ++ nNodes_t ++ ");"
      --     printf(" [%i;%i],t%i->P%i",
                    --            redistInfo.VPinterval.minimumProcessor,
                    --            redistInfo.VPinterval.maximumProcessor,
                    --            redistInfo.time, redistInfo.RPnum);	
     , "    " ++ procMap_t ++ ".update(" ++ redistInfo_t ++ ".VPinterval,"
     , "       " ++ redistInfo_t ++ ".RPnum, "
     , "       " ++ glT_t ++ "1);"
     , "    " ++ loadTime_t ++ " = " ++ getUserLoadTime_t ++ "();"
     , "  }"
     , "}"
     ]
    )



------------------------------------------------------------------------------

-- (rpMapStmts: only for mapping from "procTiles" to "bufferID")
genRPMapStmtDefines :: CompSpreadDS -> Integer -> CloogStatement -> String
genRPMapStmtDefines csDS cloogStmtNr cloogStmt@RPMapStatement {} = rpDefines
    where
    rpD = getRPdim csDS
    tmpArray_t = loopo_t ++ "tmpArray"
    tmpArrayDecl = 
        -- int* tmpArray = new int[2]; 
        ("int* " ++ tmpArray_t ++ " = new int[" ++ show rpD ++ "]; " ++
         
         -- tmpArray[0] = procTile1; tmpArray[1] = procTile2;
	 concat [tmpArray_t ++ "[" ++ show (i-1) ++ "] = " ++ procTile_t ++ show i ++ "; " |
		 i <- [1..rpD]])

    tmpNTupel = 
	("NTupel(" ++ tmpArray_t ++ ", " ++ show rpD ++ ")")

    header_t = define_t ++ " " ++ cloogStmt_t ++ show cloogStmtNr ++ "() "

    rpDefines =
        (header_t ++
         tmpArrayDecl ++
         bufferIDMap_t ++ "[" ++ tmpNTupel ++ "] = " ++ bufferID_t ++ 
         -- " % " ++ mpiSize_t ++ 
         "; " ++ bufferID_t ++ "++; ")
        
