module CompSpread.DomainGen
where

import Debug.Trace(trace)

import List
import Ratio
import Maybe(fromMaybe)
import LoopoDS.Qual
import LoopoDS.Statement
import LoopoDS.TrDep
import LMath.Types
import LMath.Algo
import Tiling(defaultTiledLoops,TileInformation(..),
              defaultTile)

import VisualPolylib(showIS2VisualPolylib)

import CompSpread.Data
import CompSpread.DomainDescriptions
import CompSpread.CloogStatements
import CompSpread.Helpers

import CodeGen(loopNest) -- debugging!

timeInfo = DimInfo Sched Nothing
procInfo = DimInfo Place Nothing
addInfo  = DimInfo Additional Nothing
isInfo   = DimInfo Additional Nothing

-- =========================================================

compStmtDomDesc =
    StmtDomDescription [ DpTile  timeInfo, DpTile  procInfo
                         -- tispc wegprojeziert:
                         -- , DpTISPC timeInfo, DpTISPC procInfo, DpTISPC addInfo
                       , DpISPC  isInfo
                       ]

writeStmtDomDesc =
    -- ACHTUNG: total veraltet! (04.06.2006)
    StmtDomDescription [ -- srcTime wegprojeziert 
                         DpTile  procInfo -- srcProc
                         -- tispc wegprojeziert:
                         -- , DpTISPC timeInfo, DpTISPC procInfo, DpTISPC addInfo
                       , DpTile  timeInfo, DpTile procInfo -- destTime, destProc
                       , DpISPC  isInfo  -- srcISPC
                       ]

sendStmtDomDesc =
    StmtDomDescription [ 
                         DpTile timeInfo -- srcTime
                       , DpTile procInfo -- srcProc
                       , DpTile timeInfo -- destTime
                       , DpTile procInfo -- destProc
                       ]

recvStmtDomDesc =
    StmtDomDescription [ 
                         DpTile timeInfo -- srcTime
                       , DpTile procInfo -- srcProc
                       , DpTile timeInfo -- destTime
                       , DpTile procInfo -- destProc
                       ]

spreadStmtDomDesc = StmtDomDescription [ DpTile timeInfo ]

rpStmtDomDesc = StmtDomDescription [ DpTile  procInfo ]

-- =========================================================

genMainCloogStatements :: CompSpreadDS -> CloogStatements
genMainCloogStatements csDS = 
    verbose "genMainCloogStatements" $
    CloogStatements { clCompStmts = genCompStatements csDS
                    , clWriteStmts = genWriteStatements csDS
                    , clUnpackStmts = genUnpackStatements csDS
                    , clSendStmts = genSendStatements csDS
                    , clRecvStmts = genRecvStatements csDS
--                    , clSpreadStmts = genSpreadStatements csDS
                    }

genCompStatements :: CompSpreadDS -> [CloogStatement]
genCompStatements csDS = compStmts
    where
    compStmts = map (\trStmtID ->
                     ComputeStatement { clStmtDomain = 
                                          [genCompStatementIS csDS trStmtID]
                                      , clStmtDomDesc = compStmtDomDesc
                                      , clStmtTrStmtID = trStmtID
                                      }
                    ) (getAllTStatementIDs csDS)

genCompStatementIS :: CompSpreadDS -> TStatementID -> IneqSystem
genCompStatementIS csDS trStmtID = 
    verbose "genCompStatementIS" $
    trace "TISPC: \n" $
    trace (showIS2VisualPolylib tispcIS (fromInteger $ tispcD)) $    
    trace "CompIS: \n" $
    trace (showIS2VisualPolylib compIS (fromInteger $ tileNrD + isD)) $    
    compIS
    where
    compIS = 
        -- project to compStmtDomDesc:
        projectDomDescriptionTo csDS fullIS fullCompStmtDomDesc compStmtDomDesc
                                (Left trStmtID)
    fullIS = 
        verbose "fullIS" $
        ineqSystem ((tileNrD + tispcD + isD, pD), 
                    myMultijoin1 [is_M, trafoEq_M, trc_M, tile_M])
    fullCompStmtDomDesc =
        StmtDomDescription [ DpTile  timeInfo, DpTile  procInfo
                           , DpTISPC timeInfo, DpTISPC procInfo, DpTISPC addInfo
                           , DpISPC  isInfo
                           ]
        
    stmtID = getStmtID csDS trStmtID

    (tr_v, tr_p, tr_c) = partsVPCih (getTrafoFct csDS trStmtID)
    (is_v, is_p, is_c) = partsVPC (getISPC csDS trStmtID)
    (trc_v, trc_p, trc_c) = partsVPC (getTRCis csDS trStmtID)
    (tile_v, tile_p, tile_c) = partsVPC tilingIS

--    (tilingIS, _) = defaultTile lattice (selfDepRelation (stmtID, emptyIS)) infos
    (tilingIS, _) = 
        trace ("tispcIS: " ++ show tispcIS) $
        trace ("selfDepRelation: " ++ show (selfDepRelation (stmtID, tispcIS))) $
        defaultTile lattice tiledDims (selfDepRelation (stmtID, tispcIS)) infos

    tispcIS = getTISPC csDS trStmtID
    lattice = csLatVects csDS
    tiledDims = csTiledDims csDS
    infos = [SrcTile,SrcCoord]
    emptyIS = ineqSystem ((tispcD,pD), zeroMatrix 0 (tispcD + pD + 1))

    tileNrD = getTilingDim csDS
    isD = getISPCdim csDS trStmtID
    tispcD = getCoordsDim csDS trStmtID
    pD = nColumns is_p

    tile_M = tile_v `conc` zeroMatrix (nRows tile_v) isD `conc` tile_p `conc` tile_c

    is_M = zeroMatrix (nRows is_v) (tileNrD + tispcD) `conc` 
           is_v `conc` is_p `conc` is_c

    trafoEq_M = trafoIneq_M `join` (-trafoIneq_M)
    trafoIneq_M = 
        (zeroMatrix (nRows tr_v) tileNrD `conc`
         (-(unitMatrix (nRows tr_v) tispcD)) `conc` 
         tr_v `conc` tr_p `conc` tr_c)

    trc_M =
        (zeroMatrix (nRows trc_v) (tileNrD + tispcD) `conc` 
         trc_v `conc` trc_p `conc` trc_c)
    

-- =========================================================

genWriteStatements :: CompSpreadDS -> [CloogStatement]
genWriteStatements csDS = map createWriteStmt $ trDeps
    where
    trDeps = csTrDeps csDS
    createWriteStmt :: TrDependence -> CloogStatement
    createWriteStmt trDep =
        verbose "createWriteStmt" $
	WriteStatement
	{ clStmtDomain = [writeIS]
	, clStmtDomDesc = writeStmtDomDesc
        , clStmtTrDepID = trDepID trDep
	} 
	where
	writeIS = finalWriteUnpackIS csDS baseIS trDep
	baseIS = baseWriteUnpackIS csDS trDep
	rpDim = getRPdim csDS

finalWriteUnpackIS :: CompSpreadDS 
                -> IneqSystem  -- ^ base IneqSystem
                -> TrDependence
                -> IneqSystem
finalWriteUnpackIS csDS baseIS trDep = 
    trace ("projectedVars: " ++ (show projectedVars)) $
    verbose "finalWriteUnpackIS" $    
    finalIS
    where
    finalIS = 
        verbose "finalIS" $
        projectToVarsMC baseIS projectedVars
              -- projectToVarsMC commConstraintsIS projectedVars
--     commConstraintsIS = 
--         ineqSystem ((tileD + srcCoordsD + procTileD + srcISdim, 
--                      pD),
--                     commConstraints_M)

    srcTStmtID = trDepSrcStmtID trDep
    srcCoordsD = getCoordsDim csDS srcTStmtID    
    srcISdim = getISPCdim csDS srcTStmtID

    tileD = getTilingDim csDS
    timeTileD = getTimeTileDim csDS
    procTileD = getProcTileDim csDS

    pD = getParamsDim csDS

    srcGlTD = 1
    destGlTD = 1

    -- vorher: srcGlobalTime | srcProcTile | srcTISPC | destGlobalTime | destProcTile | srcISPC
    -- nachher: srcGlobalTime | srcProcTile | destGlobalTime | destProcTile | srcISPC
    projectedVars =
        ([0..srcGlTD + procTileD-1] ++
         -- srcTISPC weg
	 [(srcGlTD + procTileD + srcCoordsD) .. 
	  (srcGlTD + procTileD + srcCoordsD + 
           destGlTD + procTileD + srcISdim-1)]
	)         

         
baseWriteUnpackIS :: CompSpreadDS -> TrDependence -> IneqSystem
baseWriteUnpackIS csDS trDep = 
    verbose "baseWriteUnpackIS" $
    baseIS
    where
    (tile_v, tile_p, tile_c) = partsVPC tilingIS
    (srcTr_v, srcTr_c, srcTr_p) = partsVPCih srcTrafoFct
    (srcIS_v, srcIS_p, srcIS_c) = partsVPC (getISPC csDS srcTStmtID)
    -- trc only needed for Src...
    (trc_v, trc_c, trc_p) = partsVPC (getTRCis csDS srcTStmtID)

    pD = nParams tilingIS
    tileD = getTilingDim csDS

    timeTileD = getTimeTileDim csDS
    timeTileUsedVect = genericTake timeTileD $ csTilingUsedVect csDS
    globalTiledTimeD = fromIntegral $ head $ elemIndices True timeTileUsedVect

    tileInfo = [SrcTile, SrcCoord,DestTile]
    tilingIS = 
        verbose "tilingIS" $
        fst $ defaultTile (csLatVects csDS) (csTiledDims csDS) 
                (trDepRelation trDep) tileInfo

    srcISdim = getISPCdim csDS srcTStmtID
    srcCoordsD = getCoordsDim csDS srcTStmtID
    srcTStmtID = trDepSrcStmtID trDep
    destTStmtID = trDepDestStmtID trDep

    srcTrD = nRows srcTr_v

    srcTrafoFct = getTrafoFct csDS srcTStmtID
    destTrafoFct = getTrafoFct csDS destTStmtID

    tile_M = (tile_v `conc` zeroMatrix (nRows tile_v) srcISdim `conc`
              tile_p `conc` tile_c)

    srcTr_M = (srcTrafoIS_M `join` (-srcTrafoIS_M))
    srcTrafoIS_M = (zeroMatrix srcTrD tileD `conc` (-(unitMatrix srcTrD srcCoordsD)) `conc`
                    zeroMatrix srcCoordsD tileD `conc`
                    srcTr_v `conc` srcTr_p `conc` srcTr_c)

    srcIS_M = (zeroMatrix (nRows srcIS_v) (tileD + srcCoordsD + tileD) `conc`
               srcIS_v `conc` srcIS_p `conc` srcIS_c)

    trc_M = (zeroMatrix (nRows trc_v) (tileD + srcCoordsD + tileD) `conc`
             trc_v `conc` trc_p `conc` trc_c)

    full_M = 
        myMultijoin1 [ 
                      tile_M
                     , srcTr_M
                     , srcIS_M
--                     , trc_M
                     ]
    fullIS = ineqSystem ((tileD+srcCoordsD+tileD+srcISdim, 
                          pD), full_M)

    projectedVars =        
        (concat $
         -- liefert: srcGlobalTime | srcProcTile | srcTISPC | destGlobalTime | destProcTile | srcISPC
         [[globalTiledTimeD]                          -- srcGlobalTime
         ,[timeTileD .. tileD-1]                      -- srcProcTile
         ,[tileD .. tileD + srcCoordsD-1]                     -- srcTISPC
         ,[(tileD + srcCoordsD) + globalTiledTimeD]   -- destGlobalTime
         ,[(tileD + srcCoordsD + timeTileD) .. 
           (tileD + srcCoordsD + tileD-1)]            -- destProcTile 
         ,[(tileD + srcCoordsD + tileD) ..
           (tileD + srcCoordsD + tileD + srcISdim-1)] -- srcISPC
         ]
        )

    baseIS = 
        trace ("projected vars: " ++ (show projectedVars)) $
        verbose "baseIS" $
        projectToVarsMC fullIS projectedVars


-- =========================================================

genUnpackStatements :: CompSpreadDS -> [CloogStatement]
genUnpackStatements csDS = map createUnpackStmt $ trDeps
    where
    trDeps = csTrDeps csDS
    createUnpackStmt :: TrDependence -> CloogStatement
    createUnpackStmt trDep =
	UnpackStatement
	{ clStmtDomain = [unpackIS]
	, clStmtDomDesc = writeStmtDomDesc
        , clStmtTrDepID = trDepID trDep
	} 
	where
	unpackIS = finalWriteUnpackIS csDS baseIS trDep
	baseIS = baseWriteUnpackIS csDS trDep
	rpDim = getRPdim csDS


-- =========================================================

genSendStatements :: CompSpreadDS -> [CloogStatement]
genSendStatements csDS = 
    [SendStatement
     { clStmtDomain = sendISs
     , clStmtDomDesc = sendStmtDomDesc
     } 
    ]
    where
    sendISs = map (sendRecvIS csDS) trDeps
    rpDim = getRPdim csDS
    trDeps = csTrDeps csDS

genRecvStatements :: CompSpreadDS -> [CloogStatement]
genRecvStatements csDS = 
    [RecvStatement
     { clStmtDomain = recvISs
     , clStmtDomDesc = sendStmtDomDesc
     } 
    ]
    where
    trDeps = csTrDeps csDS
    recvISs = map (sendRecvIS csDS) trDeps
    rpDim = getRPdim csDS

sendRecvIS :: CompSpreadDS -> TrDependence -> IneqSystem
sendRecvIS csDS trDep = 
    verbose "sendRecvIS" $
    trace ("sendReceiveIS: projectedVars=" ++ (show projectedVars)) $
    sendRecvIS
    where
    tileD = getTilingDim csDS

    timeTileD = getTimeTileDim csDS
    timeTileUsedVect = genericTake timeTileD $ csTilingUsedVect csDS
    globalTiledTimeD = fromIntegral $ head $ elemIndices True timeTileUsedVect

    tileInfo = [SrcTile, DestTile]
    tilingIS = 
        fst $ defaultTile (csLatVects csDS) (csTiledDims csDS) 
                (trDepRelation trDep) tileInfo

    projectedVars =
        (concat $
         -- liefert: srcGlobalTime | srcProcTile | destGlobalTime | destProcTile
         [[globalTiledTimeD]                          -- srcGlobalTime
         ,[timeTileD .. tileD-1]                      -- srcProcTile
         ,[tileD + globalTiledTimeD]                  -- destGlobalTime
         ,[tileD + timeTileD .. tileD + tileD-1]      -- destProcTile 
         ]
        )

    sendRecvIS = projectToVarsMC tilingIS projectedVars


-- =========================================================

{-- 28.08.2006: * nicht mehr benoetigt
 --             * durch eingefuegten statement-code ersetzt
genSpreadStatements :: CompSpreadDS -> [CloogStatement]
genSpreadStatements csDS = 
    verbose "genSpreadStatements" $
    [SpreadStatement { clStmtDomain = spreadDom
                     , clStmtDomDesc = spreadStmtDomDesc
                     }
    ]
    where
    allIDs = getAllTStatementIDs csDS
    compStmtDom = map (genCompStatementIS csDS) allIDs
    spreadDom = zipWith (genSpreadStmtIS csDS) compStmtDom allIDs
--}

{-- 25.08.2006: nicht mehr benoetigt
genCommStmtIS :: CompSpreadDS -> IneqSystem -> TStatementID -> IneqSystem
genCommStmtIS = genSpreadStmtIS
--}


{-- 28.08.2006: * nicht mehr benoetigt
 --             * durch eingefuegten statement-code ersetzt
genSpreadStmtIS :: CompSpreadDS -> IneqSystem -> TStatementID -> IneqSystem
genSpreadStmtIS csDS compIS trStmtID = 
    trace "spreadStmt: " $
    trace (showIS2VisualPolylib spreadIS (nVars spreadIS)) $
    trace "\nskewedIS: " $
    trace (showIS2VisualPolylib skewedIS (nVars skewedIS)) $
    trace "==============================================================\n" $
    trace "sortiertes skewedIS" $
    trace (loopNest skewedIS) $
    trace (show skewedIS) $
    trace "timeProcIS" $
    trace (loopNest timeProcIS) $
    trace (show timeProcIS) $
    trace "==============================================================\n" $
--    trace (showIS2VisualPolylib (projectToVarsMC skewedIS [0,1]) 2) $
    verbose "genSpreadStmtIS" $
    spreadIS
    where
--     timeProcDesc = StmtDomDescription [ DpTile timeInfo, DpTile procInfo ]
--     timeProcIS = 
--         projectDomDescriptionTo csDS compIS compStmtDomDesc timeProcDesc (Left trStmtID)
    globalTiledTimeDim = fromIntegral $ head $ elemIndices True timeTileUsedVect

    globalTimeDim = getGlobalTimeDim csDS
    rpDim = getRPdim csDS

    timeTileDim = getTimeTileDim csDS
    timeTileUsedVect = genericTake timeTileDim $ csTilingUsedVect csDS

    timeProcIS =
        projectToVarsMC compIS globalDims

    -- project to only global time and rp dims (for skewing)...
    globalDims = [globalTiledTimeDim] ++ [timeTileDim .. timeTileDim + rpDim-1]

    skewedIS = 
        verbose "skewedIS" $
        trace ("nVars timeProcIS: " ++ (show $ nVars timeProcIS)) $
        trace ("nVars (getSkewingFct csDS): " ++ (show $ nVars (getSkewingFct csDS))) $
        image (getSkewingFct csDS) timeProcIS

    spreadIS = 
--         projectDomDescriptionTo csDS skewedIS timeProcDesc spreadStmtDomDesc 
--                                     (Left trStmtID)
         projectToVarsMC skewedIS [0..globalTimeDim-1]

    skewedDim = nVars skewedIS
--}    


getSkewingFct :: CompSpreadDS -> AffineFunction
getSkewingFct csDS = 
    verbose "getSkewingFct" $
    affineFunction ((timeD + rpD, pD), skew_M) Homogeneous
    where
    timeD = getGlobalTimeDim csDS
    rpD = getRPdim csDS
    pD = getParamsDim csDS
    skew_M =
        myMultijoin1 
	[ (unitMatrix timeD timeD `conc` 
           (rowMatrix $ genericReplicate rpD (1%1)) `conc` zeroMatrix timeD (pD + 1))
        , (zeroMatrix rpD timeD `conc` unitMatrix rpD rpD `conc` 
           zeroMatrix rpD (pD + 1))
        , (zeroMatrix (pD+1) (timeD + rpD) `conc` unitMatrix (pD+1) (pD+1))
        ]

-- =========================================================

genRPMapStatements :: CompSpreadDS -> [CloogStatement]
genRPMapStatements csDS = [RPMapStatement { clStmtDomain = rpMapDom
                                          , clStmtDomDesc = rpStmtDomDesc
                                          }
                          ]
    where
    allIDs = getAllTStatementIDs csDS
    compStmtDom = map (genCompStatementIS csDS) allIDs
    rpMapDom = zipWith (projectToRPs csDS) compStmtDom allIDs


projectToRPs :: CompSpreadDS -> IneqSystem -> TStatementID -> IneqSystem
projectToRPs csDS compIS trStmtID = 
    projectDomDescriptionTo csDS compIS compStmtDomDesc rpStmtDomDesc (Left trStmtID)
