module CompSpread.Names
where

import List
import LMath.Types
import CompSpread.Data
import CompSpread.DomainDescriptions
import CompSpread.CloogStatements
import CompSpread.Helpers

getIteratorNames :: CompSpreadDS -> [String]
getIteratorNames csDS = iteratorNames
    where
    iteratorNames = ["iter" ++ show i | i <- [1..numTotalVars]]
    numTotalVars = sum [glTD, procTileD, glTD, procTileD, maxISPCD]
    glTD = getGlobalTimeDim csDS
    procTileD = getProcTileDim csDS
    maxISPCD = getMaxISPCdim csDS

getUsedIteratorNames :: CompSpreadDS -> CloogStatement -> [String]
getUsedIteratorNames csDS clStatement =
    genericTake vars iteratorNames
    where
    domain = clStmtDomain clStatement
    vars = if not $ null domain
           then nVars (head domain)
           else error "getUsedIteratorNames: domain should not be empty!"
    iteratorNames = getIteratorNames csDS

getScatterNames :: CompSpreadDS -> [String]
getScatterNames csDS = scatterNames
    where
    commonQualVect = getCommonQualVect (csQualVects csDS)
    allQualVects = snd $ unzip $ csQualVects csDS
    maxQualVect = argMax genericLength allQualVects
    maxCoordsD = getMaxCoordsDim csDS
    commonCoordsD = getCommonCoordsDim csDS
    maxAccD = getArrAccessMaxDim csDS
		  
    globalTimeD = getGlobalTimeDim csDS
    rpD = getRPdim csDS
    vTimeD = getSchedDim csDS
    vProcD = getPlaceDim csDS
    vAddedD = maxCoordsD - commonCoordsD
    arrNrD = 1
    
    scatterNames = 
	concat $
	[[glT_t ++ (show i) | i <- [1..globalTimeD]]
	,[stmtType_t]
	,[destProc_t ++ (show i) | i <- [1..rpD]]
        ,[srcTime_t ++ (show i) | i <- [1..globalTimeD]] 
	,[srcProc_t ++ (show i) | i <- [1..rpD]]
        ,[inner_t ++ (show i) | i <- [1..(max maxAccD maxCoordsD)]]
--	,[arrNr_t ++ (show i) | i <- [1..arrNrD]]
	]

loopo_t = "loopo___"

glT_t       = "glT"
stmtType_t  = "stmtType"
destProc_t        = "destProc"
srcTime_t   = "srcTime"
srcProc_t   = "srcProc"
-- vTime_t     = "vT"
-- vProc_t     = "vP"
-- vAdded_t    = "vAdd"
--arrNr_t     = "arrNr"
inner_t     = "coords"

srcRP_t = "srcRP"
destRP_t = "destRP"

-- for rpMap-defines:
procTile_t = "procTile"
bufferIDMap_t = loopo_t ++ "rpMap"
bufferID_t = loopo_t ++ "bufferID"

-- for spread-statement:
determineLoad_t = "determineLoad"
loadTime_t = "loadTime"
loadInfo_t = "loadInfo"
lbpServerID_t = "LBP_SERVER_ID"
lbpMessageTag_t = "LBP_MESSAGE_TAG"
request_t = "request"
recvBuffer_t = "recvBuffer"
redistInfo_t = "redistInfo"
recvBufferOffset_t = "recvBufferOffset"
nVP_t = "nVP"
nNodes_t = "nNodes"
procMap_t = "pMap" 
getUserLoadTime_t = "getcputime" -- in comp_spread_lib.h


-- general defines:
define_t = "#define"
cloogStmt_t = "S"

lookupRP_t  = "getRP"
mpiSize_t   = "mpiSize"
nodeRank_t   = "rankLBP"
mpiStatus_t = "mpiStatus"
mpiCount_t  = "mpiCount"

sendVector_t = "sendVect"
recvVector_t = "recvVect"

rdispls_t = "rdispls"
