  
  
struct ProcessorInterval {
    int minimumProcessor;
    int maximumProcessor;
};
   
/**
 * Abstract class ProcessorMap
 * Provides an interface for the following mappings:
 * (VP,t) -> RP
 * RP -> [VP]
 * As both methods are called frequently, performance is a major issue here.
 */   
class ProcessorMap {
   
   protected: 
   
   int VPcount; //number of virtual processors in program
   int RPcount; //number of real processors; may change during runtime 
   
   public:
     
     /**
      * Non-virtual standard constructors
      */
     //ProcessorMap();
     //ProcessorMap(const ProcessorMap&);
    

     /**
      * Extracts the RP which performed calculations of the specified VP at
      * the specified time. 
      */
     virtual int getRP(const int VPnum, const int timeStep) = 0;
     
     /**
      * Extracts the numbers of all VPs presently belonging to a RP.
      * @note: There is no access to former VPs of an RP. 
      * @return: result[0] element [0;inf[ means number of succeeding elements;
      *          The first resultVP is located at result[1] and so on; allocated
      *          with 'new'
      */
     virtual int *getVPs(const int RPnum) const = 0;
     
     /**
      * Transfers an interval of VPs to the responsibility of an RP. 
      * This routine is called whenever distribution of virtual processors
      * is bound to change. 
      * @param VPinterval: VPs to move to another RP
      * @param RP: RP to which load is transferred. 
      * @param time: From this time on (including) the interval belongs to the 
      *              new RP.
      * @param timeStep: current Time at which changes take effect. 
      *        Beware: be u and u' two different calls to 'update'; 
      *                u.timestep <= u'.timestep <- u was before u'
      */
     virtual void update(const ProcessorInterval VPinterval, 
                         const int RPnum, const int time) = 0;
    
   };
