module CompSpread.ScatterFunctions
where

import Debug.Trace(trace)

import List
import Ratio
import LoopoDS.Qual
import LoopoDS.Statement
import LoopoDS.TrDep
import LoopoDS.IntegralID
import LoopoDS.Access
import LMath.Types
import LMath.Algo

import CompSpread.Data
import CompSpread.DomainDescriptions
import CompSpread.CloogStatements
import CompSpread.Helpers



genCompStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genCompStmtScatterFct csDS ComputeStatement { clStmtTrStmtID = trStmtID } = comp_M
    where
    timeTileD = getTimeTileDim csDS
    procTileD = getProcTileDim csDS

    rpD = getRPdim csDS    
    srcRPD = getRPdim csDS
    isD = getISPCdim csDS trStmtID
    coordsD = getMaxCoordsDim csDS
    addD = getMaxCoordsDim csDS - getCoordsDim csDS trStmtID
    pD = getParamsDim csDS

    glTD = 1
    srcTimeD = 1
    stmtTypeD = 1 -- helper dimension for "scattering"

    -- arrayNr should be derived from the statement number (dependence Nr)
    -- arrayNrD  = 1 -- scatter dimension holding the array number

    stmtTypeNr = getStmtTypeNr ComputeStatement { clStmtDomain   = e
                                                , clStmtDomDesc  = e
                                                , clStmtTrStmtID = e
                                                }
          where
          e = error "genCompStmtScatterFct: illegal access to labeled fields!"

    trafoFct = getTrafoFct csDS trStmtID
    fullTrafoFct = affineFunction ((isD, pD),
                                   getMatrixIH trafoFct `join`
                                   zeroMatrix addD (isD + pD + 1)
                                   ) Inhomogeneous

    timeTileUsedVect = genericTake timeTileD $ csTilingUsedVect csDS
    timeTileRHS_M = rowMatrix $ map (\d -> if d then (-1%1) else 0) timeTileUsedVect
    skew_M = rowMatrix $ genericReplicate rpD (-1%1)
    fillVarsD = (max coordsD (getArrAccessMaxDim csDS)) - coordsD

    comp_M = myMultijoin1 [ glT_M, stmtType_M, rp_M, srcTime_M, srcRP_M, coords_M, fillVars_M ]
             
    glT_M =
        -- scatter-dims:
        (unitMatrix glTD glTD `conc` zeroMatrix glTD stmtTypeD `conc`
         skew_M `conc` zeroMatrix glTD (srcTimeD + srcRPD + coordsD) `conc`
        -- domain-dims:
         timeTileRHS_M `conc` zeroMatrix glTD rpD `conc`
         zeroMatrix glTD (isD + pD + 1))

    stmtType_M =
        -- scatter-dims:
        (zeroMatrix stmtTypeD glTD `conc` unitMatrix stmtTypeD stmtTypeD `conc`
         zeroMatrix stmtTypeD (rpD + srcTimeD + srcRPD + coordsD + 
        -- domain-dims:
                               timeTileD + procTileD + isD + pD) `conc`
         (-(matrix [[stmtTypeNr%1]])))

    rp_M = 
        (zeroMatrix rpD (glTD + stmtTypeD) `conc` unitMatrix rpD rpD `conc`
         zeroMatrix rpD (srcTimeD + srcRPD + coordsD + timeTileD) `conc`
         (-(unitMatrix rpD procTileD)) `conc` 
         zeroMatrix rpD (isD + pD + 1))

    srcTime_M = 
        (zeroMatrix srcTimeD (glTD + stmtTypeD + rpD) `conc` unitMatrix srcTimeD srcTimeD `conc`
         skew_M `conc`
         zeroMatrix srcTimeD (coordsD + timeTileD) `conc`
         (-(unitMatrix srcTimeD srcTimeD)) `conc` 
         zeroMatrix srcTimeD (isD + pD + 1))

    srcRP_M =
        (zeroMatrix srcRPD (glTD + stmtTypeD + rpD + srcTimeD) `conc`
         unitMatrix srcRPD srcRPD `conc`
         zeroMatrix srcRPD (coordsD + timeTileD + procTileD + isD + pD + 1)
        )
         

    coords_M =
        (zeroMatrix coordsD (glTD + stmtTypeD + rpD + srcTimeD + srcRPD) `conc`
         unitMatrix coordsD coordsD `conc`
         zeroMatrix coordsD (timeTileD + procTileD) `conc`
         (-(getMatrixIH fullTrafoFct)))

    fillVars_M =
        (zeroMatrix fillVarsD (glTD + stmtTypeD + rpD + srcTimeD + srcRPD + 
                               coordsD) `conc`
         unitMatrix fillVarsD fillVarsD `conc`
         zeroMatrix fillVarsD (timeTileD + procTileD + isD + pD + 1)
        )

genBufferStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genBufferStmtScatterFct csDS clStmt = 
    verbose "genBufferStmtScatterFct" $
    scatter_M
    where
    trDepID = clStmtTrDepID clStmt
    srcTStmtID = trDepSrcStmtID (getTrDep csDS trDepID)

    timeTileD = getTimeTileDim csDS
    procTileD = getProcTileDim csDS
    timeTileUsedVect = genericTake timeTileD $ csTilingUsedVect csDS

    destRPD = getRPdim csDS    
    srcRPD = getRPdim csDS
    isD = getISPCdim csDS srcTStmtID
    coordsD = getMaxCoordsDim csDS
    srcCoordsD = getCoordsDim csDS srcTStmtID   
    addD = getMaxCoordsDim csDS - getCoordsDim csDS srcTStmtID
    pD = getParamsDim csDS
    accMaxD = getArrAccessMaxDim csDS
    accD = targetDimIH accFct
    accAddD = accMaxD - accD
    accFct = accIndexFct srcAcc
    srcAcc = getSrcAcc csDS  trDepID
    -- arrayNr = integerFromID $ accArrayID srcAcc
    fullAccFct =
        trace (show $ getMatrixIH accFct) $        
        verbose "fullAccFct" $
        affineFunction ((isD, pD),
                        (trace (show $ nColumns (getMatrixIH accFct)) $
                         trace "bla" $
                         getMatrixIH accFct) `join`
                        (trace (show $ isD + pD + 1) $
                         zeroMatrix accAddD (isD + pD + 1)
                        )
                       ) Inhomogeneous

    trafoFct = getTrafoFct csDS srcTStmtID
--     fullTrafoFct = 
--         verbose "fullTrafoFct" $
--         affineFunction ((isD, pD),
--                         getMatrixIH trafoFct `join`
--                         zeroMatrix addD (isD + pD + 1)
--                        ) Inhomogeneous
    glTD = 1
    srcTimeD = 1
    stmtTypeD = 1 -- helper dimension for "scattering"
    arrayNrD  = 1 -- scatter dimension holding the array number

    fillVarsD = (max coordsD (getArrAccessMaxDim csDS)) - accMaxD

    scatter_M = 
        verbose "scatter_M" $
        myMultijoin1
        [ glT_M, stmtType_M, rp_M, srcTime_M, srcRP_M, acc_M, fillVars_M]

    skew_M = rowMatrix $ genericReplicate destRPD (-1%1)
--    timeTileRHS_M = rowMatrix $ map (\d -> if d then (-1%1) else 0) timeTileUsedVect

    glT_M = 
        verbose "glT_M" $
        (unitMatrix glTD glTD `conc` 
         -- globalTime actually is destTileTime, so skew by destRP:
         zeroMatrix glTD stmtTypeD `conc`
         skew_M `conc`
         zeroMatrix glTD (srcTimeD + srcRPD)  `conc`
         zeroMatrix glTD (accMaxD + fillVarsD +
                          -- domain-dims:
                          srcTimeD + procTileD) `conc`
--         timeTileRHS_M `conc`
         (-(unitMatrix glTD glTD)) `conc`
         zeroMatrix glTD (procTileD + isD + pD + 1)
        )

    stmtType_M =
        verbose "stmtType_M" $
        (zeroMatrix stmtTypeD glTD `conc` unitMatrix stmtTypeD stmtTypeD `conc`
         zeroMatrix stmtTypeD (destRPD + srcTimeD + srcRPD + accMaxD + fillVarsD +
                               -- domain-dims:
                               srcTimeD + procTileD + glTD + procTileD + isD + 
                               pD) `conc`
         (matrix [[-(getStmtTypeNr clStmt)%1]])
        )

    rp_M =
        verbose "rp_M" $
        (zeroMatrix destRPD (glTD + stmtTypeD) `conc`
         unitMatrix destRPD destRPD `conc`
         zeroMatrix destRPD (srcTimeD + srcRPD + accMaxD + fillVarsD +
                             srcTimeD) `conc`
         (-(unitMatrix destRPD procTileD)) `conc`
         zeroMatrix destRPD (glTD + procTileD + isD + pD + 1)
        )

    srcTime_M =
        verbose "rp_M" $
        (zeroMatrix srcTimeD (glTD + stmtTypeD + destRPD) `conc`
         unitMatrix srcTimeD srcTimeD `conc`
         skew_M `conc`
         zeroMatrix srcTimeD (accMaxD + fillVarsD) `conc`
         (-(unitMatrix srcTimeD srcTimeD)) `conc`
         zeroMatrix srcTimeD (procTileD + glTD + procTileD + isD + pD + 1)
        )

    srcRP_M =
        verbose "srcRP_M" $
        (zeroMatrix srcRPD (glTD + stmtTypeD + destRPD + srcTimeD) `conc`
         unitMatrix srcRPD srcRPD `conc`
         zeroMatrix srcRPD (accMaxD + fillVarsD +
                            srcTimeD + procTileD + glTD) `conc`
         (-(unitMatrix srcRPD procTileD)) `conc`
         zeroMatrix srcRPD (isD + pD + 1)
        )

--     coords_M =
--         verbose "coords_M" $
--         (zeroMatrix coordsD (glTD + stmtTypeD + destRPD + srcTimeD + srcRPD) `conc`
--          unitMatrix coordsD coordsD `conc` 
--          zeroMatrix coordsD (srcTimeD + procTileD + glTD + procTileD) `conc`
--          (-(getMatrixIH fullTrafoFct))
--         )

    acc_M =
        verbose "acc_M" $
        (zeroMatrix accMaxD (glTD + stmtTypeD + destRPD + srcTimeD + srcRPD) `conc`
         unitMatrix accMaxD accMaxD `conc` 
         zeroMatrix accMaxD (fillVarsD +
                             srcTimeD + procTileD + glTD + procTileD) `conc`
         (-(getMatrixIH fullAccFct))
        )


--     arrNr_M =
--         verbose "arrNr_M" $
--         (zeroMatrix arrayNrD (glTD + stmtTypeD + destRPD + srcRPD + coordsD) `conc`
--          unitMatrix arrayNrD arrayNrD `conc`
--          zeroMatrix arrayNrD (procTileD + timeTileD + procTileD + 
--                               isD + pD) `conc`
--          (-(matrix [[arrayNr%1]]))         
--         )
         
    fillVars_M =
        (zeroMatrix fillVarsD (glTD + stmtTypeD + destRPD + srcTimeD + srcRPD + 
                               accMaxD) `conc`
         unitMatrix fillVarsD fillVarsD `conc`
         zeroMatrix fillVarsD (srcTimeD + procTileD + glTD + procTileD +
                               isD + pD + 1)
        )  
         
                         

genWriteStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genWriteStmtScatterFct csDS stmt@WriteStatement { clStmtTrDepID = trDepID } =
    genBufferStmtScatterFct csDS stmt
    

genUnpackStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genUnpackStmtScatterFct csDS stmt@UnpackStatement { clStmtTrDepID = trDepID } =
    genBufferStmtScatterFct csDS stmt

genSendStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genSendStmtScatterFct csDS stmt@SendStatement {} =
    genCommStmtScatterFct csDS stmt

genRecvStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genRecvStmtScatterFct csDS stmt@RecvStatement {} =
    genCommStmtScatterFct csDS stmt


genCommStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genCommStmtScatterFct csDS clStmt = 
    verbose "genCommStmtScatterFct" $
    scatter_M
    where
    trDepID = clStmtTrDepID clStmt
    srcTStmtID = trDepSrcStmtID (getTrDep csDS trDepID)

    timeTileD = getTimeTileDim csDS
    procTileD = getProcTileDim csDS
    timeTileUsedVect = genericTake timeTileD $ csTilingUsedVect csDS

    destRPD = getRPdim csDS    
    srcRPD = getRPdim csDS
    pD = getParamsDim csDS

    glTD = 1
    srcTimeD = 1
    stmtTypeD = 1 -- helper dimension for "scattering"
    arrayNrD  = 1 -- scatter dimension holding the array number

    accMaxD = getArrAccessMaxDim csDS
    coordsMaxD = getMaxCoordsDim csDS
    fillVarsD = max coordsMaxD accMaxD

    scatter_M = 
        verbose "scatter_M" $
        myMultijoin1
        [ glT_M, stmtType_M, rp_M, srcTime_M, srcRP_M, fillVars_M]

    skew_M = rowMatrix $ genericReplicate destRPD (-1%1)

    glT_M = 
        verbose "glT_M" $
        (unitMatrix glTD glTD `conc` 
         -- globalTime actually is destTileTime, so skew by destRP:
         zeroMatrix glTD stmtTypeD `conc`
         skew_M `conc`
         zeroMatrix glTD (srcTimeD + srcRPD) `conc`
         zeroMatrix glTD (fillVarsD +
                          -- domain-dims:
                          srcTimeD + procTileD) `conc`
--         timeTileRHS_M `conc`
         (-(unitMatrix glTD glTD)) `conc`
         zeroMatrix glTD (procTileD + pD + 1)
        )

    stmtType_M =
        verbose "stmtType_M" $
        (zeroMatrix stmtTypeD glTD `conc` unitMatrix stmtTypeD stmtTypeD `conc`
         zeroMatrix stmtTypeD (destRPD + srcTimeD + srcRPD + fillVarsD +
                               -- domain-dims:
                               srcTimeD + procTileD + glTD + procTileD + 
                               pD) `conc`
         (matrix [[-(getStmtTypeNr clStmt)%1]])
        )

    rp_M =
        verbose "rp_M" $
        (zeroMatrix destRPD (glTD + stmtTypeD) `conc`
         unitMatrix destRPD destRPD `conc`
         zeroMatrix destRPD (srcTimeD + srcRPD + fillVarsD +
                             srcTimeD + procTileD + 
                             glTD) `conc`
         (-(unitMatrix destRPD procTileD)) `conc`
         zeroMatrix destRPD (pD + 1)
        )

    srcTime_M =
        verbose "rp_M" $
        (zeroMatrix srcTimeD (glTD + stmtTypeD + destRPD) `conc`
         unitMatrix srcTimeD srcTimeD `conc`
         skew_M `conc`
         zeroMatrix srcTimeD (fillVarsD) `conc`
         (-(unitMatrix srcTimeD srcTimeD)) `conc`
         zeroMatrix srcTimeD (procTileD + glTD + procTileD + pD + 1)
        )

    srcRP_M =
        verbose "srcRP_M" $
        (zeroMatrix srcRPD (glTD + stmtTypeD + destRPD + srcTimeD) `conc`
         unitMatrix srcRPD srcRPD `conc`
         zeroMatrix srcRPD (fillVarsD +
                            srcTimeD) `conc`
         (-(unitMatrix srcRPD procTileD)) `conc`
         zeroMatrix srcRPD (glTD + procTileD + pD + 1)
        )

         
    fillVars_M =
        (zeroMatrix fillVarsD (glTD + stmtTypeD + destRPD + srcTimeD + srcRPD) `conc`
         unitMatrix fillVarsD fillVarsD `conc`
         zeroMatrix fillVarsD (srcTimeD + procTileD + glTD + procTileD +
                               pD + 1)
        )  

genSpreadStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genSpreadStmtScatterFct csDS SpreadStatement { } = scatter_M
    where
    scatter_M = myMultijoin1
                [ glT_M, stmtType_M, fillVars_M ] -- rp_M, srcTime_M, otherRP_M, coords_M
    fillVarsD = destRPD + srcTimeD + srcRPD + (max coordsD accMaxD)
    stmtNr = getStmtTypeNr (SpreadStatement { clStmtDomain   = e
                                            , clStmtDomDesc  = e
                                            })
        where
        e = error "genSpreadStmtScatterFct: illegal access to labeled fields!"

    destRPD = getRPdim csDS    
    srcRPD = getRPdim csDS
    coordsD = getMaxCoordsDim csDS
    accMaxD = getArrAccessMaxDim csDS
    glTD = 1
    srcTimeD = 1
    stmtTypeD = 1 -- helper dimension for "scattering"
    -- arrayNrD  = 1 -- scatter dimension holding the array number
    pD = getParamsDim csDS

    glT_M = (unitMatrix glTD glTD `conc`
             zeroMatrix glTD (stmtTypeD + fillVarsD) `conc`
             (-(unitMatrix glTD glTD)) `conc`
--             zeroMatrix glTD (pD + 1)
             zeroMatrix glTD pD `conc`
             (-(matrix [[1%1]]))
            )
    
    stmtType_M = (zeroMatrix stmtTypeD glTD `conc`
                  unitMatrix stmtTypeD stmtTypeD `conc`
                  zeroMatrix stmtTypeD (fillVarsD + glTD + pD) `conc`
                  (-matrix [[stmtNr%1]])
                 )
    fillVars_M = (zeroMatrix fillVarsD (glTD + stmtTypeD) `conc`
                  unitMatrix fillVarsD fillVarsD `conc`
                  zeroMatrix fillVarsD (glTD + pD + 1)
                 )
    

genRPMapStmtScatterFct :: CompSpreadDS -> CloogStatement -> Matrix
genRPMapStmtScatterFct csDS stmt = rpMapping_M
    where
    rpD = getRPdim csDS
    pD = getParamsDim csDS
    rpMapping_M =
        myMultijoin1
        -- kein skewing noetig, da nur auf realprozessor-dimensionen projeziert wird!
        [(unitMatrix rpD rpD `conc`
          (-(unitMatrix rpD rpD)) `conc`
          zeroMatrix rpD (pD + 1))]
