#include <iostream>

#include <assert.h>
#include <stdlib.h>
#include "CppSide.h"

Pair mkPair(void *a, void *b) {
    void **p = (void **)malloc(2 * sizeof(void *));
    if (p == NULL)
	oomInLoopo();
    p[0] = a;
    p[1] = b;
    return p;
}

// out of memory in Loopo
void oomInLoopo() {
    std::cerr << "out of memory in (the C++ side of) Loopo" << std::endl;
    exit(1);
}


/*
 * Currently, GHC uses GMP to implement its Integer data type;
 * therefore, GMP's allocation functions must be set to
 * malloc etc. (with gmp_protect()) before calling PolylibGMP
 * functions and reset to GHC's own functions before returning
 * to the Haskell side (with gmp_unprotect()).
 */
#if POLYLIB_BITS==gmp

#include <gmp.h>

static void *(*alloc_func) (size_t);
static void *(*realloc_func) (void *, size_t, size_t);
static void (*free_func) (void *, size_t);

static void oomInGMP() {
    std::cerr << "Out of memory in GMP allocation" << std::endl;
    exit(1);
}

static void *my_alloc(size_t n) {
    void *p = malloc(n);
    if (p == NULL)
	oomInGMP();
    return p;
}

static void *my_realloc(void *p, size_t oldsize, size_t newsize) {
//    fprintf(stderr, "my_realloc(%p,%d,%d)\n", p, (int)oldsize, (int)newsize);
    p = realloc(p, newsize);
    if (p == NULL)
	oomInGMP();
    return p;
}

static void my_free(void *p, size_t size) {
    free(p);
}

void gmp_protect() {
//    fprintf(stderr, "gmp_protect\n");
    mp_get_memory_functions (&alloc_func, &realloc_func, &free_func);
    mp_set_memory_functions (my_alloc, my_realloc, my_free);
}

void gmp_unprotect() {
//    fprintf(stderr, "gmp_unprotect\n");
    mp_set_memory_functions (alloc_func, realloc_func, free_func);
}
#endif
