
#include "barvinokpolylib.h"

// defined in polylib.cc
extern Value *copy_value(Value &);

//
// Ehrhart
//

static Enumeration *enumerateUnion(Polyhedron *phUnion, Polyhedron *context) {
#ifdef HAVE_BARVINOK
    evalue *sum;
    Enumeration *enu;

    printf("barvinok\n");

    sum = evalue_zero();

    Polyhedron *phDisjUnion = Disjoint_Domain(phUnion, 0, 0);

    Polyhedron *p=phDisjUnion;
    while (p != NULL) {
	Polyhedron *q = p;
	p = p->next;
	q->next = NULL;
	evalue *ev = barvinok_enumerate_ev(q, context, 0);
	eadd(ev, sum);
	evalue_free(ev);
    }

    // printf("barvinok_enumerate_union: ");
    // sum = barvinok_enumerate_union(phUnion, context, 0);
    enu = partition2enumeration(sum); // frees 'sum'
/*
    if (value_sign(sum->d) == 0 && sum->x.p->type == partition) {
	printf("partition\n");
	enode *en = sum->x.p;
	int size = en->size;
	int i;
	for (i=0; i<size/2; i++) {
	    Enumeration *e2 = (Enumeration *)malloc(sizeof(Enumeration));
	    e2->ValidityDomain = en->arr[2*i].x.D;
	    e2->EP = en->arr[2*i+1];
	    e2->next = NULL;
	    if (eStart == NULL)
		eStart = e2;
	    else 
		e->next = e2;
	    e = e2;
	}
	free(en);
	free(sum);
    } else {
	printf("universe\n");
	e = (Enumeration *)malloc(sizeof(Enumeration));
	e->ValidityDomain = Universe_Polyhedron(context->Dimension);
	e->EP = *sum;
	e->next = NULL;
	free(sum);
	eStart = e;
    }
*/
    return enu;
#else
    return Domain_Enumerate(phUnion, context, 0, NULL);
#endif
}

Pair *countIntegralPoints_c(Polyhedron *phUnion, Polyhedron *context) {
    Polyhedron *p, *ctx;
    Enumeration *e, *e2;
    Pair *res;
    int len, i;

    gmp_protect();

    assert(phUnion != NULL);
    assert(context != NULL);

    e = enumerateUnion(phUnion, context);

    len=0;
    for (e2=e; e2 != NULL; e2=e2->next)
	len++;

    res = (Pair *)malloc(sizeof(Pair) * (len+1));
    for (i=0,e2=e; e2 != NULL; e2=e2->next) {
	evalue *ev;
	Polyhedron *fis;
	fis = e2->ValidityDomain;
	ev = (evalue *)malloc(sizeof(evalue));
	if (ev == NULL)
	    oomInLoopo();
//	printf("ev = %p\n", ev);
	*ev = e2->EP;
	res[i++] = mkPair(fis, ev);
//	char *paramNames[] = { "m", "n", "tt", "pp" };
//	print_evalue(stdout, ev, paramNames);
    }
    res[len] = NULL;

    while (e!=NULL) {
	Enumeration *f = e;
	e = e->next;
	free(f);
    }

    gmp_unprotect();

    return res;
}

void evalueFree_c(evalue *ev) {
//    printf("evalueFree_c(%p)\n", ev);
    gmp_protect();
//    free_evalue_refs(ev);
    free(ev);
    gmp_unprotect();
}

Value* evalueGetD_c(evalue *ev) {
//    printf("evalueGetD_c(%p)\n", ev);
    Value *d;
    gmp_protect();
    d = copy_value(ev->d);
    gmp_unprotect();
    return d;
}

Value* evalueGetN_c(evalue *ev) {
//    printf("evalueGetN_c(%p)\n", ev);
    Value *n;
    gmp_protect();
    n = copy_value(ev->x.n);
    gmp_unprotect();
    return n;
}

enode* evalueGetP_c(evalue *ev) {
//    printf("evalueGetP_c(%p)\n", ev);
    return ev->x.p;
}

int enodeIsPeriodic_c(enode *en) {
//    printf("enodeIsPeriodic_c(%p)\n", en);
    return en->type == periodic;
}

int enodeGetParameterPos_c(enode *en) {
//    printf("enodeGetParameterPos_c(%p)\n", en);
    return en->pos;
}

evalue** enodeGetValues_c(enode *en) {
    int s = en->size;
    evalue **vs = (evalue**)malloc(sizeof(evalue*) * (s+1));
    int i;

//    printf("enodeGetValues_c(%p)\n", en);

    if (vs == NULL)
	oomInLoopo();
    for (i=0; i<s; i++)
	vs[i] = &(en->arr[i]);
    vs[s] = NULL;
    return vs;
}

