#include <iostream>
#include <string>
#include <string.h>

#include <fmaths.hh>
#include <Dialogs.h>
#include <scanpars2/Parser.h>

#include "flmaths.h"
extern "C" {
#include <assert.h>
}

std::string TempDir;

//#define debug(x) { std::cerr x; }
#define debug(x) {}


Fraction *newFraction_c(long long *p) {
    debug( << "newFraction_c " << p[0] << " " << p[1] << std::endl);
    return new Fraction(p[0], p[1]);
}


void deleteFraction_c(Fraction *f) {
    delete f;
}

long long *decompFraction_c(Fraction *f) {
    long long *p = (long long *)malloc(2 * sizeof(long long));
    p[0] = f->get_num();
    p[1] = f->get_denom();
    debug( << "decompFraction_c " << *f << std::endl);
    return p;
}

FMatrix *newFMatrix_c(int rows, int cols, Fraction **coeffs) {
    FMatrix *fm = new FMatrix(rows, cols);
    for (int r=0; r<rows; r++) {
	for (int c=0; c<cols; c++)
	    fm->set_entry(r, c, **coeffs++);
    }
    debug( << "newFMatrix_c " << rows << " " << cols << std::endl << *fm << std::endl);
    return fm;
}
    
void deleteFMatrix_c(FMatrix *m) {
    debug( << "deleteFMatrix_c" << std::endl);
    delete m;
}

int rowsFMatrix_c(const FMatrix *fm) {
    debug( << "rowsFMatrix_c" << std::endl);
    return fm->get_rows();
}

int colsFMatrix_c(const FMatrix *fm) {
    debug( << "colsFMatrix_c" << std::endl);
    return fm->get_cols();
}

Fraction **entriesFMatrix_c(const FMatrix *fm) {
    const int rows = fm->get_rows();
    const int cols = fm->get_cols();
    Fraction **entries = (Fraction **)malloc(sizeof(Fraction *)*rows*cols);
    Fraction **p = entries;
    for (int r=0; r<rows; r++) {
	for (int c=0; c<cols; c++) {
	    debug( << r << " " << c << std::endl);
	    *p++ = new Fraction(fm->get_entry(r,c));
	}
    }
    debug( << "entriesFMatrix_c" << std::endl);
    return entries;
}

FIneq_system *newFIneqSys_c(int vars, int params, FMatrix *fm) {
    debug( << "newFIneqSys_c " << vars << " " << params << " " << std::endl);
    debug( << *fm << std::endl);
    return new FIneq_system(vars, params, fm);
}

void deleteFIneqSys_c(FIneq_system *fsys) {
    debug(<< "deleteFIneqSys_c" << std::endl);
    delete fsys;
}

int varsFIneqSys_c(FIneq_system *fsys) {
    debug(<< "varFIneqSys_c" << std::endl);
    return fsys->get_vars();
}

int paramsFIneqSys_c(FIneq_system *fsys) {
    debug(<< "paramsFIneqSys_c" << std::endl);
    return fsys->get_params();
}

FMatrix *matrixFIneqSys_c(FIneq_system *fsys) {
    debug(<< "matrixFIneqSys_c" << std::endl);
    return new FMatrix(fsys);
}



// FEqSystem

FEq_system *newFEqSys_c(FMatrix *fm) {
    debug(<< "newFEqSys_c" << std::endl);
    return new FEq_system(fm);
}

void deleteFEqSys_c(FEq_system *fes) {
    debug(<< "deleteFEqSys_c" << std::endl);
    delete fes;
}

FMatrix *matrixFEqSys_c(FEq_system *fes) {
    debug(<< "matrixFEqSys_c" << std::endl);
    return new FMatrix(fes);
}

// FMatrix functions

Fraction *determinant_c(FMatrix *fm) {
    debug(<< "determinant_c" << std::endl);
    return new Fraction(fm->determinant());
}

FMatrix *kernel_c(FMatrix *fm) {
    debug( << "kernel_c" << std::endl << *fm << std::endl
	   << FMatrix(Matrix(*fm).compute_kernel()) << std::endl);
    return new FMatrix(Matrix(*fm).compute_kernel());
}

FIneq_system *sortIneqs_c(const FIneq_system *fis_in,
			  int method) {
    debug(<< "sortIneqs_c" << std::endl
	  << *fis_in << std::endl);
    FIneq_system *fis = new FIneq_system(fis_in);
    switch (fis->sortIneqs(IneqTool::NoRest,
			   IneqToolsInterface::ToolMethod(method))) {
    case IneqTool::Ok: break;
    case IneqTool::NoSolution: 
	delete fis;
	fis = NULL;
	break;
    case IneqTool::Error:
	cerr << "sorting the inequality system" << std::endl;
	cerr << *fis_in << endl << "failed" << std::endl;
	exit(1);
    }
    return fis;
}

Pair gaussJordan_c(FMatrix *fm, FMatrix *res) {
    int *arbit;
    Matrix *solution = Matrix(*fm).gauss_jordan(Matrix(*res), arbit);
    if (solution != NULL) {
	debug(<< "gaussJordan_c " << *fm << *res << *solution << std::endl);
	debug(<< arbit[0]);
	for (int i=1;i<=arbit[0]; i++)
	    debug(<< " " << arbit[i]);
	debug(<<std::endl);
	FMatrix *fsol = new FMatrix(*solution);
	delete   solution;
	int *arbitrary;
	arbitrary = (int *) malloc(sizeof(int) * (arbit[0]+1));
	memcpy(arbitrary, arbit, sizeof(int) * (arbit[0]+1));
	delete[] arbit;
	return mkPair(fsol, arbitrary);
    } else
	return NULL;
}

Pair echelonReduction_c(FMatrix *fm) {
    Matrix *r1;
    FMatrix *res1 = new FMatrix(Matrix(*fm).echelon_reduction(&r1));
    FMatrix *res2 = new FMatrix(*r1);
    delete r1;
    return mkPair(res1, res2);
}

FMatrix *inverse_c(FMatrix *fm) {
    return new FMatrix(fm->inverse());
}


FIneq_system *eliminateRedundant_c(FIneq_system * fsys,
				   FIneq_system * fcontext,
				   int method) {
    Ineq_system sys = Ineq_system(*fsys);
    Ineq_system c;
    if (fcontext != NULL)
	c = Ineq_system(*fcontext);
    Ineq_system *context = (fcontext == NULL) ? NULL : &c; 
    sys.eliminateRedundant(context, IneqTool::NoRest,
			   IneqToolsInterface::ToolMethod(method));
    return new FIneq_system(sys);
}

int isIneqRedundant_c(FIneq_system *fis, int ineq,
		      FIneq_system *fcontext,
		      int method) {
    Ineq_system sys  = Ineq_system(*fis);
    Ineq_system c;
    if (fcontext != NULL)
	c = Ineq_system(*fcontext);
    Ineq_system *context = (fcontext == NULL) ? NULL : &c;
    return sys.isIneqRedundant(ineq, context, IneqTool::NoRest,
			       IneqToolsInterface::ToolMethod(method));
}



// Symbol table


char* getIndex_c(char* tempDir, char* projFile, 
		 int indexNum) {
  char* lexem;

  TempDir = tempDir;

  lexem = getIdentifierHelper(projFile, 
			      indexNum, Index);
  return lexem;
}

char* getBlob_c(char* tempDir, char* projFile, 
		int blobNum) {
  char* lexem;

  TempDir = tempDir;

  lexem = getIdentifierHelper(projFile, 
			      blobNum, Blob);
  return lexem;
}

char* getArray_c(char* tempDir, char* projFile, 
		int arrayNum) {
  char* lexem;

  TempDir = tempDir;

  lexem = getIdentifierHelper(projFile, 
			      arrayNum, Array);
  return lexem;
}

char* getIdentifierHelper(char* projFile, 
			  int num, int type) {
  using namespace std;

  ifstream is;
  Parser* parser;
  Entry entry;  

  char* lexem;

  const char* fullParserPath = 
    (string(TempDir) + "/" + string(projFile)).c_str();

  is.open(fullParserPath);
  parser = new Parser();
  parser->Load(is);
  
  entry = parser->FindSymbol(num, type);
  if (entry != NULL) {
    lexem = strdup(entry->LexemOf());
  } else {
    lexem = NULL;
  }
  delete parser;
  return lexem;
}



// other LooPo functions

void setTempDir_c(const char *tempDir) {
    TempDir = tempDir;
}

void appendSessionLog_c(const char *str) {
    sessionlog log;
    log.append(str);
}
