#ifndef _FLMATHS_H
#define _FLMATHS_H

#ifndef _IN_GHC_
#include "fmaths.hh"
#else
typedef void Fraction;
typedef void FMatrix;
typedef void FIneq_system;
typedef void FEq_system;
#endif

#include "CppSide.h"

#ifdef __cplusplus
extern "C" {
#endif

void setInitialRaySpace_c(int);

Fraction * newFraction_c(long long *num);
void deleteFraction_c(Fraction *f);
long long *decompFraction_c(Fraction *f);
FMatrix *newFMatrix_c(int rows, int cols, Fraction **coeffs);
void deleteFMatrix_c(FMatrix *m);
int rowsFMatrix_c(const FMatrix *fm);
int colsFMatrix_c(const FMatrix *fm);
Fraction **entriesFMatrix_c(const FMatrix *fm);
FIneq_system *newFIneqSys_c(int vars, int params, FMatrix *fm);
void deleteFIneqSys_c(FIneq_system *fsys);

FEq_system *newFEqSys_c(FMatrix *fm);
void deleteFEqSys_c(FEq_system *fes);
FMatrix *matrixFEqSys_c(FEq_system *fes);

int varsFIneqSys_c(FIneq_system *fsys);
int paramsFIneqSys_c(FIneq_system *fsys);
FMatrix *matrixFIneqSys_c(FIneq_system *fsys);
Fraction *determinant_c(FMatrix *fm);
FMatrix *kernel_c(FMatrix *fm);
FIneq_system *sortIneqs_c(const FIneq_system *fis_in,
			  int method);
Pair gaussJordan_c(FMatrix *fm, FMatrix *res);
Pair echelonReduction_c(FMatrix *fm);
FMatrix *inverse_c(FMatrix *fm);
FIneq_system *eliminateRedundant_c(FIneq_system * fsys,
				   FIneq_system * fcontext,
				   int method);
int isIneqRedundant_c(FIneq_system *fis, int ineq,
		      FIneq_system *fcontext,
		      int method);

char* getIndex_c(char* tempDir, char* projFile, 
		 int indexNum);
char* getBlob_c(char* tempDir, char* projFile, 
		int blobNum);
char* getArray_c(char* tempDir, char* projFile,
		 int arrayNum);
char* getIdentifierHelper(char* projFile, 
			  int num, int type);

// other LooPo functions
void setTempDir_c(const char *);
void appendSessionLog_c(const char *str);

#ifdef __cplusplus
}
#endif


/* !def _FLMATHS_H */
#endif
