#include <stdio.h>
#include <assert.h>
#include "fmaths.hh"
#include <piplib/piplib64.h>

#include "CppSide.h"
#include "piplib.h"


/* This is almast the same as in polylib.cc, but Polylib's and Piplib's
 * value_* macros are different and Piplib uses "Entier" instead of
 * "Value", so we cannot reuse the same code.
 */
/*
static long long value_to_ll(Entier &vv) {
    long long l = 0;
    Entier v, v2, mask, div;
    int s = 0, neg = 0;

    value_init(v);
    if (value_neg_p(vv)) {
        value_oppose(v, vv);
        neg = 1;
    } else
        value_assign(v, vv);

    value_init(v2);
    value_init(mask);
    value_init(div);
    value_set_si(mask, 0xFFFF);
    value_set_si(div, 0x10000);
    while (value_notzero_p(v)) {
        value_andto(v2, v, mask);
        l |= ((long long)VALUE_TO_INT(v2)) << s;
        value_division(v, v, div);
        s = s + 16;
//      printf("s = %d\n", s);
    }

    return neg ? -l : l;
}
*/


/*
 * Convert an FIneq_system to an inequality system in
 * Piplib format and add a big parameter as parameter #0
 * and another parameter (as parameter #1) to help represent
 * negative parameter values.
 */
static PipMatrix *ineqs2pipmatrix(FIneq_system *sys_in) {
    PipMatrix *m;
    Ineq_system is(*sys_in);
    Matrix *sys = is.get_c_matrix();
    const int nV = sys_in->get_vars(), nP = sys_in->get_params();
    const int bp = nV+1; /* position of the big parameter */

    m = pip_matrix_alloc(sys->get_rows(), 1+nV+2+nP+1);
    for (int r=0; r<sys->get_rows(); r++) {
	value_init(m->p[r][bp]);
	value_set_si(m->p[r][bp], 0);
	value_init(m->p[r][bp+1]);
	value_set_si(m->p[r][bp+1], 0);

	value_init(m->p[r][0]);
        value_set_si(m->p[r][0], 1);

	for (int c=0; c<nV; c++) {
            const int v = sys->entry(r,c);
	    value_init(m->p[r][c+1]);
            value_set_si(m->p[r][c+1], v);
	    value_subtract(m->p[r][bp], m->p[r][bp], m->p[r][c+1]);
        }
        for (int c=0; c<nP; c++) {
            const int v = sys->entry(r,nV+c);
	    value_init(m->p[r][bp+2+c]);
            value_set_si(m->p[r][bp+2+c], v);
	    value_subtract(m->p[r][bp+1], m->p[r][bp+1], m->p[r][bp+2+c]);
	}
	value_init(m->p[r][bp+2+nP]);
	value_set_si(m->p[r][bp+2+nP], sys->entry(r,nV+nP));
    }
    return m;
}


// Armin Groesslinger, 2001-09-06:
// We _have_ to set
//     simple = 1
// below (which is PIP's "-z" command line switch), because
// if we don't optimize the quast, a legal result of PIP's
// computation is something like
//     (if #[ ... ]
//       ()
//       ()
//     )
// (i.e. a quast with IFs, but all leaves are "()") which leads
// to a non-NULL pip_quast_root in sol.c:sol_edit, which is
// then interpreted (in the function "pip" below) as PIP_OK instead
// of the correct PIP_NOSOL !
// Therefore we have to optimize the solution to get a simple
//     ()
// in such cases, which is then interpreted as PIP_NOSOL since we get
// pip_quast_root == NULL back from sol_edit.
// See also section 2.3.2 ("Simplifying the solution") in
// the manual for PIP D.1.
PipQuast *pip_c(FIneq_system *ineqs, FIneq_system *constraints,
		int integer_sol)
{
    PipOptions *opt;
    PipMatrix *domain;
    PipMatrix *context;
    PipQuast *sol;
    int big_par_pos = ineqs->get_vars() + 1;

    gmp_protect();

    FIneq_system univCtx(0, ineqs->get_params());

    //domain = pip_matrix_read(stdin) ;
    //context = pip_matrix_read(stdin) ;
    opt = pip_options_init();
    opt->Nq = integer_sol ? 1 : 0;;
    opt->Simplify = 1;
    opt->Verbose = -1;
    opt->Urs_unknowns = 0;
    opt->Urs_parms = 0;
    opt->Maximize = 0;

    domain  = ineqs2pipmatrix(ineqs);
    context = ineqs2pipmatrix(constraints == NULL ? &univCtx : constraints);

//    pip_matrix_print(stdout, domain);
//    pip_matrix_print(stdout, context);

//    printf("vor pip_solve\n");
    sol = pip_solve(domain, context, big_par_pos, opt);
//    printf("nach pip_solve\n");
//    pip_quast_print(stdout, sol, 0);
 
    pip_matrix_free(context);
    pip_matrix_free(domain);
    pip_options_free(opt);

    gmp_unprotect();

    return sol;
}

void freeQuast_c(PipQuast *q) {
    gmp_protect();
    pip_quast_free(q);
    gmp_unprotect();
}

/*
 * Convert a PipVector to an FVector and thereby remove
 * the big parameter and the additional parameter for the
 * negative parameter values.
 * Check that the coefficient of the big parameter
 * is 1 if bpIsOne is true (for a leaf vector) 
 * or 0 if bpIsOne is false (for a condition or NewParam in a PipQuast).
 */
static FVector pipvector2fv(PipVector *pv, bool bpIsOne) {
    FVector v(pv->nb_elements-2);
    int bpNum   = VALUE_TO_INT(pv->the_vector[0]);
    int bpDenom = VALUE_TO_INT(pv->the_deno[0]);

    if (bpIsOne)
	assert(bpNum == bpDenom && bpDenom != 0);
    else
	assert(bpNum == 0 && bpDenom != 0);

    for (int i=2; i<pv->nb_elements; i++) {
	// printf("i=%d, num=%d, denom=%d\n", i, VALUE_TO_INT(pv->the_vector[i]), VALUE_TO_INT(pv->the_deno[i]));
	v[i-2] = Fraction(VALUE_TO_INT(pv->the_vector[i]),
			  VALUE_TO_INT(pv->the_deno[i]));
    }

    return v;
}

enum PvFinity { PV_NEG_INFTY, PV_FINITE, PV_POS_INFTY };

/*
 * Determine the whether a PipVector represents
 * a finite value or a neg/pos infinity.
 * A value is finite, if the coefficient of the big parameter
 * is 1. It is positive infinity, if the coefficient is > 1 and
 * negative infinity, if the coefficient is < 1.
 */
enum PvFinity pipvectorFinity(PipVector *pv) {
    enum PvFinity sign;
    int bpNum   = VALUE_TO_INT(pv->the_vector[0]);
    int bpDenom = VALUE_TO_INT(pv->the_deno[0]);

    // just to be on the safe side
    if (bpDenom < 0) {
	bpDenom = -bpDenom;
	bpNum   = -bpNum;
    }

    if (bpNum == bpDenom)
	sign = PV_FINITE;
    else if (bpNum > bpDenom)
	sign = PV_POS_INFTY;
    else if (bpNum < bpDenom)
	sign = PV_NEG_INFTY;
    else
	assert(0);

    return sign;
}

/*
 * Usually, piplib represents an infinity by a vector with
 * all denominators 0. But since we apply the big parameter
 * transformation, (in)finity is determined by the coefficient
 * of the big parameter (cf. pipvectorFinity).
 * We return a 0xN FMatrix to represent a negative infinity,
 * a 2xN FMatrix to represent a positive infinity, and
 * a 1xN FMatrix to represent a finite value.
 */ 
FMatrix **getLeaf_c(PipQuast *q) {
    PipList *l;
    int len, i;

    len = 0;
    for (l=q->list; l != NULL; l=l->next)
	len++;

    if (len == 0)
	return NULL;

    /* Special case to take care of:
     * A system with 0 variables has a list with one non-existent
     * vector at a leaf to repesent feasibility (and the empty tuple
     * for the variables).
     * E.g.
     *    pipLexmin (ineqSystem ((0,1), zeroMatrix 0 2)) Nothing True
     * has
     *   (list 
     *   )
     * as solution, and the length of the list is 1 and not 0,
     * and the list has no vector.
     */
    if (len == 1 && q->list->vector == NULL) {
	FMatrix **m = (FMatrix **)malloc(sizeof(FMatrix*));
	m[0] = NULL;
	return m;
    }

    gmp_protect();

    // printf("Anfang getLeaf_c, len=%d\n", len);
    FMatrix **m = (FMatrix **)malloc(sizeof(FMatrix*) * (len+1));
    for (l=q->list,i=0; l != NULL; l=l->next,i++) {
	PipVector *pv = l->vector;
	switch (pipvectorFinity(pv)) {
	case PV_FINITE:
	    m[i] = new FMatrix(0, pv->nb_elements - 2);
	    m[i]->insert_row(pipvector2fv(pv, true), 0);
	    break;
	case PV_POS_INFTY:
	    m[i] = new FMatrix(2, pv->nb_elements - 2);
	    break;
	case PV_NEG_INFTY:
	    m[i] = new FMatrix(0, pv->nb_elements - 2);
	    break;
        default:
	    assert(0);
	}
    }
    m[len] = NULL;
    // printf("Ende getLeaf_c\n");

    gmp_unprotect();
    return m;
}

PipQuast *getThen_c(PipQuast *q) {
    return q->next_then;
}

PipQuast *getElse_c(PipQuast *q) {
    return q->next_else;
}

FMatrix *getCond_c(PipQuast *q) {
    gmp_protect();
    FMatrix *m = new FMatrix(1, pipvector2fv(q->condition, false));
    gmp_unprotect();
    return m;
}

void **getNewParams_c(PipQuast *q) {
    void **triple;
    int *ranks;
    FMatrix **defs;
    long long *denoms;

    int count = 0;

    gmp_protect();

    for (PipNewparm *np=q->newparm; np!=NULL; np=np->next)
	count++;

    triple = (void **)malloc(sizeof(void*) * 3);
    ranks = (int *)malloc(sizeof(int) * (count+1));
    defs = (FMatrix **)malloc(sizeof(FMatrix **) * (count+1));
    denoms = (long long *)malloc(sizeof(long long) * (count+1));
    triple[0] = ranks;
    triple[1] = defs;
    triple[2] = denoms;

    ranks[0] = count;
    denoms[0] = count;
    defs[count] = NULL;

    PipNewparm *np = q->newparm;
    for (int i=0; i<count; i++) {
	ranks[i+1] = np->rank - 2;
	denoms[i+1] = VALUE_TO_INT(np->deno);
	defs[i] = new FMatrix(1, pipvector2fv(np->vector, false));
	np=np->next;
    }

    gmp_unprotect();

    return triple;
}

int isList_c(PipQuast *q) {
    return q->condition == NULL;
}
