#ifndef _CPPSIDE_POLYLIB_H
#define _CPPSIDE_POLYLIB_H

#ifndef _IN_GHC_
/* Workaround for name clashes with Polylib:
 *   `Matrix' is declared in lmaths.h, and
 *   `vector' is used by the STL
 */
#define Matrix PolyMatrix
#define vector _vector
#if POLYLIB_BITS==64
#include <polylib/polylib64.h>
#else
#include <polylib/polylibgmp.h>
#endif
#undef vector
#undef Matrix


Value *copy_value(Value &);

#else

typedef void Value;
typedef void evalue;
typedef void enode;
typedef void PolyMatrix;
typedef void Polyhedron;

/* !def _IN_GHC_ */
#endif

#ifdef __cplusplus
extern "C" {
#endif

// Polylib

Value *polylibValueFromLimbs_c(const unsigned int *limbs);
unsigned int *polylibValueToLimbsAndFree_c(Value *v);
void polylibValueFree_c(Value *v);
PolyMatrix *polylibMatrixNew_c(int rows, int columns, Value **entries);
int polylibMatrixRows_c(PolyMatrix *pm);
int polylibMatrixColumns_c(PolyMatrix *pm);
Value *polylibMatrixEntry_c(int r, int c, PolyMatrix *pm);
void polylibMatrixFree_c(PolyMatrix *pm);
PolyMatrix* polylibMatrixCopy_c(PolyMatrix *pm);
Polyhedron *polylibPolyhedronNew_c(PolyMatrix *pm);
void polylibPolyhedronFree_c(Polyhedron *p);
PolyMatrix *polylibPolyhedronMatrix_c(Polyhedron *p);
Polyhedron *polylibPolyhedronNext_c(Polyhedron *p);
void polylibPolyhedronSetNext_c(Polyhedron *p, Polyhedron *q);

/*
 * Z-polyhedra functions
 */

Pair *zpDifference_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2);
Pair *zpIntersection_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2);
Pair *zpUnion_c(Polyhedron **ph1, PolyMatrix **latt1, Polyhedron **ph2, PolyMatrix **latt2);
Pair *zpUnionDifference_c(Polyhedron **phIn, PolyMatrix **lattIn, Polyhedron **phOut, PolyMatrix **lattOut);

int zpIncludes_c(Polyhedron *p1, PolyMatrix *latt1, Polyhedron *p2, PolyMatrix *latt2);

/*
 * Polyhedra
 */

Polyhedron *simplifyDomain_c(Polyhedron *context, Polyhedron *sys);

/* old stuff */

FIneq_system *verticesAsRowVectorsToIneqs_c(FMatrix *fm);
FIneq_system *convexUnionAll_c(FIneq_system **systems, int count);
FMatrix* eqsIneqs2rays_c(FEq_system *eqs, FIneq_system *ineqs);
int polyhedronIncludes_c(FIneq_system *sys1, FIneq_system *sys2);
FIneq_system **union_c(FIneq_system **input);
FIneq_system** disjUnionAll_c(FIneq_system **systems);


FIneq_system **disjointUnionDifference_c(FIneq_system *context, FIneq_system **fisIn, FIneq_system **fisOut);
FIneq_system *simplifyIneqs_c(FIneq_system *context, FIneq_system *sys);
FIneq_system **intersectDomains_c(FIneq_system ***sys);

// taken from Polylib5.11.1:
//Polyhedron *Disjoint_Domain( Polyhedron *P, int flag, unsigned NbMaxRays );

#ifdef __cplusplus
}
#endif


/* !def _CPPSIDE_POLYLIB_H */
#endif
