module ISS.Debug.CodeGenerator(generateCode) where

import ScannerParser.ParseTree
import ScannerParser.ScannerParser
import ScannerParser.SymbolTable

getType :: LambdaNode -> String
getType (Skip {}) = "Skip"
getType a@(FuncDecl {}) = "FuncDecl " ++ show (name a)
getType (ProcCall {}) = "ProcCall"
getType (ExprFuncCall {}) = "ExprFuncCall"

generateCode :: SymTab -> Int -> LambdaTree -> String
generateCode _ _ [] = ""
generateCode st ind (x:xs) = if xs == [] then
                              generateCode' ind x
                             else
                              generateCode' ind x ++ indNewLine ind ++ generateCode st ind xs where
                               generateCode' ind node@(ForLoop {}) = "for (" ++ symTableLookup st LNS_Index (name node) ++
                                                                     " = " ++ generateCodeMaybe (-1) (loopLowerBound node)
                                                                     ++ "; i <= " ++ generateCodeMaybe (-1) (loopUpperBound node)
                                                                     ++ "; i += " ++ generateCodeMaybe (-1) (loopStride node) ++ ") {"
                                                                     ++ indNewLine (ind + 4) ++ generateCode st (ind + 4) (loopBody node)
                                                                     ++ indNewLine ind ++ "}"
                               generateCode' ind node@(Expression {}) = showOperator (generateCodeMaybe (-1) (condition node))
                                                                        (generateCodeMaybe (-1) (firstOperand node)) (operator node)
                                                                        (generateCodeMaybe (-1) (secondOperand node))
                               generateCode' ind node@(ExprIndBlobConst {}) = symTableLookup st (state node) (name node)
                               generateCode' ind node@(Assign {}) = symTableLookup st LNS_Array (name node) ++
                                                                    generateIndices (indices node) ++
                                                                    " = " ++ generateCode' (-1) (firstExpr node) ++ ";"
                               generateCode' ind node@(ExprRead {}) = symTableLookup st LNS_Array (name node) ++
                                                                      generateIndices (indices node)
                               generateCode' ind node@(WhileLoop {}) = "while (" ++ generateCodeMaybe (-1) (condition node) ++
                                                                       ") {" ++ indNewLine (ind + 4) ++ generateCode st (ind + 4) (loopBody node) ++
                                                                       indNewLine ind ++ "}"
                               generateCode' ind node@(If {}) = "if (" ++ generateCodeMaybe ind (condition node) ++ ") {" ++ indNewLine (ind + 4) ++
                                                                generateCode st (ind + 4) (thenbranch node) ++ indNewLine ind ++ "}" ++
                                                                if elsebranch node /= [] then
                                                                    " else {" ++ indNewLine (ind + 4) ++
                                                                    generateCode st (ind + 4) (elsebranch node) ++ indNewLine ind ++ "}"
                                                                else
                                                                    ""
                               generateCode' ind node@(Label {}) = lexem node
                               generateCode' ind node@(GoTo {}) = "goto " ++ lexem node ++ ";"
                               generateCode' _ node = getType node
                               generateCodeMaybe ind Nothing  = ""
                               generateCodeMaybe ind (Just x) = generateCode' ind x
                               generateIndices []      = ""
                               generateIndices indices = "[" ++ foldl1 (\x -> \y -> (x ++ ", " ++ y)) (map (generateCode' (-1)) indices) ++ "]"
                               showOperator _    fst LNOp_Plus    snd = fst ++ " + " ++ snd
                               showOperator _    fst LNOp_Minus   snd = fst ++ " - " ++ snd
                               showOperator _    fst LNOp_Times   snd = fst ++ " * " ++ snd
                               showOperator _    fst LNOp_Div     snd = fst ++ " / " ++ snd
                               showOperator _    fst LNOp_Mod     snd = fst ++ " % " ++ snd
                               showOperator _    fst LNOp_And     snd = fst ++ " && " ++ snd
                               showOperator _    fst LNOp_Or      snd = fst ++ " || " ++ snd
                               showOperator _    fst LNOp_GE      snd = fst ++ " > " ++ snd
                               showOperator _    fst LNOp_GEQ     snd = fst ++ " >= " ++ snd
                               showOperator _    fst LNOp_LE      snd = fst ++ " < " ++ snd
                               showOperator _    fst LNOp_LEQ     snd = fst ++ " <= " ++ snd
                               showOperator _    fst LNOp_EQ      snd = fst ++ " == " ++ snd
                               showOperator _    fst LNOp_NEQ     snd = fst ++ " != " ++ snd
                               showOperator _    _   LNOp_Not     snd = "!" ++ snd
--                             showOperator _    fst LNOp_Bracket _   = "(" ++ fst ++ ")"
                               showOperator cond fst LNOp_IIf     snd = cond ++ " ? " ++ fst ++ " : " ++ snd

symTableLookup :: SymTab -> LambdaNodeState -> Int -> String
symTableLookup _ LNS_Constant number = show number
symTableLookup st state number = let Just result = findSymbol (number, state) st in sym_lexem result

indNewLine :: Int -> String
indNewLine (-1) = ""
indNewLine n = '\n':indentation n

indentation :: Int -> String
indentation 0 = ""
indentation n = ' ':indentation (n-1)
