module ISS.Language.LanguageDE (languageText, languageParseError, languageBuildText, multiLineBoxWrap, multiLineBox, boxWrap, box) where

import ISS.Language.LanguageTools


-- |Retrieves a certain language string (identified by a string)
languageText :: String -> String
languageText "title"           = boxWrap 80 "Index Set Splitting" ++ "\n"
languageText "about"           = "Implementierung von Index Set Splitting in Haskell\n" ++
                                 "Programmierpraktikum im Hauptstudium\n" ++
                                 "Martin Steghoefer\n" ++
                                 "Version vom 17.07.2007"
languageText "options"         = "\nProgramm-Aufruf:\n\n  iss [Optionen]\n\nMoegliche Optionen:\n\n" ++
    "  -inputdir    [dir]                Verzeichnis, in dem die Input-Dateien liegen\n" ++
    "  -outputdir   [dir]                Verzeichnis, in das die Output-Dateien\n" ++
    "                                    geschrieben werden sollen\n" ++
    "  -ispc        [inFile] [outFile]   Ein- und Ausgabedatei der Index-Spaces\n" ++
    "                                    (.out.ispc)\n" ++
    "  -deps        [inFile] [outFile]   Ein- und Ausgabedatei der Abhängigkeiten\n" ++
    "                                    (.out.dep)\n" ++
    "  -parsetree   [inFile] [outFile]   Ein- und Ausgabedatei des Parse-Baums (.out)\n" ++
    "  -accesses    [inFile] [outFile]   Ein- und Ausgabedatei der Accesses (.acc)\n" ++
    "  -accesspairs [inFile] [outFile]   Ein- und Ausgabedatei der Accesses (.acc)\n" ++
    "  -groups      [inFile] [outFile]   Ein- und Ausgabedatei der Statement-Groups\n" ++
    "                                    (.grp)\n" ++
    "  -surrloops   [outFile]            Ausgabedatei der Surrounding Loops (.slp)\n" ++
    "  -surrifs     [outFile]            Ausgabedatei der Surrounding Ifs (.sif)\n" ++
    "  -project [inDir] [outDir] [name]  Verarbeitung eines kompletten Projekts; als\n" ++
    "                                    Dateinamen für alle Ein- und Ausgabedateien\n" ++
    "                                    werden die Loopo-Standard-Namen verwendet,\n" ++
    "                                    z.B. \"[name].out.ispc\" für die Index-Spaces;\n" ++
    "                                    mit dieser Option werden die Optionen\n" ++
    "  -showinputs                       Debug-Option: Zeigt eingelesene Daten an\n" ++
    "  -showresult                       Debug-Option: Zeigt den resultierenden\n" ++
    "                                    Parse-Baum als C-Code an\n" ++
    "  -showsplit                        Debug-Option: Zeigt den Split-Binärbaum an\n" ++
    "  -debug                            Debug-Option: Setzt alle 3 Debug-Flags\n" ++
    "                                    (showinputs, showresult und showsplit)\n" ++
    "  -about                            Zeigt Informationen über das Programm an\n" ++
    "  -usage                            Zeigt diese Optionen-Übersicht an\n" ++
    "  -propagate                        Schaltet das Propagieren ein\n" ++
    "  -propdepth [num]                  Legt die maximale Anzahl Iterationen zum\n" ++
    "                                    Ermitteln der transitiven Hülle beim\n" ++
    "                                    Propagieren fest\n" ++
    "\n" ++
    "Die Reihenfolge der Optionen spielt dabei keine Rolle.\n" ++
    "\n" ++
    "Das Programm benötigt mindestens die Angabe von Ein- und Ausgabeverzeichnis\n" ++
    "sowie der Dateien für Index-Spaces, Abhängigkeiten und Parse-Baum. Eine minimale\n" ++
    "Kombination gültiger Optionen sind also die ersten 5 Optionen (inputdir,\n" ++
    "outputdir, ispc, deps, parsetree)\n" ++
    "\n" ++
    "  iss -inputdir [inDir] -outputdir [outDir] -ispc [inFile outFile]\n" ++
    "      -deps [inFile] [outFile] - parsetree [inFile] [outFile]\n" ++
    "\n" ++
    "oder alternativ eines ganzen Projekts (-project):\n" ++
    "\n" ++
    "  iss -project [inDir] [outDir] [name]\n"
languageText "dependence"      = "Abhaengigkeit"
languageText "dependences"     = "Abhaengigkeiten"
languageText "statementSpaces" = "Index-Spaces"
languageText "accesses"        = "Zugriffe"
languageText "accesspairs"     = "Zugriffs-Paare"
languageText "groups"          = "Groups"
languageText "surrloops"       = "\"Umgebende Schleifen\""
languageText "surrifs"         = "\"Umgebende Ifs\""
languageText "tryUsageAbout"   = "\nFuer moegliche Optionen bitte \"iss -usage\" aufrufen, " ++
                                 "fuer einen Infotext\n\"iss -about\"!"
languageText "missingParam-indir"  = "Fehlender Parameter \"Eingabe-Verzeichnis\":\n"
                                     ++ "Bitte mit der Option -inputdir ein Verzeichnis angeben!\n" ++ languageText "tryUsageAbout"
languageText "missingParam-outdir" = "Fehlender Parameter \"Ausgabe-Verzeichnis\":\n"
                                     ++ "Bitte mit der Option -outputdir ein Verzeichnis angeben!\n" ++ languageText "tryUsageAbout"
languageText "missingParam-deps"   = "Fehlender Parameter \"Abhängigkeiten\":\n"
                                     ++ "Bitte mit der Option -deps eine Datei angeben!\n" ++ languageText "tryUsageAbout"
languageText "missingParam-tree"   = "Fehlender Parameter \"Parse-Baum\":\n"
                                     ++ "Bitte mit der Option -parsetree eine Datei angeben!\n" ++ languageText "tryUsageAbout"
languageText "missingParam-ispc"   = "Fehlender Parameter \"IndexSpaces\":\n"
                                     ++ "Bitte mit der Option -ispc eine Datei angeben!\n" ++ languageText "tryUsageAbout"
languageText "readingDeps"     = "Abhaengigkeiten einlesen..."
languageText "readingTree"     = "Parse-Baum einlesen..."
languageText "readingSpaces"   = "Statement-Index-Spaces einlesen..."
languageText "inputSpaces"     = "\n\n" ++ boxWrap 80 "Eingelesene Index-Spaces" ++ "\n"
languageText "inputDeps"       = "\n\n" ++ boxWrap 80 "Eingelesene Abhängigkeiten" ++ "\n"
languageText "inputPTree"      = "\n\n" ++ boxWrap 80 "Eingelesener Parse-Baum" ++ "\n"
languageText key               = error ("Deutschen Text fuer \"" ++ key ++ "\" nicht gefunden!")


-- |Retrieves a certain language string (identified by a string) and inserts
--  some given strings
languageBuildText :: String -> [String] -> String
languageBuildText "invalidParameter" [parameter] = "Der Parameter \"" ++ parameter ++ "\" ist unbekannt. " ++
                                                   "Fuer moegliche Optionen bitte \"iss -usage\" aufrufen, " ++
                                                   "fuer einen Infotext \"iss -about\"!"
languageBuildText key _                          = error ("Deutschen Text fuer \"" ++ key ++ "\" nicht gefunden!")


-- |Constructs a parse error message out of additional information about the
--  file, in which parsing failed, the reason of the error and the line in the
--  input file, at which the error occurred
languageParseError :: String -> String -> Integer -> String
languageParseError parsing reason line = "\n" ++ boxWrap 80 (parsing ++ "-Parsen ist leider aus folgendem Grund nicht geglueckt:\n"
                                                             ++ "Fehler in Zeile " ++ show line ++ ":\n" ++ reason)
