module ISS.Language.LanguageTools (multiLineBoxWrap, multiLineBox, boxWrap, box, wrapLine, indent) where

import ISS.Util

multiLineBoxWrap :: Integer -> [String] -> String
multiLineBoxWrap lineLen = multiLineBox . (foldl (++) []) . (map (wrapLine (lineLen - 4)))
                         

multiLineBox :: [String] -> String
multiLineBox texts = let lenMax = fromIntegral (maximum (map length' texts))
                         starLine = take' (lenMax + 4) infStars
                     in starLine ++ "\n" ++ foldl (++) "" (map (prepareLine lenMax) texts) ++ starLine
                     where
                         prepareLine lenMax text = "* " ++ fillUp lenMax text ++ " *\n"

boxWrap :: Integer -> String -> String
boxWrap lineLen = (multiLineBoxWrap lineLen) . (splitAtCharacter '\n')

box :: String -> String
box = multiLineBox . (splitAtCharacter '\n')

wrapLine :: Integer -> String -> [String]
wrapLine lineLen text = --split the text into a list of words an let the
                        --"buildLines" function put it together to complete
                        --lines
                        let words = splitAtCharacter ' ' text
                        in  buildLines lineLen words
                        where
                            --fill lines until all words in the list are spent
                            buildLines lineLen []    = []
                            buildLines lineLen words = let (line, restWords) = buildLine lineLen words in line:buildLines lineLen restWords
                            --fill a single line with words from the list
                            buildLine lineLen []           = ("", [])
                            buildLine lineLen (word:words) = let (line, restWords) = buildLine (lineLen - length' word - 1) words
                                                             in  if length' word <= lineLen - 1 then 
                                                                     (word ++ " " ++ line, restWords)
                                                                 else
                                                                     ("", word:words)

splitAtCharacter :: Char -> String -> [String]
splitAtCharacter _    ""          = [""]
splitAtCharacter char (head:tail) = let (first:rest) = splitAtCharacter char tail in
                                    if head == char then
                                        "":(first:rest)
                                    else
                                        (head:first):rest

indent :: Integer -> String -> String
indent num = concat . (map indentSingleLine) . (splitAtCharacter '\n') where
             indentSingleLine line = take' num infSpaces ++ line ++ "\n"

infStars :: String
infStars = '*':infStars

infSpaces :: String
infSpaces = ' ':infSpaces

concatLine :: String -> String -> String
concatLine textA textB = textA ++ "\n" ++ textB

fillUp :: Integer -> String -> String
fillUp max ""     = take' max infSpaces
fillUp max (x:xs) = x:fillUp (max - 1) xs

