module ISS.SplittingAlgorithm.SplitInitial(splitInitial) where

import ISS.DataSplitting.Info
import ISS.SplittingAlgorithm.ToolsGeneral

import LMath.Algo.AffineImage
import LMath.Logic.Domain
import LoopoDS.Dependence

-- |Calculates the initial split based on given dependences and statements
splitInitial :: IsDomain d t => [GDependenceD d t] -> [GSplittedStatement d t] -> [GSplittedStatement d t]
splitInitial deps = map (splitStatement deps)

-- |Calculates the initial split of one single statement based on given dependences
splitStatement :: IsDomain d t => [GDependenceD d t] -> GSplittedStatement d t -> GSplittedStatement d t
splitStatement dependences statement = foldl splitStatementWithDependence statement dependences

-- |Calculates the initial split of one single statement w.r.t. one single dependence
splitStatementWithDependence :: IsDomain d t => GSplittedStatement d t -> GDependenceD d t -> GSplittedStatement d t
splitStatementWithDependence statement dependence = if fst (originalStatement statement) == depDDestStmtID dependence then
                                                        let depImage = (dInverseImage hTrafo sourceSpace) `dIntersect` targetSpace
                                                            hTrafo = depDHTrafo dependence
                                                            sourceSpace = depDSrcDomain dependence
                                                            targetSpace = depDDestDomain dependence
                                                        in  splitStatementWithDomain statement depImage
                                                    else
                                                        statement
