module ISS.Util(throwErrorIfWrong, find, createFullFileName, take', length') where

import LP

-- |Throws an error with a given message function (from language file), if
--  parsing has failed. Otherwise return the parse result.
throwErrorIfWrong :: (String -> Integer -> String) -> LPResult a -> a
throwErrorIfWrong _            (ParseOk a)           = a
throwErrorIfWrong errorMessage (ParseFailed details) = (error . uncurry errorMessage) details

-- |Finds an entry in a given map (represented as list of tupels of key and value)
find :: Eq a => [(a,b)] -> a -> Maybe b
find []        _ = Nothing
find ((a,b):t) c = if c == a then Just b else find t c

-- |Adds path information to a given relative filename
createFullFileName :: String -> String -> String
createFullFileName dir file = if length dir > 0 && dir!!(length dir - 1) == '/' then dir ++ file else dir ++ "/" ++ file

-- |A "take" function like the one in the Prelude, but with "Integer" instead of "Int"
take' :: Integer -> [a] -> [a]
take' 0 _ = []
take' n (x:xs) | n > 0 = x:(take' (n-1) xs)

-- |A "length" function like the one in the Prelude, but with "Integer" instead of "Int"
length' :: [a] -> Integer
length' [] = 0
length' (x:xs) = 1 + length' xs
