
module Input4PL where 

import InputLoopo

import LoopoDS
import Loopo
import List 
import IOExts
import LMath

import Data.FiniteMap
import Maybe
import Char
import Ratio
--
-- Following functions convert the delivered 
-- indexspace matrix into the d-matrix, e.g negativ, 
-- transposed and with an added row 0 0 .. 1 

-- adds a zero row to the indexspace matrix  
--
ispcMtrxWithZeroRow  :: Matrix -> Matrix
ispcMtrxWithZeroRow mtrx = insertZeroRows mtrx pos 1
    where pos = nRows mtrx 

-- sets the last entry on 1 
--
lastEntryIsOne :: Matrix -> Matrix
lastEntryIsOne mtrx = setEntry mtrx (maxRow mtrx) (maxColumn mtrx) 1

-- composition of ispcMatrixZeroRow and lastEntryIsOne
-- to add the 0 0 .. 1 row to the indexspace matrix
--
addRow = lastEntryIsOne . ispcMtrxWithZeroRow

-- second component of (stmtNr,indexspace-matrix) is transformed into 
-- negativ, transposed matrix with additional row and returned as list of
-- (stmtNr, d-matrix)-tuples 
-- 
convert2Dmatrx2 :: [(Integer,Matrix)]->[(Integer,Matrix)]
convert2Dmatrx2 indexspaceClean = [(stmtNr,-(mtranspose (addRow mtrx))) |(stmtNr,mtrx) <- indexspaceClean]
    

-- defintion for testing  
--converted2 = convert2Dmatrx2 indexspace

--------------------------------------------------------------------
-- functions to calculate the offsets of the different matrix groups
-- (indenty matrix, access matrix, indexspace-matrix (d-matrix))
--------------------------------------------------------------------

-- map from statment number of (position, width) with offset value 0 
-- from the left, used to construct the identity matrix group
-- NOT CLEAR if the matrices have already to be transposed
-- 
stmt2posAndWidth :: [(Integer, Matrix)] -> [(Integer,(Integer,Integer))]
stmt2posAndWidth = paw 0
   where
   paw offset [] = []
   paw offset ((stmtNr,ispc):rest) = (stmtNr,(offset,width)):paw(offset+width) rest
       where
       width = nColumns ispc

-- test definition 
--stmt2pos = stmt2posAndWidth indexspace 


-- clean wrong dimensions and transposes access-matrices 
-- within the delivered list
--
accTranspose ::  [(Integer,AccessType,Integer,Integer,Matrix)] ->  [(Integer,AccessType,Integer,Integer, Matrix)]
accTranspose accs = [(a,b,c,d,mtranspose(x)) | (a,b,c,d,x) <- accs]


-- calculates the position of the access-matrices in the final 
-- matrix by the transformed access-matrix-list
-- it is crucial that acc2posAndWidth is applied on the acc-list,
-- which matrixes have been transposed before. 
--  
acc2posAndWidth :: [(Integer,AccessType,Integer,Integer,Matrix)] ->(Integer, FiniteMap Integer Integer)
acc2posAndWidth = help 0 emptyFM 
   where 
   help :: Integer -> FiniteMap Integer Integer->[(Integer,AccessType,Integer,Integer,Matrix)]->(Integer,  FiniteMap Integer Integer)
   help off fm [] = (off, fm)
   help off fm ((arrayId,accType,laNode,stmtNr,indexFun):xs)
          -- if id already in finite map continue with rest
       | elemFM arrayId fm = help off fm xs 
       -- update offset, insert arrayId in finite map
       | otherwise = help (off+width) (addToFM fm arrayId (off)) xs 
       where 
       width = nColumns indexFun

-- test definitions 
--acc2pos= acc2posAndWidth
--accsTr = accTranspose accsClean
--accPositions =  acc2pos accsTr 

-- calculates position offsets for the transformed indexspace matrixes
-- by given transformed access-matrix-list and transformend? (not sure, actually not necessary!)
-- indexspace-matrix-list 
--
ispc2position :: [(Integer,Matrix)]-> [(Integer,AccessType,Integer,Integer,Matrix)]->[(Integer, Integer, (Integer, Integer))]
ispc2position indexspaceList = paw 0 
    where 
    paw offset [] = []
    paw offset ((arrayId,accType,node,stmtNr,indexFct):rest) = (stmtNr,node,(offset,width)): paw (offset + width) rest
	where
	width = nColumns (search stmtNr dMatrixList)
	    where
	    dMatrixList = convert2Dmatrx2 indexspaceList 

-- test output
--dees = ispc2position indexspace accsClean 
--dwidths list = [x | (a,b,(c,x)) <- list]


-- presentation of infinite map on the screen
--
instance (Show key, Show elt) => Show (FiniteMap key elt) where
    show fm = "FiniteMap " ++ show (fmToList fm)

-- constructs the part which contains the identity matrix
-- of the final matrix. Parameters are stmtNr and 
-- the result of stmt2position 
--
identityGroups :: Integer -> [(Integer,(Integer,Integer))] -> Matrix 
identityGroups  stmt posWidthList
       -- if matrix ends before total width 
       -- then filled up with zero columns
       | (p+w) < total = insertZeroColumns leftMatrix (p+w) (total-leftWidth)
       -- no zero columns on the right side 
       | otherwise =  leftMatrix 
       where
       -- width of the current id-matrix 
       w = snd (posAndWidth)
       -- position/ offset of current id-matrix
       p = fst (posAndWidth)
       -- second component, contains offset and width 
       posAndWidth = search stmt posWidthList
       -- the total width of this group 
       total = sum[ x |(_,(_,x)) <- posWidthList]
       -- part from the left: null columns and id-matrix 
       leftMatrix =  insertZeroColumns (unitMatrix w w) 0 p
       -- width of the access matrix plus their left zero's
       leftWidth = nColumns(leftMatrix) 

-- test definitions 
--
--idtest1 = identityGroups 1 stmt2pos 
--idtest2 = identityGroups 2 stmt2pos 
--idtest3 = identityGroups 3 stmt2pos



-- constructs the part which contains the transformed
-- acess-matrices and their zero-groups
-- parameters are the offset list for the access-matrixes
-- and the list with transformend acess-matrixes and other 
-- essential information
--    
accessGroup :: (Integer,FiniteMap Integer Integer)->(Integer,AccessType,Integer,Integer,Matrix)->Matrix
accessGroup (totalWidth,accpositions) (arrayId,accType,laNode,stmtNr,accMtrxTr)
    -- if leftMatrix ends before totalWidth then it is 
    -- filled up with zero columns
    -- last column position is calculated by subtracting 
    -- the width of the current left matrix from totalWidth  
    | (pos + width) < total = insertZeroColumns leftMatrix (pos + width) (total-nColumns(leftMatrix)) 
    | otherwise = leftMatrix 
      where 
      -- the total width of this submatrix  
      total = totalWidth
      -- position/offset indexed by arrayId 
      pos = fromJust(lookupFM accpositions arrayId )
      -- width of the current transformd access-matrix 
      width = nColumns accMtrxTr
      -- matrix from the left: possible zero columns, access-matrix 
      leftMatrix = insertZeroColumns accMtrxTr 0 pos

-- test output to check the acc group matrices 
--accGroups = [accessGroup accPositions a | a <- accsTr]
--rowsOfAccGroups = [nRows x | x <- accGroups]



-- constructs the part which contains the transformed indexspace-matrix
-- plus the zero columns 
--  parameters: tuple with stmtNr and transformed indexspace-matrix,
--  index2pos-list and total width of the indexspace-submatrix 
--   
indexspaceGroup :: (Integer,Matrix) -> (Integer,Integer,(Integer,Integer))-> Integer  -> Matrix 
indexspaceGroup (stmt,dmatrx) (stmtNr,node,(offset,width)) totalWidth 
    | (pos+width) < total = insertZeroColumns leftMatrix (pos+width)(total-leftWidth) 
    | otherwise           = leftMatrix 
                          where 
                          leftMatrix = insertZeroColumns dmatrx 0 pos
		          leftWidth = nColumns leftMatrix 
                          total = totalWidth
                          pos = offset


-- calculates the total width of the indexspace part, 
-- gets as parameter the the result of ispc2position
--
totalWidthOf  :: [(Integer,Integer,(Integer,Integer))] -> Integer
totalWidthOf indexSpaceOffsets = addPair (trd(last(indexSpaceOffsets)))    
      where 
      trd (a,b,c)= c
      addPair (x,y)= x+y

-- search function for stmt2pos-Liste 
-- returns only the second component 
--
search :: Integer -> [(Integer,a)] -> a
search n [] = error (" Position and width not found") 
search n (x:xs) = if n == fst(x) then snd(x)
      else search n xs 

-- search function
-- returns the whole tuple 
--
search1 :: Integer -> [(Integer,a)] -> (Integer,a)
search1 n [] = error (" Position and width not found") 
search1 n (x:xs) = if n == fst(x) then x 
      else search1 n xs 

-- search function for list of quadruples
-- returns the found quadrupel
-- 
search2 :: Integer -> [(Integer,Integer,(Integer,Integer))] -> (Integer,Integer,(Integer,Integer))
search2 n [] = error (" Position and width not found") 
search2 n ((a,node,(offset,width)):xs) = if n == node then (a,node,(offset,width)) 
      else search2 n xs 


-- constructs the rows of final farkas matrix 
-- parameters: indexspace-list and access-matrix list
--
farkasRow :: [(Integer,Matrix)]->[(Integer,AccessType,Integer,Integer, Matrix)] -> [Matrix]
farkasRow ispc acc = map f preparedAccs
    where
    -- CAUTION: gets here acc
    preparedAccs = accTranspose acc
   
    -- the actual list comprehension
    f (arrayId,accType,node,stmtNr,accMtrx) 
    -- distinguishs if read or write access              
	| accType == Read = multiconc1[idGr,-(accGr),-dGr]
	| accType == Write =multiconc1[-(idGr),accGr,-dGr]
        where
        dGr = indexspaceGroup ispcTuple accTuple totalWidth 
	-- the suitable tuple of the transformed
	-- ispc list 
	ispcTuple = search1 stmtNr convertedIndexspace
	-- the tuple from the access-matrix list
	accTuple = search2 node dmatrixPos
	-- the total width of this d-matrix group
        totalWidth = totalWidthOf dmatrixPos
        -- positions of the d-matrices 
        dmatrixPos = ispc2position ispc acc
	-- transformed ispc matrices in tuple list 
	convertedIndexspace = convert2Dmatrx2 ispc
	accGr = accessGroup accMtrxsPos (arrayId,accType,node,stmtNr,accMtrx)
	-- positions of the access matrixes 
	accMtrxsPos = acc2posAndWidth preparedAccs
	idGr  = identityGroups stmtNr ematrixPos
		    where
		    -- CAUTION: gets here indexspace
		    -- the offsets of the d-matrices 
		    ematrixPos  = stmt2posAndWidth ispc



-- adds a zerorow to the matrix of joined matrices 
-- given by the output of farkasRow
--
farkasWithRightZerorow :: [Matrix]->Matrix
farkasWithRightZerorow farkasrows = insertZeroColumns submatrix pos 1
             where
    -- joins the matrices in the list
    submatrix = (multijoin1 farkasrows)
    -- width of the submatrix
    -- is position to insert 
    pos = nColumns submatrix 



-- function composition of farkasRow and 
-- farkasWithRightZerorow 
-- 
farkasMatrix :: [(Integer,Matrix)]->[(Integer,AccessType,Integer,Integer, Matrix)] ->  Matrix
farkasMatrix ispcList accList = farkasWithRightZerorow farkas
    where farkas = farkasRow ispcList accList



-- constructs the farkasMatrix just by the given path
--  
farkas:: String -> Matrix 
farkas path = farkasMatrix ispcsClean accsClean 
      where 
      ispcsClean = clean ispc slps
      accsClean  = cleanAccs accs slps
      ispc = unsafePerformIO( readispcfile path ) 
      accs  = unsafePerformIO( readaccfile path  )
      slps = unsafePerformIO (readslpfile path)

-- 
--umfang m  = ((nRows m),nColumns m )
--fm = farkas bsp

-- constructs the matrix representing the lambdas
-- for the polylib. Paramters are ispc list 
-- and acc list
-- 
lambdaMatrix2 ::[(Integer,Matrix)]->[(Integer,AccessType,Integer,Integer, Matrix)]-> Matrix
lambdaMatrix2 ispcList accList = multiconc1 [zeroMtrx, unitMtrx, zeroCol]
    where
    zeroMtrx = zeroMatrix numLambdas (widthFarkas - numLambdas - 1)
    -- to represent the lambdas 
    unitMtrx = unitMatrix numLambdas numLambdas
    -- the added zeroColumn as above 
    zeroCol  = zeroMatrix numLambdas 1
    -- width of the matrix, retrieved from constructed 
    -- farkas matrix  
    widthFarkas = nColumns farkas1
    farkas1 = farkasMatrix ispcList accList
    -- the width of the identity matrix to present lambdas
    numLambdas = totalWidthOf dmatrixPos
	where
	dmatrixPos = ispc2position ispcList accList

-- helper for extracting pi components and lamdba components
-- 
numberOfLambdas :: [(Integer, Matrix)] -> [(Integer,AccessType,Integer,Integer,Matrix)] -> Integer
numberOfLambdas ispc acc =  totalWidthOf dMatrixPositions
    where
    dMatrixPositions = ispc2position ispc acc 

--lm = lambdaMatrix2 indexspace accsClean

-- error occurs: Conc: Not the same number of rows
--lm2 = lambdaMatrix2 indexe accesses
--      where 
--      indexe = [(a,(getMatrix x))|(a,x)<- testIspcs]
--      accesses = [(a,b,c,d,(getMatrix e))| (a,b,c,d,e)<-accs]

--fm2 = farkasMatrix indexe accesses
      
--indexe = [(a,(getMatrix x))|(a,x)<- testIspcs]
--accesses = [(a,b,c,d,(getMatrix e))| (a,b,c,d,e)<-accs]