module InputLoopo (clean,cleanAccs,slps,readispcfile,readaccfile,readslpfile,
bsp,testIspcs,accs,loops,indexspace,accsClean) where 


import LoopoDS
import Loopo

import List 
import IOExts
import Ratio
import Maybe
import Char


-- path of the current test example
-- delivering the ispc, acc and slp files 
--
--bsp = /tmp/loopo1128/gausUeberfall.src.out"

bsp = "/tmp/loopo1048/example10_2.src.out"

-- test variables read out of the several files
-- (*.acc, *.ispc, *.slp) 

-- indexspaces, full-dimensional
testIspcs = unsafePerformIO (readispcfile bsp)

-- access matrices, full-dimensional
accs = unsafePerformIO (readaccfile bsp)


-- surrounding loops
loops = unsafePerformIO (readslpfile bsp)

-- dimensions are cleaned
-- 
indexspace = clean (unsafePerformIO (readispcfile bsp)) loops
accsClean  = cleanAccs (unsafePerformIO (readaccfile bsp)) loops 

----------------------------------------------------------------------
-- get input data from the ispc, acess and slp file 
---------------------------------------------------------------------

load :: (FilePath -> IO (LPResult a)) -> FilePath -> IO a
load f file =
    do res <- f file
       case res of
           ParseOk x    -> return x
	   ParseFailed f -> error ("Fehler: " ++ show f)

-- read the file, where the index space matrix is stored
--
readispcfile:: String -> IO ([(Integer, IneqSystem)])
readispcfile file =
    do  resispc<- load readISPC (file ++ ".ispc") 
        return resispc
     
-- surrounding loops 
--
readslpfile :: String -> IO ([(Integer,[Integer])])
readslpfile datei = 
     do resslps <- load readSLP (datei ++ ".slp")
	return resslps

-- read accesses 
--
readaccfile :: String -> IO ([(Integer, AccessType, Integer, Integer, AffineFunction)])
readaccfile datei = 
     do resaccs <- load readACC (datei ++ ".acc")
        return resaccs

-- shows the index space matrix for the respective statement
--
domainOfStmt :: Integer -> [(Integer,IneqSystem)] -> IneqSystem
domainOfStmt stmntid  ispcs =  
    case ( find (\(s,d)-> stmntid==s) ispcs ) of
         Nothing    -> error("statement not found")
         Just (s,d) -> d

-- helper function to clean the indexspace matrices
-- in case of non existing loops 
-- 
keep :: Matrix->[Integer]-> Integer->Matrix
keep m [] offset = m 
keep m (l:loops) offset = if (l==0) then eliminate (deleteColumns m offset 1) loops (offset + 1)
		   else keep m loops (offset+1)

eliminate m [] off = m 
eliminate m (l:xs) off
    | l == 0 = eliminate m' xs off
    | l == 1 = eliminate m xs (off+1)
    | otherwise = error "Weird surrounding loop information"
    where 
    m' = deleteColumns m off 1
				     
sus1 = getMatrix(snd(testIspcs!!0))
loo1 = snd(head(loops))

badboy  = (testIspcs!!2)
suspect = getMatrix (snd badboy)
badloop = loops!!2
lolly   = snd badloop 

sunny = deleteColumns suspect 5 2 
schauen = [nColumns (getMatrix(snd(t))) |t <-  testIspcs]


-- preprocessing of the indexspace matrices
-- by looking up loop information in the .slp file
--
clean :: [(Integer, IneqSystem)]->[(Integer,[Integer])] -> [(Integer,(Matrix))]
clean [] [] = []
clean (i:ispc) (l:loops) = if fst(i)==fst(l) 
			   then (fst(i),(keep (getMatrix (snd(i))) (snd(l)) 0)) : clean ispc loops
			   else error "Wrong identification during preprocessing"

-- test matrix 
maxe = unitMatrix 4 4 

-- deletes wrong dimensions in the access matrices 
--
cleanAccs ::[(Integer,AccessType,Integer,Integer, AffineFunction)]->[(Integer,[Integer])] ->[(Integer,AccessType,Integer,Integer, Matrix)]
cleanAccs [] lps = []
cleanAccs ((a,b,c,d,e):ispc) lps = (a,b,c,d, (keep (getMatrix e) (slps d lps) 0)) : cleanAccs ispc lps

-- shows the the list pf surounding loops for the 
-- specified statement (helper for cleanAccs)
--
slps :: Integer -> [(Integer,[Integer])] -> [Integer]
slps stmntId  loops =  
    case ( find (\(s,d)-> stmntId==s) loops) of
         Nothing    -> error("surrounding loops information not found")
         Just (s,d) -> d
