-- -*- haskell -*-
{
module InsertIfs.Scanner (scan) where

import InsertIfs.DS
import Data.Char (isDigit, chr)
import Numeric   (readOct)
}
 
%wrapper "posn"
 
@escapes = \\ ([n t \\ \"] | [0-3][0-7]{2})

tokens :-
 
"/*".*"*/"               { mkComment }
$white+                  { mkWhitespaces }
"{"                      { mkT TBraceBegin }
"}"                      { mkT TBraceEnd }
"("                      { mkT TParenBegin }
")"                      { mkT TParenEnd }
"="                      { mkT TAssign }
";"                      { mkT TSemi }
","                      { mkT TComma }
if                       { mkT TIf }
for                      { mkT TFor }
[A-Za-z\_][A-Za-z0-9\_]* { mkIdent }
.                        { mkUnknownChar }
 
{
mkComment :: AlexPosn -> String -> (Token, SourcePos)
mkComment (AlexPn _ line col) str = (TComment str, SourcePos line col)

mkWhitespaces :: AlexPosn -> String -> (Token, SourcePos)
mkWhitespaces (AlexPn _ line col) str = (TWhitespaces str, SourcePos line col)

mkT :: Token -> AlexPosn -> String -> (Token, SourcePos)
mkT t (AlexPn _ line col) _ = (t, SourcePos line col)

mkIdent :: AlexPosn -> String -> (Token, SourcePos)
mkIdent (AlexPn _ line col) str = (TID str, SourcePos line col)

mkUnknownChar :: AlexPosn -> String -> (Token, SourcePos)
mkUnknownChar (AlexPn _ line col) [ch] = (TUnknownChar ch, SourcePos line col)
mkUnknownChar _ str = error ("Internal error: mkUnknownChar received " ++ str)

scan :: String -> [(Token,SourcePos)]
scan = alexScanTokens
}
