module LMath.Algo.DTree
    ( topdownSimplifyDTreeS, topdownSimplifyDTree
    )

where

import Data.Maybe
import Data.Ratio

import LMath.Types

-- |
-- Simplify a decision tree top-down.
-- 
topdownSimplifyDTreeS ::
       (Num a, Ord a, Num poly)
    => (poly -> Maybe a)
    -> (b ->
	[QfFormula poly] ->
	poly ->
	(Signum,b)
       )
    -> (b ->
	[QfFormula poly] ->
	QfFormula poly ->
	(Bool,b)
       )
    -> (b ->
	[QfFormula poly] ->
        QfFormula poly   ->
        QfFormula poly
       )
    -> b
    -> [QfFormula poly]
    -> DTree poly val
    -> DTree poly val
topdownSimplifyDTreeS isConstPoly impliesSignum implies simplifyQf specialContext preconds dtree =
    simp specialContext preconds $
    simplifyDTreeMbOrd (fmap (`compare`0) . isConstPoly) dtree
    where
    simp sctx ctx l@(DLeaf _) = l
    simp sctx ctx (SCond p tm t0 tp)
	| signumGe sgn {-implies ctx (QfAt (AtGe p))-}  =
	   if      signumGt sgn then simp newSctx ctx tp
	   else if signumEq sgn then simp newSctx ctx t0
	   else EqCond p (simp newSctx (QfAt (AtEq p) : ctx) t0)
		         (simp newSctx (QfAt (AtGt p) : ctx) tp)
	| signumLe sgn =
	   if signumLt sgn
	      then simp newSctx ctx tm
	      else EqCond p (simp newSctx (QfAt (AtEq p)          : ctx) t0)
		            (simp newSctx (QfAt (AtGt (negate p)) : ctx) tm)
	| signumNeq sgn {-implies ctx (QfAt (AtNeq p))-} =
	   GeCond False p (simp newSctx (QfAt (AtGt p)          : ctx) tp)
		          (simp newSctx (QfAt (AtGt (negate p)) : ctx) tm)
	| otherwise = 
	    SCond p (simp newSctx (QfAt (AtGt (negate p)) : ctx) tm)
	            (simp newSctx (QfAt (AtEq p)          : ctx) t0)
		    (simp newSctx (QfAt (AtGt p)          : ctx) tp)
	where
	(sgn, newSctx) = impliesSignum sctx ctx p
    simp sctx ctx (EqCond p t0 tpm)
	| signumEq sgn  = simp newSctx ctx t0
	| signumNeq sgn = simp newSctx ctx tpm
	| otherwise            =
	    EqCond p (simp newSctx (QfAt (AtEq  p) : ctx) t0)
		     (simp newSctx (QfAt (AtNeq p) : ctx) tpm)
	where
	(sgn, newSctx) = impliesSignum sctx ctx p
    simp sctx ctx (GeCond leOk p t0p tm)
	| signumGe sgn         = simp newSctx ctx t0p
	| leOk && signumLe sgn = simp newSctx ctx tm
	| signumLt sgn         = simp newSctx ctx tm
	| otherwise            =
	    GeCond leOk p (simp newSctx (QfAt (AtGe p)          : ctx) t0p)
		          (simp newSctx (QfAt (AtGt (negate p)) : ctx) tm)
	where
	(sgn, newSctx) = impliesSignum sctx ctx p
    simp sctx ctx (FCond f tt tb)
	| impl_g     = simp newSctx_g     ctx tt
	| impl_not_g = simp newSctx_not_g ctx tb
	| otherwise =
	    FCond g (simp newSctx_g     (g       : ctx) tt)
		    (simp newSctx_not_g (QfNot g : ctx) tb)
	where
	(impl_g, newSctx_g)         = implies sctx ctx g
	(impl_not_g, newSctx_not_g) = implies sctx ctx (QfNot g)
	g = simplifyQf sctx ctx f


-- |
-- Top-down decision tree simplifier without special context information.
--
topdownSimplifyDTree :: (Num a, Ord a, Num poly)
		     => (poly -> Maybe a)
                     -> ([QfFormula poly] -> poly -> Signum)
                     -> ([QfFormula poly] -> QfFormula poly -> Bool)
                     -> ([QfFormula poly] -> QfFormula poly -> QfFormula poly)
                     -> [QfFormula poly]
                     -> DTree poly val 
		     -> DTree poly val
topdownSimplifyDTree isConst implS implQf simplify preconds dtree =
    topdownSimplifyDTreeS isConst
			  (\sc c f -> (implS  c f, undefined))
                          (\sc c p -> (implQf c p, undefined))
			  (\sc c f -> simplify c f)
			  undefined
			  preconds
			  dtree
