{-# OPTIONS -fwarn-incomplete-patterns #-}
-- |
-- Interface to Logic and Quantifier Elimination procedures.
--
module LMath.Logic
  ( -- * Quantifier Elimination and related Functions
    module LMath.Logic.QE,

    -- * DTreeContexts related to the Logic modules
    module LMath.Logic.DTreeContexts,

    -- * Sorting Inequality Systems
    module LMath.Logic.Sort,

    -- * Domain package
    module LMath.Logic.Domain,

    -- * Armin's experimental functions
    optBounds
  )
where

import Data.Ratio
import Data.List
import Data.Maybe
import Debug.Trace

import LMath.Algo
import LMath.Logic.Domain
import LMath.Logic.DTreeContexts
import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Types


-- entfernt ueberfluessige Grenzen,
optBounds :: Integral a =>
	     [QfFormula (Polynomial (Ratio a))] ->
	     DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a)))) ->
	     DTree (Polynomial (Ratio a)) (Maybe (SortedSystem (FracPoly (Ratio a))))
optBounds ctx (SCond p tm t0 tp)  =
    SCond p (optBounds (QfAt (AtGt (-p)) : ctx) tm)
            (optBounds (QfAt (AtEq    p) : ctx) t0)
	    (optBounds (QfAt (AtGt    p) : ctx) tp)
optBounds ctx (EqCond p t0 tpm) =
    EqCond p (optBounds (QfAt (AtEq  p) : ctx) t0)
             (optBounds (QfAt (AtNeq p) : ctx) tpm)
optBounds ctx (GeCond leOk p t0p tm) =
    GeCond leOk p (optBounds (QfAt (AtGe    p) : ctx) t0p)
                  (optBounds (QfAt (AtGt (-p)) : ctx) tm)
optBounds ctx (FCond qf tt tb) =
    FCond qf (optBounds (qf       : ctx) tt)
             (optBounds (QfNot qf : ctx) tb)
optBounds ctx (DLeaf Nothing) = DLeaf Nothing
optBounds ctx (DLeaf (Just ss)) =
    DLeaf $ Just $ optBounds' ctx ss
{-
    DLeaf $ Just $ optBnds mbNp (map (fmap $ liftPolyN ctxVars) ctx) ss
    where
    ctxVars = case ss of
		      (Equation vs : _)      -> vectorLength vs - 1
		      (LowerUpper ls us : _) -> vectorLength (head (ls++us)) - 1
    mbNp = 2
-}

optBounds' :: Integral a => [QfFormula (Polynomial (Ratio a))] -> SortedSystem (FracPoly (Ratio a)) -> SortedSystem (FracPoly (Ratio a))
optBounds' ctx = ob []
    where
    ob newSS [] = newSS
    ob newSS (e@(Equation _) : rest) = ob (newSS ++ [e]) rest
    ob newSS (LowerUpper low upp : rest) =
	ob (newSS ++ [LowerUpper lowNew uppNew]) rest
	where
	(lowNew, uppNew) = check low upp [] []

        check [] [] kls kus = (kls, kus)
        check [] (u:us) kls kus 
	    | impl (newSS ++ [LowerUpper kls (us++kus)]) (insertEntries u 0 [-1]) =
		check [] us kls kus
	    | otherwise = check [] us kls (u:kus)
        check (l:ls) us kls kus 
	    | impl (newSS ++ [LowerUpper (ls++kls) (us++kus)]) (insertEntries (negate l) 0 [1]) =
		check ls us kls kus
	    | otherwise = check ls us (l:kls) kus

--        nV = vectorLength (head (l++us))

    impl ss bound = {- trace (show (sys,redundant)) $ -} -- noch nicht richtig
	case redundant of
	  DLeaf True -> True
	  _          -> False
	where
	redundant =
--	    combineSubtrees $
	    fmap unBound $
--	    combineSubtrees $
	    t

	t = projectToVarsPss' sortIneqsPss ctx sys Nothing Nothing [0]

	sys = sortedSystemToIneqSystem (vectorLength bound) 0 ss
	      `join`
	      ineqSystem ((vectorLength bound,0),
			  rowMatrix (-1 : vectorEntries (negate bound))
			 )

        unBound Nothing             = True
        unBound (Just [Equation e]) =
	     impliesQf ctx (QfAt $ AtGe $ negate $ prod $ vectorEntry e 0)
	unBound (Just [LowerUpper ls us]) =
	     impliesQf ctx (disjunctionQf [QfAt $ AtGe $ negate $ prod $ vectorEntry u 0 | u <- us])
	unBound _ = error "optBounds: the 'impossible' happened ?!"

	prod fp = polyNumerator fp * polyDenominator fp


optBnds :: Integral a => Integer -> [QfFormula (Polynomial (Ratio a))] -> SortedSystem (FracPoly (Ratio a)) -> SortedSystem (FracPoly (Ratio a))
optBnds _ ctx [] = []
optBnds nInd ctx (e@(Equation _) : rest) =
    e : optBnds nInd (boundsToQfs nInd e ++ ctx') rest
    where
    ctx' = map (fmap liftPoly) ctx
optBnds nInd ctx (LowerUpper lower upper : rest) =
    LowerUpper (map fst newLs) (map fst newUs)
    : optBnds nInd ctx'' rest
    where
    (newLs, newUs) = opt lower upper
    ctx'' = map snd newLs ++ map snd newUs ++ ctx'
    ctx' = map (fmap liftPoly) ctx
--    opt :: (Polynomial (Ratio a) -> QfFormula (Polynomial (Ratio a))) ->
--	   [GVector (FracPoly (Ratio a))] -> [GVector (FracPoly (Ratio a))]
    opt ls us = optH ls' us' [] []
	where
	ls' = map (\b -> (b, QfAt $ AtGe $ listToPolynomial nInd (1:vectorEntries (negate b)))) ls
	us' = map (\b -> (b, QfAt $ AtGe $ listToPolynomial nInd (-1 : vectorEntries b))) us

        optH []     [] kls kus = (kls, kus)
        optH [] (x:xs) kls kus
	    | {- trace (show (map snd (xs++kls++kus) ++ ctx', snd x)) $ -}
	      impliesQf (map snd (xs++kls++kus) ++ ctx') (snd x) = optH [] xs kls kus
	    | otherwise = optH [] xs kls (x:kus)
        optH (x:xs) uus kls kus
	    | impliesQf (map snd (xs++uus++kls++kus) ++ ctx') (snd x) = optH xs uus kls kus
	    | otherwise = optH xs uus (x:kls) kus
{-
    opt f [] = []
    opt f (b:bs)
	| and [check c b | c <- cs]  =  b : cs
	| otherwise                  = cs
	where
	cs = [c | c <- opt f bs, check b c]
	check a b
	    | all (==0) xs  =  not $ impliesQf ctx (f x)
	    | otherwise     =  True
	    where
	    v = a - b
	    (x':xs) = reverse $ vectorEntries v
	    x = polyNumerator x' * polyDenominator x'
-}

boundsToQfs :: Integral a
	    => Integer
	    -> Bounds (GVector (FracPoly (Ratio a)))
	    -> [QfFormula (Polynomial (Ratio a))]
boundsToQfs nIndet (Equation a) =
    [QfAt $ AtEq $ listToPolynomial nIndet $ vectorEntries a]
boundsToQfs nIndet (LowerUpper ls us) =
    [QfAt $ AtGe $ listToPolynomial nIndet (1 : vectorEntries (negate l)) | l <- ls] ++
    [QfAt $ AtGe $ listToPolynomial nIndet (-1 : vectorEntries u) | u <- us]

unfracV :: Integral a => [FracPoly (Ratio a)] -> [Polynomial (Ratio a)]
unfracV v =
    map (\fp -> polyNumerator fp * (l `quotPoly` polyDenominator fp)) v
    where
    l = foldl lcmPoly 1 denoms
    denoms = map polyDenominator v


listToPolynomial :: Integral a => Integer -> [FracPoly (Ratio a)] -> Polynomial (Ratio a)
listToPolynomial nIndet v =
    sum $
    zipWith (*) (map (liftPolyN (genericLength v-1)) v') ([indeterminate (genericLength v+nIndet-1) i | i <- [0..genericLength v-2]] ++ [1])
    where
    v' = unfracV v

