module LMath.Logic.DTreeContexts
    ( fracPolyDTreeContext, polyDTreeContext,
      -- * Experimental functions
      fcDTreeContext
      -- , polyXDTreeContext
    )
where

import Data.Ratio

import Config

-- import LMath.Algo.IntZeros
import LMath.Logic.QE
import LMath.Types
import LMath.Tools


polyDTreeContext :: Integral a
		     => [QfFormula (Polynomial (Ratio a))]  -- ^ initial context
                     -> DTreeContext (Polynomial (Ratio a)) (Polynomial (Ratio a))
polyDTreeContext ctx =
    dtc { dtcSimplifyIsWeak = not $ externalToolAvailable SLFQ }
    where
    dtc = newDTreeContext ctx
		    (\ctx f -> impliesQf  ctx f)
		    (\ctx p -> impliesSignum ctx p)
		    (\ctx f -> simplifyQf ctx f)
		    (\p -> p)
                    (\p -> fmap (`compare` 0) $ isConstPoly p)
-- |
-- A 'DTreeContext' for @RealFrac a => FracPoly a@
--
fracPolyDTreeContext :: RealFrac a
		     => [QfFormula (Polynomial a)]  -- ^ initial context
                     -> DTreeContext (FracPoly a) (Polynomial a)
fracPolyDTreeContext ctx =
    dtc { dtcSimplifyIsWeak = not $ externalToolAvailable SLFQ }
    where
    dtc = newDTreeContext ctx
		    (\ctx f -> impliesQf  ctx f)
		    (\ctx p -> impliesSignum ctx p)
		    (\ctx f -> simplifyQf ctx f)
		    (\fp -> simp (polyNumerator fp * polyDenominator fp))
                    (\p -> fmap (`compare` 0) $ isConstPoly p)
    simp p = mapPoly fromIntegral (constPoly co * pp)
        where
        pp = product [f^e | (f,expo) <- fs, let e = if odd expo then 1 else 2]
        (co,fs) = factoriseP $ snd $ ratPoly2intPoly p

fcDTreeContext :: Integral a
               => [QfFormula (IFracPolyFC a)]  -- ^ initial context
               -> DTreeContext (IFracPolyFC a) (IFracPolyFC a)
fcDTreeContext ctx =
    newDTreeContext ctx
		    (\ctx f -> impliesQfFC ctx f)
		    (\ctx p -> impliesSignumFC ctx p)
		    (\ctx f -> f)
		    (\p -> p)
                    (fmap (`compare` 0) . isConstFC)

{-
-- maybe later...
polyXDTreeContext :: Integral a
               =>   -- ^ arbitrary user-supplied value
               -> [QfFormula (X a)]  -- ^ initial context
               -> DTreeContext (X a) (X a)
polyXDTreeContext ctx =
    newDTreeContext ctx
		    (\sc ctx f -> (impliesXQf ctx f, sc))
		    (\sc ctx p -> (impliesXSignum ctx p, sc))
		    (\sc ctx f -> f)
		    (\p -> p)
-}