module LMath.Logic.Examples.CloogExample3 where

import Data.Ratio
import Data.Maybe
import Data.List (genericLength, genericTake, maximumBy, intersperse)
import Debug.Trace

import LMath.Logic.QE
import LMath.Logic.Sort
import LMath.Tools
import LMath.Types
import LMath.Logic.DTreeContexts

import LMath.Logic.Domain
import LMath.Logic.DomainContext
import LMath.Logic.DomainTree
import LMath.Logic.Examples.Quillere
import VisualPolylib

----------------------------------------------------------------------
-- Beispiel aus dem Benutzerhandbuch von CLooG
----------------------------------------------------------------------

-- parameters:
n_ :: Polynomial Rational
n_ = indeterminate nP_ 0

params_ :: [Polynomial Rational]
params_ = [n_]

nP_, nV_ :: Integer
nP_ = genericLength params_
nV_ = 2

--
-- The context which should be used within this module.
--
context_ = --Nothing
           mkDC n_ge_25_context_

-- context: n >= 25
n_ge_25_context_ = [QfAt $ AtGe (n_ - 25)]

-- context: n == 25
n_eq_25_context_ = [QfAt $ AtEq (n_ - 25)]

-- context: n >= 15
n_ge_15_context_ = [QfAt $ AtGe (n_ - 15)]

polynomial_tree_context_ =
    polyDTreeContext () (fromDomainContext' context_)
fracPoly_tree_context_ =
    fracPolyDTreeContext () (fromDomainContext' context_)

s1_domain :: (Domain (Polynomial Rational), String)
s1_domain = (d, "S1")
    where
    d = ineqToDomain2 context_ (nV_, nP_) ineqSys
    ineqSys = ineqSystem ((nV_, 0), matrix m)
    --    i    j      1
    m = [[1,   0,     0 ] -- i >= 0
	,[-1,  0,     n_] -- i <= n
	,[0,   1,     0 ] -- j >= 0
	,[0,  -1,  n_-15] -- j <= n - 15
	]

s2_domain :: (Domain (Polynomial Rational), String)
s2_domain = (d, "S2")
    where
    d = ineqToDomain2 context_ (nV_, nP_) ineqSys
    ineqSys = ineqSystem ((nV_, 0), matrix m)
    --    i    j      1
    m = [[1,   0,    -15] -- i >= 15
	,[-1,  0,     n_] -- i <= n
	,[0,   1,    -10] -- j >= 10
	,[0,  -1,     n_] -- j <= n
	]

allDomains = [s1_domain, s2_domain]

main = do
       putStrLn "s1_domain:"
       putStrLn (show s1_domain)
       putStrLn "s2_domain:"
       putStrLn (show s2_domain)
       printLine
       putStrLn "calculating projections:"
       putStrLn "s1_project_1:"
       putStrLn (show s1_project_1)
       putStrLn "s2_project_1:"
       putStrLn (show s2_project_1)
       printLine
       putStrLn "calculating partitions_1:"
       putStrLn (show partitions_1)
       printLine
       putStrLn "calculating topsort_1:"
       putStrLn (show topsort_1)
       printLine
       putStrLn "calculating projections_2:"
       putStrLn (show projections_2)
       printLine
       putStrLn "calculating partitions_2"
       putStrLn (show partitions_2)
       printLine
       putStrLn "calculating topsort_2"
       putStrLn (show topsort_2)
    where
    printLine = putStrLn (replicate 70 '-')

--
-- Helper functions for the projection phase:
--
proj1 (d,l) = fmap (\x -> (x,l))
	      (projectToVarsD polynomial_tree_context_ context_ [0] d)
proj12 (d,l) = fmap (\x -> (x,l))
	       (projectToVarsD polynomial_tree_context_ context_ [0,1] d)


--
-- Project on the 1. dimension (which is "t")
--
s1_project_1, s2_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
s1_project_1 = proj1 s1_domain
s2_project_1 = proj1 s2_domain

--
-- Partitionate the results of the projections and
-- find the overlapping parts.
--
partitions_1 :: DTree (Polynomial Rational)
		[(Domain (Polynomial Rational), [String])]
partitions_1 = lPartitionDT context_ projections
               --fmap (map (\(x,y) -> (ineqToDomain2 context_ (nV_,nP_) x, y))) $
	       --lPartitionDT2 context_ projections
    where
    projections = mergeListOfDTrees polynomial_tree_context_ (++) $
		  map (fmap (\x -> [x])) projList
    projList = [s1_project_1, s2_project_1]

--
-- Topologically sort the partitions.
--
topsort_1 :: DTree (Polynomial Rational)
	     [(Domain (Polynomial Rational), [String])]
topsort_1 = lTopSortDT context_ 0 partitions_1


test_topsort_1 = catchToolIOError (print topsort_1) print

--
-- Project on the 1st and 2nd dimension
--
s1_project_12 = proj12 s1_domain
s2_project_12 = proj12 s2_domain

projections_2 :: DTree (Polynomial Rational)
                 [(OuterLoop, [(Domain (Polynomial Rational), String)])]
projections_2 = fmapDTree_with_context initContext proj2 topsort_1
    where
    initContext = fromDomainContext' context_
    proj2 c ps = [(OuterLoop d, [(simplifyD (mkDC c) ((getDomain l) `intersectD` d), l) | l <- ls])
		  | (d,ls) <- ps]
    -- Wichtig:
    -- An dieser Stelle muesste eigentlich die Projektion auf die 1+2
    -- Komponente von S1,..,S4 gemacht werden. Weil wir in diesem Beispiel
    -- aber nur 2 Dimensionen haben, habe ich das hier weggelassen.
    -- Im Allgemeinen muesste man das aber schon machen und bekommt dann
    -- nicht einfache Domains, sondern Baeume von Domains.
    -- Es ist zwar moeglich, diese zusaetzliche Ebene wieder zu
    -- entfernen, aber das ist eine recht laestige Arbeit...
    --
    -- Fuer unsere Zwecke sollte es OK sein, wenn man das hier weglaesst.
    --
    getDomain l = fst $ head $ filter (\(_,l') -> l == l') allDomains

partitions_2 :: DTree (Polynomial Rational)
                [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
partitions_2 = fmapDTree_with_context initContext (\c -> map (partition c)) projections_2
    where
    initContext = fromDomainContext' context_
    partition c (outLoop, ds) = (outLoop, lPartitionD (mkDC c) ds)

topsort_2 :: DTree (Polynomial Rational)
             [(OuterLoop, [(Domain (Polynomial Rational), [String])])]
topsort_2 = simplifyDTree (fromDomainContext' context_) $
	    flatten $ fmapDTree_with_context initContext
	    (\c -> (map (\(outLoop, xs) -> fmap (\x -> (outLoop, x)) (lTopSortD (mkDC c) 1 xs)
                        )
                   )
            )
	    partitions_2
    where
    initContext = fromDomainContext' context_
    flatten1 dt = fmap (\t -> if null t
			      then DLeaf []
                              else mergeListOfDTrees polynomial_tree_context_ mergeOp $
                                   map (fmap (\x -> [x])) t
		       ) dt
    flatten2 = flattenDTree
    flatten = flatten2 . flatten1
    mergeOp = (++)

polyLibRep topsort_result = -- z.B. (subst_topsort_2 [Just 99] Nothing)
    putStrLn $ unlines $
    map (\(d,l) -> "\nLabel " ++ show l ++ ":\n" ++
		   showDomain2VisualPolylib d (nVarsD d) []) $
    concat $ head $ leavesDTree topsort_result

--
-- Replace the parameter @m@ by a numerical constant.
--
subst_context_ = mkDC n_eq_25_context_
simpl_domain_  = (simplifyD subst_context_) .
		 (substituteAllConstantsD subst_context_)
simpl_dtree_   = simplifyDTree (fromDomainContext' subst_context_)

--
-- Helper functions (for the substitution and simplification):
--
simpl_proj1_ = simpl_dtree_ . fmap (\(x,y) -> (simpl_domain_ x, y))
simpl_partitions1_ =
    simpl_dtree_ .
    fmap (filter (not.isAlwaysEmptyD'.fst) . (map (\(x,y) -> (simpl_domain_ x, y))))
simpl_topsort1_ = simpl_partitions1_

--
-- Substituted projections on the 1. dimension.
--
subst_s1_project_1, subst_s2_project_1 ::
    DTree (Polynomial Rational) (Domain (Polynomial Rational), String)
subst_s1_project_1 = simpl_proj1_ s1_project_1
subst_s2_project_1 = simpl_proj1_ s2_project_1

--
-- Substituted partitions.
--
subst_partitions_1 :: DTree (Polynomial Rational)
		      [(Domain (Polynomial Rational), [String])]
subst_partitions_1 = simpl_partitions1_ partitions_1

--
-- Substituted topologically sorted partitions.
--
subst_topsort_1 :: DTree (Polynomial Rational)
		   [(Domain (Polynomial Rational), [String])]
subst_topsort_1 = simpl_topsort1_ topsort_1

subst_topsort_2 params context =
    let subst x = substituteParamsD x params
    in fmapDTree_with_context (fromDomainContext' context)
	   (\c -> map (\(OuterLoop outLoop, xs) -> (OuterLoop (simplifyD (mkDC c) (subst outLoop)),
						    map (\(x,y) -> (simplifyD (mkDC c) x,y)) xs))) $
       fmap (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (subst x,y)) xs))) topsort_2

subst_n25_topsort_2 = simplifyAllDomains $
		      simplifyDTree n_eq_25_context_ $
			     subst_topsort_2 [Just 5, Just 10, Just 10] context_
    where
    simplifyAllDomains = fmapDTree_with_context n_eq_25_context_
			 (\context -> (map (\(outLoop, xs) -> (outLoop, map (\(x,y) -> (simplifyD (mkDC context) x, y)) xs)) ))

-- Zum Anzeigen:
-- putStrLn $ concat $ concatMap (map (\(x,y) -> (showDomain2VisualPolylib x 2 [] ++ show y ++ "\n" ++ replicate 70 '-' ++ "\n"))) $ head $ leavesDTree subst_p5_q10_r10_topsort_2
