module LMath.Numbers
    ( primes, factorise )
where

import Control.Exception as CE

primes :: [Integer]
primes = sieve [2..]
    where
    sieve :: [Integer] -> [Integer]
    sieve (x:xs) = x : sieve [y | y <- xs, y `rem` x /= 0]


-- |
-- Factorise a non-negative integer.
--
-- @
--   let (sgn,facts) = factorise n
--   in  sgn * product [f^e | (f,e) <- facts]  ==  n
-- @
--
factorise :: Integral b => b -> (b, [(b,b)])
factorise 0 = (0, [])
factorise n =
    (signum n
    , map (\(f,e) -> (fromIntegral f,fromIntegral e)) $
      f primes (abs $ fromIntegral n)
    )
    where
    f ps 1     = []
    f (p:ps) n
      | d > 0     = (p,d) : f ps qu
      | otherwise = f ps n
	where
	(qu,d) = divs n 0
	divs x c
	    | r == 0    = divs q (c+1)
	    | otherwise = (x,c)
	    where
	    (q,r) = x `quotRem` p
