load_package assist;
load_package redlog;
off rlverbose;
rlset ofsf;
on ratarg;

% for 'resultant'
% on bezout;

% try to avoid linebreaks in numbers
linelength(100000)$

% print a result and return the empty string(!);
% printResult is used in a statement
%  write "&", printResult(...), "#";
%
procedure printResult(func, r);
begin
  map(func, {r});
  return "";
end$

% prints quantifier-free formulas and terms in HsLooPo's notation 
procedure printFormula(ex,varN);
begin scalar oper, p1, p2, len, junc;
  if ex eq true then
    write "QfAt AtTrue"
  else if ex eq false then
    write "QfAt AtFalse"
  else if numberp(ex) then
  begin
    write "Const (";
    if varN < 0 then
      write "AtLeast ", -(varN+1)
    else
      write "Exactly ", varN;
    write "," , ex, ")";
  end
  else if symbolic idp(ex) then
  begin
    write "Var (";
    if varN < 0 then
      write "AtLeast ", -(varN+1)
    else
      write "Exactly ", varN;
    write ",", detidnum(ex), ")";
  end
  else
  begin
    oper := part(ex, 0);
    if oper eq minus then
    begin
      write "Neg (";
      printFormula(part(ex,1), varN);
      write ")";
    end
    else if oper eq not then
    begin
      write "QfNot (";
      printFormula(part(ex,1), varN);
      write ")";
    end
    else if oper eq expt then
    begin
      write "Pwr (";
      printFormula(part(ex,1), varN);
      write ") ", part(ex,2);
    end
    else
    begin
      len := arglength(ex);
      p1 := part(ex,1);
      p2 := part(ex,2);
      junc := (if      oper eq plus  then "Add "
               else if oper eq times then "Mul "
               else if oper eq and   then "QfAnd "
	       else if oper eq or    then "QfOr "
               else if oper eq greaterp then << p2 := nil; "QfAt (AtGt" >>
               else if oper eq geq      then << p2 := nil; "QfAt (AtGe" >>
               else if oper eq lessp    then << p1 := -p1; p2 := nil; "QfAt (AtGt" >>
               else if oper eq leq      then << p1 := -p1; p2 := nil; "QfAt (AtGe" >>
               else if oper eq equal    then << p2 := nil; "QfAt (AtEq"  >>
               else if oper eq neq      then << p2 := nil; "QfAt (AtNeq" >> );
      if len>2 then
      begin
        for p := 2:len do
          write "(", junc;
        write "(";
        printFormula(part(ex,1), varN);
        write ")";
        for p := 2:len do
        begin
          write "(";
          printFormula(part(ex,p), varN);
          write "))";
        end;
      end
      else
      begin
        write junc;
        write "(";
        printFormula(p1, varN);
        write ")";
        if p2 then
        begin
          write " (";
          printFormula(p2, varN);
        end;
        write ")";
      end;
    end;
  end;
end$

procedure printQuo(q, nPar);
begin
  write "(";
  printFormula(num(q), nPar);
  write ",";
  printFormula(den(q), nPar);
  write ")";
end$

procedure printBounds(bnds, curVar, nPar);
begin scalar isFirst, cv, co;
  write "[";
  isFirst := t;
  for each bnd in bnds do
  begin
    if isFirst then isFirst:=nil else write ",";
    write "[";
    for i:=0 : (curVar-1) do
    begin
      co := coeffn(bnd, mkid(x,i), 1);
      printQuo(co, nPar);
      bnd := sub({mkid(x,i)=0}, bnd);
      write ",";
    end;
    printQuo(bnd, nPar);
    write "]";
  end;
  write "]";
end$

procedure printSorted(res, curVar, nPar);
begin scalar isFirst;
  if res eq sol_ok then
    write "SolOk"
  else
  begin
    write "Sorted [";
    isFirst := t;
    for each tup in res do
    begin
      if isFirst then isFirst:=nil else write ",";
      write "(";
      printFormula(first(tup), nPar);
      write ",";
      printBounds(second(tup), curVar, nPar);
      write ",";
      printBounds(third(tup), curVar, nPar);
      write ",";
      printSorted(part(tup,4), curVar+1, nPar);
      write ")";
    end;
    write "]";
  end;
end$

procedure printList(thelist, printer);
begin scalar isFirst;
  write "[";
  isFirst := t;
  for each anelement in thelist do
  begin
    if isFirst then isFirst := nil else write ",";
    map(printer, {anelement});
  end;
  write "]";
end;

procedure printTuple(tuple, printer);
begin scalar isFirst;
  write "(";
  isFirst := t;
  for each compo in tuple do
  begin
    if isFirst then isFirst := nil else write ", ";
    map(printer, {compo});
  end;
  write ")";
end;

procedure printPair(tuple, printerA, printerB);
begin
  write "(";
  map(printerA, {part(tuple,1)});
  write ",";
  map(printerB, {part(tuple,2)});
  write ")";
end;

procedure print1Lexminmax(lm,nPar);
begin scalar isFirst;
  write "(";
  printFormula(first(lm), nPar);
  write ",[";
  isFirst := t;
  for each l in second(lm) do
  begin
    if isFirst then isFirst := nil else write ",";
    printQuo(l, nPar);
  end;
  write "])";
end;


% a>=0 and b<=0 and c=0 and d>0 and e<0  -->   {a,-b,c,-c,d-1,-e-1}
% NOTE: "false" as input formula is not handled correctly
%       (should never occur when conjToBounds is called for
%       the result of a disjointUnion).
procedure conjToBounds(formula);
begin scalar oper, tm;
  return
  for each atf in rlatl(formula) join
    << if atf eq true then {}
       else << oper := part(atf,0);
               tm := lhs(atf);
               if      oper eq geq      then {tm}
               else if oper eq leq      then {-tm}
               else if oper eq greaterp then {tm-1}
               else if oper eq lessp    then {-tm-1}
               else if oper eq equal    then {tm,-tm}
            >>
    >>
end;

%ineqs := x1 >= 0 and x1 >= p and x2 >= 0 and x2 <= -x1+10;

%ineqs2 := x1 >= 0 and x1 >= p and q*x2 >= 0 and x2 <= -x1+10;
%ineqs3 := x1 >= 0 and q*x2 >= 0 and x2 <= -x1+10;

%ineqsT1 := x1 >= 0 and x1 <= n and o1 >= 0 and o1 <= s and t1*(s+1)+o1=x1;
varsT1 := {t1,o1,x1};

% ex({a,b},all({x1,x2},ineqs impl x2 >= b*x1+a) and ex({x1},sub({x2=b*x1+a},ineqs)))

%ineqsM := 1*x5+0 >= 0 and 1*x6+0 >= 0 and -1*x5+-1*x6+15 >= 0 and 1*x11+0 >= 0 and 1*x12+0 >= 0 and -1*x11+-1*x12+15 >= 0 and -3*x1+-3*x2+1*x6+0 >= 0 and 3*x1+3*x2+-1*x6+2 >= 0 and 3*x1+1*x5+-1*x6+0 >= 0 and -3*x1+-1*x5+1*x6+2 >= 0 and -3*x7+-3*x8+1*x12+0 >= 0 and 3*x7+3*x8+-1*x12+2 >= 0 and 3*x7+1*x11+-1*x12+0 >= 0 and -3*x7+-1*x11+1*x12+2 >= 0 and 3*x2+1*x3+-1*x5+0 >= 0 and 3*x1+3*x2+1*x4+-1*x6+0 >= 0 and -3*x2+-1*x3+1*x5+0 >= 0 and -3*x1+-3*x2+-1*x4+1*x6+0 >= 0 and 3*x8+1*x9+-1*x11+0 >= 0 and 3*x7+3*x8+1*x10+-1*x12+0 >= 0 and -3*x8+-1*x9+1*x11+0 >= 0 and -3*x7+-3*x8+-1*x10+1*x12+0 >= 0 and -1*x5+1*x11+0 >= 0 and -1*x6+1*x12+-1 >= 0 and 1*x5+-1*x11+0 >= 0 and 1*x6+-1*x12+1 >= 0;
varsM:={x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12};

%ineqsM2 := p*x5+0 >= 0 and 1*x6+0 >= 0 and -1*x5+-1*x6+15 >= 0 and 1*x11+0 >= 0 and 1*x12+0 >= 0 and -1*x11+-1*x12+15 >= 0 and q*x1+-3*x2+1*x6+0 >= 0 and 3*x1+3*x2+-1*x6+2 >= 0 and 3*x1+1*x5+-1*x6+0 >= 0 and -3*x1+-1*x5+1*x6+2 >= 0 and -3*x7+-3*x8+1*x12+0 >= 0 and 3*x7+3*x8+-1*x12+2 >= 0 and 3*x7+1*x11+-1*x12+0 >= 0 and -3*x7+-1*x11+1*x12+2 >= 0 and 3*x2+1*x3+-1*x5+0 >= 0 and 3*x1+3*x2+1*x4+-1*x6+0 >= 0 and -3*x2+-1*x3+1*x5+0 >= 0 and -3*x1+-3*x2+-1*x4+1*x6+0 >= 0 and 3*x8+1*x9+-1*x11+0 >= 0 and 3*x7+3*x8+1*x10+-1*x12+0 >= 0 and -3*x8+-1*x9+1*x11+0 >= 0 and -3*x7+-3*x8+-1*x10+1*x12+0 >= 0 and -1*x5+1*x11+0 >= 0 and -1*x6+1*x12+-1 >= 0 and 1*x5+-1*x11+0 >= 0 and 1*x6+-1*x12+1 >= 0;

%ineqsM3 := p*x5+0 >= 0 and 1*x6+0 >= 0 and -1*x5+-1*x6+15 >= 0 and 1*x11+0 >= 0 and 1*x12+0 >= 0 and (p-1)*x11+q*x12+15 >= 0 and -3*q*x1+-3*x2+1*x6+0 >= 0 and 3*x1+3*x2+-1*x6+2 >= 0 and 3*x1+1*x5+-1*x6+0 >= 0 and -3*x1+-1*x5+1*x6+2 >= 0 and -3*x7+-3*x8+1*x12+0 >= 0 and 3*x7+3*x8+-1*x12+2 >= 0 and 3*x7+1*x11+-1*x12+0 >= 0 and -3*x7+-1*x11+1*x12+2 >= 0 and 3*x2+1*x3+-1*x5+0 >= 0 and 3*x1+3*x2+1*x4+-1*x6+0 >= 0 and -3*x2+-1*x3+1*x5+0 >= 0 and -3*x1+-3*x2+-1*x4+1*x6+0 >= 0 and 3*x8+1*x9+-1*x11+0 >= 0 and 3*x7+3*x8+1*x10+-1*x12+0 >= 0 and -3*x8+-1*x9+1*x11+0 >= 0 and -3*x7+-3*x8+-1*x10+1*x12+0 >= 0 and -1*x5+1*x11+0 >= 0 and -1*x6+1*x12+-1 >= 0 and 1*x5+-1*x11+0 >= 0 and 1*x6+-1*x12+1 >= 0;


% rlqe(ex(x2, sub({x1=m},ineqs)) and all({x1,x2},ineqs impl m<=x1));

procedure dbg(a);
begin
  write(a);
end;

procedure runIt(sys,vars,theo);
begin scalar lm, lm2, cs;
  lm := min1(sys,vars,theo);
  lm2 := removeSpecializations(lm,theo,{});
  cs := findCompats(lm,theo,t);
  return cs;
end;

procedure sortIneqs(sys,vars,theo);
begin scalar b, lbss, ubss;
  b := lowerAndUpperBounds(sys,vars,theo);
  lbss := first(b);
  ubss := second(b);
  return makeTree(theo, for each th in theo mkand th, lbss, ubss);
end;

procedure lowerAndUpperBounds(sys,vars,theo);
begin scalar lbs, ubs, lbss, ubss, ctxvars;
  if vars={} then return {{sys,{}}};

  ctxvars := {};
  lbss := {};
  ubss := {};
  while length(vars) > 0 do
  begin
%dbg({first(vars)});
    lbs := append (lowerBounds(sys,vars,ctxvars,theo),
                   lowerUnbounded(sys,vars,ctxvars,theo));
    ubs := append (upperBounds(sys,vars,ctxvars,theo),
                   upperUnbounded(sys,vars,ctxvars,theo));
    lbss := lbs . lbss;
    ubss := ubs . ubss;
    ctxvars := first(vars) . ctxvars;
    vars    := rest(vars);
  end;

  return {reverse(lbss), reverse(ubss)};
end;

procedure makeTree(theo, ctx, lbss, ubss);
begin scalar res, lbc, lbEqs, lbIneqs, ubc, ubEqs, ubIneqs, bc, ctx2;
  if length(lbss) = 0 then return SOL_OK;
%dbg({enter,length(lbss)});
  res := {};
  for each lb in first(lbss) do
  begin
    for each ub in first(ubss) do
    begin
      lbc     := first(lb);
      lbEqs   := second(lb);
      ubc     := first(ub);
      ubEqs   := second(ub);
      bc := rlgsn(lbc and ubc, theo);
      if rlqe(rlex(ctx and bc), theo) neq false then
      begin
        if rlqe(rlall(ctx impl bc), theo) eq true then bc := true;
        ctx2    := rlsimpl(ctx and bc, theo);
        res := {bc,map(rhs,lbEqs),map(rhs,ubEqs),
                  makeTree(theo,ctx2,rest(lbss),rest(ubss))
               } . res;
      end;
    end;
  end;
%dbg({exit,length(lbss)});
  return res;
end;

procedure upperUnbounded(sys,vars,ctxvars,theo);
begin scalar ub;
  ub := unbounded(sys,vars,theo,nil);
  ub := rlqe(ex(ctxvars, ub), theo);
  return (if ub eq false then {} else {{ub,{}}});
end;

procedure lowerUnbounded(sys,vars,ctxvars,theo);
begin scalar ub;
  ub := unbounded(sys,vars,theo,t);
  ub := rlqe(ex(ctxvars, ub), theo);
  return (if ub eq false then {} else {{ub,{}}});
end;

procedure upperBounds(sys,vars,ctxvars,theo);
begin
  return findCompats(upperBoundsWC(sys,vars,ctxvars,theo),theo,nil);
end;

procedure upperBoundsWC(sys,vars,ctxvars,theo);
begin
  return weakCond(removeSpecializations(max1(sys,vars,theo),ctxvars,theo),nil,sys,vars,ctxvars,theo);
end;

%  ... -> {{cond',{x1=...}}, ...}
procedure lowerBounds(sys,vars,ctxvars,theo);
begin
  return findCompats(lowerBoundsWC(sys,vars,ctxvars,theo),theo,t);
end;

% ... -> {{cond,cond',{x1=...}}, ...}
procedure lowerBoundsWC(sys,vars,ctxvars,theo);
begin
  return weakCond(removeSpecializations(min1(sys,vars,theo),ctxvars,theo),t,sys,vars,ctxvars,theo);
end;


procedure findCompats(lm,theo,isMinimum);
begin scalar compats, cl, c, borders, newcond, cond, m, comp, incomp;
  compats := {};

  for each mm in lm do
  begin
    cond := first(mm);
    m := first(second(mm));
    cl := compats;
    compats := {};
    incomp := true;
    while length(cl) > 0 do
    begin
      c := first(cl);
      cl := rest(cl);
      cond2 := first(c);
      borders := second(c);

      newcond  := rlsimpl(cond and cond2, theo);
      newcond2 := rlsimpl((not cond) and cond2, theo);
      if rlqe(rlex(newcond)) neq false then
        compats := {newcond,m.borders} . compats;
      if rlqe(rlex(newcond2)) neq false then
        compats := {newcond2,borders} . compats;
      incomp := incomp and not(cond2);
    end;
    incomp := rlsimpl(cond and incomp, theo);
    if incomp neq false then
      compats := {incomp,second(mm)} . compats;
  end;
  return compats;
  
end;

procedure weakCond(lm,isMinimum,sys,vars,ctxvars,theo);
begin
  return (map (m => {weakC(rhs(m),isMinimum,sys,vars,ctxvars,theo),
                     {m}},
               everyBoundOnce(lm)));
end;

procedure weakC(bound,isMinimum,sys,vars,ctxvars,theo);
begin scalar outF, cond;
  outF := if isMinimum
	     then goodFormula2(first(vars)-bound,lessp)
             else goodFormula2(first(vars)-bound,greaterp);
  cond := rlqe(all(ctxvars, not(ex(vars, sys and first(outF)))), theo)
          and second(outF);
  return rlgsn(mydecdeg(cond), theo);
end;

procedure removeSpecializations(lm,ctxvars,theo);
begin scalar m, lm2, lm3, ok, res;
  lm2 := {};
  while length(lm) > 0 do
  begin
    m  := first(lm);
    lm := rest(lm);
    ok := t;
    if rlqe(all(ctxvars, not(first(m))), theo) eq true then
      ok := nil
    else if rlqe(all(ctxvars, first(m)), theo) neq true then
    begin
      lm3 := append(lm,lm2);
      while length(lm3) > 0 and ok do
      begin
        mm2 := first(lm3);
        lm3 := rest(lm3);
        if rlqe(all(ctxvars, first(m) impl first(mm2)),theo) eq true then
	  ok:=nil;
      end;
    end;
%    write ok;
    if ok then 
      lm2 := m . lm2;
  end;
  return lm2;
end;

% {{..., {x1=a}}, {...,{x1=a}}, {...,{x1=b}}} -> {x1=a, x1=b}
procedure everyBoundOnce(answer);
begin scalar res;
  return map (a=>first(a),frequency(map(first,map(second,answer))));
end;

procedure max1(sys,vars,theo);
begin
  return minmax1(sys,vars,theo,nil);
end;

procedure min1(sys,vars,theo);
begin
  return minmax1(sys,vars,theo,t);
end;

procedure minmax1(sys,vars,theo,isMinimum);
begin scalar v1, vrest, m, sysM, lexcond, res1;
  v1    := first vars;
  vrest := rest vars;
  m     := mkidnew();
  sysM := sub({v1=m}, sys);
  lexcond := if isMinimum then v1<=m else v1>=m;
%on rlqegsd;
  res1 := rlqe(ex(vrest, sys and all(m.vrest,sysM impl lexcond)), theo);
%off rlqegsd;
  res1 := mydecdeg(res1);
  return rlqeaTS(ex(v1,res1), theo);
end;

% hier kommen ziemlich viele epsilons und infinities heraus
procedure minmax1lin(sys,vars,ctxvars,theo,isMinimum);
begin scalar coeffs, linexp, cond, formula, res, res2;
  coeffs := map(v => mkidnew(), 1 . ctxvars);
  linexp := 0;
  for each vc in pair(coeffs, 1 . ctxvars) do
    linexp := linexp + first(vc)*second(vc);
  cond := if isMinimum then first(vars) >= linexp else first(vars) <= linexp;
  formula := all(append(ctxvars,vars), sys impl cond)
             and ex(append(ctxvars,vars), sys and first(vars)=linexp);
  write {coeffs,formula};
  res := rlqeaT(ex(coeffs, rlqe(formula,theo)), theo);
  write res;
  res2 := {};
  for each r in res do
  begin
    if not hasInfOrEps(second(r)) then
      res2 := r . res2;
  end;
  return map (r => {first(r), first(vars) = sub(second(r), linexp)}, res2);
end;


procedure unbounded(sys,vars,theo,isMinimum);
begin scalar m, v1, vrest, ucond;
  m     := mkidnew();
  v1    := first vars;
  vrest := rest  vars;
  ucond := if isMinimum then v1<=m else v1>=m;
  return rlqe(ex(m, all(v1, ucond impl ex(vrest,sys))),theo);
end;

procedure goodFormula2(expre,p);
begin
  return (if arglength(expre) < 2 then
            {p(expre,0), true}
          else if part(expre,0) eq quotient then
            {p(part(expre,2)*part(expre,1),0), part(expre,2)<>0}
          else
            {p(expre,0), true}
         );
end;

procedure goodFormula(expre,p);
begin
  return (if arglength(expre) < 2 then
            p(expre,0)
          else if part(expre,0) eq quotient then
            p(part(expre,2)*part(expre,1),0)
          else
            p(expre,0)
         );
end;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%procedure isPolyhedron(sys,vars);
%begin
%  x1 := mkidnew();
%  x2 := mkidnew();%
%end;



procedure checkS(sys,vars,theo);
begin
  return equivalent(sys,vars,theo,sortIneqs(sys,vars,theo));
end;

procedure equivalent(sys,vars,theo,sorted);
begin scalar ctx;
  ctx := for each th in theo mkand th;
  return rlqe(rlall(ctx impl (sys equiv sortedToFormula(sorted,vars))),theo);
end;

procedure sortedToFormula(sorted,vars);
begin scalar lbs, ubs, p, formula, subf;
  formula := false;

  for each p in sorted do
  begin
%    write p;
     lbs := for each lb in second(p) mkand goodFormula(first(vars)-lb,geq);
     ubs := for each ub in third(p) mkand goodFormula(first(vars)-ub,leq);
     subf := if part(p,4) eq SOL_OK
               then true
               else sortedToFormula(part(p,4), rest(vars));
      formula := formula or (first(p) and lbs and ubs and subf);
  end;
  return formula;
end;


procedure rlqeaT(f,theo);
begin scalar res, theoC;
  res := {};
  theoC := for each f in theo mkand f;
  for each r in rlqea(f, theo) do
  begin
    if rlqe(rlex(first(r) and theoC)) eq true then
      res := r . res;
  end;
  return res;
end;

procedure rlqeaTS(f,theo);
begin 
  return filterOutSpec(rlqeaT(f, theo));
end;

procedure filterOutSpec(xs);
begin
  return (if length(xs) = 0
             then {}
             else filterOutSpec1(first(xs), filterOutSpec(rest(xs)))
         );
end;

procedure filterOutSpec1(x,xs);
begin scalar h;
  return
  (if length(xs) = 0 then {x}
   else
   << h := first(xs);
      if rlqe(rlall(first(h) impl first(x))) eq true then filterOutSpec1(x,rest(xs))
      else if rlqe(rlall(first(x) impl first(h))) eq true then filterOutSpec1(h,rest(xs))
      else (h . filterOutSpec1(x,rest(xs)))
   >>)
end;





%%%%% convex union

procedure convexUnion(systems, vars, ctx);
begin scalar f;
  f := convexUnionFormula(systems, vars);
  f := rlgsn(mydecdeg(rlqe(f, ctx)));
  return sortIneqs(f, vars, ctx);
end;

procedure convexUnionFormula(systems, vars);
begin scalar rsAndSubs, hvars, eqs, ineqs, oneCond, geCond, formula;
  rsAndSubs := map (s => {mkidnew(), map(v=>v=mkidnew(),vars), s}, systems);
  hvars := append( map (first,rsAndSubs),
                   for each ras in rsAndSubs join map(rhs,second(ras))
                 );

  eqs := for each v in vars mkand
	   (v = for each ras in rsAndSubs sum
	           (first(ras) * sub(second(ras), v))
	   );

  ineqs := for each ras in rsAndSubs mkand
	     sub(second(ras), third(ras));

  oneCond := (1 = (for each ras in rsAndSubs  sum  first(ras)));

  geZeroCond := for each ras in rsAndSubs  mkand  (first(ras) >= 0);

  formula := ineqs and eqs and geZeroCond and oneCond;

  return ex(hvars, formula);
end;

procedure convexUnion2(sys1, sys2, vars);
begin scalar r, subs1, subs2, vars1, vars2, eqs, formula;
  r := mkidnew();
  subs1 := map (v => v=mkidnew(), vars);
  subs2 := map (v => v=mkidnew(), vars);
  sys1 := sub(subs1, sys1);
  sys2 := sub(subs2, sys2);
  vars1 := map (rhs,subs1);
  vars2 := map (rhs,subs2);

  eqs := for each vvv in pair(vars,pair(vars1,vars2)) mkand
            first(vvv) = first(second(vvv)) * r + second(second(vvv)) * (1-r);

  formula := ex(r, 0<=r and r<=1 and
                   ex(append(vars1,vars2), sys1 and sys2 and eqs));

  return formula;
end;



% a > b   -->     a >= b+1
% a < b   -->   a+1 <= b
procedure makeIneqWeak(ineq);
begin scalar l, r, oper;
  oper := part(ineq, 0);
  r := rhs(ineq);
  l := lhs(ineq);

  if oper eq greaterp then
    ineq := l >= r+1
  else if oper eq lessp then
    ineq := l+1 <= r;

  return ineq;
end;

%% 'forumla' is a quantifier free formula 
% procedure makeAllIneqsWeak(formula);
% begin scalar p0;
%   if arglength(formula) >= 0 then
%   begin
%     p0 := part(formula, 0);
%     if p0 eq or then
%       formula := for p:=1:arglength(formula) mkor
% 	            makeAllIneqsWeak(part(formula,p))
%     else if p0 eq and then
%       formula := for p:=1:arglength(formula) mkand
% 	            makeAllIneqsWeak(part(formula,p))
%     else if p0 eq impl then
%       formula := makeAllIneqsWeak(part(formula,1)) impl
%                  makeAllIneqsWeak(part(formula,2))
%     else if p0 eq repl then
%       formula := makeAllIneqsWeak(part(formula,1)) repl
%                  makeAllIneqsWeak(part(formula,2))
%     else if p0 eq equiv then
%       formula := makeAllIneqsWeak(part(formula,1)) equiv
%                  makeAllIneqsWeak(part(formula,2))
%     else if p0 eq not then
%       formula := not(makeAllIneqsWeak(part(formula,1)))
%     else
%       formula := makeIneqWeak(formula);
%   end;
% 
%   return formula;
% end;

procedure findDisjuncts(formula,ctx);
begin scalar fs, res;
  return filter(f=>rlqe(rlex(ctx and f)), extract(rldnf(formula),or));
end;

procedure extract(formula,constr);
begin
  return (if formula eq false then
            {}
          else if formula eq true then
            {true}
          else if part(formula, 0) eq constr then
            for p:=1:arglength(formula) join extract(part(formula,p),constr)
          else
            {formula}
         );
end;

% procedure disjointUnionW(formula,ctx);
% begin
%   return map(makeAllIneqsWeak, disjointUnion(formula,ctx));
% end;

%
% Given a conjunction of atomic formulas,
% return a list of formulas such that the disjunction
% of the formulas in the result is equivalent to the given
% formula and no result formula has "<>" as relation
% (it is replaced by a disjunction of ">0" and "<0").
%
procedure replaceNeq(formula);
begin scalar al, p0, f2;
  al := arglength(formula);
  % Note: arglength(true) = -1
  if al leq 0 then
    return { formula };

    p0 := part(formula,0);
    return
      (if p0 eq and then
         <<
            % rlsimpl to remove the initial  "true and ..."
            % created by "mkand"
            f2 := replaceNeq(rlsimpl(for i := 2:al mkand part(formula,i)));
            for each a in replaceNeq(part(formula,1)) join
              map(f => a and f, f2)
         >>
       else if p0 eq neq then
         {part(formula,1) < 0, part(formula,1) > 0}
       else
         { formula }
      );

end;

%
% Compute a disjoint union.
% The result does not contain "<>" as relation operator.
%
procedure disjointUnion(formula,ctx);
begin scalar disj, ddisj, anti, res, res1;
  disj := findDisjuncts(formula,ctx);
  if length(disj) = 1 then
    return  for each d in map(replaceNeq,disj) join d;

  ddisj := {};
  anti := true;
  for each d in disj do
  begin
    ddisj := (d and anti) . ddisj;
    anti := anti and not(d);
  end;

  res1 := for each l in ddisj join disjointUnion(l,ctx);

  res := for each c in res1 join replaceNeq(c);

  return res;
end;


%%%%% Lexicographic minimum

procedure lexle(xs,ys);
begin
  return (if length(xs) = 1 then
            first(xs) <= first(ys)
          else
            first(xs) < first(ys) or
            (first(xs) = first(ys) and lexle(rest(xs),rest(ys)))
         );
end;

procedure lexmin(sys,vars,theo);
begin scalar ms, subs, s1, res;
  ms := for i:=1:length(vars) collect mkidnew();
  subs := map(mv => first(mv)=second(mv), pair(vars,ms));
  on rlqegsd;
  s1 := rlqe( rlsimpl (all(ms,sub(subs,sys) impl lexle(vars,ms))));
  off rlqegsd;
  s1 := rlqeaTS(ex(vars, sys and s1), theo);

  res := {};
  for each s in s1 do
  begin
    res := {first(s), map(v => sub(second(s), v), vars)} . res;
  end;
  return res;
%  return map(s => {first(s), map(v => sub(second(s), v), vars)}, s1);
end;



procedure hasInfOrEps(x);
begin scalar l, res, p;
  if symbolic idp(x) then
    res := symbolic isInfOrEps(x)
  else
  begin
    l   := arglength(x);
    res := nil;
    for i := 1 : l do
    begin
      p := hasInfOrEps(part(x,i));
      res := symbolic(res or p);
    end;
  end;
  return res;
end;

symbolic procedure isInfOrEps x;
 isPrefixOf('infinity, x) or isPrefixOf('epsilon, x);

symbolic procedure isPrefixOf(x,y);
 isPrefixOfL(explode x, explode y);

symbolic procedure isPrefixOfL(x,y);
 if null x
    then t
    else if null y
            then nil
            else (car x = car y) and isPrefixOfL(cdr x, cdr y);


%%%%% Helper functions

procedure mydecdeg(formula);
begin scalar res, hd, fs, a, b;
  if arglength(formula) > 1 then
  begin
    hd := part(formula,0);
    if (hd eq and) or (hd eq or) or (hd eq impl) or (hd eq repl) or (hd eq equiv) then
      res := hd(map(mydecdeg,formula))
    else if (hd eq all) or (hd eq ex) then
      res := hd(part(formula,1), mydecdeg(part(formula,2)))
    else if (hd eq lessp) or (hd eq greaterp) or (hd eq leq) or (hd eq geq) or (hd eq equal) or (hd eq neq) then
    begin
      fs := factorize(part(formula,1));
      if length(fs) > 1 then
      begin
        fs := map(f => (first(f) ^ second(f)), fs);
        a  := first(fs);
        b  := for each f in rest(fs) product f;
        res := if (hd eq leq) or (hd eq geq) then
                 (hd(a,0) and mydecdeg(geq(b,0))) or (hd(0,a) and mydecdeg(geq(0,b)))
               else if (hd eq equal) then
                 (hd(a,0) or mydecdeg(equal(b,0)))
               else if (hd eq neq) then
                 (hd(a,0) and mydecdeg(neq(b,0)))
               else 
                 (hd(a,0) and mydecdeg(greaterp(b,0))) or (hd(0,a) and mydecdeg(greaterp(0,b)));
      end
      else
	res := formula;
    end
    else
      res := formula;
  end
  else
    res := formula;

  return res;
end$


procedure filter(predic,xs);
begin
  return (if length xs = 0 then
             {}
          else if first(map(predic,{first(xs)})) = true then
	      first(xs) . filter(predic, rest(xs))
          else
             filter(predic, rest(xs))
        );
end;


;END;


% rlatab(((p - 10 <= 0 and p >= 0) impl (p - x1 = 0 and x1 = 0)) and (p <= 0 impl (p - x1 = 0 and x1 = 0)));
% ***** Bus Error in ofsf_smcpknowl

%load_package redlog;
%rlset ofsf;
%rlatab((a>=0 and a<=10 impl a=b) and a<=0);
%***** Bus Error in ofsf_smcpknowl
