module LMath.Types.Misc
    ( unfracIneqSystem,
      unfracEqSystem
    )
where

import Data.Ratio

import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.AffineExp
import LMath.Types.IneqSystem
import LMath.Types.EqSystem
import LMath.Types.Polynomial
import LMath.Types.Signum

-- |
-- Convert an inequality system with fractional coefficients to an
-- equivalent system with polynomial coefficients.
-- If no function to decide the sign of polynomials is given,
-- the inequalities are multiplied by the square of
-- the denominators.
--
unfracIneqSystem :: (Real a, Fractional a) =>
		    (Maybe (Polynomial a -> Signum)) ->
		    GIneqSystem (FracPoly a) ->
		    GIneqSystem (Polynomial a)
unfracIneqSystem decideSgn sys =
    fromAffineExp $ unfracAffineExp decideSgn $ toAffineExp sys

unfracEqSystem :: (Real a, Fractional a)
	       => GEqSystem (FracPoly a)
	       -> GEqSystem (Polynomial a)
unfracEqSystem eqs = fromAffineExp $ unfracAffineExp Nothing $ toAffineExp eqs

unfracAffineExp :: (Real a, Fractional a) =>
		   (Maybe (Polynomial a -> Signum)) ->
		   GAffineExp (FracPoly a) ->
		   GAffineExp (Polynomial a)
unfracAffineExp mbImpliesSgn expr =
    case unjoin expr of
      [] -> affineExp ((nV,nP), zeroMatrix 0 (nV+nP+1)) 
      es -> multijoin1 $ map unfrac es
    where
    nV = nVars expr
    nP = nParams expr
    unfrac e = fmap (\f -> polyNumerator f * (fact `quotPoly` polyDenominator f)) e
	where
	nums   = map polyNumerator   (entries e)
	denoms = map polyDenominator (entries e)
	fact'  = foldl lcmPoly 1 denoms
	fact =
	    case mbImpliesSgn of
	      Just impliesSgn
		  | signumGe sgn -> fact'
		  | signumLe sgn -> - fact'
		  | otherwise    -> fact' * fact'
		  where
		  sgn = impliesSgn fact'
	      Nothing  ->  fact'

