-- |
-- Data structures and functions for multivarite polynomials.
-- Special functions for univariate polynomials are defined in
-- "LMath.Types.Polynomial.Univariate". Root isolation for univariate real
-- polynomials can be found in "LMath.Types.Polynomial.Roots", algebraic
-- number arithmetic in "LMath.Types.Polynomial.AlgNum", and
-- "LMath.Types.Polynomial.ModArith" provides arithmetic in
-- residue class fields.
--
module LMath.Types.Polynomial
  ( -- * Polynomial Data Type and Fundamental Functions
    Polynomial,
    indeterminate, indeterminateNI,
    constPoly, constPolyN, constPolyNI, constPolyNp,
    isZeroPoly, isConstPoly,
    nIndeterminates, nIndeterminatesN, ensureNIndets, ensureEndNIndets,
    deleteIndeterminates, deleteEndIndeterminates, swapIndeterminates,
    leadingCoeff, constantCoeff, collectCoeffs, totalDegree, degree, degreeIn,
    liftPoly, liftPolyN, appendIndeterminates, insertIndeterminates,
    reverseIndeterminates,
    usedIndeterminates, levelPoly, invlevelPoly,
    coefficients, linearCoeffs,
    splitPoly, mapPoly,
    monic, signMonic, constPolySignum,
    showU,

    -- ** Number of Indeterminates
    NIndeterminates(..),
    unNIndeterminates, addNIndeterminates, addToNIndeterminates, canExtendToNI,

    -- ** Some more functions
    renamePolyVariable, polyToList,
    createPolyFromList, indeterminateToFront,
    affinePolyFromList,

    -- ** Polynomial Arithmetics
    negPoly, addPoly, multPoly, quotPoly, remPoly, quotRemPoly, potPoly, pmul,
    gcdPoly, lcmPoly, euclid, extEuclid,
    content, derivative,    
    reductum, reducta,
    sylvester, resultant, subresultants,
    determinantP, gcdFree,
	    
    -- ** Evaluating and Substituting into Polynomials
    evalPoly, replacePoly, replacePoly1, reducePoly,
    
    -- ** Fractions of Polynomials
    FracPoly, (%%), fromPoly, polyNumerator, polyDenominator,
    mapFP, applyFP, intFracPoly, isConstFP, evalFP,

    -- ** Conversion Functions
    ratPoly2intPoly, univariate, flatPoly, affineToPoly,
    module LMath.Types.Polynomial.Conversion,

    -- * Functions for Univariate Polynomials, Root Isolation, and
    --   Algebraic Number Arithmetic
    module LMath.Types.Polynomial.Univariate,
--    module LMath.Types.Polynomial.Roots,
--    module LMath.Types.Polynomial.AlgNum,
--    module LMath.Types.Polynomial.ModArith,

    -- * Fractions of Polynomials with Integral Coefficients (EXPERIMENTAL)
    IFracPoly, (%%%), ifpNumerator, ifpDenominator,
    fracPolyToIFracPoly, ifracPolyToFracPoly, flatIFracPoly,
    applyIFP, mapIFP, nIndeterminatesIFP,
    isConstIFP,

    -- * Fractions with Polynomials and Ceil\/Floor Operators (EXPERIMENTAL)
    module LMath.Types.Polynomial.FracPolyFC,

    -- * Show Functions
    module LMath.Types.Polynomial.Show
  )
where

import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Conversion
import LMath.Types.Polynomial.Univariate
-- import LMath.Types.Polynomial.Roots
-- import LMath.Types.Polynomial.ModArith
-- import LMath.Types.Polynomial.AlgNum
import LMath.Types.Polynomial.Show
import LMath.Types.Polynomial.FracPolyFC
import LMath.Types.Polynomial.Bareiss
import LMath.Types.Polynomial.Resultant
