module LMath.Types.Polynomial.Bareiss
    ( determinantP, bareiss )
where

import Control.Monad.ST
import Data.Array.ST
import Data.List

import LMath.Types.Matrix
import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Univariate

-- |
-- Calculate the determinant of a matrix.
-- The determinant of a matrix with (multivariate) polynomials
-- as entries is calculated using Bareiss\' method.
--
determinantP :: Fractional a =>	GMatrix (Polynomial a) -> Polynomial a
determinantP = bareiss quotPoly


bareiss :: Num a => (a -> a -> a) -> GMatrix a -> a
bareiss quotFn (matr :: GMatrix entries)
    | n == 0     =  1
    | otherwise  =  runST (mkut matr)
    where
    mkut :: GMatrix entries -> ST s entries
    mkut m =
	do a <- newArray ((0,0),(n-1,n-1)) undefined :: ST s (STArray s (Int,Int) b)
	   sequence_ [writeArray a (r,c) (entry m r c) | (r,c) <- indexes m]
	   ut a 0

    ut a k
	| k >= n    =  readArray a (n-1,n-1)
	| otherwise  =
--	    trace (show k ++ " von " ++ show n) $
	    sequence [readArray a (r,k) | r <- [k..n-1]] >>= \cs ->
	    case findIndex (not.(==0)) cs of
	        Nothing -> return 0
		Just i  -> do exch k (k + fromIntegral i)
			      kk  <- readArray a (k,k)
			      kk2 <- (if k > 0
				       then readArray a (k-1,k-1)
				       else return 1)
			      sequence_ [bar kk kk2 i j
					| i <- [k+1..n-1], j <- [k+1..n-1]]
			      ut a (k+1)
	where
	bar kk kk2 i j =
	    do ij <- readArray a (i,j)
	       ik <- readArray a (i,k)
	       kj <- readArray a (k,j)
	       writeArray a (i,j) ((kk*ij - ik*kj) `quotFn` kk2)

	exch r1 r2 
	    | r1 /= r2  =
		sequence_ [readArray  a (r1,c) >>= \v1 ->
			   readArray  a (r2,c) >>= \v2 ->
			   writeArray a (r2,c) v1 >>
			   writeArray a (r1,c) v2 
			  | c <- [0..n-1]]
	    | otherwise = return ()

    n = nRows matr
