{-# OPTIONS -fwarn-incomplete-patterns #-}
-- |
-- This module is very experimental.
-- The interface is subject to change without warning.
--
module LMath.Types.Polynomial.FracPolyFC
    ( -- * Data type and standard functions
      IFracPolyFC,
      mkFC, mkFCi, mkFloorFC, mkFloorFCi,
      floorFC, ceilFC,
      isConstFC, isIntegralFC,

      -- * Esoteric functions (you should know what you are doing when you use them)
      pseudoExpression, unifyPseudos, isFloorMultipleFC
    )
where

import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Show
import LMath.Types.Signum

import Data.List
import Data.Ratio
import Data.Maybe
import System.IO.Unsafe
import System.Mem.StableName

import Debug.Trace



type StableFC a = StableName (FEMap a)

type FEMap a = [IFracPoly (Polynomial a)]
type BackOffset = Int

data Integral a => IFracPolyFC a = FC !(StableFC a) (FEMap a) (IFracPoly a)


makeFC :: Integral a => FEMap a -> IFracPoly a -> IFracPolyFC a
makeFC feMap fp
--    | length feMap == 2 && nIndeterminatesIFP fp == Exactly 3  =
--	error ("makeFC: " ++ show (feMap, fp))
    | otherwise  = FC (unsafePerformIO $ makeStableName feMap) feMap fp

instance (Show a,Integral a) => Show (IFracPolyFC a) where
    showsPrec prec (FC _ fe fp) =
	showParen (prec>=10) (showString "FC [" .
			      foldl (.) id (intersperse (showString ", ") $ map showIFP fe) .
			      showString "] " . showsPrec 11 fp)
	where
	showIFP ifp
	    | den == 1  =  sp num
	    | otherwise =  sp num . showString "/" . sp den
	    where
	    sp p = showString (showPolyExt ['u' : show i | i <- [0..]] 
			              (showPolyStd ['r' : show i | i <- [0..]])
			              ((<0).leadingCoeff)
			              p
			      )
	    num = ifpNumerator   ifp
	    den = ifpDenominator ifp


liftFEMap :: Integral a => FEMap a -> FEMap a
liftFEMap = map (applyIFP liftPoly)

insertFL :: Integral a
         => ((IFracPoly (Polynomial a)), FEMap a)
         -> FEMap a
         -> (FEMap a, BackOffset)
insertFL (ifpp,fe1) oldFE =
    case findIndex (==(newNum%%%newDen)) newFE2 of
        Just idx -> (newFE2, length newFE2 - idx)
	Nothing  -> (liftFEMap ((newNum%%%newDen) : newFE2), length newFE2 + 1)
    where
    num = ifpNumerator   ifpp
    den = ifpDenominator ifpp
    usedNum = usedIndeterminates num
    usedDen = usedIndeterminates den
    numFLs = map (fe1!!) usedNum
    denFLs = map (fe1!!) usedDen

    (newFE1, newNumFLsOff) = mapAccumL (\oFE fl -> insertFL (fl,fe1) oFE) oldFE numFLs
    (newFE2, newDenFLsOff) = mapAccumL (\oFE fl -> insertFL (fl,fe1) oFE) newFE1 denFLs

    newNum = replacePoly1 (\es -> product [(indeterminate (length newFE2) (length newFE2 - off))^(es!!u) | u <- usedNum | off <- newNumFLsOff]) num
    newDen = replacePoly1 (\es -> product [(indeterminate (length newFE2) (length newFE2 - off))^(es!!u) | u <- usedDen | off <- newDenFLsOff]) den

    mkRat  = mapPoly (mapFP (% 1))
    map2 f (a,b) = (f a, f b)

unifyFEs :: Integral a => FEMap a -> FEMap a -> (FEMap a, [BackOffset])
unifyFEs fe1 fe2 = mapAccumR (\oFE nd -> insertFL (nd,fe2) oFE) fe1 fe2


-- |
-- Unify the \"pseudos\" (indeterminates introduced to refer to floor
-- expressions) in the given 'IFracPolyFC' expression. I.e.,
-- compute a mapping @f@ such that when @f@ is applied to all the
-- given 'IFracPolyFC' expressions, the pseudos are the same in
-- all the resulting 'IFracPolyFC's.
--
unifyPseudos :: Integral a
             => [IFracPolyFC a] -> ([IFracPoly a], IFracPolyFC a -> IFracPolyFC a)
unifyPseudos []  = ([],id)
unifyPseudos fcs =
    -- trace (show fcs) $
    (denoms, unifyP)
    where
    denoms = map (\fe -> genPseudoExpression newFE
		               ((flatPoly $ ensureEndNIndets newFElength $ ifpDenominator fe) %%% 1))
	         newFE

    unifyP (FC _ fe fp) =
	case findIndex (eqFE fe) allFEs of
	    Nothing -> error "unifyPseudos: result function applied to \
			     \unseen IFracPolyFC"
            Just idx ->
--		trace (show (nIndeterminates $ ifpNumerator fp, nIndeterminates $ ifpDenominator fp, nIndeterminates $ repl num, nIndeterminates $ repl den, FC newSn newFE (repl num %%% repl den))) $
		FC newSn newFE (repl num %%% repl den)
		where
		repl = replacePoly1 (\es -> product [indeterminateNI {-(addToNIndeterminates newFElength nIndet)-} nIndet ind ^ e | e <- es | off <- (backOffs!!idx)++[0,-1 .. ], let ind = newFElength - off])
		nIndet = addToNIndeterminates (newFElength - length fe) $
			 nIndeterminatesIFP fp
		num = ifpNumerator   fp
		den = ifpDenominator fp

    newFElength = length newFE
    backOffs = [length fe1,length fe1-1 .. 1] : backOffs'
    (newFE,backOffs') = mapAccumL unifyFEs fe1 fes
    newSn = unsafePerformIO $ makeStableName newFE
{-
    nIndet = maximum $
	     map (\(FC _ fe fp) -> addToNIndeterminates (negate $ length fe) $
				   max (nIndeterminates $ ifpNumerator fp)
		                       (nIndeterminates $ ifpDenominator fp) )
		                   
	         fcs
-}
    allFEs@(fe1:fes) = nubBy eqFE $ map (\(FC _ fe _) -> fe) fcs

    -- if the length of 2 FEMap's is different, the
    -- number of indeterminates in the IFracPoly's is also different,
    -- so compare them only when the lengths are equal!
    eqFE xs ys = length xs == length ys && xs == ys


-- |
-- Replace every occurence of a variables X
-- which stands for floor(a\/b) by the expression
-- (a-X)\/b, i.e., introduce the \"epsilons\"
-- for the conservative approximation of formulas.
--
pseudoExpression :: Integral a => IFracPolyFC a -> IFracPoly a
pseudoExpression (FC _ femap fp) = genPseudoExpression femap fp

genPseudoExpression :: Integral a => FEMap a -> IFracPoly a -> IFracPoly a
genPseudoExpression (femap :: FEMap b)  fp  =
--    trace (show $ map lookup [4,2..0]) $
--    trace (show (femap, fp, nIndeterminatesIFP fp)) $
    ({-(\x -> trace ("a: " ++ show x) x) $ -} reducePoly (+) mon 0 (ifpNumerator fp)) /
    ({-(\x -> trace ("b: " ++ show x) x) $ -} reducePoly (+) mon 0 (ifpDenominator fp))
    where
    mon :: b -> Maybe [Int] -> IFracPoly b
    mon co (Just es) = (constPoly co %%% 1) * product [lookup i ^ e | e <- es | i <- [0..]]
    mon co Nothing   = constPoly co %%% 1

    l = length femap

    lookup :: Int -> IFracPoly b
    lookup i
	| i >= length femap  =  pseudo %%% 1
	| (Just n, Just d) <- (isConstPoly num, isConstPoly den) =
            (liftPolyN l n - pseudo) %%% liftPolyN l d
	| otherwise = (num' - (pseudo%%%1)) / den'
	where
	repl :: Polynomial (Polynomial b) -> IFracPoly b
	repl = reducePoly (+) mono 0 . flatPoly
	    where
	    mono co (Just es) =
		(constPoly co %%% 1) * product [lookup j ^ e | e <- es | j <- [0..]]
	    mono co Nothing   = constPoly co %%% 1

	num' = repl num
	den' = repl den
	num = ifpNumerator (femap!!i)
	den = ifpDenominator (femap!!i)

        pseudo :: Polynomial b
	pseudo = indeterminateNI (nIndeterminatesIFP fp) i
	   


unFC :: Integral a => IFracPolyFC a -> IFracPoly a
unFC (FC _ _ fp) = fp

-- |
-- Check whether the argument represents a constant
-- term and return the constant if so.
--
isConstFC :: Integral a => IFracPolyFC a -> Maybe (Ratio a)
isConstFC (FC _ _ fp) = fmap (\(a,b) -> a % b) $ isConstIFP fp

-- |
-- Make a 'IFracPolyFC' which represents the same
-- value as the given 'FracPoly'.
--
mkFC :: Integral a => FracPoly (Ratio a) -> IFracPolyFC a
mkFC = mkFCi . fracPolyToIFracPoly

-- |
-- Make a 'IFracPolyFC' which represents the same
-- value as the given 'IFracPoly'.
--
mkFCi :: Integral a => IFracPoly a -> IFracPolyFC a
mkFCi fp = makeFC [] fp

-- |
-- Make a 'IFracPolyFC' which represents the floor of
-- the value of the given 'FracPoly'.
--
mkFloorFC :: Integral a => FracPoly (Ratio a) -> IFracPolyFC a
mkFloorFC = mkFloorFCi . fracPolyToIFracPoly

-- |
-- Make a 'IFracPolyFC' which represents the floor of
-- the value of the given 'IFracPoly'.
--
mkFloorFCi :: Integral a => IFracPoly a -> IFracPolyFC a
mkFloorFCi fp
    | Just (n,d) <- isConstIFP fp  =  mkFCi (constPoly (floor (n % d)) %%% 1)
    | otherwise                    =  makeFC ps (pseudo %%% 1)
    where
    pseudo = indeterminateNI (addToNIndeterminates 1 $ nIndeterminatesIFP fp) 0
    ps = [constPolyN 1 (ifpNumerator fp) %%% constPolyN 1 (ifpDenominator fp)]

-- |
-- Compute the floor of the given 'IFracPolyFC'.
--
floorFC :: Integral a => IFracPolyFC a -> IFracPolyFC a
floorFC fc@(FC sn fe fp)
    | isIntegralFC fc  =  fc
    | otherwise        =  makeFC newFE newFp
    where
    num = splitPoly (length fe) $ ifpNumerator fp
    den = splitPoly (length fe) $ ifpDenominator fp

    newFE  = liftFEMap ((num%%%den) : fe)
    newFp  = indeterminateNI (addToNIndeterminates 1 $ nIndeterminatesIFP fp) 0 %%% 1


-- |
-- Compute the ceil of the given 'IFracPolyFC'.
--
ceilFC :: Integral a => IFracPolyFC a -> IFracPolyFC a
ceilFC = negate . floorFC . negate

arithOp :: Integral a
        => (IFracPoly a -> IFracPoly a -> IFracPoly a)
        -> IFracPolyFC a
	-> IFracPolyFC a
	-> IFracPolyFC a
arithOp f a@(FC sn1 fe1 fp1) b@(FC sn2 fe2 fp2)
    | sn1 == sn2  =
--	trace "fast arithmetic" $
--	(let res = f fp1 fp2 in trace (show ((a, nIndeterminatesIFP fp1, b, nIndeterminatesIFP fp2), fe1, res, nIndeterminatesIFP res))) $
	FC sn1 fe1 $ f fp1 fp2
    | otherwise =
--	trace "slow arithmetic" $
--	trace (show ((a, nIndeterminatesIFP fp1, b, nIndeterminatesIFP fp2), newFE, res, nIndeterminatesIFP res)) $
--	(if q (nIndeterminatesIFP newFp1) (nIndeterminatesIFP newFp2) then trace ("xxx: " ++ show (newFp1,nIndeterminatesIFP newFp1, newFp2, nIndeterminatesIFP newFp2)) else id) $
--	(if length newFE==2 && unNIndeterminates (nIndeterminatesIFP res) == 3 then error ("XXXXXXXXXXXXXXXXXXXXX: " ++ show (newFE,res)) else id) $
	makeFC newFE res
    where
    q (Exactly x) (Exactly y) = x /= y
    q _ _ = False

    res = f newFp1 newFp2

    (newFE,offs) = mapAccumR (\oFE f -> insertFL (f,fe2) oFE) fe1 fe2
    newFp1 = applyIFP (liftPolyN (length newFE-length fe1)) fp1

    nInd2 = length newFE + max (unNIndeterminates (nIndeterminatesIFP fp1) - length fe1) (unNIndeterminates (nIndeterminatesIFP fp2) - length fe2)

    newFp2 = applyIFP (replacePoly1 (\es -> product [indeterminate nInd2 ind ^ e | e <- es | o <- offs ++ [0,-1 .. ], let ind = length newFE - o])) fp2



-- |
-- Check whether the given 'IFracPolyFC' is integral.
--
isIntegralFC :: Integral a => IFracPolyFC a -> Bool
isIntegralFC (FC _ _ fp)  =  ifpDenominator fp == 1


-- |
-- Check whether the given 'IFracPolyFC' represents a multiple of
-- of a power of a floor expression, i.e., it is of the form
-- @a*floor(b)^e@. Return @Just ([(b,e)],a)@ if so and @Nothing@ otherwise.
--
isFloorMultipleFC :: (Integral a, Integral e)
                  => IFracPolyFC a
		  -> Maybe ([(IFracPolyFC a,e)], IFracPoly a)
isFloorMultipleFC (FC sn fe fp) =
    case (num, denom) of
      ([(n,nes)], [(d,des)])
	  | all (==0) des, Just firstNN <- findIndex (/=0) nes, firstNN < length fe,
	    all (==0) $ drop (firstNN+1) nes  ->  -- n*(fe!!firstNN)^(nes!!firstNN)/d
		Just ([(FC sn fe (applyIFP flatPoly (fe!!firstNN)), fromIntegral (nes!!firstNN))], n %%% d)
      _ -> Nothing
    where
    num   = monomials $ splitPoly (length fe) $ ifpNumerator   fp
    denom = monomials $ splitPoly (length fe) $ ifpDenominator fp


instance Integral a => Eq (IFracPolyFC a) where
    a == b  =  x==0
	where
	FC _ _ x = a-b

instance Integral a => Num (IFracPolyFC a) where
    (+) = arithOp (+)
    (*) = arithOp (*)
    negate (FC sn fe fp) = FC sn fe $ negate fp
    fromInteger = makeFC [] . fromInteger

    signum = error "signum: undefined for Integral a => Num (IFracPolyFC a)"
    abs    = error "abs: undefined for Integral a => Num (IFracPolyFC a)"

instance Integral a => Fractional (IFracPolyFC a) where
    (/) = arithOp (/)
    recip (FC sn fe fp) = FC sn fe $ recip fp
    fromRational = makeFC [] . fromRational


{-
nPseudosFC :: Integral b => FracPolyFC a -> b
nPseudosFC (FC _ ps _) = genericLength ps

-}
