module LMath.Types.Polynomial.Resultant
    ( resultant, subresultants, sylvester,
      resultant2
    )
where

import LMath.Types.LClasses
import LMath.Types.Matrix
import LMath.Types.Polynomial.Base
import LMath.Types.Polynomial.Bareiss
import LMath.Types.Polynomial.Univariate

import Debug.Trace

mytrace a b = b

-- |
-- Compute the resultant of two polynomials.
--
resultant :: Fractional a => Polynomial a -> Polynomial a -> Polynomial a
resultant a b = -- trace (show (degree a, degree b)) $
                determinantP $ sylvester b a

resultant2 :: RealFrac a => Polynomial a -> Polynomial a -> Polynomial a
resultant2 f g
    | any (==Nothing) $ entries syl'  =
        trace "non-const case" $
        determinantP syl
    | otherwise  =
        trace "const case" $        
        constPoly $ fromRational $ determinant syl''
    where
    syl   = sylvester f g
    syl'  = fmap isConstPoly syl
    syl'' = fmap (\(Just r) -> toRational r) syl'

sylvester :: Fractional a =>
	     Polynomial a -> Polynomial a -> GMatrix (Polynomial a)
sylvester a b = mtranspose syl
    where
    m = degree a
    n = degree b
    coeffsA = coefficients a
    coeffsB = coefficients b

    syl =
	matrix (m+n,m+n,
		concat (replicate (n-1) (coeffsA ++ replicate n zero)) ++
		(if n >= 1 then coeffsA else []) ++
		concat (replicate (m-1) (coeffsB ++ replicate m zero)) ++
		(if m >= 1 then coeffsB else []))
    zero = 0 --ZeroPolynomial (nIndeterminates a - 1)

-- |
-- Compute the subresultants of two polynomials.
-- The i-th element in the list returned by @subresultants p q@
-- is the i-th scalar subresultant of @p@ and @q@.
--
subresultants :: Fractional a => Polynomial a -> Polynomial a -> [Polynomial a]
subresultants a b =
    map sub [0 .. fromIntegral (min m n) - 1] ++ [lc ^ abs (m-n) | n>=0, m>=0]
    where
    sub k = mytrace (show $ sylvester k) $ determinantP (sylvester k)
	    --foldl1 (*) [entry ut i i | i <- [0..fromIntegral(m+n-1)-2*k]]
	    --`quotPoly` multiplier
	--where
	--(multiplier,ut) = utGauss (trace (show $ sylvester k) $ sylvester k)
    lc = head (if m < n then coeffsA else coeffsB)
    m = degree a
    n = degree b
    coeffsA = reverse $ coefficients a
    coeffsB = reverse $ coefficients b

    sylvester k =
	subMatrix 
	(mtranspose $
	 matrix (n,m+n,
		concat (replicate (n-1) (coeffsA ++ replicate n zero)) ++
		(if n >= 1 then coeffsA else [])))
	0 0 (fromIntegral (m+n)-2*k-1) (fromIntegral n-k-1)
	`conc`
	subMatrix
	(mtranspose $
	 matrix (m,m+n,
		concat (replicate (m-1) (coeffsB ++ replicate m zero)) ++
		(if m >= 1 then coeffsB else [])))
	0 0 (fromIntegral (m+n)-2*k-1) (fromIntegral m-k-1)
    zero = 0 -- ZeroPolynomial (nIndeterminates a - 1)
