module LoopoDS.Groups
    ( readGroups, writeGroups
    )

where

import Data.List

import LP

-- Read a ".grp" file from disk.
-- Returns the number of statements and a list of groups (which are
-- represented as lists of statement numbers).
readGroups :: FilePath -> IO (LPResult (Integer, [[Integer]]))
readGroups filename =
    do contents <- readFile filename
       return $ runLP parseGroups contents

parseGroups :: Parser (Integer, [[Integer]])
parseGroups =
    do nGroups <- parseInteger
       nStmts  <- parseInteger
       groups <- parseN nGroups (parseInteger >>= \n -> parseIntegers n)
       return (nStmts, groups)

writeGroups :: (Integer, [[Integer]]) -> FilePath -> IO ()
writeGroups grps filename = writeFile filename (groupsToString grps)

groupsToString :: (Integer, [[Integer]]) -> String
groupsToString (nStmts, groups) = (show . length) nonEmptyGroups
    ++ "\n" ++ show nStmts ++ "\n"
    ++ (concat . map groupToString) nonEmptyGroups where
        nonEmptyGroups = filter (\x -> length x > 0) groups

groupToString :: [Integer] -> String
groupToString group = (show . length) group ++ "\n" ++ (concat . intersperse " " . map show) group ++ "\n"
