module LoopoDS.Placement
    ( Placement(..)
    )
where

-- |
-- PRELIMINARY placement data type.
--
--newtype Ord a => Placement a = Placement (FiniteMap a Matrix)

data Placement a = Placement (FiniteMap a (Quast Rational))

parseSchedOrStmtPlm :: HasIntegralID a => Parse (Placement a)
parseSchedOrStmtPlm =
    do nStmts <- parseInteger
       nV     <- parseInteger
       nP     <- parseInteger
       pl <- many $ do stmt  <- parseInteger
		       quast <- parseCommentQuast
		       char '#'
		       spaces
		       return (integerToID stmt,quast)
       return $ listToFM pl


stmtPlacementFmt :: Integer -> Integer -> Placement StatementID -> String
stmtPlacementFmt nLoopIndices nP (Placement alloc) =
    unlines ([show (length stmtIDs), show nLoopIndices, show nP] ++ concatMap fmt stmtIDs)
    where
    fmt stmtID = [show (integerFromID stmtID),
		  placementFmt alloc stmtID ("FCO placement for " ++ show stmtID),
		  "#"
		 ]
    stmtIDs = keysFM alloc

dataPlacementFmt :: Integer
		 -> Integer
		 -> Integer
		 -> Placement ArrayID
		 -> String
dataPlacementFmt nArraysTotal nLoopIndices nP (Placement alloc) =
    unlines ([show nArraysTotal, show (length arrayIDs), show nLoopIndices, show nP] ++
	     concatMap fmt arrayIDs
	    )
    where
    fmt aID = [show (integerFromID aID),
	       placementFmt alloc aID ("FCO placement for " ++ show aID),
	       "#"
	      ]
    arrayIDs = keysFM alloc    

placementFmt :: (Ord a, Show a) => FiniteMap a Matrix -> a -> String -> String
placementFmt alloc pid comment =
    "( ( " ++ comment ++ " )\n" ++ allocFmt (lookupWithDefaultFM alloc errmsg pid)
    ++ "\n)"
    where
    errmsg = error ("placementFmt: looking up " ++ show pid ++ " faild (not in placement)")


allocFmt :: Matrix -> String
allocFmt m = "( list " ++ concat (intersperse " " (map vectorFmt $ rowVectors m)) ++ " )"

vectorFmt :: Vector -> String
vectorFmt v = "#[" ++ concat (intersperse " " (map ratFmt $ vectorEntries v)) ++ "]"

ratFmt :: Rational -> String
ratFmt r
    | denominator r == 1  =  show (numerator r)
    | otherwise           =  show (numerator r) ++ "/" ++ show (denominator r)

