module LoopoDS.Qual
    ( QualEntry(..),
      readQUAL
    )
where

import LoopoDS.IntegralID
import LoopoDS.Statement
import ParseBase

import Text.ParserCombinators.Parsec

data QualEntry = Sched | Place | Additional
	       deriving (Show, Eq)


readQUAL :: FilePath -> IO (LPResult [(TStatementID, [QualEntry])])
readQUAL filename = loadAndParse filename parseQUALs
    where
    parseQUALs :: Parser [(TStatementID, [QualEntry])]
    parseQUALs = 
	do nStmt <- parseInteger
	   parseN nStmt parseQualVect

    readQualEntry :: Char -> QualEntry
    readQualEntry 's' = Sched
    readQualEntry 'p' = Place
    readQualEntry 'a' = Additional

    parseQualVect :: Parser (TStatementID, [QualEntry])
    parseQualVect =
	do stmtNr <- parseInteger
	   qualLength <- parseInteger
	   qualVect <-parseN qualLength parseQualEntry
	   return (integerToID stmtNr, qualVect)
			
    parseQualEntry :: Parser QualEntry
    parseQualEntry = 
	do 
	   qualChar <- parseChar "spa"
	   spaces
	   return $ readQualEntry qualChar
	   
