module MultiDim where 

import OneDim
import InputLoopo 
import Input4PL 

import LoopoDS
import Loopo
import List 

import IOExts
import LMath

import Data.FiniteMap
import Maybe
import Char
import Ratio


-- 
-- multi dimensional part
--

-- auxiliary function 
-- calculates the number of structure parameters
-- (in multidim1 replaced by nParams) 

numSPar :: [(Integer, Matrix)]->[(Integer,[Integer])]->Integer
numSPar ixsp slp =(total - vars)
    where 
    total = maxList [(nColumns x) | (_,x) <-ixsp] -- width ispc
    vars  = maxList [sum xs |(_,xs)<- slp ]       -- number of 
    maxList :: [Integer] -> Integer                -- loop vars
    maxList (x:[]) = x
    maxList (x:xs) 
	| x >= head(xs) = maxList (x:tail(xs))
	| otherwise = maxList xs


p = numSPar indexspace loops 
-- returns pi-placements out of list of ranked placements (fco ...) 
-- result has structure [((widths, pi-part),(complete placement))]
--
getPlcs2::[((Int,Integer),[Integer],[Integer],Placement)]->[(([Integer],[Rational]),[Rational])]
getPlcs2 xs = [((a,x),(x++y))|(_,a,_,(x,y))<-xs]

-- returns the allocations out of the list provided by 
-- getPlcs2, result is ([[allocation]],[complete placement])
--
cutPlcs2 :: [(([Integer],[Rational]),[Rational])]->[([[Rational]],[Rational])]
cutPlcs2 xs = [(cut pl b, rest)| ((b,pl),rest)<-xs]


-- cuts away entries which do not belong to the part of 
-- the loop indeces within the allocation 
--    
variables :: [Rational]-> Integer -> [Rational]
variables allocation paramPart = if (a <=  varPart) 
				 then genericTake varPart allocation
				 else allocation  
    where varPart = ((genericLength allocation)-paramPart)
          a       = genericLength allocation 

dog = ranking bsp 
cat = head (cutPlcs2 (getPlcs2(dog)))
mouse = fst cat 
sweety = mouse!!4


testVariables xs p = [variables x p| x<-xs]  
-- same behavior like "variables", but 
-- in case that nParams is used
--      
variables1 :: [Rational]-> Integer -> [Rational]
variables1 allocation numSPar = genericTake numVar allocation
    where numVar = ((genericLength allocation)-numSPar)-1


-- tests if a matrix is linearly independent
-- (auxilliary function for check)
--
lu :: Matrix -> Bool 
lu mat = if ((length(kernel mat))==((nColumns mat)-(nRows mat))) then True
	 else False 

-- test 
isLI = matrix [[1%1,0%1],[0%1,1%1]]

-- ckecks a list of allocations on l.i., p is the number
-- of structure parameters
-- if the constructed matrix has more rows then columns it
-- true too, as ... 
--
check :: [[[Rational]]] -> Integer -> Bool
check [] p = True 
check (x:plcs) p = if ((nRows m >= length justVar)||(lu m)) then check plcs p
			 else False
			     where 
			     -- just the part of variables is important
			     m = matrix justVar
			     justVar = [variables y p| y <- x]


-- same like "check" but in combination with nParams
--
check1 :: [[[Rational]]] -> Integer -> Bool
check1 [] p = True 
check1 (x:plcs) p = if lu (matrix justVar) then check1 plcs p
		    else False 
			where 
			-- just the part of variables is important
			justVar = [variables1 y p| y <- x]



-- checks if list of placements are l.i. to each other by 
-- using the helper check
-- AGAIN: p is number of structure parameters
totalCheck :: [[[Rational]]]->Integer-> Bool  			    
totalCheck xs p = check prepared p  
		 where
                 -- x'th allocation from pl1 combined with
		 -- x'th allocations of other placements   
		 prepared = transpose xs 

-- same like "totalCheck", but for the use of "nPrarams"
--  
totalCheck1 :: [[[Rational]]]->Integer-> Bool  			    
totalCheck1 xs p = check1 prepared p  
		 where
                 -- x'th allocation from pl1 combined with
		 -- x'th allocations of other placements   
		 prepared = transpose xs 


-- guard to detect allocations whose complete 
-- variable part consists of zeros (zero dimensional)
--  
noZeroDim :: [[Rational]]->Integer->Bool		   
noZeroDim allocs p = or (checkNonZero allocs p)
    where
    checkNonZero :: [[Rational]] -> Integer -> [Bool]
    checkNonZero [] p = [] 
    checkNonZero (x:allocs) p =  ((not(and[(v==0 % 1) |v <-varPart]))||(l<=1)): checkNonZero allocs p 
	where 
	varPart = variables x p
	l       = length varPart


-- generates the candidates for k multidimensions 
--
candidate 0 [] = [[]]
candidate k [] = []
candidate k (x:xs) = [x:ys|ys<- (candidate (k-1) xs)] ++ candidate k xs
  


-- calculates the multidimensional placements,
-- paramters are ranked placements(list), # of dimensions,
-- cleaned indexspace, surrounding loops.   
--
multidim :: [((Int,Integer),[Integer],[Integer],Placement)]-> Int -> [(Integer, Matrix)]->[(Integer,[Integer])] -> [Matrix]

multidim pl d isp l 
    | (d == 0)  = [ rowMatrix r| r <- zeroPlacements] -- zero-dim. 
    | (d == 1)  = res4OneDim                          -- one-dim.
    | otherwise = res4MultiDim                        -- multi-dim.   
    where
    res4OneDim   = [rowMatrix(fromJust(lookupFM table c))|c <- cand1]
    res4MultiDim = [matrix [fromJust(lookupFM table key)|key<-c]|c<-candM]
    cand1 = [x | x <- allocats, noZeroDim x p] -- just allocations, 
                                               -- zero-dim's filtered out
    candM = [x|x<- candidate d allocats , totalCheck x p] -- checked on l.i.
    p = numSPar isp l                                      
    zeroPlacements = [ p | (_,p)<- placements] -- just the placements  
    allocats = [ a | (a,_) <-placements]       -- just the allocations 
    placements = cutPlcs2(getPlcs2 pl) -- ([allocats],[whole placement])
    table  = listToFM placements -- hashtable is filled   

-- IDEE!! :: p direkt, spare isp und l ...

-- test definition
multidim1 pl d disp -- the list with iscp-system, uncleaned !!! 
    | (d == 0)  = [ rowMatrix r| r <- zeroPlacements] 
    | (d == 1)  = res4OneDim 
    | otherwise = res4MultiDim
    where
    res4OneDim   = [rowMatrix(fromJust(lookupFM table c))|c <- cand1]
    res4MultiDim = [matrix [fromJust(lookupFM table key)|key<-c]|c<-candM]
    cand1 = [x | x <- allocats, noZeroDim x p]
    candM = [x|x<- candidate d allocats , totalCheck1 x p]
    p = nParams (snd(head disp))
    zeroPlacements = [ p | (_,p)<- placements] -- the placements  
    allocats = [ a | (a,_) <-placements]       -- just the allacations 
    placements = cutPlcs2(getPlcs2 pl) -- ([allocats],[whole placement])
    table  = listToFM placements -- hashtable is filled   
		   
-- claculates discrete distance / communication distance of 
-- one placement  
-- 
ddistance mdpl ispcCleaned accsCleaned = checkZeroColumns accesses
    where
    accesses = cut(columns'(mtranspose result)) dpartWidths 
    result   = (dpartMatrix * lambdaMatrix)
    lambdaMatrix  = (matrix(drop nonLambdaPart (columns' mdpl)))
    nonLambdaPart = fromIntegral (nColumns mdpl-widthLpart)
    widthLpart= sum [width | (_,_,(_,width)) <- posAndWidthDmatrix]
    -- width information of the transposed and prepared indexspace
    -- matrices
    posAndWidthDmatrix = ispc2position ispcCleaned accsCleaned 
    dpartWidths = fst dPartTuple 
    dpartMatrix = snd dPartTuple 
    dPartTuple  = createDmatrix ispcCleaned accsCleaned
     --checkZeroColumns :: [[Rational]]->[Ratio Integer]
    checkZeroColumns ys = length (snd partitioned)
	where
        partitioned = partition (==0%1) outer
        outer = [sum i | i <- inner]
	inner = [[sum x|x<- xs]|xs<-ys]

-- test

testDDistance mdpls ispcCleaned accsCleaned = list 
    where
    list = [ddistance mdpl ispcCleaned accsCleaned |mdpl<-mdpls]

-- calculates the communication volume just for one placement,
-- result shows how many accesses are not zero 
--
volumeM mdpl ispcC accsC slp = sum  [v |(v,r)<-(fst discreteDistances)]
    where
    discreteDistances = partition (distanceNotZero) numberOfSlps
    distanceNotZero :: (Integer,Rational)-> Bool
    distanceNotZero t = if snd(t)==(0%1) 
			then False
			else True
    numberOfSlps = [(sum(slps s slp), d) | (s,d)<-withStmtNumbers]
    withStmtNumbers = zip stmtNumbers checkedOnZero
    checkedOnZero  = checkZeroColumns accesses 
    stmtNumbers = stmtNrs accsC
    accesses = cut(columns'(mtranspose result)) dpartWidths 
    result   = (dpartMatrix * lambdaMatrix)
    lambdaMatrix  = (matrix(drop nonLambdaPart (columns' mdpl)))
    nonLambdaPart = fromIntegral (nColumns mdpl-widthLpart)
    widthLpart= sum [width | (_,_,(_,width)) <- posAndWidthDmatrix]
    -- width information of d-matrices 
    posAndWidthDmatrix = ispc2position ispcC accsC 
    dpartTuple =  createDmatrix ispcC accsC
    dpartMatrix = snd dpartTuple
    dpartWidths = fst dpartTuple
    -- test on zeroMatrices 
    checkZeroColumns ys = sumColumns
	where
	sumColumns = [sum i | i <- sumEntries]
	sumEntries = [[sum x|x<- xs]|xs<-ys]
   
-- test definitions 
-- udo = volumeM sweety indexspace accsClean loops 
testVolumeM mdpls ispcCleaned accsCleaned slps= list 
    where
    list = [volumeM  mdpl ispcCleaned accsCleaned slps|mdpl<-mdpls]

-- kyo = testVolumeM cat indexspace accsClean loops

-- complete function for multidimensional placements 

fco5:: String-> Int-> [((Int,Integer),[[Integer]],[Integer],Matrix)]
fco5 path dim = rankingPlacements res   
    where 
    res =[((ddistance pl ixs axs , volumeM pl ixs axs slps),bw,wL,pl)|pl <- plcsMulti]
    plcsMulti  = multidim plcsSimple dim ixs slps       
    plcsSimple = fco4 path
    --d = createDmatrix ixs axs 
    ixs = clean ispcDirty slps
    axs = cleanAccs acc slps 
    ispcDirty = unsafePerformIO( readispcfile path) 
    acc  = unsafePerformIO( readaccfile path)
    slps = unsafePerformIO ( readslpfile path) 
    bw   = [widthIdMatrices, widthAccMatrices, widthOfNulldimArray]
    --             
    -- widths of the unit matrices 
    widthIdMatrices = [ nColumns x | (_,x) <-ixs]
    -- width of the null dimensional array
    widthOfNulldimArray = [fst(accessPosition)-sum(widthAccMatrices)]
    -- width of the lambda part 
    wL = [width | (_,_,(_,width)) <- posAndWidthDmatrix]
    -- width information of the transposed and prepared indexspace
    -- matrices
    posAndWidthDmatrix = ispc2position ixs axs
    -- acc list with transposed acc matrices 
    accTr = accTranspose axs
    -- position information of the different access matrices in the matrix
    accessPosition = (acc2posAndWidth (accTr))
    -- width of the access matrices / array accesses 
    widthAccMatrices = arrayAllok acc
	where
	-- helper to determine the array allocations 
	-- and where they are supposed to cot off
	-- 
	arrayAllok :: [(Integer,AccessType,Integer,Integer, AffineFunction)] -> [Integer]
	arrayAllok acc = [width | x <- numberOfAccesses] 
	    where 
	    -- CAUTION: Assume here that all acc matrices have the same width 
	    width =  head[nColumns (x)|(_,_,_,_,x) <- accTr ]
	    numberOfAccesses  = init(fmToList (snd(accessPosition)))


-- sorts the results of fco5
-- according to less communication distance (highest # of zeros)
-- and highest communication volumina 
--
betterPlacement:: ((Int,Integer),[[Integer]],[Integer],Matrix)->((Int,Integer),[[Integer]],[Integer],Matrix)-> Ordering
betterPlacement a b 
	   |fst (first a) > fst(first b) = LT
           |fst (first a) == fst (first b) = if snd ( first a) > snd (first b) 
			     then LT 
			     else GT
	   |otherwise = GT
	   where
	   first ::(a,b,c,d)->a
           first (a,b,c,d)= a 


-- ranking functions of the placements 
-- 
rankingPlacements  = sortBy betterPlacement


watchDim n = [length (fco5 bsp x)| x <-[0..n]]
