module NewCommCode (main, insertIfsMain, unsafeGetCloogIneqs, parseCommandLine, CloogStatement(..), CloogStmtInfo(..), CommSide(..), CloogData(..), unsafeGetCloogStmtInfos, unsafeCreateCommCodeDS,CommCodeDS(..), CommandLine(), TilingOptions(..),
CloogStatementIneqs(..)) 
-- MG: 24 nov 2006: most of the structures are needed in TaskData
where


import LMath
import DepAcc (srcAcc,acpsToAccs,getSrcAndDstAccs)
import Tiling(defaultTiledLoops,TileInformation(..),
              defaultTile,
              test_l -- for testing
             )
import LoopoDS
import System.IO.Unsafe(unsafePerformIO)
import Debug.Trace(trace)
-- import CodeGen(loopNest, getBoundText)
import IO(openFile, hGetContents, IOMode(..))
import System(getArgs, exitFailure)
import System.Cmd(system)
import System.Time
import System.Console.GetOpt
import GHC.IOBase(ExitCode(..))
import GHC.Base(assert)
import CppInterface
import Foreign.C.String
import Monad (when)
import Maybe
import Ix
import List
import Ratio
import Control.Exception
import InsertIfs.Parser(parseSource)
import InsertIfs.ParserDefs(ParserState(..))
import TrDepDirections(relevantTrDeps, relevantTiledTrDeps,
                       trDepsCommunicationDirection)
import VisualPolylib(showIS2VisualPolylib)

import ScannerParser.SymbolTable(getIndexLexems, getBlobLexems, getArrayLexems)
import ScannerParser.ScannerParser(openParseFile)

enableDebugMessages = False

myTrace :: String -> a -> a
myTrace | enableDebugMessages = trace
        | otherwise           = flip const


{- parsec
import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language (javaStyle)

lexer        = P.makeTokenParser javaStyle
parens       = P.parens lexer
braces       = P.braces lexer
whiteSpace   = P.whiteSpace lexer
myTry = Text.ParserCombinators.Parsec.try
-}

-------------------------------------------------------------------------------
------------------ data structures: -------------------------------------------

-- |
-- Most necessary information about the project and settings are stored
-- in this data type.
--
data CommCodeDS = CommCodeDS 
    { ccTempDir          :: String
    , ccProjFile         :: String
    , ccTargoutExec      :: FilePath
    , ccCloogExec        :: FilePath
    , ccCloogOptions     :: String
    , ccISPCs            :: [Statement]
    , ccTISPCs           :: [TStatement]
    , ccDeps             :: [Dependence]
    , ccTrDeps           :: [TrDependence]
    , ccTrafos           :: [(TStatementID, AffineFunction)]
    , ccTRCs             :: [TRC]
    , ccQualVects        :: [(TStatementID, [QualEntry])]
    , ccIndexNames       :: [String]
    , ccBlobNames        :: [String]
    , ccArrayNames       :: [String]
    , ccACPs             :: [AccessPair]
    , ccAccesses         :: [Access]
    , ccSLPs             :: SurroundingLoops
    , ccTilingOptions    :: TilingOptions
    , ccUseCommunication :: Bool
    , ccCyclicMapping    :: Bool
    , ccUseWrapper       :: Bool
    , ccMergeReceive     :: Bool
    , ccModifyDims       :: [(TStatementID, (Integer, [Vector]))]
    , ccMinBlobSize      :: Integer
    } deriving Show
			

-- |
-- Contains options for tiling.
--
data TilingOptions = TilingOptions 
    { tiOptUseTiling :: Bool
    , tiOptLatVects :: [Vector]
    , tiOptTiledDims :: [Integer]
    , tiOptTilingUsedVect :: [Bool]
    , tiOptTimeTiling :: Bool
    , tiOptSpaceTiling :: Bool
    } deriving Show


-- |
-- Contains different types of identifiers that may occur in the original
-- program. (These identifiers are retrieved by the $targout$ tool.)
--
data Identifiers = Identifiers 
    { idIndices :: [String]
    , idBlobs :: [String]
    , idArrays :: [String]
    } deriving Show
				

-- |
-- Contains all necessary information for one statement that is
-- to be generated by CLooG.
-- Note: its index space can be described by a union of multiple polytopes
--
data CloogStatement = CloogStatement
    { cloogStmtIS :: [IneqSystem]
    , cloogStmtInfo :: CloogStmtInfo
    } deriving Show

-- |
-- Serves for distinguishing the different types of cloog statements.
--
data CloogStmtInfo = ComputeStmt { siTrStmtNr :: TStatementID }
		   | PrepareBufferStmt { siCommSide :: CommSide
				       , siTrDepNr :: TrDependenceID
				       , siArrayNr :: ArrayID
				       , siSrcCoordsDim :: Integer
				       , siDestCoordsDim :: Integer
				       }
		   | CommunicateStmt
		   | RPMappingStmt
		     deriving Show

-- |
-- Representation of the sender or receiver side of a communication.
--
data CommSide = SendSide | RecvSide
		deriving (Show, Eq)

-- |
-- Contains data for describing all cloog statements (and scatter names).
--
data CloogData = CloogMainData
                 { cloogDataCompStmts :: [CloogStatement]
		 , cloogDataSendStmts :: [CloogStatement]
		 , cloogDataRecvStmts :: [CloogStatement]
		 , cloogDataCommStmts :: [CloogStatement]
		 , cloogDataScatterNames :: [String]
		 } 
	       | CloogRPMappingData 
		 { cloogDataRPMappingStmts :: [CloogStatement] }
		 deriving Show


-- |
-- Contains all command line options, where the following options
-- are used to modify aspects of the generated program:
--   clUseCommunication: use @True@ for MPI communication code generation,
--                       @False@ for OpenMP code generation
--   clCyclicMapping: determines distribution of processor tiles onto 
--                    real processors (ignored for OpenMP mode)
--   clUseWrapper: a special wrapper program is generated, which will start
--                 the computation with an adjusted number of MPI processes
--                 (ignored for OpenMP mode)
--   clMergeReceive: If @True@, tries to \"overlap\" computation with the 
--                   unpacking of data from the receiver side's buffer.
--                   Otherwise, the unpacking takes place right after the
--                   alltoall_v statement is finished.
--
data CommandLine = CommandLine
    { clTempDir :: String
    , clProjFile :: String
    , clTargoutExec :: FilePath
    , clCloogExec :: FilePath
    , clCloogOptions :: String
      -- clTilingOptions :: TilingOptions, -- evtl. clLatticeVects :: [Vector]?
    , clLatVects :: [Vector]
    , clTiledDims :: [Integer]
    , clTilingUsedVect :: TilingUsedVector
    , clUseCommunication :: Bool
    , clCyclicMapping :: Bool
    , clUseWrapper :: Bool
    , clMergeReceive :: Bool
    , clModifyDims :: [(Integer, Integer, [[Integer]])]
    , clMinBlobSize :: Integer
    } deriving Show

-- |
-- Options about how CLooG is used for different runs.
--
data CloogRunOption = CloogMainProg
		    | CloogRealProcMapping { cloogRunRPdim :: Integer,
					     cloogRunCyclicMapping :: Bool
					   }

newtype TilingUsedVector = TilingUsedVector [Bool]		     
    deriving Show

data CloogStatementIneqs = 
    CloogStatementIneqs { ineqComputes     :: [IneqSystem]
                        , ineqWrites       :: [IneqSystem]
                        , ineqUnpacks      :: [IneqSystem]
                        , ineqCommunicates :: [IneqSystem]
                        }

{- parsec:
type MyParser a = GenParser Char ParserState a

data ParserState = ParserState
    { psRealProcName :: String
    , psIfStmt :: String
    , psMatchedRP :: Bool
    }	 
-}

-- data StmtType = Receive | Compute | Send | Communicate | RPMapping
-- 	      deriving (Show, Eq, Enum)

newtype StmtDomDescription = StmtDomDescription [StmtDomPart]
    deriving Show

data StmtDomPart = DpTile  DimInfo
		 | DpTISPC DimInfo
		 | DpISPC  DimInfo
		   deriving (Show, Eq)
			    
data DimInfo = DimInfo { dimQual :: QualEntry 
		       , dimRole :: Maybe CommSide
		       }
	       deriving (Show, Eq)


-------------------------------------------------------------------------------
----------------------------- main functions: ---------------------------------

main :: IO ()
main = -- error "TODO: main" 
       do 
       inputStr <- getArgs
       let commandLine = parseCommandLine inputStr       
           ccDS = unsafeCreateCommCodeDS commandLine          
	   mainRun = CloogMainProg
	   rpMapRun = 
	       CloogRealProcMapping { cloogRunRPdim = getRPdim ccDS,
				      cloogRunCyclicMapping = 
				        ccCyclicMapping ccDS}
           loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
	   rpMappingData = 
	       genCloogData ccDS rpMapRun
	   mainData = 
	       genCloogData ccDS mainRun
	   cloogRPMappingInput = genCloogInput ccDS rpMappingData
           cloogMainInput = genCloogInput ccDS mainData
       putStrLn "writing decls..."
       writeDeclarations ccDS
       putStrLn "writing inits..."
       writeInitialisations ccDS
       --putStrLn cloogMainInput
       putStrLn "writing rpMap defines..."
       writeDefinesFile ccDS rpMappingData
       putStrLn "writing main defines..."
       writeDefinesFile ccDS mainData
       putStrLn "writing rpMapInput..."			
       writeFile (loopoFile ++ 
		  cloogInputSuffix rpMapRun) cloogRPMappingInput
       putStrLn "writing cloogInput..."
       writeFile (loopoFile ++
		  cloogInputSuffix mainRun) cloogMainInput
       runCloog ccDS rpMapRun
       runCloog ccDS mainRun
       insertIFs ccDS
--        if (ccUseCommunication ccDS) 
-- 	  then do insertIFs ccDS
-- 	  else return ()

insertIfsMain :: IO ()
insertIfsMain =
       do 
       inputStr <- getArgs
       let commandLine = parseCommandLine inputStr
           ccDS = unsafeCreateCommCodeDS commandLine  
       insertIFs ccDS

---
-- Returns all inequality systems for all cloog statements.
-- (in the order: computation-, send-, recv-, communication-
--  statements)
-- Expects a CommandLine (see "parseCommandLine")
-- (use "showIS2cloog" for printing in Polylib format)
--
unsafeGetCloogIneqs :: CommandLine -> CloogStatementIneqs
unsafeGetCloogIneqs cl =
    CloogStatementIneqs { ineqComputes     = concatMap cloogStmtIS $ cloogDataCompStmts mainData
			, ineqWrites       = concatMap cloogStmtIS $ cloogDataSendStmts mainData
			, ineqUnpacks      = concatMap cloogStmtIS $ cloogDataRecvStmts mainData
			, ineqCommunicates = concatMap cloogStmtIS $ cloogDataCommStmts mainData
			}
    where
    ccDS = unsafeCreateCommCodeDS cl
    mainData = genCloogData ccDS CloogMainProg
    


unsafeGetCloogStmtInfos :: CommandLine -> CloogData 
unsafeGetCloogStmtInfos cl = genCloogData ccDS CloogMainProg
    where
    ccDS = unsafeCreateCommCodeDS cl
    


genCloogInput :: CommCodeDS -> CloogData -> String
genCloogInput ccDS cloogData = verbose "generateCloogInput" $
			       cloogInput
    where			
    cloogInput = 
	header ++ 
	stmtDescriptions ++ 
	scatterFctsDescriptions
    header = 
	verbose "CloogInput: header" $
	genCloogHeader ccDS

    stmtDescriptions = 
	verbose "CloogInput: stmtDescriptions" $
	genStmtDescriptions ccDS cloogData

    scatterFctsDescriptions = 
	verbose "CloogInput: scatterFctsDescriptions" $
	genScatterFctsDescriptions ccDS cloogData


genCloogData :: CommCodeDS -> CloogRunOption -> CloogData
genCloogData ccDS CloogMainProg = 
    verbose "genCloogData" $
    CloogMainData 
    { cloogDataCompStmts = createCompStmts ccDS
    , cloogDataSendStmts = (if useComm 
                            then createSendStmts ccDS
			    else [])      
    , cloogDataRecvStmts = (if useComm 
                            then createRecvStmts ccDS
			    else [])
    , cloogDataCommStmts = (if useComm 
                            then createCommStmts ccDS
			    else [])
    , cloogDataScatterNames = getScatterNames ccDS
    }
    where
    useComm = ccUseCommunication ccDS
genCloogData ccDS CloogRealProcMapping {} =
    CloogRPMappingData { cloogDataRPMappingStmts = createRPMappingStmts ccDS }
				

genCloogHeader :: CommCodeDS -> String
genCloogHeader CommCodeDS { ccBlobNames = blobs
                          , ccMinBlobSize = minBlobSize } = 
    verbose "genCloogHeader" $
    "# created: " ++ getTime ++ "\n" ++
    "# ---------------------- CONTEXT ----------------------\n" ++
    "c # language is C\n\n" ++
    (if (numBlobs > 0) 
     then "1 # Context\n" ++
          (showIS2cloog $ ineqSystem ((0, numBlobs), 
				      unitMatrix numBlobs numBlobs `conc` 
				      blobSize_M)) ++ "\n\n"
     else "1 # Context (no parameter, so always true)\n" ++
          "1 2       # 1 line and 2 columns  \n" ++
          "#  1 \n" ++
          "1  0   # 0 >= 0, always true\n\n") ++
          "1 # set parameter names\n" ++
    (concat $ intersperse " " blobs) ++ "\n\n"
    where
    numBlobs = length blobs
    blobSize_M = (columnMatrix (genericReplicate numBlobs (-minBlobSize%1)))

genStmtDescriptions :: CommCodeDS -> CloogData -> String
genStmtDescriptions ccDS CloogMainData { cloogDataCompStmts = comps
				       , cloogDataSendStmts = sends
				       , cloogDataRecvStmts = recvs
				       , cloogDataCommStmts = comms
				       } 
    = verbose "genStmtDescriptions" $
      "# --------------------- STATEMENTS --------------------\n" ++
      show numTotalStmts ++ "\n\n" ++
      allDescriptions ++
      "\n1 # set the iterator names\n" ++ 
      (concat $ intersperse " " $ getIteratorNames ccDS) ++ "\n\n"      
    where
    allDescriptions = 
	compStmtsDescription ++
	if (ccUseCommunication ccDS) 
	then sendStmtsDescription ++ 
	     recvStmtsDescription ++ 
	     commStmtsDescription
	else []

    numTotalStmts = if (ccUseCommunication ccDS)
		    then length (comps ++ sends ++ recvs ++ comms)
		    else length comps

    futureOptions = "0 0 0 # for future options...\n"
    compStmtsDescription =
	"# " ++ (show $ length comps) ++ " computation stmts:\n" ++
	concat (map domainDesc comps)
    sendStmtsDescription =
	"# " ++ (show $ length sends) ++ " prepare-send stmts:\n" ++
	concat (map domainDesc sends)
    recvStmtsDescription =
	"# " ++ (show $ length recvs) ++ " prepare-receive stmts:\n" ++
	concat (map domainDesc recvs)
    commStmtsDescription =
	"# " ++ (show $ length comms) ++ " communication stmts:\n" ++
	concat (map domainDesc comms)

genStmtDescriptions CommCodeDS { ccCyclicMapping = cyclicMapping } 
		    CloogRPMappingData { cloogDataRPMappingStmts = rpMapStmts }
    = "# -----  rp-mapping: ----- \n" ++
      (show (length rpMapStmts)) ++ " # rp-mapping statement(s)\n" ++
      concatMap domainDesc rpMapStmts ++
      "\n0 # don't set the iterator names\n\n" 

domainDesc :: CloogStatement -> String
domainDesc stmt = 
    show (length (cloogStmtIS stmt)) ++ " # domains per statement\n" ++
    concatMap (\is -> showIS2cloog is ++ "\n\n") 
	    (cloogStmtIS stmt) ++
    "0 0 0 # for future options...\n\n\n"
    

genScatterFcts :: CommCodeDS -> CloogData -> String
genScatterFcts ccDS CloogMainData { cloogDataCompStmts = comps
				  , cloogDataSendStmts = sends
				  , cloogDataRecvStmts = recvs
				  , cloogDataCommStmts = comms
				  }  =
    verbose "genScatterFcts (CloogMainData)" $
    concat $ scatterFcts
    where
    scatterFcts = 
	myTrace (show compFcts) $
	compFcts ++ sendFcts ++ recvFcts ++ commFcts
    compFcts = 
	verbose "compFcts" $
	map (genCompScatterFct ccDS) comps
    sendFcts =
	verbose "sendFcts" $
	if ccUseCommunication ccDS 
	then map (genPrepareBufferScatterFct ccDS) sends
	else []
    recvFcts = 
	verbose "recvFcts" $
	if ccUseCommunication ccDS
	then map (genPrepareBufferScatterFct ccDS) recvs
	else []
    commFcts = 
	verbose "commFcts" $
	if ccUseCommunication ccDS 
	then [genCommScatterFct ccDS (head comms)]
	else []

genScatterFcts ccDS CloogRPMappingData {} = 
    concatMap (genRPMappingScatterFct ccDS) [1..numRPMappingStmts]
    where
    numRPMappingStmts = if ccCyclicMapping ccDS then 1 else 2

genRPMappingScatterFct :: CommCodeDS -> Integer -> String
genRPMappingScatterFct ccDS rpMapStmtNr =
    verbose "genScatterFcts (CloogRPMappingData)" $
    showScatterMatrix2Cloog rpMapping_M
    where
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS
    timeTiling = tiOptTimeTiling $ ccTilingOptions ccDS
    procTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
 
    cyclicMapping = ccCyclicMapping ccDS
    helpDim = if cyclicMapping then 0 else 1

    commonCoordsDim = getCommonCoordsDim ccDS
    tilingDim = getTilingDim ccDS
    timeDim = getSchedDim ccDS
    timeTileDim = getTimeTileDim ccDS
    procDim = getPlaceDim ccDS
    procTileDim = getProcTileDim ccDS
    params = getParamsDim ccDS
	      
    rpDim = if procTiling then procTileDim else procDim

    rpMapping_M = 
        
	myTrace (show [rpDim, procTileDim, timeTileDim, timeDim, tilingDim, params]) $
        myTrace (show helpDim_M ++ show rpDim_M) $
	myMultijoin1 [rpDim_M, helpDim_M]
                     
    rpDim_M =
        (-- kein skewing, da nur auf realprozessor-dimensionen projeziert wird!
         (if procTiling 
          then 
	  --   [helpDim] | procDims ||  tileDims   | commonCoords | params | const
	  --             |          || time | proc | time | proc  |        |   
	  --   _________ |__________||______|______|______|_______|________|______
	  --      0      |     1    ||  0   |  -1  |   0  |  0    |   0    |   0 
	  myTrace ("genScatterFcts (rpMapping): procTileDim = " ++ show procTileDim) $
	  (zeroMatrix procTileDim helpDim `conc`
           unitMatrix procTileDim procTileDim `conc`
	   zeroMatrix procTileDim timeTileDim `conc`
	   (-(unitMatrix procTileDim procTileDim)) `conc`
	   zeroMatrix procTileDim (commonCoordsDim + params + 1)
	  )
	  else 
	  --   [helpDim] | procDims ||  tileDims   | commonCoords | params | const
	  --             |          || time |(proc)| time | proc  |        |   
	  --   _________ |__________||______|______|______|_______|________|______
	  --       0     |     1    ||  0   |  --  |   0  |  -1   |   0    |   0 
	  (zeroMatrix procDim helpDim `conc`
	   unitMatrix procDim procDim `conc`
	   zeroMatrix procDim timeTileDim `conc`
	   zeroMatrix procDim timeDim `conc`
	   (-(unitMatrix procDim procDim)) `conc`
	   zeroMatrix procDim (params + 1))
	 )
        )

    helpDim_M =    
	  -- additional help-dimension to schedule 2 loop nests in the case of block mapping
	 (--   [helpDim] | procDims ||  tileDims   | commonCoords | params | const
	  --             |          || time |[proc]| time | proc  |        |   
	  --   _________ |__________||______|______|______|_______|________|______
	  --       1     |     0    ||  0   |   0  |   0  |   0   |   0    | -rpMapStmtNr
	  myTrace "rpMappingStmts: helpDim" $
          (unitMatrix helpDim helpDim `conc`
	   zeroMatrix helpDim (rpDim + tilingDim + commonCoordsDim + params) `conc`
	   if cyclicMapping
	   then zeroMatrix helpDim 1
	   else (-(matrix [[rpMapStmtNr%1]])))
	 )
	

genCompScatterFct :: CommCodeDS -> CloogStatement -> String
genCompScatterFct ccDS CloogStatement 
		       { cloogStmtInfo = 
			   stmtInfo@ComputeStmt { siTrStmtNr = trStmtNr }
		       } = 
        verbose "genCompScatterFct" $
	--trace ("\naccDim: " ++ show arrAccessDim ++ "\n") $
        showScatterMatrix2Cloog comp_M
    where
    useCommunication = ccUseCommunication ccDS
    tiOpts = ccTilingOptions ccDS

    timeTiling = tiOptTimeTiling tiOpts
    procTiling = tiOptSpaceTiling tiOpts
    timeTileUsedVect = 
        genericTake timeTileDim $ tiOptTilingUsedVect tiOpts

    ispcDim = getISPCdim ccDS trStmtNr
    coordsDim = getCoordsDim ccDS trStmtNr
    maxCoordsDim = getMaxCoordsDim ccDS
    commonCoordsDim = getCommonCoordsDim ccDS

    coordsTimeDim = getSchedDim ccDS
    coordsProcDim = getPlaceDim ccDS

    timeTileDim = getTimeTileDim ccDS
    procTileDim = getProcTileDim ccDS
    tilingDim = getTilingDim ccDS
		  
    vTimeDim = if timeTiling then coordsTimeDim else 0
    vProcDim = if procTiling then coordsProcDim else 0
    -- max number of dimensions added by targgen:
    vAddedDim = maxCoordsDim - commonCoordsDim

    -- number of remaining added dims for this TrStatement:
    additionalRemainingDim = maxCoordsDim - coordsDim

    params = getParamsDim ccDS

    globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS
    blockNrDim = if useCommunication then 1 else 0
    stmtTypeDim = if useCommunication then 1 else 0
    arrNrDim = 1
    blockNr = getBlockNr ccDS stmtInfo
    stmtType = getStmtTypeNr stmtInfo
    arrAccessDim = getArrAccessMaxDim ccDS
    trafoFct = getTrafoFct ccDS trStmtNr
    trafoFctAdded_ = 
	verbose ("trafoFctAdded: added rows=" ++ show additionalRemainingDim) $
	affineFunction ((ispcDim, params), 
			getMatrixIH trafoFct `join` 
			zeroMatrix additionalRemainingDim (ispcDim + params + 1)
		       ) Inhomogeneous
    trafoFctAdded =
        case lookup trStmtNr (ccModifyDims ccDS) of
           Nothing -> trafoFctAdded_
           Just (dimsToRetain, newTrafoPart)
             | dimsToRetain+genericLength newTrafoPart <= coordsDim  -> 
               trace ("replacing tstmt " ++ show trStmtNr) $
               affineFunction ((ispcDim, params)
                              ,multijoin1 [subMatrixHor trafoFctAdded_ 0 (dimsToRetain-1)
                                          ,rowVectorsToMatrix ispcDim newTrafoPart
                                          ,subMatrixHor trafoFctAdded_ (dimsToRetain+genericLength newTrafoPart-1) (nRows trafoFctAdded_-1)
                                          ]
                              )
                              Homogeneous
             | otherwise  ->  error ("too many dimensions in replacement for statement " ++ show trStmtNr ++ "\n" ++
                                    show (trafoFct,trafoFctAdded_,coordsDim))

{-
    armin_replace :: [(TStatementID, (Integer, [Vector]))]
    armin_replace = [(integerToID 1, (0, [vector [2,1,0,0,0],
                                          vector [1,1,0,0,0]
                                         ]
                                     )
                     )
                    ]
-}
    
--    (trafo_v, trafo_p, trafo_c) = 
--      partsVPCih trafoFct

    -- the part of the trafo-matrix that results in time coordinates:
    timeTrafo_M = 
	verbose "@timeTrafo_M" $
	myTrace ("(0, coordsTimedim-1) = (" ++ 
	        show 0 ++ ", " ++ show (coordsTimeDim - 1) ++ 
	        ")") $
	subMatrixHor (getMatrixIH trafoFctAdded) 0 (coordsTimeDim-1)
    procTrafo_M =
	verbose "@procTrafo_M" $
	myTrace ("(coordsTimeDim, coordsTimeDim+coordsProcDim-1) = (" ++ 
	        show coordsTimeDim ++ ", " ++ show (coordsTimeDim+coordsProcDim - 1) ++ 
	        ")") $
	subMatrixHor (getMatrixIH trafoFctAdded) 
		     coordsTimeDim (coordsTimeDim+coordsProcDim-1)
    -- remaining dimensions:
    addTrafo_M =
	verbose "@remaining dims" $
	myTrace ("(commonCoordsDim, maxCoordsDim-1) = (" ++ 
	        show commonCoordsDim ++ ", " ++ show (maxCoordsDim - 1) ++ 
	        ")") $
	subMatrixHor (getMatrixIH trafoFctAdded) 
			 (commonCoordsDim) (maxCoordsDim-1)

    fillVarsDim = 
	if useCommunication 
	then (rpDim + arrAccessDim + arrNrDim)
	else 0
		     
    useSkewing = {- useCommunication && -} timeTiling

    -- skew_M = getSkewingMatrixRPpart rpDim timeTileUsedVect
    skew_M = rowMatrix $ genericReplicate rpDim (-1%1)
    
    timeTilePart_M = getTimeTileRHS timeTileUsedVect


    comp_M =
	myTrace ("skew_M: " ++ show skew_M) $
	myTrace ("trafoFctAdded: " ++ show trafoFctAdded) $
	myTrace ("getMatrixIH trafoFctAdded: " ++ (show $ getMatrixIH trafoFctAdded)) $
	myTrace ("addTrafo_M: " ++ show addTrafo_M) $
	if useCommunication
	then myMultijoin1 
		 [ globalTime_M, blockNr_M, rp_M, vTime_M, vProc_M, vAdded_M, stmtType_M,
		   fillVars_M
		 ]
	else myMultijoin1
		 [ globalTime_M, rp_M, vTime_M, vProc_M, vAdded_M]

    globalTime_M = 
	-- scatter-part:
	----------------
	unitMatrix globalTimeDim globalTimeDim `conc`
	zeroMatrix globalTimeDim (blockNrDim) `conc`
	(if useSkewing then skew_M else zeroMatrix globalTimeDim rpDim) `conc`
	zeroMatrix globalTimeDim (vTimeDim + vProcDim + vAddedDim + 
				  stmtTypeDim + fillVarsDim) `conc`
	-- domain-part:
	---------------
        (if timeTiling 
	 -- if time-tiling is used, tiled time is our global time:
	 then timeTilePart_M
	 else zeroMatrix globalTimeDim timeTileDim) `conc`
	(zeroMatrix globalTimeDim procTileDim) `conc`

	(if not timeTiling
	 -- if no time-Tiling is used, then coordsTime becomes our global time
	 -- for this purpose, timeTrafo is used to transform the ispc dimensions:
	 then -(timeTrafo_M)
	 else (zeroMatrix globalTimeDim (ispcDim + params + 1))
	)

    blockNr_M = 
	zeroMatrix blockNrDim globalTimeDim `conc`
	unitMatrix blockNrDim blockNrDim `conc`
	zeroMatrix blockNrDim (rpDim + vTimeDim + vProcDim + vAddedDim + 
			       stmtTypeDim + fillVarsDim + tilingDim + 
			       ispcDim + params) `conc`
	(-(matrix [[blockNr%1]]))

    rp_M = 
	zeroMatrix rpDim (globalTimeDim + blockNrDim) `conc`
	unitMatrix rpDim rpDim `conc`
	zeroMatrix rpDim (vTimeDim + vProcDim + vAddedDim + 
			  stmtTypeDim + fillVarsDim +
			  timeTileDim) `conc`
	(if procTiling
	 then -(unitMatrix rpDim procTileDim)
	 else zeroMatrix rpDim procTileDim) `conc`
	(if not procTiling
	 then -(procTrafo_M)
	 else zeroMatrix rpDim (ispcDim + params + 1)
	)
	
    vTime_M = 
	(zeroMatrix vTimeDim (globalTimeDim + blockNrDim +
			      rpDim) `conc`
	 unitMatrix vTimeDim vTimeDim `conc`
         zeroMatrix vTimeDim (vProcDim + vAddedDim +
			      stmtTypeDim + fillVarsDim + 
			      timeTileDim + procTileDim) `conc`
	 (if timeTiling
	  then -(timeTrafo_M)
	  else zeroMatrix vTimeDim (ispcDim + params + 1))
	)

    vProc_M = 
	(zeroMatrix vProcDim (globalTimeDim + blockNrDim +
			      rpDim + vTimeDim) `conc`
	 unitMatrix vProcDim vProcDim `conc`
	 zeroMatrix vProcDim (vAddedDim +
			      stmtTypeDim + fillVarsDim + 
			      timeTileDim + procTileDim) `conc`
	 (if procTiling
	  then -(procTrafo_M)
	  else zeroMatrix vProcDim (ispcDim + params + 1))
	)

    vAdded_M = 
	zeroMatrix vAddedDim (globalTimeDim + blockNrDim +
			      rpDim + vTimeDim + vProcDim) `conc`
	unitMatrix vAddedDim vAddedDim `conc`
	zeroMatrix vAddedDim (stmtTypeDim + fillVarsDim + 
			      timeTileDim + procTileDim) `conc`
	(-(addTrafo_M)) 

    stmtType_M = 
	zeroMatrix stmtTypeDim (globalTimeDim + blockNrDim +
				rpDim + vTimeDim + vProcDim + vAddedDim) `conc`
	unitMatrix stmtTypeDim stmtTypeDim `conc`
	zeroMatrix stmtTypeDim (fillVarsDim + 
				timeTileDim + procTileDim +
				ispcDim + params) `conc`
	(-matrix [[stmtType%1]])

    fillVars_M =
	zeroMatrix fillVarsDim (globalTimeDim + blockNrDim + rpDim + 
				vTimeDim + vProcDim + vAddedDim + 
				stmtTypeDim) `conc`
	unitMatrix fillVarsDim fillVarsDim `conc`
	zeroMatrix fillVarsDim (timeTileDim + procTileDim +
				ispcDim + params + 1)
	

genPrepareBufferScatterFct :: CommCodeDS -> CloogStatement -> String
genPrepareBufferScatterFct ccDS CloogStatement 
		       { cloogStmtIS = cloogIS
		       , cloogStmtInfo = stmtInfo@PrepareBufferStmt
			                 { siCommSide = commSide
					 , siTrDepNr = trDepNr
					 , siArrayNr = arrID
					 , siSrcCoordsDim = srcCoordsDim
					 , siDestCoordsDim = destCoordsDim
					 }
								     
		       } = verbose ("genPrepareBufferScatterFct(" ++
			            show commSide ++ ")") $
				    myTrace (show (nColumns (head cloogIS)) ++ ", [tilingDim,destDim,srcDomDim,destDomDim,params,1]" ++ show [tilingDim,destDim,srcDomDim,destDomDim,params,1]) $
			   if (nColumns (head cloogIS) ==
			       (tilingDim + destDim + 
				srcDomDim + destDomDim +
				params + 1))
			   then showScatterMatrix2Cloog scatter_M
			   else error ("genPrepareBufferScatterFct: " ++
				       "cloogIS doesn't match" ++
				       "scatterFct" ++
				       "nColumns cloogIS: " ++ 
				       show (nColumns (head cloogIS)) ++
				    ", [tilingDim,destDim,srcDomDim,destDomDim,params,1]" ++ show [tilingDim,destDim,srcDomDim,destDomDim,params,1])
    where

    isSender = (commSide == SendSide)

    scatter_M = myMultijoin1
	     [ globalTime_M, blockNr_M, rp_M, 
	       vTime_M, vProc_M, vAdded_M, stmtType_M, 
	       -- otherTime_M, 
	       otherRP_M,
	       arrAcc_M, arrAccAdd_M, arrNr_M, fillVars_M ]

    arrNr = integerFromID arrID
    trDep = getTrDep ccDS trDepNr

    tiOpts = ccTilingOptions ccDS
    timeTiling = tiOptTimeTiling tiOpts
    procTiling = tiOptSpaceTiling tiOpts
    timeTileUsedVect = genericTake timeTileDim $ tiOptTilingUsedVect tiOpts

    maxCoordsDim = getMaxCoordsDim ccDS
    commonCoordsDim = getCommonCoordsDim ccDS
    coordsTimeDim = getSchedDim ccDS
    coordsProcDim = getPlaceDim ccDS

    tilingDim = getTilingDim ccDS
    timeTileDim = getTimeTileDim ccDS
    procTileDim = getProcTileDim ccDS
		  
    
    destAddDim = destCoordsDim - commonCoordsDim

    vTimeDim = if timeTiling then coordsTimeDim else 0
    vProcDim = if procTiling then coordsProcDim	else 0
    -- max number of dimensions added by targgen:
    vAddedDim = maxCoordsDim - commonCoordsDim

    -- number of remaining dims to add for this TrStatement:
    additionalRemainingDim = maxCoordsDim - srcCoordsDim

    params = getParamsDim ccDS
    globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS

    -- the corresponding side's dimensionality:
    otherRPdim = rpDim

    destDim = 
	if procTiling 
	then procTileDim 
	else coordsProcDim + destAddDim

    blockNrDim = 1
    stmtTypeDim = 1
    arrNrDim = 1

    -- mergeReceive == True: merge receiving of data with other operations 
    --                       of the next global. May be incorrect for 
    --                       complex cases ("crossing dependences")
    -- mergeReceive == False: receive data after alltoall, before the next
    --                        global time step.
    mergeReceive = ccMergeReceive ccDS

    blockNr = 
        if (isSender || mergeReceive)
        then getBlockNr ccDS stmtInfo -- receive data merged with other operations
        else 2 -- receive data after alltoall
    stmtType = getStmtTypeNr stmtInfo
    arrAccessMaxDim = getArrAccessMaxDim ccDS
    accDim = targetDimIH $ accIndexFct srcAcc
    accAddDim = arrAccessMaxDim - accDim		      

    srcAcc = getSrcAcc ccDS trDepNr
    (accFct_v, accFct_p, accFct_c) = 
	--or should it be: partsVPC ?
	partsVPCih $ accIndexFct srcAcc

    srcTrStmtID = trDepSrcStmtID trDep
    trafoFct = getTrafoFct ccDS srcTrStmtID
    trafoFctAdded = 
	myTrace "trafoFctAdded" $
	affineFunction ((srcDomDim, params), 
			getMatrixIH trafoFct `join`
			zeroMatrix additionalRemainingDim (srcDomDim + params + 1)
		       ) Inhomogeneous
    
    makeVPCih m = partsVPCih $ affineFunction ((srcDomDim, params), m) Inhomogeneous

    (trafo_v, trafo_p, trafo_c) = 
      partsVPCih trafoFctAdded

    -- the part of the trafo-matrix that results in time coordinates:
    (timeTrafo_v, timeTrafo_p, timeTrafo_c) =
	verbose "timeTrafo_vpc" $
	makeVPCih $
	subMatrixHor (getMatrixIH trafoFctAdded) 0 (coordsTimeDim-1)
    (procTrafo_v, procTrafo_p, procTrafo_c) =
	verbose "procTrafo_vpc" $
	makeVPCih $
	subMatrixHor (getMatrixIH trafoFctAdded) 
		     coordsTimeDim (coordsTimeDim+coordsProcDim-1)
    -- remaining dimensions:
    (addTrafo_v, addTrafo_p, addTrafo_c) =
	verbose "addTrafo_vpc" $
	makeVPCih $
	assert (commonCoordsDim + vAddedDim - 1 <= nRows (getMatrixIH trafoFctAdded)) $
	subMatrixHor (getMatrixIH trafoFctAdded) 
		     (commonCoordsDim) (commonCoordsDim + vAddedDim-1)

    (srcDepDomain, destDepDomain) = getDepDomains ccDS trDep
    srcDomDim = nVars srcDepDomain
    destDomDim = 
	-- if proc tiling is used, we won't have destCoords,
	-- so we don't need to use the destDependenceDomain
	-- and it's space-time-transformation:
	if procTiling 
	then 0 
	else nVars destDepDomain

    fillVarsDim = 0 -- bei compStmt: otherRPdim + arrAccessMaxDim + arrNrDim

    useSkewing = timeTiling

    skew_M = rowMatrix $ genericReplicate rpDim (-1%1)

    timeTilePart_M = getTimeTileRHS timeTileUsedVect


    -- only needed for receive-stmt:
    delayByOne_M = 
	verbose ("delay by one...") $
	columnMatrix $
	if timeTiling 
	then map (\isTiled -> if isTiled then 1%1 else 0) timeTileUsedVect
	else -- without time tiling, we have to increase the first entry in the 
	     -- constant column of the timeTrafo matrix (representing the outermost
             -- time dimension):
	     head $ columns' $ 
	     myTrace ("oldEntry: " ++ show oldEntry) $
	     setEntry timeTrafo_c 0 0 (oldEntry + 1)
		 where
		 oldEntry = entry timeTrafo_c 0 0

    globalTime_M = 
	verbose "globalTime_M" $
	unitMatrix globalTimeDim globalTimeDim `conc`
	zeroMatrix globalTimeDim (blockNrDim) `conc`
        (if (useSkewing && (isSender || not mergeReceive)) 
	 then skew_M 
	 else zeroMatrix globalTimeDim rpDim) `conc`
	zeroMatrix globalTimeDim (vTimeDim + vProcDim + vAddedDim + 
				  stmtTypeDim) `conc`
        (if (useSkewing && not isSender && mergeReceive) 
	 -- for the receiving side, we have to skew by otherRP,
	 -- because "rp" and "otherRP" have changed...
	 then skew_M 
	 else zeroMatrix globalTimeDim otherRPdim) `conc`
        zeroMatrix globalTimeDim (arrAccessMaxDim + arrNrDim + 
				  fillVarsDim) `conc`
	-- SrcTile:
	(if timeTiling 
	 then timeTilePart_M 
	 else zeroMatrix globalTimeDim timeTileDim) `conc`
	zeroMatrix globalTimeDim procTileDim `conc`
	zeroMatrix globalTimeDim destDim `conc`
        (if timeTiling 
	 then zeroMatrix globalTimeDim srcDomDim 
	 else (-timeTrafo_v)
	) `conc`
	 zeroMatrix globalTimeDim (destDomDim) `conc`
	(if (isSender || not mergeReceive)
	 then (if timeTiling 
	       then zeroMatrix globalTimeDim (params + 1)
	       else (-timeTrafo_p) `conc` (-timeTrafo_c))
	 else -- we have to delay the receiver side by one global time step
	      (if timeTiling 
	       then (zeroMatrix globalTimeDim params)
	       else (-timeTrafo_p)) `conc`
	 -- the receiving side is delayed by one global time step
	 -- (relative to the sender global time)
	 (-delayByOne_M)
	)

    blockNr_M = 
	verbose "blockNr_M" $
	zeroMatrix blockNrDim globalTimeDim `conc`
	unitMatrix blockNrDim blockNrDim `conc`
	zeroMatrix blockNrDim (rpDim + vTimeDim + vProcDim + vAddedDim + 
			       stmtTypeDim + otherRPdim +
			       arrAccessMaxDim + arrNrDim + fillVarsDim +
			       tilingDim + destDim + srcDomDim + destDomDim +
			       params) `conc`
	(-(matrix [[blockNr%1]]))

    rp_M = 
	verbose ("rp_M" ++ " isSender: " ++ show isSender ++ " procTiling: " ++ show procTiling)$
	zeroMatrix rpDim (globalTimeDim + blockNrDim) `conc`
	unitMatrix rpDim rpDim `conc`
	zeroMatrix rpDim (vTimeDim + vProcDim + vAddedDim + stmtTypeDim + 
			  otherRPdim +
			  arrAccessMaxDim + arrNrDim + fillVarsDim + 
			  timeTileDim) `conc`
	if (isSender || not mergeReceive)
	-- the sender real processor is our src-Processor-part 
	-- (of the domain describing inequality system):
	then ((if procTiling
	       then -(unitMatrix rpDim procTileDim)
	       else zeroMatrix rpDim (procTileDim)
	      ) `conc`
	      zeroMatrix rpDim destDim `conc`
	      (if not procTiling
	       then ((-procTrafo_v) `conc`
		     zeroMatrix rpDim destDomDim `conc`
		     (-procTrafo_p) `conc` (-procTrafo_c)
		    )
	       else zeroMatrix rpDim (srcDomDim + destDomDim +
				      params + 1))
	     )
	else -- the receiverer real processor is our dest-Processor-part 
	     -- (of the domain describing inequality system):
	     (zeroMatrix rpDim procTileDim `conc`
	      (if procTiling
	       then (-(unitMatrix rpDim procTileDim))
	       else ((-(unitMatrix rpDim coordsProcDim)) `conc`
		     zeroMatrix rpDim destAddDim)
	      ) `conc`
	      zeroMatrix rpDim (srcDomDim + destDomDim +
				params + 1)
	     )
	
    vTime_M =
	verbose "vTime_M" $
	zeroMatrix vTimeDim (globalTimeDim + blockNrDim +
			     rpDim) `conc`
	unitMatrix vTimeDim vTimeDim `conc`
	zeroMatrix vTimeDim (vProcDim + vAddedDim + stmtTypeDim + 
			     otherRPdim +
			     arrAccessMaxDim + arrNrDim +
			     fillVarsDim + timeTileDim + procTileDim +
			     destDim) `conc`
        (if timeTiling
	 then (-timeTrafo_v) `conc` 
	      zeroMatrix vTimeDim destDomDim `conc`
	      (-timeTrafo_p) `conc` (-timeTrafo_c)
	 else zeroMatrix vTimeDim srcDomDim `conc`
	      zeroMatrix vTimeDim destDomDim `conc`
	      zeroMatrix vTimeDim (params + 1)
	)

    vProc_M =
	verbose "vProc_M" $
	zeroMatrix vProcDim (globalTimeDim + blockNrDim +
			     rpDim + vTimeDim) `conc`
	unitMatrix vProcDim vProcDim `conc`
	zeroMatrix vProcDim (vAddedDim + stmtTypeDim + 
			     otherRPdim +
			     arrAccessMaxDim + arrNrDim +
			     fillVarsDim + timeTileDim + procTileDim +
			     destDim) `conc`
        (if procTiling
	 then (-procTrafo_v) `conc` 
	      zeroMatrix vProcDim destDomDim `conc`
	      (-procTrafo_p) `conc` (-procTrafo_c)
	 else zeroMatrix vProcDim srcDomDim `conc`
	      zeroMatrix vProcDim destDomDim `conc`
	      zeroMatrix vProcDim (params + 1)
	)

    vAdded_M =
	verbose "vAdded_M" $
	zeroMatrix vAddedDim (globalTimeDim + blockNrDim +
			      rpDim + vTimeDim + vProcDim) `conc`
	unitMatrix vAddedDim vAddedDim `conc`
	zeroMatrix vAddedDim (stmtTypeDim + 
			      otherRPdim +
			      arrAccessMaxDim + arrNrDim +
			      fillVarsDim + timeTileDim + procTileDim +
			      destDim) `conc`
	(-addTrafo_v) `conc`
	zeroMatrix vAddedDim destDomDim `conc`
	(-addTrafo_p) `conc` (-addTrafo_c)

    stmtType_M = 
	verbose "stmtType_M" $
	zeroMatrix stmtTypeDim (globalTimeDim + blockNrDim +
				rpDim + vTimeDim + vProcDim + vAddedDim) `conc`
	unitMatrix stmtTypeDim stmtTypeDim `conc`
	zeroMatrix stmtTypeDim (otherRPdim +
				arrAccessMaxDim + arrNrDim + fillVarsDim +
				timeTileDim + procTileDim +
				destDim + 
				srcDomDim + destDomDim +
				params) `conc`
	(-matrix [[stmtType%1]])

    otherRP_M = 
	verbose "otherRP_M" $
	zeroMatrix rpDim (globalTimeDim + blockNrDim +			  
			  rpDim + vTimeDim + vProcDim + vAddedDim + 
			  stmtTypeDim) `conc`
	unitMatrix rpDim rpDim `conc`
	zeroMatrix rpDim (arrAccessMaxDim + arrNrDim + fillVarsDim) `conc`
	if (isSender || not mergeReceive)
	then (zeroMatrix rpDim (timeTileDim + procTileDim) `conc`
	      (-(unitMatrix rpDim rpDim)) `conc`
	      zeroMatrix rpDim ((destDim - rpDim) +
				srcDomDim + destDomDim +
				params + 1)
	     )
	-- is receiver:
	else (if procTiling 
	      then (zeroMatrix rpDim timeTileDim `conc`
		    (-(unitMatrix rpDim rpDim)) `conc`
		    zeroMatrix rpDim (destDim +
				      srcDomDim + destDomDim +
				      params + 1))
	      else (zeroMatrix rpDim (timeTileDim + procTileDim + 
				      destDim) `conc`
		    (-procTrafo_v) `conc`
		    zeroMatrix rpDim destDomDim `conc`
	            (-procTrafo_p) `conc` (-procTrafo_c)
		   )
	     )		    		    

    arrAcc_M = 	
        myTrace ("srcDomDim: " ++ (show srcDomDim) ++ " nVars accFct: " ++ show (nColumns accFct_v)) $
        myTrace ("params: " ++ (show params) ++       " nParams accFct: " ++ show (nColumns accFct_p)) $
        myTrace ("const: " ++ (show 1) ++             " nConst accFct: " ++ show (nColumns accFct_c)) $
	verbose "arrAcc_M" $
	(zeroMatrix accDim (globalTimeDim + blockNrDim + rpDim + 
			    vTimeDim + vProcDim + vAddedDim + stmtTypeDim +
			    otherRPdim) `conc`
	 unitMatrix accDim accDim `conc`
	 zeroMatrix accDim (accAddDim + arrNrDim + fillVarsDim + 
			    timeTileDim + procTileDim + 
			    destDim) `conc`
	 (-accFct_v) `conc`
	 zeroMatrix accDim destDomDim `conc`
	 (-accFct_p) `conc`
	 (-accFct_c)
	)
	 
    arrAccAdd_M =
	verbose "arrAccAdd_M" $
	 (zeroMatrix accAddDim (globalTimeDim + blockNrDim + rpDim + 
				vTimeDim + vProcDim + vAddedDim + stmtTypeDim +
				otherRPdim + accDim) `conc`
	  unitMatrix accAddDim accAddDim `conc`
	  zeroMatrix accAddDim (arrNrDim + fillVarsDim +
				timeTileDim + procTileDim +
				destDim + srcDomDim + destDomDim +
				params + 1)
	 )

    arrNr_M = 
	verbose "arrNr_M" $
	zeroMatrix arrNrDim (globalTimeDim + blockNrDim + rpDim + 
			     vTimeDim + vProcDim + vAddedDim + stmtTypeDim + 
			     otherRPdim + arrAccessMaxDim) `conc`
	unitMatrix arrNrDim arrNrDim `conc`
	zeroMatrix arrNrDim (fillVarsDim + tilingDim + 
			     destDim + srcDomDim + destDomDim +
			     params) `conc`
	(-matrix [[arrNr%1]])

    fillVars_M =
	verbose "fillVars_M" $
	zeroMatrix fillVarsDim (globalTimeDim + blockNrDim + rpDim + 
				vTimeDim + vProcDim + vAddedDim + 
				stmtTypeDim + otherRPdim +
				arrAccessMaxDim + arrNrDim) `conc`
	unitMatrix fillVarsDim fillVarsDim `conc`
	zeroMatrix fillVarsDim (timeTileDim + procTileDim +
				destDim + 
				srcDomDim + destDomDim +
				params + 1)
	

genCommScatterFct :: CommCodeDS -> CloogStatement -> String
genCommScatterFct ccDS CloogStatement 
		       { cloogStmtIS = is
		       , cloogStmtInfo = CommunicateStmt
		       } = 
    verbose "genCommScatterFct" $
    showScatterMatrix2Cloog comm_M
    where
    
    tiOpts = ccTilingOptions ccDS
    timeTiling = tiOptTimeTiling tiOpts
    procTiling = tiOptSpaceTiling tiOpts

    globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS
    otherRPdim = rpDim -- (destDim)
    arrAccMaxDim = getArrAccessMaxDim ccDS
    arrNrDim = 1
    stmtTypeNrDim = 1
    blockNrDim = 1
    blockNr = getBlockNr ccDS CommunicateStmt
    vTimeDim = if timeTiling then getSchedDim ccDS else 0
    vProcDim = if procTiling then getPlaceDim ccDS else 0
    vAddedDim = maxCoordsDim - commonCoordsDim

    maxCoordsDim = getMaxCoordsDim ccDS
    commonCoordsDim = getCommonCoordsDim ccDS

    fillVarsDim = rpDim + vTimeDim + vProcDim + vAddedDim + stmtTypeNrDim + 		  
		  otherRPdim + arrAccMaxDim + arrNrDim
    paramsDim = getParamsDim ccDS

    comm_M = myMultijoin1 [ globalTime_M, blockNr_M, fillVars_M ]

    globalTime_M = 
        unitMatrix globalTimeDim globalTimeDim `conc`
	zeroMatrix globalTimeDim (blockNrDim + fillVarsDim) `conc`
	(-(unitMatrix globalTimeDim globalTimeDim)) `conc`
	-- (skewing has already been applied 
	--  to the domain describing ineqSystem...)
	zeroMatrix globalTimeDim (paramsDim + 1)

    blockNr_M =
	zeroMatrix blockNrDim globalTimeDim `conc`
	unitMatrix blockNrDim blockNrDim `conc`
	zeroMatrix blockNrDim (fillVarsDim + globalTimeDim +
			       paramsDim) `conc`
	matrix [[-(blockNr%1)]]
	
    fillVars_M =
        zeroMatrix fillVarsDim (globalTimeDim + blockNrDim) `conc`
	unitMatrix fillVarsDim fillVarsDim `conc`
	zeroMatrix fillVarsDim (globalTimeDim + 
				paramsDim + 1)


getSkewingMatrixRPpart :: Integer -> [Bool] -> Matrix
getSkewingMatrixRPpart rpDim timeTilingUsedVect = 
    verbose "getSkewingMatrixRPpart" $
    assert (not $ null timeTilingUsedVect) $
    myMultijoin1 $ 
    map (\isTiled -> 
	 if isTiled 
	 -- If time-tiling is used for one dimension, 
	 -- we have to skew that tiled time-dimension:
	 then 
	 -- we skew by the sum of all real processors:
	 -- (in case that we don't have space-tiling,
	 -- our real processor dimension is the placement
	 -- dimension)
	 rowMatrix $ genericReplicate rpDim (-1%1)
	 else zeroMatrix 1 rpDim
	)
    timeTilingUsedVect

getTimeTileRHS :: [Bool] -> Matrix
getTimeTileRHS timeTileUsedVect = 
    rowMatrix $ map (\d -> if d then (-1%1) else 0)
              timeTileUsedVect

getScatterNames :: CommCodeDS -> [String]
getScatterNames ccDS = 
    verbose "getScatterNames" $
    if useComm 
    then commScatterNames
    else nonCommScatterNames		       
    where
    useComm = ccUseCommunication ccDS
    tiOpts = ccTilingOptions ccDS
    timeTiling = tiOptTimeTiling tiOpts
    procTiling = tiOptSpaceTiling tiOpts
    commonQualVect = getCommonQualVect (ccQualVects ccDS)
    allQualVects = snd $ unzip $ ccQualVects ccDS
    maxQualVect = argMax genericLength allQualVects
		  
    sched_t = "s"
    place_t = "p"
    add_t   = "a"

    globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS
    vTimeDim = if timeTiling then getSchedDim ccDS else 0
    vProcDim = if procTiling then getPlaceDim ccDS else 0
    vAddedDim = maxCoordsDim - commonCoordsDim

    maxCoordsDim = getMaxCoordsDim ccDS
    commonCoordsDim = getCommonCoordsDim ccDS

    arrAccDim = getArrAccessMaxDim ccDS
    arrNrDim = 1

    localDim = max (arrAccDim + arrNrDim) vProcDim
    
    nonCommScatterNames =
	-- TODO: nicht synchroner fall...?
	concat $
	[[globalTime_t ++ (show i) | i <- [1..globalTimeDim]]
	,[rp_t ++ (show i) | i <- [1..rpDim]]
	,[vTime_t ++ (show i) | i <- [1..vTimeDim]]
	,[vProc_t ++ (show i) | i <- [1..vProcDim]]
	,[vAdded_t ++ (show i) | i <- [1..vAddedDim]]
	]

    commScatterNames = 
	concat $
	[[globalTime_t ++ (show i) | i <- [1..globalTimeDim]]
	,[blockNr_t]
	,[rp_t ++ (show i) | i <- [1..rpDim]]
	,[vTime_t ++ (show i) | i <- [1..vTimeDim]]
	,[vProc_t ++ (show i) | i <- [1..vProcDim]]
	,[vAdded_t ++ (show i) | i <- [1..vAddedDim]]
	,[stmtType_t]
	,[otherProc_t ++ (show i) | i <- [1..rpDim]]
	,[arrAcc_t ++ (show i) | i <- [1..arrAccDim]]
	,[arrNr_t ++ (show i) | i <- [1..arrNrDim]]
	]


qualVect2Strings :: [QualEntry] -> [String]
qualVect2Strings qualVect = map qualEntry2String qualVect
    where
    qualEntry2String Sched = "s"
    qualEntry2String Place = "p"
    qualEntry2String Additional = "a"

genScatterFctsDescriptions :: CommCodeDS -> CloogData -> String
genScatterFctsDescriptions ccDS clData@CloogMainData
			       { cloogDataCompStmts = comps
			       , cloogDataSendStmts = sends
			       , cloogDataRecvStmts = recvs
			       , cloogDataCommStmts = comms
			       , cloogDataScatterNames = scatterNames
			       } =
    verbose "genScatterFctsDescriptions (CloogMainData)" $
    "# --------------------- SCATTERING --------------------\n" ++
    (show numTotalStmts) ++ " # Scattering functions\n" ++
    scatterFcts ++
    "\n1 # we set the scattering dimension names\n" ++
    (concat $ intersperse " " scatterNames) ++ "\n\n"    
    where
    scatterFcts = 
	-- verbose "scatter functions for descriptions" $
	genScatterFcts ccDS clData
    numTotalStmts = 
	(length $ comps ++ sends ++ recvs ++ comms)
    
genScatterFctsDescriptions ccDS clData@CloogRPMappingData {} = 
    verbose "genScatterFctsDescriptions (CloogRPMappingData)" $
    "# --------------------- SCATTERING --------------------\n" ++
    (show numRPMappingStmts) ++ " # Scattering function\n" ++
    scatterFcts ++
    "\n1 # we set the scattering dimension names\n" ++
    scatterNames ++ "\n\n"    
    where    
    numRPMappingStmts = if ccCyclicMapping ccDS then 1 else 2
    scatterFcts = genScatterFcts ccDS clData
    scatterNames = 
	if ccCyclicMapping ccDS
	then rpNames
	else helperDim_t ++ " " ++ rpNames

    rpNames = 
	concat $ intersperse " " $
        [rp_t ++ show i | i <- [1.. getRPdim ccDS]]

createCompStmts :: CommCodeDS -> [CloogStatement]
createCompStmts ccDS = verbose "createCompStmts" $
		       map (createCompStmtHelper ccDS) $ getAllTStatementIDs ccDS

createCompStmtHelper :: CommCodeDS -> TStatementID -> CloogStatement
createCompStmtHelper ccDS trStmtID = 
    verbose "createCompStmtHelper" $
    myTrace (showIS2VisualPolylib projectedIS (nVars projectedIS)) $
    CloogStatement { cloogStmtIS = [projectedIS],
		     cloogStmtInfo = 
		     ComputeStmt { siTrStmtNr = trStmtID }
		   }
    where
    projectedIS = 
	-- assert (not $ areIneqsInfeasible compIS) $
	projectToVarsMC compIS projectedVars
    compIS = compStmtIneqSystem ccDS trStmtID
    tileNrDim = getTilingDim ccDS
    isDim = getISPCdim ccDS trStmtID
    tispcDim = getCoordsDim ccDS trStmtID

--  tileNrDims   |  tispcDims  |  isDims    |   params | const
--   time | proc | time | proc | time | proc|          |   

    -- project to all vars except TISPC (18.10.2005)
    projectedVars = 
	[0..tileNrDim-1] ++ 
	-- tispc wegprojeziert
	[(tileNrDim+tispcDim) .. 
	 (tileNrDim+tispcDim+isDim-1)]
	

---
compStmtIneqSystem :: CommCodeDS -> TStatementID -> IneqSystem
compStmtIneqSystem ccDS trStmtID = verbose "compStmtIneqSystem" $
				   if null assertions
				   then fullIS
				   else error ("compStmtIneqSystem: " ++ 
					       assertions)

    where
    assertions = assertCoords -- ++ assertFeasible

    assertCoords = if allEq [tileCoordsDim, tispcDim]
		   then ""
		   else "tileCoordsDim /= tispcDim"

    assertFeasible = if not $ areIneqsInfeasible fullIS
		     then ""
		     else ("compStmtIneqSystem: fullIS is infeasible! (" ++
			   "TStatementID: " ++ show trStmtID)

    useTiling = tiOptUseTiling $ ccTilingOptions ccDS
    stmtID = getStmtID ccDS trStmtID
    trafoFct = getTrafoFct ccDS trStmtID
    trc = getTRCis ccDS trStmtID
    is = assert (isDim == nVars trc) $ getISPC ccDS trStmtID

    isDim = getISPCdim ccDS trStmtID
    tispcDim = getCoordsDim ccDS trStmtID
    numParams = nParams trafoFct
    totalVars = isDim + tispcDim + tileNrDim

    (trafo_v, trafo_p, trafo_c) = 
	myTrace ("trStmt: " ++ show trStmtID ++ ", stmtNr: " ++ show (getStmtID ccDS trStmtID) ++ ", trafoFct: " ++ show trafoFct) $
	partsVPCih trafoFct
    (is_v, is_p, is_c) = partsVPC is
    (trc_v, trc_p, trc_c) = partsVPC trc
    (tile_v, tile_p, tile_c) = 
	partsVPC tilingIS

    emptyTStatement = 
	((integerToID (-1), integerToID (-1),
	  -- verbose "emptyIS" $
	  ineqSystem ((tispcDim, numParams), 
		      zeroMatrix 0 (tispcDim + numParams + 1)))
	) :: TStatement	 
    (tilingIS, [tileNrDim, tileCoordsDim]) = 
	if useTiling
	then verbose "tiling" $
	     getTiles (ccTilingOptions ccDS) 
			  [SrcTile, SrcCoord] emptyTStatement
	else (emptyIS, [0, tispcDim])
    emptyIS = ineqSystem ((0,0), zeroMatrix 0 0)

--  tileNrDims   |  tispcDims  |  isDims    |   params | const
--   time | proc | time | proc | time | proc|          |   

    is_M = 
	-- verbose "is_M" $
	zeroMatrix (nRows is_v) (tileNrDim + tispcDim) `conc`
	is_v `conc` is_p `conc` is_c
    
    stMappingEq_M = 
	-- verbose "stMappingEq_M" $
	stMapping_M `join` (-stMapping_M)
    stMapping_M = 
	myTrace ("tispcDim: " ++ (show tispcDim) ++ " tileCoordsDim:" ++ (show tileCoordsDim)) $
	zeroMatrix (nRows trafo_v) tileNrDim `conc`
        (-(unitMatrix (nRows trafo_v) (tispcDim))) `conc`
	trafo_v `conc` trafo_p `conc` trafo_c

    tiling_M = 
	-- verbose "tiling_M" $
	tile_v `conc`
	zeroMatrix (nRows tile_v) isDim `conc` 
	tile_p `conc` tile_c

    trc_M = 
	-- verbose "trc_M" $
	zeroMatrix (nRows trc_v) (tileNrDim + tispcDim) `conc` 
	trc_v `conc` trc_p `conc` trc_c
    
    raw_M = 
	-- verbose "raw_M" $
	myMultijoin1 $ [ is_M, stMappingEq_M, trc_M ] ++
		       if useTiling then [ tiling_M ] else []

    fullIS = ineqSystem ((totalVars, numParams), raw_M)

createSendStmts :: CommCodeDS -> [CloogStatement]
createSendStmts ccDS = verbose "createSendStmts" $
		       map createSendStmt $ trDeps
    where
    trDeps = verbose "trDeps" $ 
	     ccTrDeps ccDS
    createSendStmt :: TrDependence -> CloogStatement
    createSendStmt trDep =
	verbose ("createSendStmt: trDepID=" ++ (show (integerFromID $ trDepID trDep))) $
	CloogStatement
	{ cloogStmtIS = sendIS,
	  cloogStmtInfo = 
	    PrepareBufferStmt { siCommSide = SendSide
			      , siTrDepNr = trDepID trDep
			      , siArrayNr = accArrayID $ srcAcc
			      , siSrcCoordsDim = srcCoordsDim
			      , siDestCoordsDim = destCoordsDim
			      }
	} 
	where
	sendIS = map (finalSendReceiveIS ccDS baseIS trDep) [1..rpDim]
	(baseIS, (srcCoordsDim, destCoordsDim)) = 
	    sendReceiveBaseIS ccDS trDep
	srcAcc = getSrcAcc ccDS $ trDepID trDep
	rpDim = getRPdim ccDS

finalSendReceiveIS :: CommCodeDS -> IneqSystem -> TrDependence 
		   -> Integer -- ^ @currProc@ the number of the current real processor
		   -> IneqSystem
finalSendReceiveIS ccDS baseIS trDep currProc = 
    verbose "finalSendReceiveIS" $
    myTrace ("commConstraints_M: " ++ show commConstraints_M) $
    finalIS
    where
    spaceTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
    tileDim = getTilingDim ccDS
    timeTileDim = getTimeTileDim ccDS
    procTileDim = getProcTileDim ccDS
    schedDim = getSchedDim ccDS
    placeDim = getPlaceDim ccDS
    srcTrStmtID = trDepSrcStmtID trDep
    destTrStmtID = trDepDestStmtID trDep
    srcCoordsDim = getCoordsDim ccDS srcTrStmtID
    destCoordsDim = getCoordsDim ccDS destTrStmtID
    srcAddDim = srcCoordsDim - getCommonCoordsDim ccDS
    destAddDim = 
        if spaceTiling 
        then 0 
        else destCoordsDim - getCommonCoordsDim ccDS
    srcISPCdim = getISPCdim ccDS srcTrStmtID
    destISPCdim = 
	if spaceTiling 
	then 0 
	else getISPCdim ccDS destTrStmtID

    destDim = if spaceTiling then tileDim else destCoordsDim
    destProcDim = if spaceTiling then procTileDim else destCoordsDim - schedDim

    rpDim = getRPdim ccDS
    remainProcs = rpDim - currProc

    params = getParamsDim ccDS

    numTotalVars = 
	tileDim + srcCoordsDim + destProcDim + srcISPCdim + destISPCdim
    projectedVars = 
	([0..tileDim-1] ++ 
	 -- srcTISPC weg: [tileDim..tileDim+srcCoordsDim-1] ++
	[(tileDim+srcCoordsDim) .. 
	 (tileDim+srcCoordsDim+destProcDim+srcISPCdim+destISPCdim-1)]
	)

    finalIS = 
	-- assert (not $ areIneqsInfeasible constraintIS) $
	projectToVarsMC constraintIS projectedVars

    constraintIS = (ineqSystem ((numTotalVars, params), 
				commConstraints_M))

    commConstraints_M = 
        verbose "commConstraints_M" $
{--
        trace ("tileDim: " ++ (show tileDim) ++ 
               ", srcCoordsDim: " ++ (show srcCoordsDim) ++
               ", destProcDim: " ++ (show destProcDim) ++
               ", srcISPCdim: " ++ (show srcISPCdim) ++
               ", destISPCdim: " ++ (show destISPCdim)) $
--}
	myMultijoin1 [base_M, 
		      commConstraintsIneq_M,
		      commConstraintsEqs_M, 
		      -commConstraintsEqs_M]
	 
    base_M = 
        verbose "base_M" $
        getMatrix baseIS
    commConstraintsIneq_M = 
        verbose "commConstraintsIneq_M" $
	if spaceTiling
	--  -srcProcTile + destProcTile - 1 >= 0
	then (zeroMatrix 1 (timeTileDim + (currProc-1)) `conc`
	      (-(unitMatrix 1 1)) `conc`
	      zeroMatrix 1 (remainProcs + srcCoordsDim + (currProc-1)) `conc`
	      (unitMatrix 1 1) `conc`	
	      zeroMatrix 1 (remainProcs + destAddDim + srcISPCdim + destISPCdim + 
			    params) `conc`
	      (-(unitMatrix 1 1))
	     )
	else (zeroMatrix 1 (tileDim + schedDim + (currProc-1)) `conc`
	      (-(unitMatrix 1 1)) `conc`
	      zeroMatrix 1 (remainProcs + srcAddDim + (currProc-1)) `conc`
	      (unitMatrix 1 1) `conc`	
	      zeroMatrix 1 (remainProcs + destAddDim + srcISPCdim + destISPCdim + 
			    params) `conc`
	      (-(unitMatrix 1 1))
	     )
	
    commConstraintsEqs_M =
        verbose "commConstraintsEqs_M" $
	if spaceTiling 
	-- srcProcTile == destProcTile
	then (zeroMatrix remainProcs (timeTileDim + currProc) `conc`
	      (-(unitMatrix remainProcs remainProcs)) `conc`
	      zeroMatrix remainProcs (srcCoordsDim + currProc) `conc`
	      (unitMatrix remainProcs remainProcs) `conc`	
	      zeroMatrix remainProcs (destAddDim + srcISPCdim + destISPCdim + 
				      params) `conc`
	      (columnMatrix (genericReplicate remainProcs (0%1)))
	     )
	else (zeroMatrix remainProcs (tileDim + schedDim + currProc) `conc`
	      (-(unitMatrix remainProcs remainProcs)) `conc`
	      zeroMatrix remainProcs (srcAddDim + currProc) `conc`
	      (unitMatrix remainProcs remainProcs) `conc`	
	      zeroMatrix remainProcs (destAddDim + srcISPCdim + destISPCdim + 
				      params) `conc`
	      (columnMatrix (genericReplicate remainProcs (0%1)))
	     )
    

sendReceiveBaseIS :: CommCodeDS -> TrDependence 
		  -> (IneqSystem, (Integer, Integer))
sendReceiveBaseIS ccDS trDep =
    myTrace ("srcDomain: " ++ (show srcDomain) ++"\n " ++ "destDomain: " ++ (show destDomain) ++ "\n " ++ "depHTrafo: " ++ (show (depHTrafo dep))) $
    verbose "sendReceiveBaseIS" $
    myTrace (show smallSendRecvIS) $
    if null assertions 
    then (smallSendRecvIS,
	  (srcCoordsDim, destCoordsDim))
    else trace ("warning: " ++ assertions) $
	 (smallSendRecvIS,
	  (srcCoordsDim, destCoordsDim))
    where

    origDepID = trDepOrigID trDep
    dep = 
	fromMaybe (error "getDepDomains: origDepID not found")
		  (find (\d -> depID d == origDepID) (ccDeps ccDS))

    assertions = []
    -- assertFeasible

    assertFeasible = 
	if areIneqsInfeasible sendRecvIS
	then ("sendRecvIS is infeasible! (" ++
	      "TrDependenceID: " ++ show (trDepID trDep) ++
	      ", is: " ++ show sendRecvIS)
	else ""

    tiOpts = ccTilingOptions ccDS
    useTiling = tiOptUseTiling tiOpts
    spaceTiling = tiOptSpaceTiling tiOpts
    latVects = tiOptLatVects tiOpts
    tiledDims = tiOptTiledDims tiOpts
    dummyLattices = [vector [1%1]]

    (srcDomain, destDomain) = 
	getDepDomains ccDS trDep
    srcTrStmtID = trDepSrcStmtID trDep
    destTrStmtID = trDepDestStmtID trDep
    srcTrafoFct = getTrafoFct ccDS srcTrStmtID
    destTrafoFct = getTrafoFct ccDS destTrStmtID
    (srcTrafo_v, srcTrafo_p, srcTrafo_c) = partsVPCih srcTrafoFct
    (destTrafo_v, destTrafo_p, destTrafo_c) = partsVPCih destTrafoFct
				
    srcTRC = getTRCis ccDS srcTrStmtID
    destTRC = getTRCis ccDS destTrStmtID
    (srcTRC_v, srcTRC_p, srcTRC_c) = partsVPC srcTRC
    (destTRC_v, destTRC_p, destTRC_c) = partsVPC destTRC
  
    srcISdim = nColumns srcTrafo_v
 
    -- if we use spaceTiling then we don't have destCoords, so
    -- we don't need the space-time-trafo:
    destISdim = if spaceTiling then 0 else nColumns destTrafo_v

    tileDim = getTilingDim ccDS
    destDim = if spaceTiling then tileDim else destCoordsDim
    destTimeDim = 
	if spaceTiling 
	then getTimeTileDim ccDS
	else getSchedDim ccDS
    paramsDim = getParamsDim ccDS
    totalVarsDim =         
	tileDim + srcCoordsDim + destDim + srcISdim + destISdim

    (base_v, base_p, base_c) = partsVPC baseIS

    -- project to all vars except destination-time...
    -- ...and srcTISPC (18.10.2005) resulting in:
    -- tileDim | destProcDim | srcISPC | destISPC
    projectedVars = 
	[0..tileDim-1] ++ 
	[tileDim..tileDim+srcCoordsDim-1] ++
	-- destTime weg: [(tileDim+srcCoordsDim)..
	--		  (tileDim+srcCoordsDim+destTimeDim-1)] ++
	[(tileDim+srcCoordsDim+destTimeDim) .. 
	 (tileDim+srcCoordsDim+destDim+srcISdim+destISdim-1)]

    smallSendRecvIS =
	projectToVarsMC sendRecvIS projectedVars

    sendRecvIS = 
	ineqSystem ((totalVarsDim, 
		     paramsDim), sendRecv_M)

    sendRecv_M = 
	myMultijoin1 ([base_M, srcTrafoEq_M, srcTRC_M] ++ 
		      (if spaceTiling 
		       then [] 
		       else [ destTrafoEq_M, destTRC_M ]))

    srcTrafoEq_M = srcTrafoIs_M `join` (-srcTrafoIs_M)
    destTrafoEq_M = destTrafoIs_M `join` (-destTrafoIs_M)

    base_M = 
	(base_v `conc` 
	 zeroMatrix (nRows base_v) (srcISdim + destISdim) `conc`
	 base_p `conc` base_c)

    srcTrafoIs_M = 
	(zeroMatrix srcCoordsDim tileDim `conc`
	 (-(unitMatrix srcCoordsDim srcCoordsDim)) `conc`
	 zeroMatrix srcCoordsDim destDim `conc`
	 srcTrafo_v `conc`
	 zeroMatrix srcCoordsDim destISdim `conc`
	 srcTrafo_p `conc` srcTrafo_c)

    srcTRC_M =
	(zeroMatrix (nRows srcTRC) (tileDim + srcCoordsDim + destDim) `conc`
	 srcTRC_v `conc`
	 zeroMatrix (nRows srcTRC) destISdim `conc`
	 srcTRC_p `conc` srcTRC_c)

    destTrafoIs_M = 
	(zeroMatrix destCoordsDim (tileDim + srcCoordsDim) `conc`
	 (-(unitMatrix destCoordsDim destCoordsDim)) `conc`
	 zeroMatrix destCoordsDim srcISdim `conc`
	 destTrafo_v `conc` destTrafo_p `conc` destTrafo_c)
	
    destTRC_M =
	(zeroMatrix (nRows destTRC) (tileDim + srcCoordsDim + destDim) `conc`
	 zeroMatrix (nRows destTRC) srcISdim `conc`
	 destTRC_v `conc` srcTRC_p `conc` srcTRC_c)

    baseIS = if useTiling then tilingIS else nonTilingIS

    tilingInfo = 
	((if useTiling then [SrcTile] else []) ++
	 [SrcCoord] ++
	 (if spaceTiling then [DestTile] else [DestCoord]))

--    (_, [srcCoordsDim, destCoordsDim]) =
--	defaultTile dummyLattices (trDepRelation trDep) [SrcCoord, DestCoord]
    srcCoordsDim = getCoordsDim ccDS srcTrStmtID
    destCoordsDim = getCoordsDim ccDS destTrStmtID

    (nonTilingIS, _) = 
	defaultTile dummyLattices [0] (trDepRelation trDep) tilingInfo
    (tilingIS, _) = 
	defaultTile latVects tiledDims (trDepRelation trDep) tilingInfo

createRecvStmts :: CommCodeDS -> [CloogStatement]
createRecvStmts ccDS = verbose "createRecvStmts" $
		       map createRecvStmt $ ccTrDeps ccDS
    where
    createRecvStmt :: TrDependence -> CloogStatement
    createRecvStmt trDep =     
	CloogStatement
	{ cloogStmtIS = recvIS,
	  cloogStmtInfo = 
	    PrepareBufferStmt { siCommSide = RecvSide
			      , siTrDepNr = trDepID trDep
			      , siArrayNr = accArrayID $ srcAcc
			      , siSrcCoordsDim = srcCoordsDim
			      , siDestCoordsDim = destCoordsDim
			      }
	} 
	where
	recvIS = map (finalSendReceiveIS ccDS baseIS trDep) [1..rpDim]
	(baseIS, (srcCoordsDim, destCoordsDim)) = sendReceiveBaseIS ccDS trDep
	srcAcc = getSrcAcc ccDS $ trDepID trDep
	rpDim = getRPdim ccDS

createCommStmts :: CommCodeDS -> [CloogStatement]
createCommStmts ccDS = 
    verbose "createCommStmts" $		          
    [CloogStatement { cloogStmtIS = 
		        map createCommStmtIS (getAllTStatementIDs ccDS)
		    , cloogStmtInfo = CommunicateStmt
		    }
    ]
    where
    createCommStmtIS :: TStatementID -> IneqSystem
    createCommStmtIS trStmtID = 
	myTrace ("globalIS: " ++ show globalIS ++ "\nfinalIS: " ++ show finalIS) $
        myTrace (showIS2VisualPolylib finalIS (fromInteger $ nVars finalIS)) $    
	-- assert (not $ areIneqsInfeasible finalIS) $
	finalIS
	where
	timeTiling = tiOptTimeTiling $ ccTilingOptions ccDS
	spaceTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
	useTiling = tiOptUseTiling $ ccTilingOptions ccDS
	skewingFct = getSkewingFct ccDS trStmtID
	compIS = compStmtIneqSystem ccDS trStmtID
	isDim = getISPCdim ccDS trStmtID
	-- globalTimeDim = getGlobalTimeDim ccDS
	rpDim = getRPdim ccDS
	coordsTimeDim = getSchedDim ccDS
        timeTileDim = getTimeTileDim ccDS
        timeDim = if timeTiling then timeTileDim else coordsTimeDim
        timeTileUsedVect = 
          genericTake timeTileDim $ tiOptTilingUsedVect (ccTilingOptions ccDS)

        globalTiledTimeDim = fromIntegral $ head $ elemIndices True timeTileUsedVect

        -- our compStmtIS looks like: 
        --  tileNrDims   |  tispcDims  |  isDims    |   params | const
        --   time | proc | time | proc | time | proc|          |   
	-- so we have to project to:
	globalDims = timeDims ++ rpDims
	timeDims = -- [0..timeDim-1] -- no matter wether tiling or not
                   if timeTiling
                   then [globalTiledTimeDim]
                   else [0..timeDim-1]
	rpDims = if spaceTiling 
		 then [timeTileDim .. timeTileDim+rpDim-1]
		 else [timeTileDim+coordsTimeDim .. 
		       timeTileDim+coordsTimeDim+rpDim-1]

	globalIS = 
	    -- assert (not $ areIneqsInfeasible compIS) $
	    projectToVarsMC compIS globalDims

	skewedIS = image skewingFct globalIS 

        finalIS = 
	    projectToVarsMC (if timeTiling 
			     then skewedIS
			     else globalIS) timeDims

-- skewing function only for communication statement:
getSkewingFct :: CommCodeDS -> TStatementID -> AffineFunction
getSkewingFct ccDS trStmtID = 
    verbose "getSkewingFct" $
    myTrace ("skewingFct: " ++ (show $ affineFunction ((globalTimeDim + rpDim, paramsDim), skew_M) Homogeneous)) $
    affineFunction ((globalTimeDim + rpDim, paramsDim), skew_M) Homogeneous
    where
    tiOpts = ccTilingOptions ccDS
    timeTileDim = getTimeTileDim ccDS
    timeTileUsedVect = genericTake timeTileDim $ tiOptTilingUsedVect tiOpts
    globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS
    paramsDim = getParamsDim ccDS
    rpPart_M = 
	-- verbose "rpPart_M" $
        rowMatrix $ genericReplicate rpDim (-1%1)
	-- getSkewingMatrixRPpart rpDim timeTileUsedVect

    skew_M = 
	myTrace ("glTime: " ++ (show globalTimeDim) ++ ", rpDim: " ++ (show rpDim)) $
	myMultijoin1
	[ (unitMatrix globalTimeDim globalTimeDim `conc` 
	   (-rpPart_M) `conc`
	   zeroMatrix globalTimeDim (paramsDim + 1)),
	  (zeroMatrix rpDim globalTimeDim `conc`
	   unitMatrix rpDim rpDim `conc`
	   zeroMatrix rpDim (paramsDim + 1)),
	  (zeroMatrix (paramsDim+1) (globalTimeDim + rpDim) `conc`
	   unitMatrix (paramsDim+1) (paramsDim+1))
	]
    

-- TODO:?
createRPMappingStmts :: CommCodeDS -> [CloogStatement]
createRPMappingStmts ccDS = 
    verbose "createRPMappingStmts" $
    [CloogStatement { cloogStmtIS = rpMapStmtIS
		    , cloogStmtInfo = RPMappingStmt
		    }
    ]
    where
    trIDs = getAllTStatementIDs ccDS
    rpMapStmtIS = map genRPMapStmtIS (getAllTStatementIDs ccDS)
   
    genRPMapStmtIS trID = projectToCommonDims ccDS trID compStmtIS
	where compStmtIS = compStmtIneqSystem ccDS trID
		  

-- TODO:?
projectToCommonDims :: CommCodeDS -> TStatementID -> IneqSystem -> IneqSystem
projectToCommonDims ccDS trStmtID tileIS = 
    verbose "projectToCommonDims" $
    myTrace ("project to: " ++ (show projectedDims)) $
    projectToVarsMC tileIS projectedDims
    where
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS
    -- isDim = getISPCdim ccDS trStmtID
    commonCoordsDim = getCommonCoordsDim ccDS
    ispcDim = getISPCdim ccDS trStmtID
    tilingDim = 
	if useTiling 
	then genericLength $ tiOptLatVects $ ccTilingOptions ccDS 
	else 0
    projectedDims = [0 .. tilingDim + commonCoordsDim - 1]    
    

writeDeclarations :: CommCodeDS -> IO ()
writeDeclarations ccDS =
    do
    let loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
	suffix = ".decls"
	useWrapper = ccUseWrapper ccDS
	tmpIter_t = "tmpIter"
        iteratorNames = getIteratorNames ccDS
	scatterNames = getScatterNames ccDS
	rpDim = getRPdim ccDS
        loopVarDecls = 
	    "int " ++ concat (intersperse ", " (scatterNames ++ 
						iteratorNames)) ++ ";\n"
	mpiDecls = 
	    unlines $
	    [ ""
	    , "#include <mpi.h>"
	    , "int " ++ mpiSize_t ++ ", " ++ mpiRank_t ++ ";"
	    , ""
	    ]

        allToAllDecls =
	    unlines $
	    [ ""
	    , "inline void my_Alltoall(int *sendbuf, int sendcount, MPI_Datatype sendtype,"
            , "                        int *recvbuf, int recvcnt, MPI_Datatype recvtype,"
            , "                        MPI_Comm comm) {"
	    , "  using namespace std;"
	    , "  //for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {"
	    , "    //cerr << sendbuf[i];"
	    , "  //}"
	    , "  MPI_Alltoall(sendbuf, sendcount, sendtype, recvbuf, recvcnt, recvtype, comm);"
	    , "  //for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {"
	    , "    //cerr << recvbuf[i];"
	    , "  //}"
	    , "}"
	    , ""
	    , ""
	    , "inline void my_Alltoallv(double *sendbuf, int* sendcounts, int* sdispls,"
	    , "                         MPI_Datatype sendtype,"
            , "                         double *recvbuf, int* recvcnts, int* rdispls,"
	    , "                         MPI_Datatype recvtype,"
            , "                         MPI_Comm comm) {"
	    , "  using namespace std;"
	    , "  //for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {"
	    , "    //cerr << \"rank=\" << " ++ mpiRank_t ++ " << \" send: \";"
	    , "    //for (int j = 0; j < sendcounts[i]; j++) {"
	    , "      //cerr << sendbuf[sdispls[i]+j] << \" \";"
	    , "    //}"
	    , "    //cerr << ::std::endl;"
	    , "  //}"
	    , "  MPI_Alltoallv(sendbuf, sendcounts, sdispls, sendtype, recvbuf, recvcnts, rdispls,"
	    , "                recvtype, comm);"
	    , "  //for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {"
	    , "    //cerr << \"rank=\" << " ++ mpiRank_t ++ " << \" recv: \";"
	    , "    //for (int j = 0; j < recvcnts[i]; j++) {"
	    , "      //cerr << recvbuf[rdispls[i]+j] << \" \";"
	    , "    //}"
	    , "    //cerr << endl;"
	    , "  //}"
	    , "}"
	    , ""
	    , "inline void printVector(::std::vector<double>& vect) {"
	    , "  ::std::cerr << \"printVect: \";"
	    , "  for (size_t i=0; i < vect.size(); i++) {"
	    , "    ::std::cerr << vect[i] << \"#\";"
	    , "  }"
	    , "  ::std::cerr << ::std::endl;"
	    , "}"
	    , ""
	    , "inline void break_here() {"
	    , "  ::std::cerr << \"break\" << ::std::flush;"
	    , "}"
	    , ""
	    ]
 
        useWrapperDecls =
	    [ "typedef ::std::map<NTupel, int> RPMAP;"
	    , "RPMAP " ++ rpMap_t ++ ";"
	    ]
       
        noWrapperDecls = 
	    [ "typedef ::std::pair<int, int> RP_BufID_PAIR;"
	    , "typedef ::std::map<NTupel, RP_BufID_PAIR> RPMAP;" 
	    , "RPMAP " ++ rpMap_t ++ ";"
	    , "::std::vector<int> " ++ bufferID2RP_t ++ ";"
	    , "RP_BufID_PAIR " ++ rp_bufID_pair_t ++ ";"
	    ]

        rpMapDecls = 
	    unlines $
	    [ "#include <map>"
	    , "#include <vector>"
	    , "#include <utility>"
	    , "#include <numeric>"
	    , "#include <iostream>"
	    , "#include \"NTupel.h\""
	    ] ++
            (if (useWrapper || not (ccMergeReceive ccDS))
	     then useWrapperDecls 
	     else noWrapperDecls) ++
	    [ "int " ++ helperDim_t ++ " = 0; "
	    , "int " ++ otherRP_t ++ " = 0; "
	    , "int " ++ bufferID_t ++ " = 0; "
	    , "::std::vector< ::std::vector<double> > " ++ sendVector_t ++ ";"
	    , "int " ++ sendVectorLength_t ++ " = 0;"
            , (if (useWrapper || not (ccMergeReceive ccDS))
	       then "inline int " ++ lookupRP_t ++ "("
	       else "inline RP_BufID_PAIR " ++ lookupRP_t ++ "(")
	    , (concat $ 
	       intersperse ", " ["int rp" ++ show i | i <- [1..rpDim]]) ++
	       ") {\n" ++
	      
	       --   int tmpArray[] = { rp1, rp2 };
	       "  int " ++ tmpArray_t ++ "[] = {" ++
	       (concat $ intersperse ", " ["rp" ++ show i | i <- [1..rpDim]]) ++ "};\n" ++
											  
               --   NTupel tmpVector(tmpArray, 2); 
	       "  NTupel " ++ tmpIndexVector_t ++ "(" ++ 
	       tmpArray_t ++ ", " ++ show rpDim ++ ");\n" ++

               --   RPMAP::iterator tmpIter = rpMap.find(tmpVector);
	       "  RPMAP::iterator " ++ tmpIter_t ++ " = " ++ rpMap_t ++ 
	       ".find(" ++ tmpIndexVector_t ++ ");\n" ++

               --   if (tmpIter != rpMap.end()) {
	       "  if (" ++ tmpIter_t ++ " != " ++ rpMap_t ++ ".end()) { \n" ++

	       --     return (*tmpIter).second;
	       "    return (*" ++ tmpIter_t ++ ").second;\n" ++

	       -- } else { ::std::cerr << "key not found!"; abort; }
	       "  } else { ::std::cerr << \"key not found!\"; abort(); }\n" ++

               -- }
	       "}"
	    ]
        declarations = "/* created: " ++ getTime ++ " */\n" ++
	               loopVarDecls ++ 
		       if ccUseCommunication ccDS
		       then rpMapDecls ++ mpiDecls ++ allToAllDecls
		       else []
    writeFile (loopoFile ++ suffix) declarations 

writeInitialisations :: CommCodeDS -> IO ()
writeInitialisations ccDS =
    do
    let loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
	suffix = ".inits"
	mpiInits = 
	    unlines $
	    [ "MPI_Init(&" ++ argc_t ++ ", &" ++ argv_t ++ ");"
	    , "MPI_Comm_size(MPI_COMM_WORLD, &" ++ mpiSize_t ++ ");"
	    , "MPI_Comm_rank(MPI_COMM_WORLD, &" ++ mpiRank_t ++ ");"
	    , "::std::vector<int> " ++ scounts_t ++ "(" ++ mpiSize_t ++ ", 0);"
	    , "::std::vector<int> " ++ rcounts_t ++ "(" ++ mpiSize_t ++ ", 0);"
	    , "::std::vector<int> " ++ sdispls_t ++ "(" ++ mpiSize_t ++ ", 0);"
	    , "::std::vector<int> " ++ rdispls_t ++ "(" ++ mpiSize_t ++ ", 0);"
	    , "int " ++ recvIndex_t ++ "[" ++ mpiSize_t ++ "];"
	    , "::std::vector<double> " ++ sendBuffer_t ++ "(1);"
	    , "::std::vector<double> " ++ recvBuffer_t ++ "(1);"
	    , "for (int i = 0; i < " ++ mpiSize_t ++ "; i++) {"
	    , "  " ++ recvIndex_t ++ "[i] = 0;"
	    , (if (ccUseWrapper ccDS || not (ccMergeReceive ccDS))
	       then sendVector_t ++ ".push_back(::std::vector<double>());"
	       else [])
	    , "}"
	    ]
        inits = "/* created: " ++ getTime ++ " */\n" ++
	        if ccUseCommunication ccDS
		then mpiInits
		else []
    writeFile (loopoFile ++ suffix) inits 
    

writeDefinesFile :: CommCodeDS -> CloogData -> IO ()
writeDefinesFile ccDS CloogMainData 
		     { cloogDataCompStmts = compStmts
		     , cloogDataSendStmts = sendStmts
		     , cloogDataRecvStmts = recvStmts
		     , cloogDataCommStmts = commStmts
		     , cloogDataScatterNames = scatterNames
		     } = 
    let loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
	suffix = ".defines" 
    in 
    do origStmtsCode <- getOrigStatementsCode ccDS
       let numCompStmts = genericLength compStmts
	   numSendStmts = genericLength sendStmts
	   numRecvStmts = genericLength recvStmts
	   numCommStmts = genericLength commStmts
	   sendStmtsNrs = 
	       [(numCompStmts + 1) .. (numCompStmts + numSendStmts)]
	   recvStmtsNrs = 
	       [(numCompStmts + numSendStmts + 1) ..
		(numCompStmts + numSendStmts + numRecvStmts)]
	   commStmtsNr = (numCompStmts + numSendStmts + numRecvStmts + 1)
			   
	   compStmtsDefines = 
	       genCompStmtsDefines ccDS origStmtsCode compStmts
	   
           sendStmtsDefines = 
	       unlines $ 
	       zipWith (genSendStmtDefines ccDS) sendStmtsNrs sendStmts
	   recvStmtsDefines = 
	       unlines $
	       zipWith (genRecvStmtDefines ccDS) recvStmtsNrs recvStmts
	   commStmtsDefines =
	       -- all commStmts have the same define (-> only one cloogStmtNr)
	       if ccUseCommunication ccDS
	       then genCommStmtDefines ccDS commStmtsNr (head commStmts)
	       else []

	   defines = 
	       "/* created: " ++ getTime ++ " */\n" ++
	       unlines [compStmtsDefines, sendStmtsDefines, 
			recvStmtsDefines, commStmtsDefines]
       writeFile (loopoFile ++ suffix) defines
       return ()

writeDefinesFile ccDS CloogRPMappingData 
		     { cloogDataRPMappingStmts = rpMapStmts } =
    do
    let definesFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS ++ ".rpDefines"
	numRPMapDefines = if ccCyclicMapping ccDS
			  then 1
			  else 2 -- additional enumeration of 
				 -- index space for block mapping

        defines = 
	    map (genRPMapStmtDefines ccDS) [1..numRPMapDefines]
    writeFile definesFile (unlines defines)
    return ()


defineHeader :: CommCodeDS -> Integer -> CloogStatement -> String
defineHeader ccDS cloogStmtNr cloogStmt = 
    define_t ++ " " ++ cloogStmt_t ++ (show cloogStmtNr) ++ "(" ++
    concat (intersperse ", " usedIteratorNames) ++ ") "
    where
    usedIteratorNames = getUsedIteratorNames ccDS cloogStmt

getUsedIteratorNames ccDS CloogStatement
			 { cloogStmtIS = domainIS
			 } = genericTake (nVars (head domainIS)) iteratorNames
    where
    iteratorNames = getIteratorNames ccDS


genCompStmtsDefines :: CommCodeDS -> [(StatementID, String)] -> 
		       [CloogStatement] -> String
genCompStmtsDefines ccDS origStatementsCode compStmts = 
    verbose "compStmtsDefines" $
    unlines compStmtsDefines
    where
    allTrStmtIDs = getAllTStatementIDs ccDS
    compStmtIndexMapping = map (genCompStmtIndexMappingCode ccDS) allTrStmtIDs
    compStmtsDefines = zipWith oneCompStmtDefine [1..] compStmts

    oneCompStmtDefine :: Integer -> CloogStatement -> String
    oneCompStmtDefine cloogStmtNr cloogStmt = 
	let trStmtID = siTrStmtNr $ cloogStmtInfo cloogStmt
        in concat 
	       [ defineHeader ccDS cloogStmtNr cloogStmt
	       , genCompStmtIndexMappingCode ccDS trStmtID
	       , fromJust $ 
		 lookup (getStmtID ccDS trStmtID) origStatementsCode
	       ]

genSendStmtDefines :: CommCodeDS -> Integer -> CloogStatement -> String
genSendStmtDefines ccDS cloogStmtNr 
		   cloogStmt@CloogStatement
		             { cloogStmtInfo = 
			       stmtInfo@PrepareBufferStmt { siTrDepNr = trDepNr
							  , siArrayNr = arrayID
							  }
			     } = 
    (defineHeader ccDS cloogStmtNr cloogStmt ++
     getRP_t ++ " " ++ writeBuffer_t)
    where
    arrayName = 
	genericIndex (ccArrayNames ccDS) (integerFromID arrayID - 1)
    srcAcc = getSrcAcc ccDS trDepNr
    accDim = targetDimIH $ accIndexFct srcAcc
    getRP_t = getRPlookupText ccDS SendSide trDepNr
    writeBuffer_t =
        if (ccUseWrapper ccDS || ccMergeReceive ccDS)
        then pushData_t
        else -- wenn sofort am ende des alltoalls alle daten ausgepackt werden,
             -- ist ein if-statement noetig, um lokale Kommunikation 
             -- zu verhindern:
             ("if (" ++ mpiRank_t ++ " != " ++ otherRP_t ++ ") {" ++
              pushData_t ++ 
              " } ")
    pushData_t =
	(sendVector_t ++ "[" ++ 
	 (if (ccUseWrapper ccDS || not (ccMergeReceive ccDS))
	  then otherRP_t
	  else bufferID_t) ++ 
	 "].push_back(" ++ 
	 arrayName ++
	 concat ["[" ++ arrAcc_t ++ (show i) ++ "]" | i <- [1..accDim]] ++
	 ");")


genRecvStmtDefines :: CommCodeDS -> Integer -> CloogStatement -> String
genRecvStmtDefines ccDS cloogStmtNr 
		   cloogStmt@CloogStatement
			     { cloogStmtInfo = 
			         PrepareBufferStmt { siTrDepNr = trDepNr
						   , siArrayNr = arrayID
						   }
		   } =
    (defineHeader ccDS cloogStmtNr cloogStmt ++
     getRP_t ++ " " ++ readBuffer_t)
    where
    arrayName = 
	genericIndex (ccArrayNames ccDS) (integerFromID arrayID - 1)
    srcAcc = getSrcAcc ccDS trDepNr
    accDim = targetDimIH $ accIndexFct srcAcc
    getRP_t = getRPlookupText ccDS RecvSide trDepNr
    readBuffer_t =
        if (ccUseWrapper ccDS || ccMergeReceive ccDS)
        then unpackData_t
        else -- wenn sofort am ende des alltoalls alle daten ausgepackt werden,
             -- ist ein if-statement noetig, um lokale Kommunikation 
             -- zu verhindern:
             ("if (" ++ mpiRank_t ++ " != " ++ otherRP_t ++ ") {" ++
              unpackData_t ++ 
              " } ")
    unpackData_t =
	(arrayName ++
	 concat ["[" ++ arrAcc_t ++ (show i) ++ "]" | i <- [1..accDim]] ++
	 " = " ++
	 recvBuffer_t ++ "[" ++ rdispls_t ++ "[" ++ otherRP_t ++ "] + "  ++
	 recvIndex_t ++ "[" ++ otherRP_t ++ "]++];")


genCommStmtDefines :: CommCodeDS -> Integer -> CloogStatement -> String
genCommStmtDefines ccDS cloogStmtNr cloogStmt = 
    -- "#define Sxx(x, x, x,)"
    defineHeader ccDS cloogStmtNr cloogStmt ++

     -- sendBuffer.clear(); 
    (sendBuffer_t ++ ".clear(); " ++
 
     -- for(int tmpRP = 0; tmpRP < size; tmpRP++) {
     ("for(int " ++ tmpRP_t ++ " = 0; " ++ 
      tmpRP_t ++ " < " ++ mpiSize_t ++ "; " ++ tmpRP_t ++ "++) { ") ++
     -- int scount = 0;
     ("int " ++ scount_t ++ " = 0; ") ++
	 
     (if (ccUseWrapper ccDS || not (ccMergeReceive ccDS))
      then (-- copy(sendVector[tmpRP].begin(), sendVector[tmpRP].end(), 
            --      back_inserter(sendBuffer));
            ("copy(" ++ sendVector_t ++ "[" ++ tmpRP_t ++ "].begin(), " ++ 
	     sendVector_t ++ "[" ++ tmpRP_t ++ "].end(), " ++ 
	     "back_inserter(" ++ sendBuffer_t ++ ")); ") ++ 
            -- scount = sendVector[tmpRP].size();
            (scount_t ++ " = " ++ 
	     sendVector_t ++ "[" ++ tmpRP_t ++ "].size(); ") ++
            -- sendVector[tmpRP].clear();
            (sendVector_t ++ "[" ++ tmpRP_t ++ "].clear(); ")
	   )
      else noWrapperDefines) ++

     -- scounts[tmpRP] = scount;
     (scounts_t ++ "[" ++ tmpRP_t ++ "] = " ++ scount_t ++ "; ") ++ 
		  
     -- sdispls[tmpRP] = (tmpRP>0) ? sdispls[tmpRP-1] + scounts[tmpRP-1] : 0;
     (sdispls_t ++ "[" ++ tmpRP_t ++ "] = (" ++ tmpRP_t ++ ">0) ? " ++ 
      sdispls_t ++ "[" ++ tmpRP_t ++ "-1] + " ++ scounts_t ++ 
      "[" ++ tmpRP_t ++ "-1] : 0; ") ++ 
    
     -- rcounts[tmpRP] = 0;
     (rcounts_t ++ "[" ++ tmpRP_t ++ "] = 0; ") ++

     -- } // end tmpRP
     ("} ") ++

     -- MPI_Alltoall(&scounts[0], 1, MPI_INT, 
     -- 	     &rcounts[0], 1, MPI_INT, comm);
     (mpiAlltoall_t ++ "(&" ++ scounts_t ++ "[0], 1, " ++ 
      mpiInt_t ++ ", &" ++ rcounts_t ++ "[0], 1, " ++ 
      mpiInt_t ++ ", " ++ mpiComm_t ++ ");") ++

--     -- partial_sum(rcounts.begin(), rcounts.end(), rdispls.begin());
--     ("partial_sum(" ++ rcounts_t ++ ".begin(), " ++ rcounts_t ++ ".end(), " ++
--      rdispls_t ++ ".begin()); ") ++ 

     -- for(int i=0; i<size; i++) { 
     --   rdispls[i] = (i>0) ? rdispls[i-1] + rcounts[i-1] : 0;
     --   if(rcounts[i] > 0) { recvIndex[i] = 0; }
     -- };
     ("for(int i=0; i < " ++ mpiSize_t ++ "; i++) {" ++
      rdispls_t ++ "[i] = (i>0) ? " ++ rdispls_t ++ "[i-1] + " ++ 
      rcounts_t ++ "[i-1] : 0; " ++ 
      "if(" ++ rcounts_t ++ "[i] > 0) { " ++ recvIndex_t ++ "[i] = 0;} } ") ++

     -- recvBuffer.resize(rdispls[size-1] + rcounts[size-1]);
     (recvBuffer_t ++ ".resize(" ++ rdispls_t ++ "[" ++ mpiSize_t ++ "-1] + " ++
      rcounts_t ++ "[" ++ mpiSize_t ++ "-1]);") ++

      -- MPI_Alltoallv(&sendBuffer[0], &scounts[0], &sdispls[0], MPI_DOUBLE,
      --               &recvBuffer[0], &rcounts[0], &rdispls[0], MPI_DOUBLE, 
      --               comm);
      (mpiAlltoallv_t ++ "(&" ++ sendBuffer_t ++ "[0], &" ++ scounts_t ++ 
       "[0], &" ++ sdispls_t ++ "[0], " ++ mpiDouble_t ++ ", &" ++ 
       recvBuffer_t ++ "[0], &" ++ rcounts_t ++ "[0], &" ++ rdispls_t ++ 
       "[0], " ++ mpiDouble_t ++ ", " ++ mpiComm_t ++ ");"))
    where
    tmpRP_t = loopo_t ++ "tmpRP"
    tmpBufID_t = loopo_t ++ "tmpBufID"
    scount_t = loopo_t ++ "scount"
    noWrapperDefines = 
	(-- for(int tmpBufID = 0; tmpBufID < sendVectLength; tmpBufID++) {
         ("for(int " ++ tmpBufID_t ++ " = 0; " ++ 
	  tmpBufID_t ++ " < " ++ sendVectorLength_t ++ "; " ++ tmpBufID_t ++ 
	  "++) { ") ++
	 -- if(bufferID2RP[tmpBufID] == tmpRP) {
         ("if(" ++ bufferID2RP_t ++ "[" ++ tmpBufID_t ++ "] == " ++ 
	  tmpRP_t ++ ") { ") ++
	 -- copy(sendVector[tmpBufID].begin(), sendVector[tmpBufID].end(), 
         --      back_inserter(sendBuffer));
         ("copy(" ++ sendVector_t ++ "[" ++ tmpBufID_t ++ "].begin(), " ++ 
	  sendVector_t ++ "[" ++ tmpBufID_t ++ "].end(), " ++ 
	  "back_inserter(" ++ sendBuffer_t ++ ")); ") ++ 
         -- scount += sendVector[tmpBufID].size();
         (scount_t ++ " += " ++ 
	  sendVector_t ++ "[" ++ tmpBufID_t ++ "].size(); ") ++
         -- sendVector[tmpBufID].clear();
         (sendVector_t ++ "[" ++ tmpBufID_t ++ "].clear(); ") ++
         -- } // end if
	 ("} ") ++
	 -- } // end for tmpBufID
	 ("} ")
	)
    

genRPMapStmtDefines :: CommCodeDS -> Integer -> String
genRPMapStmtDefines ccDS cloogStmtNr =     
    verbose "genRPMapStmtDefines" $
    if ccUseWrapper ccDS
    then useWrapperDefines
    else if mergeReceive
         then (if cyclicMapping
	       then cyclicMappingDefines
	       else blockMappingDefines)
         else noMergeReceiveDefines
     where
     rpDim = getRPdim ccDS
     cyclicMapping = ccCyclicMapping ccDS
     mergeReceive = ccMergeReceive ccDS
     tmpArray_t = loopo_t ++ "tmpArray"
     tmpArrayDecl = 
	 -- int* tmpArray = new int[2]; 
	 ("int* " ++ tmpArray_t ++ " = new int[" ++ show rpDim ++ "]; " ++

         -- tmpArray[0] = rp1; tmpArray[1] = rp2;
	 concat [tmpArray_t ++ "[" ++ show (i-1) ++ "] = " ++ rp_t ++ show i ++ "; " |
		 i <- [1..rpDim]])
     tmpNTupel = 
	  ("NTupel(" ++ tmpArray_t ++ ", " ++ show rpDim ++ ")")

     header_t = define_t ++ " " ++ cloogStmt_t ++ show cloogStmtNr ++ "() "

     useWrapperDefines =
	 (header_t ++
	  tmpArrayDecl ++
	  rpMap_t ++ "[" ++ tmpNTupel ++ "] = " ++ bufferID_t ++ "++; ")

     noMergeReceiveDefines =
         (header_t ++
          tmpArrayDecl ++
          rpMap_t ++ "[" ++ tmpNTupel ++ "] = " ++ bufferID_t ++ " % " ++ 
	  mpiSize_t ++ "; " ++ bufferID_t ++ "++; ")

     cyclicMappingDefines =
	 (header_t ++ 
	  tmpArrayDecl ++ 
          -- zyklische Verteilung:
          -- rpMap[tmpNTupel] = ::std::make_pair(bufferID % mpiSize, bufferID);
	  rpMap_t ++ "[" ++ tmpNTupel ++ "] = ::std::make_pair(" ++ bufferID_t ++ " % " ++ 
	  mpiSize_t ++ ", " ++ bufferID_t ++ "); " ++ 
		  
          -- bufferID2RP.push_back(bufferID % mpiSize);
	  bufferID2RP_t ++ ".push_back(" ++ bufferID_t ++ " % " ++ mpiSize_t ++ "); " ++
	  -- sendVect.push_back(::std::vector<double>());
	  sendVector_t ++ ".push_back(::std::vector<double>());" ++

          -- sendVectLength = ++bufferID;
          sendVectorLength_t ++ " = ++" ++ bufferID_t ++ ";")

     blockMappingDefines =
	 (header_t ++ 
	  if cloogStmtNr == 1 
	  -- zuerst nur "tiles zaehlen"
	  then (-- sendVect.push_back(::std::vector<double>());
		sendVector_t ++ ".push_back(::std::vector<double>());" ++
		-- sendVectLength++;
		sendVectorLength_t ++ "++;")
	  else (tmpArrayDecl ++ 
	        -- block-Verteilung:
		-- rpMap[tmpNTupel] = ::std::make_pair((bufferID * mpiSize) / sendVectLength, bufferID);
		rpMap_t ++ "[" ++ tmpNTupel ++ "] = ::std::make_pair((" ++ bufferID_t ++ 
		"*" ++ mpiSize_t ++ ") / " ++ 
		sendVectorLength_t ++ ", " ++ bufferID_t ++ "); " ++ 
		-- bufferID2RP.push_back((bufferID * mpiSize) / sendVectLength);
		bufferID2RP_t ++ ".push_back((" ++ bufferID_t ++ 
		"*" ++ mpiSize_t ++ ") / " ++ 
		sendVectorLength_t ++ "); " ++ bufferID_t ++ "++;")
	 )


getRPlookupText :: CommCodeDS -> CommSide -> TrDependenceID -> String
getRPlookupText ccDS side trDepNr = rpLookupText
    where
    trDep = getTrDep ccDS trDepNr
    srcTrStmtID = trDepSrcStmtID trDep
    destTrStmtID = trDepDestStmtID trDep
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS
    procTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
    timeDim = getSchedDim ccDS
    coordsProcDim = getPlaceDim ccDS
    tilingDim = getTilingDim ccDS
    rpDim = getRPdim ccDS
	
    destDim = if procTiling 
	      then getProcTileDim ccDS
	      else destCoordsDim - timeDim

    srcCoordsDim = getCoordsDim ccDS srcTrStmtID
    destCoordsDim = getCoordsDim ccDS destTrStmtID
    iteratorNames = getIteratorNames ccDS
    rpNames = -- genericTake rpDim $ genericDrop dropDims iteratorNames
              if (side == SendSide || ccMergeReceive ccDS)
              then [otherProc_t ++ (show i) | i <- [1..rpDim]]
              else [rp_t ++ (show i) | i <- [1..rpDim]]

    rpLookupText =
	if (ccUseWrapper ccDS || not (ccMergeReceive ccDS))
	then (otherRP_t ++ " = " ++ 
	      lookupRP_t ++ "(" ++ (concat $ intersperse "," rpNames) ++ 
	      "); ")
	else -- rp_bufID_pair = lookupRP(rpNames...);
	     (rp_bufID_pair_t ++ " = " ++
	      lookupRP_t ++ "(" ++ (concat $ intersperse "," rpNames) ++ 
	      "); " ++ otherRP_t ++ " = " ++ rp_bufID_pair_t ++ ".first; " ++
              bufferID_t ++ " = " ++ rp_bufID_pair_t ++ ".second; ")


getOrigStatementsCode :: CommCodeDS -> IO [(StatementID, String)]
getOrigStatementsCode ccDS@CommCodeDS 
			  { ccTempDir = tempDir
			  , ccProjFile = projFile
			  , ccTargoutExec = targoutExec
			  } = 
    let
    tempDir = ccTempDir ccDS
    projFile = ccProjFile ccDS
    tmpFile = projFile ++ ".stmts"
    langFile = "c_stmtsOnly.lang"
    targoutFlags = 
	" -temp " ++ tempDir ++
	" -proj " ++ tempDir ++ "/" ++ projFile ++
	" -pp " ++
	" -count " ++
	" -lang " ++ langFile ++
	" -f " ++ tmpFile
    in do
       exitCode <- do 
		   putStrLn ("calling targout for statement text...\n" ++
			     targoutExec ++ targoutFlags)
		   system (targoutExec ++ targoutFlags)
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "targout completed."
             else do error ("targout exited with error code: " ++
			    (show errCode))
       stmtsTextFile <- openFile (tempDir ++ "/" ++ tmpFile) ReadMode
       rawStmtsText <- hGetContents stmtsTextFile
       let stmtsLines = filter (isPrefixOf stmtPrefix) $
			lines rawStmtsText
       return $ map getStmtNumberAndText stmtsLines

---
-- Reads the origStmtNr and the corresponding code for a statement from a 
-- given line of the temporary file generated by "targout".
getStmtNumberAndText :: String -> (StatementID, String)
getStmtNumberAndText line =
    (integerToID (read stmtNr::Integer), unwords stmtText)
    where
    (stmtNr:stmtText) = words $ drop (length stmtPrefix) line

genCompStmtIndexMappingCode :: CommCodeDS -> TStatementID -> String
genCompStmtIndexMappingCode ccDS trStmtID = 
    concat $
    zipWith (\l c -> l ++ " = " ++ c ++ "; ") 
	    loopoIndices cloogIndices
    where
    origStmtID = getStmtID ccDS trStmtID
    tilingDim = getTilingDim ccDS
    ispcDim = getISPCdim ccDS trStmtID
    coordsDim = getCoordsDim ccDS trStmtID
    (SLP slps) = ccSLPs ccDS
    slpVector = (fromJust $ lookup origStmtID slps)::[Bool]
    slp = (elemIndices True slpVector) :: [Int]
    iteratorNames = getIteratorNames ccDS
    loopoIndices = 
	[ genericIndex (ccIndexNames ccDS) lp | lp <- slp ]
    cloogIndices = 
	genericTake ispcDim $ genericDrop (tilingDim) iteratorNames

stmtPrefix :: String
stmtPrefix = "StatementNr: "
	
-- scatter-dims:
globalTime_t :: String
globalTime_t = "glT"
rp_t = "rp"
blockNr_t = "bl"
vTime_t = "vT"
vProc_t = "vP"
vAdded_t = "addDim"
stmtType_t = "stmtType"
arrAcc_t = "arrAcc"
arrNr_t = "arrNr"
localDim_t = "local"

-- stmt-domain-dims:
-- Reihenfolge: Tile | Coords | other (addCoords / ISPC /...)
timeTile_t = "tileT"
procTile_t = "tileP"
coordsTime_t = "coordT"
coordsProc_t = "coordP"
add_t = "add"
innerVar_t = "other"
ispc_t = "ispc"

-- for the ".define" file:    
define_t = "#define"
cloogStmt_t = "S"

-- the prefix for all generated temp vars
loopo_t = "loopo___" 

argc_t = "argc"
argv_t = "argv"
rpMap_t = loopo_t ++ "rpMap"

-- for the rp-lookup:
lookupRP_t = loopo_t ++ "lookupRP"
helperDim_t = loopo_t ++ "helpDim"
tmpArray_t = loopo_t ++ "tmpArray"
tmpIndexVector_t = loopo_t ++ "NTupelRP"
sendVectorLength_t = loopo_t ++ "sendVectLength"
bufferID2RP_t = loopo_t ++ "bufferID2RP"
rp_bufID_pair_t = loopo_t ++ "rp_bufID_pair"

-- the cooresponding rp-number for a send/recv-stmt:
otherRP_t = loopo_t ++ "otherRP"

-- the corresponding globalTime and proc-number for a send/recv-stmt:
otherTime_t = "otherT"
otherProc_t = "otherP"

bufferID_t = loopo_t ++ "bufferID"
mpiRank_t = loopo_t ++ "rank"
mpiSize_t = loopo_t ++ "np"
mpiInt_t = "MPI_INT"
mpiDouble_t = "MPI_DOUBLE"
mpiComm_t = "MPI_COMM_WORLD"
mpiAlltoall_t = "my_Alltoall"
mpiAlltoallv_t = "my_Alltoallv"

sendVector_t = loopo_t ++ "sendVect"
sendBuffer_t = loopo_t ++ "sBuffer"
recvBuffer_t = loopo_t ++ "rBuffer"
recvIndex_t = loopo_t ++ "recvIndex"
scounts_t = loopo_t ++ "scounts"
rcounts_t = loopo_t ++ "rcounts"
sdispls_t = loopo_t ++ "sdispls"
rdispls_t = loopo_t ++ "rdispls"

sedSuffix_t = ".sedOutput"

getIteratorNames :: CommCodeDS -> [String]
getIteratorNames ccDS = 
    -- verbose "getIteratorNames" $
    iteratorNames
    where
    procTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
    -- globalTimeDim = getGlobalTimeDim ccDS
    rpDim = getRPdim ccDS
    tilingDim = getTilingDim ccDS
    timeTileDim = getTimeTileDim ccDS
    procTileDim = getProcTileDim ccDS
    coordsTimeDim = getSchedDim ccDS
    coordsProcDim = getPlaceDim ccDS
    addCoordsDim = getMaxCoordsDim ccDS - getCommonCoordsDim ccDS
    destDim = 
	-- neu: jetzt ohne zeit-dimension! ;-)
	if procTiling 
	then procTileDim 
	else coordsProcDim + addCoordsDim
    maxISPCdim= getMaxISPCdim ccDS
    numInnerVars = numTotalVars - numOuterVars
    numOuterVars = 
	timeTileDim + procTileDim
    numTotalVars = 
	numOuterVars + destDim + maxISPCdim + maxISPCdim

    iteratorNames = 
	[timeTile_t ++ show i | i <- [1..timeTileDim]] ++
	[procTile_t ++ show i | i <- [1..procTileDim]] ++
--	[coordsTime_t ++ show i | i <- [1..coordsTimeDim]] ++
--	[coordsProc_t ++ show i | i <- [1..coordsProcDim]] ++
	if (ccUseCommunication ccDS) 
	then [innerVar_t ++ show i | i <- [1 .. numInnerVars]]
	else [innerVar_t ++ show i | i <- [1..addCoordsDim + maxISPCdim]]

runCloog :: CommCodeDS -> CloogRunOption -> IO ()
runCloog ccDS cloogRunOption = 
    let 
    cloogOptions = ccCloogOptions ccDS     
    loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
    cloogInputFile = loopoFile ++ cloogInputSuffix cloogRunOption
    cloogOutputFile = loopoFile ++ cloogOutputSuffix cloogRunOption
    cloogFlags = getCloogFlags cloogRunOption
    cloogCommand = 
	ccCloogExec ccDS ++ " " ++ cloogFlags ++ " " ++ cloogOptions ++
        " -o " ++ cloogOutputFile ++ " " ++ cloogInputFile
    in do 
       exitCode <- do 
		   trace ("running CLooG: " ++ show cloogCommand) $ 
			 system cloogCommand
       let (ok, errCode) = checkResult exitCode
       if ok then do putStrLn "cloog completed."
             else do putStrLn ("cloog exited with error code: " ++ 
			       (show errCode))       
    
getCloogFlags :: CloogRunOption -> String
getCloogFlags CloogMainProg = "-block 1 -esp 0 -csp 0 -cpp 1"
getCloogFlags CloogRealProcMapping { cloogRunRPdim = rpDim
				   , cloogRunCyclicMapping = cyclicMapping } = 
    if cyclicMapping
    then "-block 1 -esp 0 -csp 0 -cpp 1 -stop " ++ (show (rpDim))
    -- for block mapping, an additional help dimension is used, so -stop has to be adapted:
    else "-block 1 -esp 0 -csp 0 -cpp 1 -stop " ++ (show (rpDim+1))
-- for old cloog:    "-block 1 -esp 0 -csp 0 -strides 0 -cpp 1 -stop " ++ (show (rpDim + 1))

cloogInputSuffix CloogMainProg = ".cloogInput"
cloogInputSuffix CloogRealProcMapping {} = ".rpMapInput"

cloogOutputSuffix CloogMainProg= ".cloogOutput"
cloogOutputSuffix CloogRealProcMapping {} = ".rpMapOutput"

finalOutputSuffix = ".finalOutput"


runSed :: CommCodeDS -> IO ()
runSed ccDS = 
    do 
    let
-- sed -r -e 's/(rp2=.*)\{/\1 if(bl == 1 || lookupRP(rp1, rp2) == rank) {/' exs/simple2d_1/simple2d.src.out.cloogOutput > exs/simple2d_1/simple2d.src.out.sedOutput -e 's/\{ (rp2 =.*;)/\1 if(bl == 1 || lookupRP(rp1, rp2) == rank) \{/

      loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
      cloogOutputFile =  loopoFile ++ cloogOutputSuffix CloogMainProg
      sedFile = loopoFile ++ sedSuffix_t
      rpDim = getRPdim ccDS
      commStmtBlockNr = getBlockNr ccDS CommunicateStmt {}
      rpNames = concat $ intersperse ", " [rp_t ++ show i | i <- [1..rpDim]]
      ifConstraint_t = 
         -- if(lookupRP(rp1, rp2) == mpiRank || bl == 1) {
	 " if(" ++ blockNr_t ++ " == " ++ show commStmtBlockNr ++ " || " ++
	 lookupRP_t ++ "(" ++ rpNames ++ ") == " ++ mpiRank_t ++ ") {"
      sedCommand =
  	("sed -r " ++ 
	 "-e 's/(" ++ rp_t ++ show rpDim ++ "=.*)\\{/\\1" ++
	 ifConstraint_t ++
	 "/' " ++ 
	 -- note: whitespace could change in the future!
	 "-e 's/\\{( " ++ rp_t ++ show rpDim ++ " =.*;)/\\1 " ++ 
	 ifConstraint_t ++ "/' " ++

	 -- "-e '/if.*\{.*$/ {' -e 'N' -e '/rp1 =.*;$/ {' -e 's/\{//' -e 's/rp1 =.*;/& if(bl == 1 || loopo___lookupRP(rp1) == loopo___rank) {/' -e '}' -e '}'"
	 "-e '/^[[:space:]]*if.*\\{.*$/ {' " ++ -- matche 'if...{'
	 "-e 'N' " ++ -- schaue naechste Zeile an...
         "-e '/" ++ rp_t ++ show rpDim ++ " " ++ "=.*;$/ {' " ++ 
	 "-e 's/\\{//' " ++ -- "klaue Klammer"
	 "-e 's/" ++ rp_t ++ show rpDim ++ " " ++ "=.*;/&" ++ ifConstraint_t ++
	 "/' " ++
	 "-e '}' -e '}' " ++
	 
	 cloogOutputFile ++ " > " ++ sedFile)
    exitCode <- do trace ("running sed: " ++ show sedCommand) $ 
						       system sedCommand
    let (ok, errCode) = checkResult exitCode
    if ok then do putStrLn "sed completed."
          else do putStrLn ("sed exited with error code: " ++ 
			    (show errCode))       


insertIFs :: CommCodeDS -> IO ()
insertIFs ccDS =
  do
  let baseFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
      inpFile = baseFile ++ cloogOutputSuffix CloogMainProg
      outputFile = baseFile ++ finalOutputSuffix
      rpDim = getRPdim ccDS
      globalTimeDim = getGlobalTimeDim ccDS
      rpNames = 
          (concat $ intersperse ", " 
           [rp_t ++ show i | i <- [1..rpDim]])
      otherRPnames =
          (concat $ intersperse ", " 
           [otherProc_t ++ show i | i <- [1..rpDim]])
      remVarNames = 
          (genericDrop (globalTimeDim + 1) (getScatterNames ccDS)) ++
          getIteratorNames ccDS ++
          (nub $ ccIndexNames ccDS)
          
      remVarNamesList = concat $ intersperse "," remVarNames
      commStmtBlockNr = getBlockNr ccDS CommunicateStmt {}
      rpDesc = rp_t ++ show rpDim
      rpDescOpenMP = rp_t ++ show 1
      otherRPid = otherProc_t ++ show rpDim
      ifConstraint_t = 
	  if (ccUseWrapper ccDS || not (ccMergeReceive ccDS))
	  then -- if(bl > 0 || lookupRP(rp1, rp2) == mpiRank)
	       " if(" ++ blockNr_t ++ " > 0 || " ++
	       lookupRP_t ++ "(" ++ rpNames ++ ") == " ++ mpiRank_t ++ ")"	  
	  else -- if(bl == 1 || lookupRP(rp1, rp2).first == mpiRank)
	       " if(" ++ blockNr_t ++ " == " ++ show commStmtBlockNr ++ " || " ++
	       lookupRP_t ++ "(" ++ rpNames ++ ").first == " ++ mpiRank_t ++ ")"
      openMP_t =
          -- \#pragma omp parallel for private(rp2,vT1,vP1,vP2,tileT1,tileP1,other1,other2,other3,i,j,k)
          "#pragma omp parallel for private(" ++ remVarNamesList ++ ")"
      recvIfConstraint_t =
          -- if(bl < 2 || lookupRP(otherRP1, otherRP2) == mpiRank)
	  " if(" ++ blockNr_t ++ " < 2 || " ++ 
          lookupRP_t ++ "(" ++ otherRPnames ++ ") == " ++ mpiRank_t ++ ")"	  

  input <- readFile inpFile
  putStrLn ("parsing \"" ++ inpFile ++ "\"\n")
  let useComm = ccUseCommunication ccDS
      mergeReceive = ccMergeReceive ccDS
      initState = 
          ParserState { psRPid = rpDesc
		      , psIfConstraint = ifConstraint_t
                      , psUseCommunication = useComm
		      , psCloogStmtString = cloogStmt_t
		      }
      noMergeReceiveState = 
          ParserState { psRPid = otherRPid
		      , psIfConstraint = recvIfConstraint_t
                      , psUseCommunication = useComm
		      , psCloogStmtString = cloogStmt_t
		      }

      openMPState =
          ParserState { psRPid = rpDescOpenMP
                      , psIfConstraint = openMP_t
                      , psUseCommunication = useComm
		      , psCloogStmtString = cloogStmt_t
                      }

      firstOutput = 
          myTrace "first parser pass" $
          parseSource input initState
      noMergeReceiveOutput = 
          myTrace "second parser pass" $
          parseSource firstOutput noMergeReceiveState

      openMPOutput =
          myTrace "openMP parser pass" $
          parseSource input openMPState

      output = if useComm 
               then if mergeReceive then firstOutput else noMergeReceiveOutput
               else openMPOutput
  writeFile outputFile output
  putStrLn ("parsing completed.\n")

-------------------------------------------------------------------------------
----------------------- helper functions: -------------------------------------

fst3 :: (a, b, c) -> a
fst3 (x, _, _) = x
    

argMax f = maximumBy (\x y -> f x `compare` f y)

--
-- returns the number of scatter dimensions
numScatterDims :: CommCodeDS -> CloogRunOption -> Integer
numScatterDims ccDS cloogRunOption = error "TODO: numScatterDims"

getAllTStatementIDs :: CommCodeDS -> [TStatementID]
getAllTStatementIDs ccDS = fst3 $ unzip3 $ ccTISPCs ccDS

getTrafoFct :: CommCodeDS -> TStatementID -> AffineFunction
getTrafoFct ccDS trStmtID = trafoFct
    where
    trafoFct = 
	fromMaybe (error "getTrafoFct: no trafoFct for this TStatementID") $ 
		  lookup trStmtID (ccTrafos ccDS)

getTrDep :: CommCodeDS -> TrDependenceID -> TrDependence
getTrDep ccDS aTrDepID = trDep
    where
    trDep = fromMaybe (error "getTrDep: no TrDep for this TrDependenceID") $ 
	    find (\d -> trDepID d == aTrDepID) (ccTrDeps ccDS)
		  
    

getTRCis :: CommCodeDS -> TStatementID -> IneqSystem
getTRCis ccDS trStmtID = trcConstraints trc
    where
    trc = 
	fromJust $ 
	find (\oneTRC -> (trcTargetStmtID oneTRC) == trStmtID) (ccTRCs ccDS)

---
-- Filters all TrDependences that can cause communication between
-- (logical) processors.
filterTrueFlowDeps :: [TrDependence] -> TilingOptions -> LoopoDS -> [TrDependence]
filterTrueFlowDeps trDeps tilingOptions loopoDS = 
    verbose "filterTrueFlowDeps" $
    remainingDeps
    where
    remainingDeps = -- MC 2006-03-09: tiled version may be buggy!
                    -- if useSpaceTiling
		    -- then remainingDeps_tiled
		    -- else remainingDeps_nonTiled
                    remainingDeps_nonTiled
    remainingDeps_nonTiled = filter (\d -> not $ (trDepID d) `elem` 
				           zeroCommunicationIDs_nonTiled) trueFlowDeps
    remainingDeps_tiled = filter (\d -> not (trDepID d `elem` 
				             (zeroCommunicationIDs_tiled ++ 
					      zeroCommunicationIDs_nonTiled)
				             )
				 ) trueFlowDeps
			  
    zeroCommunicationIDs_nonTiled = map trDepID zeroCommunicationTrDeps_nonTiled
    zeroCommunicationTrDeps_nonTiled = 
	filter (\d -> let direction = trDepsCommunicationDirection d
		      in myTrace ("direction: " ++ show direction) $
	                 checkDep d)
	       processorTrDeps_nonTiled

    zeroCommunicationIDs_tiled = 
	myTrace ("zeroCommunicationIDs_tiled: " ++ (show $ map trDepID zeroCommunicationTrDeps_tiled))
	map trDepID zeroCommunicationTrDeps_tiled
    zeroCommunicationTrDeps_tiled = 
        verbose ("zeroCommunicationTrDeps_tiled") $
	filter (\d -> checkDep d) processorTrDeps_tiled

    -- checks, if a dependence cannot cause communication.
    -- throws an error, if strict-FCO is violated!
    checkDep :: TrDependence -> Bool
    checkDep d = let (fstCDs,sndCDs) = unzip $ trDepsCommunicationDirection d
                 in if ((any (== False) fstCDs) && 
                        -- test for strict FCO...
                        (not ((trDepType d) `elem` [ UniformInput, UniformDummy
                                                   , NonUniformInput, NonUniformDummy
                                                   ]) 
                        )
                       )
                    then myTrace (show (fstCDs,sndCDs) ++ " " ++ show d) $
                         error "non strict FCO!"
                    else (all (== (True,True)) $ zip fstCDs sndCDs)

    processorTrDeps_nonTiled = relevantTrDeps (head placementDims) (last placementDims) loopoDS

    -- remaining (non-tiled) dependence relations are tiled and projected to proc-Tile dimensions:
    processorTrDeps_tiled = 
	relevantTiledTrDeps remainingDeps_nonTiled procTileFrom procTileTo latVects

    trueFlowDeps = filter (\d -> (trDepType d) `elem` 
                           [UniformTrue, UniformFlow, 
                            NonUniformTrue,NonUniformFlow])
		   trDeps
    qualVects = loopoQUAL loopoDS
    qualVect = getCommonQualVect qualVects
    scheduleDims = map fromIntegral $ elemIndices Sched qualVect
    placementDims = map fromIntegral $ elemIndices Place qualVect
    procTileFrom = timeDim
    procTileTo = timeDim+rpTileDim-1

    rpTileDim = tilingDim - timeDim
    timeDim = genericLength scheduleDims
    useSpaceTiling = tiOptSpaceTiling tilingOptions
    latVects =  tiOptLatVects tilingOptions
    tilingDim = genericLength latVects


getISPC :: CommCodeDS -> TStatementID -> IneqSystem
getISPC ccDS trStmtID = 
    myTrace ("\ngetISPC: " ++ show is ++ "\n") $
    is
    where
    stmtID = getStmtID ccDS trStmtID
    is = fromMaybe (error "getISPC: no ISPC for this StatementID found!") $ 
	 lookup stmtID (ccISPCs ccDS)

getParamsDim :: CommCodeDS -> Integer
getParamsDim ccDS = paramsDim
    where
    (_, ispcs) = unzip $ ccISPCs ccDS
    paramsDims = map nParams ispcs
    paramsDim = if allEq paramsDims
		then head paramsDims
		else error ("all ISPCs should have" ++
			    " the same number of parameters!")

---
-- Returns the number of variables in the corresponding original index space
-- for a given target statement ID
getISPCdim :: CommCodeDS -> TStatementID -> Integer
getISPCdim ccDS = 
    nVars . getISPC ccDS

getMaxISPCdim :: CommCodeDS -> Integer
getMaxISPCdim ccDS = maximum $ map (getISPCdim ccDS) $ getAllTStatementIDs ccDS


getTypeOfDim :: CommCodeDS -> Integer -> QualEntry
getTypeOfDim CommCodeDS { ccQualVects = qualVects } dim
    | allEq types  =  head types
    | otherwise    =  error ("getTypeOfDim: dimension " ++ show dim ++
                             "does not have the same type in all statements: "
                             ++ show types)
    where
    (_, qualEntries) = unzip qualVects
    types = map (!! fromIntegral dim) qualEntries
---
-- Counts the number of time dimensions in the target program
-- 
-- @return timeDim, the number of schedule dimensions in the target program
getSchedDim :: CommCodeDS -> Integer
getSchedDim CommCodeDS { ccQualVects = qualVects } = 
    myTrace ("\ngetSchedDim: " ++ (show timeDim) ++ "\n") $
    timeDim
    where
    (_, qualEntries) = unzip qualVects
    schedEntries = map (filter (\e -> e == Sched)) qualEntries
    timeDimNumbers = map genericLength schedEntries
    timeDim = 
	if allEq timeDimNumbers
	then head timeDimNumbers
	else error ("number of schedule dimensions not equal for " ++
                    "all target statements!")

getPlaceDim :: CommCodeDS -> Integer
getPlaceDim CommCodeDS { ccQualVects = qualVects } = 
    myTrace ("\ngetPlaceDim: " ++ (show $ head placeDims) ++ "\n") $
    if allEq placeDims
    then head placeDims
    else error ("number of placement dimensions not equal for " ++
                "all target statements!")
    where
    (_, qualEntries) = unzip qualVects
    placeEntries = map (filter (\e -> e == Place)) qualEntries
    placeDims = map genericLength placeEntries	

getCoordsDim :: CommCodeDS -> TStatementID -> Integer
getCoordsDim CommCodeDS { ccQualVects = qualVects } trStmtID = 
    myTrace ("coordsDim: trStmtID=" ++ (show trStmtID)) $
    genericLength qualVect
    where 
    qualVect = 
	fromMaybe (error $ "getCoordsDim: no quality vector" ++
		   "found for this TStatementID: " ++ (show trStmtID)) $ 
	lookup trStmtID $ qualVects

getCommonCoordsDim :: CommCodeDS -> Integer
getCommonCoordsDim CommCodeDS { ccQualVects = qualVects } = 
    if allEq coordsDims 
    then head coordsDims 
    else error ("getCommonCoordsDim: varying number " ++ 
		"of schedule and placement dimensions")
    where
    (_, qualEntries) = unzip qualVects
    coordsDims = map (genericLength .
		      filter (\e -> e == Sched || e == Place)) qualEntries

getMaxCoordsDim :: CommCodeDS -> Integer
getMaxCoordsDim CommCodeDS { ccQualVects = qualVects } = maxCoordsDim
    where
    maxCoordsDim = maximum $ map genericLength $ snd $ unzip qualVects

getCommonQualVect :: [(TStatementID, [QualEntry])] -> [QualEntry]
getCommonQualVect qualVects = head $ commonQualVects
    where
    (_, qualEntries) = unzip qualVects
    commonQualVects = map (filter (\e -> e == Sched || e == Place)) qualEntries


getGlobalTimeDim :: CommCodeDS -> Integer
getGlobalTimeDim ccDS = if timeTiling 
			then 1 -- timeTileDim
			else getSchedDim ccDS
    where
    tiOpts = ccTilingOptions ccDS
    timeTiling = tiOptTimeTiling tiOpts
    tilingDim = genericLength $ tiOptLatVects tiOpts
    timeTileDim = min tilingDim (getSchedDim ccDS)

-- returns the correct number of rp-dims in both cases of tiling or non-tiling
getRPdim :: CommCodeDS -> Integer
getRPdim ccDS = 
    myTrace ("\ngetRPdim: " ++ show rpDim ++ "\n") $
    rpDim
    where
    rpDim = if procTiling 
	    then tilingDim - timeTileDim
	    else getPlaceDim ccDS
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS
    procTiling = tiOptSpaceTiling $ ccTilingOptions ccDS
    tilingDim = getTilingDim ccDS
    timeDim = getSchedDim ccDS
    timeTileDim = min tilingDim timeDim

getTilingDim :: CommCodeDS -> Integer
getTilingDim ccDS = if useTiling 
		    then genericLength $ tiOptLatVects $ ccTilingOptions ccDS
		    else 0
    where
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS    

getTimeTileDim :: CommCodeDS -> Integer
getTimeTileDim ccDS
    | useTiling  =
        genericLength $
        filter (==Sched) $
        [getTypeOfDim ccDS d | d <- tiOptTiledDims $ ccTilingOptions ccDS]
    | otherwise  =  0
{-
    if useTiling
    then min (getTilingDim ccDS) (getSchedDim ccDS)
    else 0
-}
    where
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS    

getProcTileDim :: CommCodeDS -> Integer
getProcTileDim ccDS
    | useTiling  =
        genericLength $
        filter (==Place) $
        [getTypeOfDim ccDS d | d <- tiOptTiledDims $ ccTilingOptions ccDS]
    | otherwise  =  0
{- 
    myTrace ("\ngetProcTileDim: " ++ show procTileDim ++ "\n") $
    procTileDim
    where
    procTileDim = if useTiling
		  then max (getTilingDim ccDS - getSchedDim ccDS) 0
		  else 0
-}
    where
    useTiling = tiOptUseTiling $ ccTilingOptions ccDS    

getArrAccessMaxDim :: CommCodeDS -> Integer
getArrAccessMaxDim ccDS = maximum accessTargetDims
    where
    accesses = ccAccesses ccDS
    accessTargetDims = map (targetDimIH.accIndexFct) accesses
			  

getStmtID :: CommCodeDS -> TStatementID -> StatementID 
getStmtID CommCodeDS { ccTISPCs = tispcs } trID = stmtID
    where 
    ids' = foldr (\(t,s,ineq) ids -> if (t == trID) 
				     then s:ids 
				     else ids) [] tispcs
    stmtID = if length ids' == 1
	     then head ids' 
	     else error $ "couldn't find StatementID for TStatementID: " ++ 
		      (show trID)

getTStmtIDs :: CommCodeDS -> StatementID -> [TStatementID]
getTStmtIDs CommCodeDS { ccTISPCs = tispcs } stmtID = trStmtIDs
    where 
    trStmtIDs' = foldr (\(t,s,ineq) trIDs -> if (s == stmtID) 
					     then t:trIDs 
					     else trIDs) [] tispcs
    trStmtIDs = if length trStmtIDs' > 0 
		then trStmtIDs'
		else error $ "couldn't find TStatementIDs for StatementID: " ++ 
			 (show stmtID)

getSrcAcc :: CommCodeDS -> TrDependenceID -> Access
getSrcAcc ccDS trDepNr = fst $ getSrcAndDstAccs accs acps trDep
    where
    acps = ccACPs ccDS
    accs = ccAccesses ccDS
    trDep = getTrDep ccDS trDepNr


getDepDomains :: CommCodeDS -> TrDependence -> (IneqSystem, IneqSystem)
getDepDomains CommCodeDS { ccDeps = deps} trDep =
    -- verbose "getDepDomains" $
    myTrace (show dep) $
    (depSrcDomain dep, depDestDomain dep)
    where
    origDepID = trDepOrigID trDep
    dep = 
	fromMaybe (error "getDepDomains: origDepID not found")
		  (find (\d -> depID d == origDepID) deps)
		   
getStmtTypeNr :: CloogStmtInfo -> Integer
getStmtTypeNr PrepareBufferStmt { siCommSide = commSide } 
    | commSide == RecvSide = 0
    | commSide == SendSide = 2
    | otherwise = error "no such communication side possible!"
getStmtTypeNr ComputeStmt {}     = 1
getStmtTypeNr CommunicateStmt {} = 
    error "stmtTypeNr should not be used in a CommunicateStmt!"
getStmtTypeNr RPMappingStmt {}   =
    error "stmtTypeNr should not be used in a RPMappingStmt!"

getBlockNr :: CommCodeDS -> CloogStmtInfo -> Integer
getBlockNr _ ComputeStmt {}       = 0
getBlockNr ccDS PrepareBufferStmt { siCommSide = commSide } = 
    if (ccMergeReceive ccDS || commSide == SendSide)
    then 0 
    else 2
getBlockNr _ CommunicateStmt {}   = 1
getBlockNr _ RPMappingStmt {}     =
    error "blockNr should not be used in a RPMappingStmt!"

getIdentifiers :: String -> String -> Identifiers
getIdentifiers tempDir projFile =
    Identifiers 
    { idIndices = indices,
      idBlobs = blobs,
      idArrays = arrays
    }		 
    where
    (symtab,_,_,_) = unsafePerformIO $ openParseFile  (tempDir ++ "/" ++ projFile)
    arrays = getArrayLexems symtab
    indices = getIndexLexems symtab
    blobs = getBlobLexems symtab
{-- MC 04.12.2006: sollte jetzt durch Benjamins symbol table funktionen ersetzt sein...
    indices = 
	map fromJust $ 
	takeWhile isJust $ map (getIndex tempDir projFile) [1..]
    blobs   = 
	map fromJust $ 
	takeWhile isJust $ map (getBlob tempDir projFile) [1..]
    arrays  =
	map fromJust $ 
	takeWhile isJust $ map (getArray tempDir projFile) [1..]
--}

getTilingOptions :: [Vector]                      -- ^ The lattice vectors.
                 -> [Integer]                     -- ^ The list of tiled dimensions.
                 -> TilingUsedVector              -- ^ The tiling used-vector.
                 -> [(TStatementID, [QualEntry])] -- ^ The quality vectors for each trStatement.
                 -> TilingOptions
getTilingOptions latVects tiledDims (TilingUsedVector tileUsedVect) qualVects = 
    TilingOptions 
    { tiOptUseTiling = or tileUsedVect, 
      tiOptLatVects = latVects,
      tiOptTiledDims = tiledDims,
      tiOptTilingUsedVect = tileUsedVect,
      tiOptTimeTiling = or timeTilingUsedVect,
      tiOptSpaceTiling = or spaceTilingUsedVect
    }
    where
    ccDS = undefinedCommCodeDS { ccQualVects = qualVects }
    schedDim = getSchedDim ccDS
    placeDim = getPlaceDim ccDS
    timeTilingUsedVect = genericTake schedDim tileUsedVect
    spaceTilingUsedVect = 
	genericTake placeDim $ genericDrop schedDim tileUsedVect


getTiles :: TilingOptions -> [TileInformation] -> TStatement -> 
	    (IneqSystem, [Integer])
getTiles tiOpts infos (tsID,sID,ispc) = res
    where
    lattice = tiOptLatVects tiOpts
    tiledDims = tiOptTiledDims tiOpts
	      -- ACHTUNG: TStatement <-> Statement ?
    res = defaultTile lattice tiledDims (selfDepRelation (sID,ispc)) infos


data NewCommCodeFlags = TempDir String
                      | ProjectFile String
                      | TargoutBin String
                      | CloogBin String
                      | CloogOptions String
                      | LaticeVectors String
                      | TilingDims String
                      | TilingUsedVectorOption String
                      | UseCommunication String
                      | CyclicMapping String
                      | UseMPIWrapper String
                      | UnpackAfterAllToAll String
                      | ModifyDims String
                      | MinBlobSize String

newCommCodeOptions :: [OptDescr NewCommCodeFlags]
newCommCodeOptions =
    [ Option [] ["temp-dir"]      (ReqArg TempDir "DIR")
             "set the name of the temporary directory"
    , Option [] ["project"]       (ReqArg ProjectFile "FILE")
             "set the name of the project file (e.g., \"project.src.out\""
    , Option [] ["targout-bin"]   (ReqArg TargoutBin "BINARY")
             "set the name of the targout binary"
    , Option [] ["cloog-bin"]     (ReqArg CloogBin "BINARY")
             "set the name of the CLooG binary"
    , Option [] ["cloog-options"] (ReqArg CloogOptions "STRING")
             "optional CLooG options (default=\"\")"
    , Option [] ["latice-vectors"] (ReqArg LaticeVectors "STRING")
             "???"
    , Option [] ["tiling-dims"]    (ReqArg TilingDims "STRING")
             "???"
    , Option [] ["tiling-used-vectors"] (ReqArg TilingUsedVectorOption "STRING")
             "???"
    , Option [] ["use-communication"] (ReqArg UseCommunication "STRING")
             "???"
    , Option [] ["cyclic-mapping"] (ReqArg CyclicMapping "STRING")
             "???"
    , Option [] ["use-mpi-wrapper"] (ReqArg UseMPIWrapper "STRING")
             "???"
    , Option [] ["unpack-after-all-to-all"]  (ReqArg UnpackAfterAllToAll "STRING")
             "???"
    , Option [] ["modify-dims"]    (ReqArg ModifyDims "STRING")
             "???"
    , Option [] ["min-blob-size"]  (ReqArg MinBlobSize "STRING")
             "???"
    ]

newCommCodeOpts :: [String] -> IO ([NewCommCodeFlags], [String])
newCommCodeOpts argv = 
    case getOpt Permute newCommCodeOptions argv of
      (o,rem,[]) -> return (o,rem)
      (_,_,errs) -> ioError (userError (concat errs ++ usageInfo header newCommCodeOptions))
      where header = "Usage: [OPTION...]"

parseCommandLine_GETOPT_EXPERIMENTAL :: [String] -> IO CommandLine
parseCommandLine_GETOPT_EXPERIMENTAL argv =
    do
    (opts, rem) <- newCommCodeOpts argv
    when (not $ null rem)
         (putStrLn ("Error: Could not parse NewCommCode command line " ++
                    "(stopped at: '" ++ show rem ++ "')") >> exitFailure)
    let initCL = CommandLine
                 { clTempDir          = err "temp-dir"
                 , clProjFile         = err "project"
                 , clTargoutExec      = err "targout-bin"
                 , clCloogExec        = err "cloog-bin"
                 , clCloogOptions     = ""
                 , clLatVects         = err "latice-vectors"
                 , clTiledDims        = err "tiling-dims"
                 , clTilingUsedVect   = err "tiling-used-vect"
                 , clUseCommunication = err "use-communication"
                 , clCyclicMapping    = err "cyclic-mapping"
                 , clUseWrapper       = err "use-mpi-wrapper"
                 , clMergeReceive     = err "unpack-after-all-to-all"
                 , clModifyDims       = err "modify-dims"
                 , clMinBlobSize      = err "min-blob-size"
                 }
        cl = foldl extract initCL opts
        err s = error
                ("Error while parsing NewCommCode command line options.\n" ++
                 "Required option '--" ++ s ++ "' is missing.\n\n" ++ usageInfo')
    checkOptions cl
    where
    usageInfo' = usageInfo "NewCommCode options:\n" newCommCodeOptions
    extract cl (TempDir       str) = cl { clTempDir      = str }
    extract cl (ProjectFile   str) = cl { clProjFile     = str }
    extract cl (TargoutBin    str) = cl { clTargoutExec  = str }
    extract cl (CloogBin      str) = cl { clCloogExec    = str }
    extract cl (CloogOptions  str) = cl { clCloogOptions = str }
    extract cl (LaticeVectors str) = cl { clLatVects     = latVects str }
    extract cl (TilingDims    str) =
        cl { clTiledDims = parseOption str "tiling-dims" }
    extract cl (TilingUsedVectorOption str) =
        cl { clTilingUsedVect = TilingUsedVector $
                                parseOption str "tiling-used-vect" }
    extract cl (UseCommunication  str) =
        cl { clUseCommunication = parseBool str "use-communication" }
    extract cl (CyclicMapping str) =
        cl { clCyclicMapping = parseBool str "cyclic-mapping" }
    extract cl (UseMPIWrapper str) =
        cl { clUseWrapper = parseBool str "use-mpi-wrapper" }
    extract cl (UnpackAfterAllToAll str) =
        cl { clMergeReceive = parseBool str "unpack-after-all-to-all" }
    extract cl (ModifyDims str) =
        cl { clModifyDims = parseOption str "modify-dims" }
    extract cl (MinBlobSize str) =
        cl { clMinBlobSize = parseOption str "min-blob-size" }

    latVects str = map (vector.map (\x -> x%1)) $ latVectsInt (latVectsIntRaw str)
    latVectsInt [] = [[]]
    latVectsInt vects = vects
    latVectsIntRaw str = (read str) :: [[Integer]]

    parseBool "True"  _ = True
    parseBool "true"  _ = True
    parseBool "False" _ = False
    parseBool "false" _ = False
    parseBool str option =
        error ("Could not parse bool value: '" ++ str ++ "'\n" ++
               "(--" ++ option ++ " expects 'true' or 'false')")

    parseOption :: Read a => String -> String -> a
    parseOption str option =
        let (prefix,val) = genericSplitAt (genericLength option) str
        in case reads val of
              [(v,"")] -> v
              _        -> error $ "could not parse '" ++ val ++
                                  "' as value for --" ++ option

    checkOptions :: CommandLine -> IO CommandLine
    checkOptions cl
{-
  MISSING CHECK:
        | if ((length $ latVects latVectsStr) ==
                                  (length $ tileUsedVectBool tileUsedVectStr))
                              then tileUsedVect tileUsedVectStr
                              else error $ "wrong length of tileUsedVector!"
-}
        | otherwise = return cl


parseCommandLine :: [String] -> CommandLine
parseCommandLine args =
    trace ("Warning: NewCommCode.parseCommandLine will be replaced soon. " ++
           "The new signature is '[String] -> IO CommandLine'") $
    case args of
      [] -> error usageInfo
      [tempDir, projFile, 
       latVectsStr, tiledDimsStr, tileUsedVectStr,
       targoutExec, cloogExec,
       useCommStr, cyclicMappingStr, useWrapperStr, mergeReceiveStr,
       minBlobSizeStr,
       -- BEGIN HACK ATTENTION: command line options should be all re-written anyway
       cloogOptions] -> trace "oldParseCommandLine: HACK" $
                        parseCommandLine $
                        [tempDir, projFile, 
                         latVectsStr, tiledDimsStr, tileUsedVectStr,
                         targoutExec, cloogExec,
                         useCommStr, cyclicMappingStr, useWrapperStr, mergeReceiveStr,
                         "[]",
                         minBlobSizeStr,
                         cloogOptions]
       -- END HACK      
      [tempDir, projFile, 
       latVectsStr, tiledDimsStr, tileUsedVectStr,
       targoutExec, cloogExec,
       useCommStr, cyclicMappingStr, useWrapperStr, mergeReceiveStr,
       modifyDims,
       minBlobSizeStr,
       cloogOptions] ->
              CommandLine { clTempDir = tempDir
		          , clProjFile = projFile
		          , clTargoutExec = targoutExec
		          , clCloogExec = cloogExec
		          , clCloogOptions = cloogOptions
		          , clLatVects = latVects latVectsStr
                          , clTiledDims = (read tiledDimsStr)::[Integer]
		          , clTilingUsedVect = 
                              if ((length $ latVects latVectsStr) ==
                                  (length $ tileUsedVectBool tileUsedVectStr))
                              then tileUsedVect tileUsedVectStr
                              else error $ "wrong length of tileUsedVector!"
		          , clUseCommunication = 
                            parseBoolOption "--useCommunication=" useCommStr
                          , clCyclicMapping = 
                            parseBoolOption "--cyclicMapping=" cyclicMappingStr
		          , clUseWrapper = 
                            parseBoolOption "--useMPIWrapper=" useWrapperStr
                          , clMergeReceive = 
                            not $ parseBoolOption "--unpackAfterAlltoall=" mergeReceiveStr
                          , clModifyDims = parseOption "--modifyDims=" modifyDims
                          , clMinBlobSize = (read minBlobSizeStr) :: Integer
                          }
      _ -> error $ "Could not parse command line parameters\n" ++
                   usageInfo
    where
    usageInfo :: String
    usageInfo =
        ("Usage: \n" ++
         "codegen <tmp-path> <project.src.out> <lattice-matrix> <tiledDims>\n" ++         
	 "  <tile-used-vector> <targout-executable> <cloog-executable>\n" ++
         "  --useCommunication=<True | False> \n" ++
         "  --cyclicMapping=<True | False> \n" ++
         "  --useMPIWrapper=<True | False> \n" ++
         "  --unpackAfterAlltoall=<True | False> \n" ++ 
         "  --modifyDims=... :: [(Integer, Integer, [[Integer]])]\n" ++
         "  <minBlobSize> " ++
         "  <additional cloog parameters | \"\">\n\n" ++

         "example:\n" ++
         "codegen \"/tmp/loopo23456/\" \"example.src.out\" " ++
         "\"[[10,0,0],[0,1,0],[0,0,200]]\" \"[0,1,2]\" \"[True,False,True]\" " ++
         "\"../loopo/targout/targout\" \"../cloog/gmp-bin/cloog\" " ++
         "--useCommunication=True " ++
         "--cyclicMapping=True " ++
         "--useMPIWrapper=False " ++
         "--unpackAfterAlltoall=True " ++ 
         "--modifyDims=[] " ++
         "1000 " ++
         "\"\""
        )

    parseBoolOption :: String -> String -> Bool
    parseBoolOption option str =
        let (prefix,val) = genericSplitAt (genericLength option) str
        in case (prefix,val) of
             (option, "True") -> True
             (option, "False") -> False
             _ -> error $ "expecting: " ++ option ++ "<True|False>\n" ++ usageInfo

    parseOption :: Read a => String -> String -> a
    parseOption option str =
        let (prefix,val) = genericSplitAt (genericLength option) str
        in case reads val of
              [(v,"")] -> v
              _        -> error $ "could not parse '" ++ val ++
                                  "' as value for " ++ option
        
                                  
    tileUsedVect str = 
	verbose "parse tileUsedVectStr" $
	TilingUsedVector (tileUsedVectBool str)

    tileUsedVectBool str = (read str) :: [Bool]

    latVectsIntRaw str = 
	verbose "parse latVectsIntRaw" $
        (read str) :: [[Integer]]

    latVectsInt vects = if null vects
		  then [[]]
		  else vects
    latVects str = map (vector.map (\x -> x%1)) $ latVectsInt (latVectsIntRaw str)


unsafeCreateCommCodeDS :: CommandLine -> CommCodeDS
unsafeCreateCommCodeDS commandLine  =
    verbose "unsafeCreateCommCodeDS" $
    CommCodeDS 
    { ccTempDir = tempDir
    , ccProjFile = projFile
    , ccTargoutExec = clTargoutExec commandLine
    , ccCloogExec = clCloogExec commandLine
    , ccCloogOptions = clCloogOptions commandLine
    , ccISPCs = ispcs
    , ccTISPCs = makeProperTISPCs ispcs tispcs trafos
    , ccDeps = deps
    , ccTrDeps = trDeps
    , ccTrafos = trafos
    , ccTRCs = trcs
    , ccACPs = acps
    , ccAccesses = accs
    , ccSLPs = slp
    , ccQualVects = qualVects
    , ccIndexNames = idIndices identifiers
    , ccBlobNames = idBlobs identifiers
    , ccArrayNames = idArrays identifiers
    , ccTilingOptions = tilingOptions
    , ccUseCommunication = clUseCommunication commandLine
    , ccCyclicMapping = clCyclicMapping commandLine
    , ccUseWrapper = clUseWrapper commandLine
    , ccMergeReceive = clMergeReceive commandLine
    , ccModifyDims  = [(integerToID t,(i,map (vector . map fromIntegral) rss)) | (t,i,rss) <- clModifyDims  commandLine]
    , ccMinBlobSize = clMinBlobSize commandLine
    }
    where    
    tempDir = clTempDir commandLine
    projFile = clProjFile commandLine
    loopoFile = tempDir ++ "/" ++ projFile
    loopoDS = 
	verbose "reading loopoDS" $
	unsafeReadLoopoDS loopoFile
    deps = map (trimDependence slp) (loopoDEP loopoDS)
    trDeps = filterTrueFlowDeps allTrDeps tilingOptions loopoDS
    allTrDeps = transformDeps tispcs trafos deps
    trafos = makeProperTrafos (loopoTrafo loopoDS)
    ispcs = map (trimIndexSpace slp) (loopoISPC loopoDS)
    tispcs = loopoTISPC loopoDS
    slp = loopoSLP loopoDS
    acps = loopoACP loopoDS
    raccs = loopoACC loopoDS
    accs = acpsToAccs slp acps (loopoISPC loopoDS) allTrDeps raccs
    qualVects = loopoQUAL loopoDS
    trcs = loopoTRC loopoDS
    identifiers = getIdentifiers tempDir projFile
    latVects = clLatVects commandLine
    tiledDims = clTiledDims commandLine
    tilingUsedVect =  clTilingUsedVect commandLine
    tilingOptions = getTilingOptions latVects tiledDims tilingUsedVect qualVects


undefinedCommCodeDS :: CommCodeDS
undefinedCommCodeDS =   
    CommCodeDS { ccQualVects        = e
	       , ccTempDir          = e
               , ccProjFile         = e
	       , ccTargoutExec      = e
	       , ccCloogExec        = e
               , ccCloogOptions     = e
	       , ccISPCs            = e
	       , ccTISPCs           = e
	       , ccDeps             = e
	       , ccTrDeps           = e
	       , ccTrafos           = e
	       , ccTRCs             = e
	       , ccIndexNames       = e
	       , ccBlobNames        = e
               , ccArrayNames       = e
	       , ccACPs             = e
	       , ccAccesses         = e
	       , ccSLPs             = e
	       , ccTilingOptions    = e
	       , ccUseCommunication = e
	       , ccCyclicMapping    = e
	       , ccUseWrapper       = e
               , ccMergeReceive     = e
               , ccModifyDims       = e
               , ccMinBlobSize      = e
	       }
    where e = error "using undefinedCommCodeDS"
    

--- 
-- shows a "LooPo style" IneqSystem in Cloog-conform syntax
showIS2cloog :: IneqSystem -> String
showIS2cloog is = output
    where
    rawM = getMatrix $ fmap fromIntegral $ integralIneqSystem is
    firstColumn = columnVectorsToMatrix1 [vector (replicate (nRows rawM) (1%1))]
    finalM = firstColumn `conc` rawM
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False


--- 
-- shows a "LooPo style" IneqSystem in Cloog-conform syntax
showIS2cloogOpt :: IneqSystem -> String
showIS2cloogOpt is = output
    where
    (ineqs, eqs) = findEqualities is
    rawIneqM = getMatrix $ fmap fromIntegral $ integralIneqSystem ineqs
    rawEqM = getMatrix $ fmap fromIntegral $ integralEqSystem eqs
    firstColumn = 
	columnVectorsToMatrix1 [vector (replicate (nRows rawIneqM) (1%1) ++
					replicate (nRows rawEqM) (0%1))]
    finalM = firstColumn `conc` (rawIneqM `join` rawEqM)
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False


--- 
-- shows a scatter function matrix in Cloog-conform syntax
showScatterMatrix2Cloog :: Matrix -> String
showScatterMatrix2Cloog m = output
    where
    zeroColumn = columnVectorsToMatrix1 [vector (replicate (nRows m) (0%1))]
    finalM = zeroColumn `conc` m
    output = show (nRows finalM) ++ " " ++ show (nColumns finalM) ++ "\n" ++
	     intShowMat finalM False ++ "\n"


foreign import ccall getIndex_c :: CString -> CString -> CInt ->
				   IO CString
getIndex :: String -> String -> Integer -> Maybe String
getIndex tempDir projFile indexNum = 
    unsafeCppiConvert3 getIndex_c tempDir projFile indexNum

foreign import ccall getBlob_c :: CString -> CString -> CInt ->
				   IO CString
getBlob :: String -> String -> Integer -> Maybe String
getBlob tempDir projFile blobNum = 
    unsafeCppiConvert3 getBlob_c tempDir projFile blobNum

foreign import ccall getArray_c :: CString -> CString -> CInt ->
				   IO CString
getArray :: String -> String -> Integer -> Maybe String
getArray tempDir projFile arrayID = 
    unsafeCppiConvert3 getArray_c tempDir projFile arrayID


checkResult :: ExitCode -> (Bool, Int)
checkResult ExitSuccess = (True, 0)
checkResult (ExitFailure x) = (False, x)

getTime :: String
getTime = timeStr
    where
    clockTime = unsafePerformIO getClockTime
    calendarTime = unsafePerformIO $ toCalendarTime clockTime
    timeStr =  calendarTimeToString calendarTime


--
-- Debugging function. Can be used to display the results of finished functions.
--
verbose :: String -> a -> a
verbose str x =
    myTrace ("$? " ++ str ++ "...") $
    let skipIt = False -- suppress debugging output
    in if skipIt then x
       -- else seq x $ trace ("Verbose: " ++ str ++ ": result=" ++ show x) x
       else seq x $ myTrace ("$! " ++ str ++ " done. ") x

projectToVarsMC is vars = 
    -- verbose "projectToVarsMC" $
     fromMaybe (trace "warning: projectToVars' HsFM delivered empty IS" $
		emptyIS) projected
    where
    emptyIS = ineqSystem ((varsDim, paramsDim), 
			  empty_M)
    empty_M = 
        -- contradicting inequality: -1 >= 0
	(zeroMatrix 1 (varsDim + paramsDim) `conc`
	 matrix [[-1%1]])
    varsDim = genericLength vars
    paramsDim = nParams is			  
    projected = projectToVars' HsFM is Nothing vars

myMultijoin1 ms = 
    assert (not $ null ms) $
    if ok 
    then multijoin1 ms
    else error ("multijoin1 failed with different columns in submatrices: " ++ 
		show cols)
    where
    ok = all (== head cols) cols
    cols = map nColumns ms

allEq :: (Eq a) => [a] -> Bool
allEq [] = False
allEq (x:xs) = all (==x) xs

-- |
-- Project @oldIneqs@ described by @oldDescr@ to the
-- new domain description @newDescr@. 
--
projectDomDescriptionTo :: CommCodeDS 
			-> IneqSystem         -- ^ @oldIneqs@
			-> StmtDomDescription -- ^ @oldDescr@
			-> StmtDomDescription -- ^ @newDescr@
			-> Either TStatementID TrDependenceID
			-> IneqSystem
projectDomDescriptionTo ccDS 
			oldIneqs 
			(StmtDomDescription oldDescr) 
			(StmtDomDescription newDescr) 
			someID = 
			    assert (length (concatMap snd oldDims) == nVars oldIneqs) newIneqs
    where    
    oldDims = 
	zip oldDescr 
		(getStmtDomDims ccDS (StmtDomDescription oldDescr) someID)
    newDims = concatMap lookupPart newDescr

    newIneqs = projectToVarsMC oldIneqs newDims

    lookupPart p =
	fromMaybe (error ("projectDomDescriptionTo: " ++ 
			  "part of domain description not found: " ++ show p)) 
	              (lookup p oldDims)

-- |
-- Compute a list of dimensions for the statement description
-- given by @stmtDesc@ for either a given target statement
-- or a given target dependence.
--
getStmtDomDims :: CommCodeDS 
	       -> StmtDomDescription 
	       -> Either TStatementID TrDependenceID 
	       -> [[Integer]]
getStmtDomDims ccDS (StmtDomDescription stmtDesc) someID =
    snd $  mapAccumL (stmtDomDimHelper ccDS someID) 0 stmtDesc

stmtDomDimHelper :: CommCodeDS 
		 -> Either TStatementID TrDependenceID 
		 -> Integer 
		 -> StmtDomPart 
		 -> (Integer, [Integer])
stmtDomDimHelper ccDS someID dimAccum (DpTile dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
      where
      qual = dimQual dimInfo
      dim = case qual of
	    Sched -> getTimeTileDim ccDS
	    Place -> getProcTileDim ccDS
            Additional -> error ("stmtDomHelper: no additional tiling" ++
				 " dimension possible!")      
stmtDomDimHelper ccDS someID dimAccum (DpTISPC dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
      where
      dim = case (dimQual dimInfo) of
	    Sched -> getSchedDim ccDS
	    Place -> getPlaceDim ccDS
	    Additional -> (getMaxCoordsDim ccDS - 
			   getCoordsDim ccDS relevantStmtID)
      qual = dimQual dimInfo
      relevantStmtID = getRelevantStmtID ccDS someID dimInfo
stmtDomDimHelper ccDS someID dimAccum (DpISPC dimInfo) = 
    (dim + dimAccum, [dimAccum..dimAccum+dim-1])
    where
    relevantStmtID = getRelevantStmtID ccDS someID dimInfo
    dim = getISPCdim ccDS relevantStmtID    

getRelevantStmtID :: CommCodeDS 
		  -> Either TStatementID TrDependenceID 
		  -> DimInfo 
		  -> TStatementID
getRelevantStmtID ccDS someID dimInfo = relevantStmtID
    where
    trStmtID = either id (const $ error "TStatementID should not be accessed!") someID
    trDepID  = either (const $ error "TrDependenceID should not be accessed!") id someID
    trDep = getTrDep ccDS trDepID
    srcID = trDepSrcStmtID trDep
    destID = trDepDestStmtID trDep
    role = dimRole dimInfo
    relevantStmtID = 
	case role of
		 (Just SendSide) -> srcID
		 (Just RecvSide) -> destID
		 Nothing         -> trStmtID


------------------------------------------------------------------------
-- for testing:

tilingCL = ["exs/gauss-2/","gauss-2.src.out",
	    "[[1,0],[0,3]]", "[True, True]",
	    "True", "True","../loopo/targout/targout",
	    "../cloog/cloog-0.12.3/obj_i686-pc-linux-gnu/cloog",
	    ""]
	   
testTilingDS :: CommCodeDS
testTilingDS = unsafeCreateCommCodeDS cl
    where
    cl = parseCommandLine simpleCL

nonTilingCL = ["exs/gauss-2/","gauss-2.src.out",
	       "[]", "[]",
	       "True", "True", "True", "../loopo/targout/targout",
	       "../cloog/cloog-0.12.3/obj_i686-pc-linux-gnu/cloog",
	       ""]

testNonTilingDS :: CommCodeDS
testNonTilingDS = unsafeCreateCommCodeDS cl
    where
    cl = parseCommandLine nonTilingCL

simpleCL = ["exs/simple2d_1/","simple2d.src.out",
	    "[[2,0],[0,4]]", "[True,True]",
	    "True", "True", "True", "../loopo/targout/targout",
	    "../cloog/cloog-0.12.3/obj_i686-pc-linux-gnu/cloog",
	    ""]

    
polynom = ["exs/polynom/","polynom.src.out",
	    "[[2,0],[0,3]]", "[True,True]",
	    "True", "True", "False", "False", "../loopo/targout/targout",
	    "../cloog/cloog-0.14.0/gmp-version/cloog",
	    ""]
{- Tests vorlauefig deaktiviert
doTest :: IO ()
doTest = do
	 let ccDS = testTilingDS
	 let cloogData = genCloogData ccDS CloogMainProg
	     cloogInput = 
		 genCloogInput ccDS cloogData		 
	 writeDefinesFile ccDS cloogData
	 putStrLn cloogInput

parseSimpleFanin :: IO()
parseSimpleFanin =
    do
    commandLine <- parseCommandLine ["exs/simple2d_4_fanin/", "simple2d.src.out"
				    , "[]", "[]"
				    , "True", "True", "True", "", "", ""]
    let ccDS = unsafeCreateCommCodeDS commandLine
    insertIFs ccDS

parseSimpleTiling :: IO()
parseSimpleTiling =
    do
    commandLine <- parseCommandLine ["exs/simple2d_4/", "simple2d.src.out"
				    , "[[2,0],[0,2]]", "[True, True]"
				    , "True", "True", "True", "", "", ""]
    let ccDS = unsafeCreateCommCodeDS commandLine
    insertIFs ccDS

parseSimple2d :: IO ()
parseSimple2d =
    do
    commandLine <- parseCommandLine ["exs/simple2d_1/", "simple2d.src.out"
				    , "[[2,0],[0,4]]", "[True, True]"
				    , "True", "True", "True", "", "", ""]
    let ccDS = unsafeCreateCommCodeDS commandLine
    insertIFs ccDS

testDomDescr1 = StmtDomDescription 
		[ DpTile  (DimInfo Sched (Just SendSide))
		, DpTile  (DimInfo Place (Just SendSide))
		, DpTISPC (DimInfo Sched (Just RecvSide))
		, DpISPC  (DimInfo Sched Nothing)
		]

testDomDescr2 = StmtDomDescription 
		[ DpTISPC (DimInfo Sched (Just RecvSide))
		, DpISPC  (DimInfo Sched Nothing)
		, DpTile  (DimInfo Place (Just SendSide))
		]
getStmtDomDimsTest = getStmtDomDims ccDS testDomDescr1 someID
    where
    ccDS = testTilingDS
    someID = (Left (integerToID 1))

projectDomDescrTest = 
    projectDomDescriptionTo ccDS is testDomDescr1 testDomDescr2 someID
    where
    ccDS = testTilingDS
    is = ineqSystem ((5,1), matrix [[1%1,-2,1,3,4,0,0]
				   ,[ -1, 5,2,3,1,1,0]
				   ,[  2, 0,2,3,1,0,1]
				   ,[  1, 2,0,3,1,0,2]
				   ,[  0, 1,3,0,3,1,3]
				   ])
    someID = (Left (integerToID 1))

testIS = 
    ineqSystem ((3,1), matrix [[1%1, 0, 0,   0, -1]
			      ,[ -1, 0, 0,   1,  0]
			      ,[  0, 1, 0,   0,  2]
			      ,[  1,-1, 0, 1%2,  0]
			      ,[  0, 0, 1,-1%4,  0]
			      ,[  2, 0,-1,   0, -2]
			      ])


fullTest :: [String] -> IO ()
fullTest inputStrs = 
    do
    commandLine <- parseCommandLine inputStrs
    let ccDS = unsafeCreateCommCodeDS commandLine
	mainRun = CloogMainProg
	rpMapRun = CloogRealProcMapping { cloogRunRPdim = getRPdim ccDS
					, cloogRunCyclicMapping = 
					    ccCyclicMapping ccDS }
	loopoFile = ccTempDir ccDS ++ "/" ++ ccProjFile ccDS
	rpMappingData = 
	    genCloogData ccDS rpMapRun
	mainData = 
	    genCloogData ccDS mainRun
	cloogRPMappingInput = genCloogInput ccDS rpMappingData
	cloogMainInput = genCloogInput ccDS mainData
    -- putStrLn cloogMainInput
    writeDeclarations ccDS
    writeInitialisations ccDS
    writeDefinesFile ccDS mainData
    writeDefinesFile ccDS rpMappingData
    writeFile (loopoFile ++ 
	       cloogInputSuffix rpMapRun) cloogRPMappingInput
    writeFile (loopoFile ++
	       cloogInputSuffix mainRun) cloogMainInput
    runCloog ccDS rpMapRun
    runCloog ccDS mainRun
    if (ccUseCommunication ccDS) 
       then do runSed ccDS
       else return ()
-}

{- parsec:
    
rpDimMatched :: MyParser Bool
rpDimMatched =
    do
    state <- getState
    return (psMatchedRP state)

cloogCode :: MyParser String
cloogCode =
    do
    code <- statements
    eof
    return $ code   
	  
statements :: MyParser String
statements = 
    do
    stmts <- many oneStmt
    return (concat stmts) 
    <?> "statement"

oneStmt :: MyParser String
oneStmt = 
    do
    sp <- many space <?> "space before oneStmt"
    stmtText <- myTry (myTry ifStmt
		       <|> myTry forStmt
	               <|> myTry assignBlockStmt
		       <|> myTry comment 
	               <|> cloogStmt
		       -- <|> myTry assignStmt
		      )
    return (trace ("returning: " ++ sp ++ stmtText) $
	    stmtText)
    <?> "oneStmt"

comment :: MyParser String
comment =
     do 
     cBegin <- (--trace "comment..." $
		string "/*")
     cEnd <- manyTill anyChar (myTry (string "*/"))
     return (cBegin ++ cEnd ++ "*/")
     <?> "comment"

cloogStmt :: MyParser String
cloogStmt =
     do
     string cloogStmt_t
     stmtNr <- many1 digit
     rest <- parensCode
             -- manyTill (noneOf "{};") (myTry $ char ';')
     sp <- many space
     char ';'
     return (cloogStmt_t ++ stmtNr ++ rest ++ ";")
     <?> "cloogStmt"

assignBlockStmt :: MyParser String
assignBlockStmt =
     do
     openBrace <- (-- trace "assignBlockStmt..." $
		   char '{')
     sp1 <- many space <?> "space in assignBlockStmt"
     stmts <- myTry (do
		     rpStmt <- rpAssign
		     restStmts <- statements
		     state <- getState
		     return (rpStmt ++ " " ++ psIfStmt state ++ " { " ++ 
			     restStmts ++ " } ")
		     <?> "assignBlock: rpPart")
	      <|> 
		   (do
		    firstAssign <- assignStmt
		    otherStmts <- statements
		    return (firstAssign ++ otherStmts)
		    <?> "assignBlock: otherPart"
		   )
     char '}' <?> "closing brace of assignBlockStmt"
     return ([openBrace] ++ sp1 ++ stmts ++ "}")
     <?> "assignBlockStmt"
		       

forStmt :: MyParser String
forStmt = 
     do
     itsHead <- forHead
     sp <- many space <?> "space after forHead"
     forBody <- codeBlock
     return (itsHead ++ sp ++ forBody)
     <?> "forStmt"

forHead :: MyParser String
forHead = 
    do
    forBegin <- string "for"
    sp <- many space <?> "space after for"
    itsHead <- myTry (do
		    char '('
	            -- check for "rp1 = "
		    rp <- rpAssign
	            upperBorder <- many (noneOf ";")
	            char ';'
	            incIndex <- many (noneOf ")")
	            char ')'
	            state <- getState
	            return ("(" ++ rp ++ upperBorder ++ ";" ++ 
			    -- include the additional if-stmt
			    incIndex ++ ")" ++ (psIfStmt state))
		   )
	       <|>
	       do
	       header <- parensCode
	       return header
    return (forBegin ++ sp ++ itsHead)
    <?> "forHead"

ifStmt :: MyParser String
ifStmt =
    do
    ifStr <- (trace "ifStmt..." $
	      string "if")
    sp1 <- many space
    cond <- parensCode
    sp2 <- many space
    char '{'
    stmts <- myTry (do
		    sp <- many space <?> "space in ifStmt before rpAssign"
	            rpStmt <- rpAssign
	            restStmts <- statements
	            return (sp ++ rpStmt ++ restStmts)
		   )	     
	      <|>
	      myTry (do
		     sp <- many space <?> "space in ifStmt before assignStmt"
		     otherStmt <- assignStmt
		     restStmts <- statements
		     return (sp ++ otherStmt ++ restStmts)
		    )	   
	      <|>
	      do
	      otherStmts <- statements <?> "other stmts in ifStmt"
	      return otherStmts
    char '}'
    return (--trace ("ifStmt: " ++ ifStr ++ sp1 ++ cond ++ sp2 ++ "{" ++ stmts ++ "}") $
	    ifStr ++ sp1 ++ cond ++ sp2 ++ "{" ++ stmts ++ "}") 
    <?> "ifStmt"
    

parensCode :: MyParser String
parensCode = 
     do 
     char '('
     inner <- many (parensCode
		    <|> many1 (noneOf "(){}")
		   )
     char ')'
     return ("(" ++ concat inner ++ ")")
     <?> "parensCode"

codeBlock :: MyParser String
codeBlock =
     do     
     char '{'
     stmts <- statements
     char '}'
     return ("{" ++ stmts ++ "}")
     <?> "codeBlock"
	   
rpAssign :: MyParser String
rpAssign =
     do
     state <- getState
     rpStr <- string (psRealProcName state)
     sp2 <- many space
     char '='
     rSide <- manyTill (noneOf ";") (char ';')
     return (rpStr ++ sp2 ++ "=" ++ rSide ++ ";")
     <?> "rpAssign"

assignStmt :: MyParser String
assignStmt =
     do
     lSide <- manyTill (noneOf "{}=") (char '=')
     rSide <- manyTill (noneOf ";") (char ';')
     return (lSide ++ "=" ++ rSide ++ ";")
     <?> "assignStmt"			       
		    
       
testCloogParser :: IO ()
testCloogParser =
    do
    let simpleCL = ["exs/simple2d_4_fanin/","simple2d.src.out",
	    "[]", "[]",
	    "True", "True","../loopo/targout/targout",
	    "../cloog/cloog-0.12.3/obj_i686-pc-linux-gnu/cloog",
	    ""]
	ccDS = unsafeCreateCommCodeDS $ parseCommandLine simpleCL
    insertIFs ccDS
    return ()

-}
