module OneDim where 

import InputLoopo 
import Input4PL

import LoopoDS
import Loopo
import List 

import IOExts
import LMath

import Data.FiniteMap
import Maybe
import Char
import Ratio


-- calculates the fco placements 
-- parameters: ispc list, acc list
--
fco :: [(Integer, Matrix)] -> [(Integer,AccessType,Integer,Integer,Matrix)] -> [([Rational],[Rational])] 
fco ispc acc= precis pisAndLambdas -- multiple pi-parts are wrapped up to the smallest 
    where
    pisAndLambdas = [splitVector pos vec | vec <- linesAndRays ]
    -- lines and rays are preprocessed out of the LRV 
    linesAndRays  = (lrvLines lrv) ++ (lrvRays lrv) 
    -- the result of the Polilyb
    lrv = dualOfSystem eqSys ineq
    -- for the function dualOfSystem 
    -- equation system: farkas matrix 
    eqSys   = eqSystem ((v,0),finalMatrix) 
    -- inequation system: lambda matrix 
    ineq = ineqSystem ((v,0), lambdaMatrx) 
    -- the width of the farkas matrix minus one column 
    v = (nColumns finalMatrix) - 1
    -- the final farkas matrix 
    finalMatrix = farkasMatrix ispc acc
    -- the  lambda matrix showing that all all lambdas are > 0 
    lambdaMatrx = lambdaMatrix2 ispc acc 
    -- position of seperation between Pi vars and lambda vars
    pos = v - (numberOfLambdas ispc acc)

--linesAndRays :: [(Integer, Matrix)] -> [(Integer,AccessType,Integer,Integer,Matrix)] -> [([Rational],[Rational])] 
fcoLR ispc acc= linesAndRays
    where
    pisAndLambdas = [splitVector pos vec | vec <- linesAndRays ]
    -- lines and rays are preprocessed out of the LRV 
    linesAndRays  = lines ++ rays
    lines = (lrvLines lrv) 
    rays = (lrvRays lrv) 
    -- the result of the Polilyb
    lrv = dualOfSystem eqSys ineq
    -- for the function dualOfSystem 
    -- equation system: farkas matrix 
    eqSys   = eqSystem ((v,0),finalMatrix) 
    -- inequation system: lambda matrix 
    ineq = ineqSystem ((v,0), lambdaMatrx) 
    -- the width of the farkas matrix minus one column 
    v = (nColumns finalMatrix) - 1
    -- the final farkas matrix 
    finalMatrix = farkasMatrix ispc acc
    -- the  lambda matrix showing that all all lambdas are > 0 
    lambdaMatrx = lambdaMatrix2 ispc acc 
    -- position of seperation between Pi vars and lambda vars
    pos = v - (numberOfLambdas ispc acc)




-- new type definition for the results of dualOfSystem 
--
type Placement = ([ Rational], [Rational])

-- calculates fco placements
-- parameters: coded path information
--   
fco2:: String -> [Placement]
fco2 path = fco ispcClean accsCleaned 
      where
      ispcClean = clean ispc slps
      accsCleaned = cleanAccs acc slps 
      slps = unsafePerformIO (readslpfile path)
      ispc = unsafePerformIO( readispcfile path) 
      acc  = unsafePerformIO( readaccfile path)

-- helper for separation of pi and lambda part
splitVector ::Integral a => a -> Vector -> ([ Rational], [Rational])
splitVector position vec = genericSplitAt position (vectorEntries vec)


-- to cancel down the fractions
-- contained in the list 
-- 
cancel :: [Rational] -> [Rational]
cancel xs = map (/ factor) xs
   where
   factor = calculatedGCD % calculatedLCM 
   calculatedGCD = (foldl gcd 0 (filter (/=0)(map numerator xs))) 
   calculatedLCM = (foldl lcm 1 ( map denominator xs))


-- testing 
eins = [1 % 1,2 % 1,3 % 1] 
zwei = [2 % 1,4 % 1,6 % 1] 
drei = [3 % 1,6 % 1,9 % 1]
vier = [4 % 1,4 % 1,4 % 1] 
brueche = [zwei,drei]

-- function cancel applied on the first component 
-- of the tupel 
--
cancelPis :: Placement ->  Placement
cancelPis (pis, lambdas)  =  (cancel(pis),lambdas) 

-- testing


pl1,pl2, pl3, pl4 :: Placement 
pl1 = (eins, [2%3])
pl2 = (zwei, [3%2])
pl3 = (drei, [1%2 ])
pl4 = (vier, [1%1])


-- precises the list containing ([pi's],[lambdas]):
-- elements with multiple pi-part are deleted, 
-- the smallest remains. [1,2,3] remains, [2,4,6] deleted 
--
precis :: [([Rational],[Rational])] -> [([Rational],[Rational])]
precis = nubBy(\x y -> fst(x)==fst(y)) . map(cancelPis)

-- testing 
placs = [pl1,pl2,pl3,pl4]
--placies = fco2 bsp




-- cuts a list in different pieces
-- length of the pieces is provided by a list of Int's 
-- 
cut :: Integral b =>[a] -> [b] -> [[a]] 
cut [] [] = [] 
--cut list []= [list]
cut list (w:widths) = slice : cut cutList widths  
     where 
     slice = fst resultOfSplit
     cutList =  snd resultOfSplit 
     resultOfSplit  = genericSplitAt w list 



 
-- separates the pi-part into the the allocations 
-- according to the width of matrices in the ispc-list 
-- and the acc-matrix-list 
--
separateAllocations ispcCleaned accsCleaned placements = [( cut (fst(x)) width, snd(x)) | x <- placements  ] 
   where
   width = widthIdMatrices ++ widthAccMatrices
   -- widths of the unit matrices 
   widthIdMatrices = [ nColumns x | (_,x) <-ispcCleaned]
   -- acc list with transposed acc matrices 
   accTr = accTranspose accsCleaned
   -- position information of the different access matrices in the matrix
   accessPosition = (acc2posAndWidth (accTr))
   -- width of the access matrices / array accesses 
   widthAccMatrices = accWidths
   -- to get the number and the ids of the arrays (TOO COMPLICATED)
   arrayIdsAndOffsets  = init(fmToList (snd(accessPosition)))
   -- calculates the width according to the provided id 
   accWidths =  [widthOfAcc id accTr|(id,off)<-arrayIdsAndOffsets]
       where
       -- helper: estimates the width of the specified transposed matrix
       widthOfAcc :: Integer -> [(Integer,AccessType,Integer,Integer,Matrix)] -> Integer
       widthOfAcc arrayId accTr =  
	   case ( find (\(id,a,b,c,matrx )-> arrayId==id)accTr ) of
	   Nothing    -> error("width information for acc matrix not found")
	   Just (id,a,b,c,matrx) -> nColumns matrx		    
 
-- divides pi-part into allocations,
-- but does not take lambdas with it 
--
allocat ispcCleaned accsCleaned placements = [cut (fst(x)) width | x <- placements  ] 
   where
   width = widthIdMatrices ++ widthAccMatrices
   -- widths of the unit matrices (identity matrices) 
   widthIdMatrices = [ nColumns x | (_,x) <- ispcCleaned]
   -- acc matrices are transposed 
   accTr = accTranspose accsCleaned 
   -- position information of the different access matrices in the matrix
   accessPosition = (acc2posAndWidth (accTr))
   -- width of the access matrices / array accesses 
   widthAccMatrices = accWidths
   -- to get the number and the ids of the arrays (TOO COMPLICATED)
   arrayIdsAndOffsets  = init(fmToList (snd(accessPosition)))
   -- calculates the width according to the provided id 
   accWidths =  [widthOfAcc id accTr|(id,off)<-arrayIdsAndOffsets]
       where
       -- helper: estimates the width of the specified transposed matrix
       widthOfAcc :: Integer -> [(Integer,AccessType,Integer,Integer,Matrix)] -> Integer
       widthOfAcc arrayId accTr =  
	   case ( find (\(id,a,b,c,matrx )-> arrayId==id)accTr ) of
	   Nothing    -> error("width information for acc matrix not found")
	   Just (id,a,b,c,matrx) -> nColumns matrx 




-- help to see the allocation better on the screen 
see xs = pretty [makeUp x | x <- xs]  

-- test 
--table = see bank 

-- helper for see
pretty list = putStr $ concat $ map (\x -> x ++"\n") list 

-- helper for see
makeUp list = concat [(show x) ++ "\t" | x <- list]

-- helper for printPlacements
makeUp2 list = concat [(show x)++ "\n"| x <- list]

-- to read better the placements 
printPlacements list = putStr $ makeUp2 list 
-- test definition 
--fcoResult  = fco indexspace accsClean
--allocations = separateAllocations indexspace accsClean fcoResult
--llocations2 = allocat indexspace acc fcoResult  

isZero :: [Integer] -> Integer -> Bool 
isZero [] i = False 
isZero xs 0 = True 
isZero (x:xs) i = if (x == 0) 
  then (True && (isZero xs (i-1)))
  else False  

isZero2 :: [[Integer]] -> Integer -> Integer -> Bool 
isZero2 [] i j = False 
isZero2 xs 0 j= False  
isZero2 (x:xs) i j= if (isZero x j) 
  then (True || (isZero2 xs (i-1) j))
  else False || (isZero2 xs (i-1) j) 



-- takes the string of the hard coded path
-- and returns the placements
--   
--fco3:: String -> [([Integer],Placement)]
fco3 path =  [(blockwidth, placement)|placement <- placementList]
	     where      
	     placementList = fco ispcCleaned accsCleaned
	     accsCleaned = cleanAccs acc slps 
             ispcCleaned = clean ispcDirty slps 
	     ispcDirty= unsafePerformIO( readispcfile path) 
	     acc  = unsafePerformIO( readaccfile path)
             slps = unsafePerformIO (readslpfile path)

             blockwidth = (widthIdMatrices,accWidths,sum(widthLambdas))  
	     --blockwidth = widthIdMatrices ++ accWidths ++ widthLambdas 
	     --             
	     -- widths of the unit matrices 
	     widthIdMatrices = [ nColumns x | (_,x) <-ispcCleaned]
	     -- width of the lambda part 
	     widthLambdas= [width | (_,_,(_,width)) <- posAndWidthDmatrix]
             -- width information of the transposed and prepared indexspace
	     -- matrices
	     posAndWidthDmatrix = ispc2position ispcCleaned accsCleaned 
	     -- acc list with transposed acc matrices 
	     accTr = accTranspose accsCleaned
	     -- position information of the different access matrices in the matrix
	     accessPosition = (acc2posAndWidth (accTr))
	     -- to get the number and the ids of the arrays (TOO COMPLICATED)
	     arrayIdsAndOffsets  = init(fmToList (snd(accessPosition)))
             -- calculates the width according to the provided id 
	     accWidths =  [widthOfAcc id accTr|(id,off)<-arrayIdsAndOffsets]
		 where
		 -- helper: estimates the width of the specified transposed matrix
		 widthOfAcc :: Integer -> [(Integer,AccessType,Integer,Integer,Matrix)] -> Integer
		 widthOfAcc arrayId accTr =  
		     case ( find (\(id,a,b,c,matrx )-> arrayId==id)accTr ) of
		     Nothing    -> error("width information for acc matrix not found")
		     Just (id,a,b,c,matrx) -> nColumns matrx 

--
-- better representation by separation between
-- lambda and pi widths 

-- takes the string of the hard coded path
-- and returns the placements
--   
fco4:: String -> [((Int,Integer),[Integer],[Integer],Placement)]
fco4 path =  [((distance pl d, volume pl d slps accsCleaned),blockwidth,widthLambdas, pl)|pl <- placementList]
	     where      
	     placementList = fco ispc accsCleaned
	     d = createDmatrix ispc accsCleaned 
	     ispc = clean ispcDirty slps
	     accsCleaned = cleanAccs acc slps 
	     ispcDirty = unsafePerformIO( readispcfile path) 
	     acc  = unsafePerformIO( readaccfile path)
	     slps = unsafePerformIO ( readslpfile path) 
	     blockwidth = widthIdMatrices ++ widthAccMatrices ++ widthOfNulldimArray
	     --             
	     -- widths of the unit matrices 
	     widthIdMatrices = [ nColumns x | (_,x) <-ispc]
	     -- width of the null dimensional array
	     widthOfNulldimArray = [fst(accessPosition)-sum(widthAccMatrices)]
	     -- width of the lambda part 
	     widthLambdas= [width | (_,_,(_,width)) <- posAndWidthDmatrix]
             -- width information of the transposed and prepared indexspace
	     -- matrices
	     posAndWidthDmatrix = ispc2position ispc accsCleaned
	     -- acc list with transposed acc matrices 
	     accTr = accTranspose accsCleaned
	     -- position information of the different access matrices in the matrix
	     accessPosition = (acc2posAndWidth (accTr))
	     -- width of the access matrices / array accesses 
	     widthAccMatrices = arrayAllok acc
		 where
		 -- helper to determine the array allocations 
		 -- and where they are supposed to cot off
		 -- 
		 arrayAllok :: [(Integer,AccessType,Integer,Integer, AffineFunction)] -> [Integer]
		 arrayAllok acc = [width | x <- numberOfAccesses] 
		     where 
		     -- CAUTION: Assume here that all acc matrices have the same width 
		     width =  head[nColumns (x)|(_,_,_,_,x) <- accTr ]
		     numberOfAccesses  = init(fmToList (snd(accessPosition)))

-- testing
--howmanyplacements = length (fco4 bsp)



-- helper function for a better presentation
-- of the placements 
--
printPlacements2  path  = putStr $ concat $ [(show b) ++ "\t\t" ++ (show p) ++ "\n\n"|(b,p)<-list]
    where list = fco3 path 

-- helper function to clean the indexspace matrices
-- in case of non existing loops 
-- 
--keep :: Matrix->[Integer]-> Integer->Matrix
--keep m [] offset = m 
--keep m (l:loops) offset = if l==0 then keep ( deleteColumns m offset 1) loops (offset + 1)
	--	   else keep m loops (offset+1)

--
--        discrete distance
--


-- parameters: indexspace-list and access-matrix list
--
dGroups :: [(Integer,Matrix)]->[(Integer,AccessType,Integer,Integer, Matrix)] -> [Matrix]
dGroups ispcs accs = map f preparedAccs
    where
    -- CAUTION: gets here acc
    preparedAccs = accTranspose accs
   
    -- the actual list comprehension
    f (arrayId,accType,node,stmtNr,accMtrx) 
    -- distinguishs if read or write access
	| accType == Read = dGr
	| accType == Write =dGr
        where
        dGr = indexspaceGroup ispcTuple accTuple totalWidth 
	-- the suitable tuple of the transformed
	-- ispc list 
	ispcTuple = search1 stmtNr convertedIndexspace
	-- the tuple from the access-matrix list
	accTuple = search2 node dmatrixPos
	-- the total width of this d-matrix group
        totalWidth = totalWidthOf dmatrixPos
        -- positions of the d-matrices
        dmatrixPos = ispc2position ispcs accs
	-- transformed ispc matrices in tuple list 
	convertedIndexspace = convert2Dmatrx2 ispcs
	accGr = accessGroup accMtrxsPos (arrayId,accType,node,stmtNr,accMtrx)
	-- positions of the access matrixes 
	accMtrxsPos = acc2posAndWidth preparedAccs
	
-- helper to create the part with the d-groups 
-- 
createDmatrix ispcs accs = (widths,(multijoin1 dGroupxs))
			   where
			   dGroupxs = (dGroups ispcs accs)
			   widths = [nRows dg | dg <- dGroupxs]

-- test definition 
--dPart = createDmatrix indexspace accsClean
--lambdas = [ snd(snd x) | x <- (fco3 bsp)]

--breiten = fst (dPart)


-- calculates the distance vector by multiplicating the 
-- lambda part of a placement with the dPart of the farkas
-- matrix 
--
calcDistance lambdas dpartInfo = [mulMV dpart (vector l) | l <- lambdas]
			     where 
			     dpart = snd(dpartInfo)

-- to calculate the discrete distance just for one placement
-- the result shows how many accesses are not zero 
--
distance placement dpartInfo = (length (fst partioned))
			     where
			     partioned = partition (==0%1) wrapped
			     wrapped = [sum a | a <- accesses]
                             total = length accesses
                             accesses = cut result widths
			     widths = fst(dpartInfo)
			     result = vectorEntries(mulMV dpart (vector lambdas))
				 where
				 dpart = snd(dpartInfo)
				 lambdas = snd (placement)

-- to calculate the communication volume just for one placement
-- the result shows how many accesses are not zero 
--
volume pl dpartInfo loops axs= sum  [v |(v,r)<-(fst notZero)]
			     where
                             notZero = partition (distanceNotZero) withSlps
                             distanceNotZero :: (Integer,Rational)-> Bool
                             distanceNotZero t = if snd(t)==(0%1) 
					       then False
					       else True
			     withSlps = [(sum(slps s loops), d) | (s,d)<-stats]
                             stats = zip stmts checked
			     checked = [sum a | a <- accesses]
                             stmts = stmtNrs axs
                             accesses = cut result widths
			     widths = fst(dpartInfo)
			     result = vectorEntries(mulMV dpart (vector lambdas))
				 where
				 dpart = snd(dpartInfo)
				 lambdas = snd (pl)


-- helper 
stmtNrs :: [(Integer,AccessType,Integer,Integer,Matrix)]->[Integer]
stmtNrs axs = [stmtnr | (_,_,_,stmtnr,_) <- axs]		            

--place = head placies

-- separates the result vector according to the widths of the
-- d-matrices 
--
commDis :: [GVector Rational]->[Integer]->[[[Rational]]]
commDis  distVectors widths = [ cut d widths| d <- dists]
    where
    dists = [vectorEntries distVec | distVec <- distVectors]




--check1 comdis = [sum xs | xs <- comdis] 

--check2 xs = [check1 x | x <- xs]


numberZeroDistance xs = [zeros | (zeros,_) <- partitioned]
		     where 
		     partitioned = [partition (==0%1) x | x <- xs]
		     total = length xs 

-- helper for function ranking: 
-- sorts the results of fco4
-- according to less communication distance (highest # of zeros)
-- and highest communication volumina 
--
better:: ((Int,Integer),[Integer],[Integer],Placement)->((Int,Integer),[Integer],[Integer],Placement)-> Ordering
better a b 
	   |fst (first a) > fst(first b) = LT
           |fst (first a) == fst (first b) = if snd ( first a) > snd (first b) 
			     then LT 
			     else GT
	   |otherwise = GT
	   where
	   first ::(a,b,c,d)->a
           first (a,b,c,d)= a 


-- ranking functions of the placements 
-- 
ranking :: String -> [((Int,Integer),[Integer],[Integer],Placement)]
ranking str = sortBy better (fco4 str)



