module Rabltest
where

import LoopoDS
import LMath
import List

bsp = LoopoDS.unsafeReadLoopoDS "/home/cl/rosentha/testdata/matmult-imper.src.out"
ispcs=List.map Prelude.snd (LoopoDS.loopoISPC bsp)
hdispc = Prelude.head ispcs
 
myVol :: IneqSystem -> DTree (Polynomial Rational) (FracPoly Rational)
myVol ies = makeDTree (fracPolyDTreeContext () []) (volume (getMatrix (stdIneq2FPIneq ies)))



--stdIneq2FPIneq :: IneqSystem -> GIneqSystem (FracPoly (Ratio b))

--(MGen [FracPoly a] (FracPoly (Ratio b)),  GHC.Real.Integral b) =>
stdIneq2FPIneq (ies :: IneqSystem ) =
    let mat :: GMatrix Rational
        mat = getMatrix ies
	matFracPoly :: GMatrix (FracPoly Rational)
	matFracPoly = convertMatrixEntries convertToFracPoly mat
	varMat :: GMatrix (FracPoly Rational)
	varMat = subMatrixVer matFracPoly 0 (nVars ies-1)
        paramList :: [FracPoly Rational]
	paramList = [fromPoly (indeterminate p i) | 
		     let p = LMath.nParams ies, i<-[0..p-1]] ++
		    [convertToFracPoly 1]
	paramConv :: [Rational] -> FracPoly Rational
	paramConv paramPartCoeff = 
	    sum (zipWith (*) (map fromRational paramPartCoeff) paramList)
	paramMat :: GMatrix Rational 
	paramMat = subMatrixVer mat (nVars ies) (nColumns mat)  -- -1
	convertedParamMat :: GMatrix (FracPoly Rational)
	convertedParamMat = columnMatrix (map paramConv (rows' paramMat))
	combinedResMat ::  GMatrix (FracPoly Rational)
	combinedResMat = varMat `conc` convertedParamMat
	resIes :: GIneqSystem (FracPoly Rational)
	resIes = ineqSystem((nVars ies,0),combinedResMat)
    in resIes


testTiling1 =
  let
    p = fromPoly (indeterminate 2 0)
    n = fromPoly (indeterminate 2 1)

    --              i  t    1
    hyp = matrix [[ 1, 0,   0]
                 ,[ 1,-p,   0]
                 ,[-1, 0,   n]
                 ,[-1, p, p-1]
                 ]
    result = -- chernikova hyp
             volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

test0 =
  let
    a = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],[0,1,0],[-1,0,a],[0,-1,a]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

testVol0 =
  let
    a = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],[0,1,0],[-1,0,a],[0,-1,a]]
    result = volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

testVol6 =
  let
    a = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0,0],[0,1,0,0],[0,0,1,0],[-1,0,0,a],[0,-1,0,a],[0,0,-1,a],[1,-1,0,0],[1,0,-1,0]]
    result = volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

test1 =
  let
    a = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,-a]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

test2 = 
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],
                  [-1,0,4],
                  [0,1,0],
                  [0,-1,4],
                  [p,-1,2-(2*p)]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

test3 =
  let
    hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,1]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

test4 =
  let
    hyp = matrix [[1,0,2],[0,1,2],[-1,0,2],[0,-1,2]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

test5 = 
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],
                  [0,1,0],
                  [0,-1,4],
                  [p,-1,2-(2*p)]]
    result = chernikova hyp
  in makeDTree (fracPolyDTreeContext () []) result

testVol1 = 
  let
    a = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,6],[0,-1,5],[1,-1,-a]]
    vol = volume hyp
  in makeDTree (fracPolyDTreeContext () []) vol

testVol2 = 
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],
                  [-1,0,4],
                  [0,1,0],
                  [0,-1,4],
                  [p,-1,2-(2*p)]]
    vol = volume hyp
  in makeDTree (fracPolyDTreeContext () []) vol

testVol5 =
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],
                  [0,1,0],
                  [0,-1,4],
                  [p,-1,2-(2*p)]]
    vol = volume hyp
  in makeDTree (fracPolyDTreeContext () []) vol

testSimpleN1 =
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,-1],[0,1,-1],[-1,0,p],[0,-1,p]]
    result = volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

testSimpleN0 =
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],[0,1,0],[-1,0,p],[0,-1,p]]
    result = volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

testSimpleN_10 =
  let
    p = fromPoly (indeterminate 1 0)
    hyp = matrix [[1,0,0],[0,1,0],[-1,0,p-1],[0,-1,p-1]]
    result = volume hyp
  in makeDTree (fracPolyDTreeContext () []) result

