module ScannerParser.FineCode (printFineCode, renameSTInds) where

import ScannerParser.ParseTree
import ScannerParser.SymbolTable
import ScannerParser.LambdaExpression
import ScannerParser.PrintCode 
import Data.List
import LoopoDS.Qual

-- | for displaying syntax tree in C code.

-- | Prints C-code, but without useless expressions
-- (no 1*x, but just x)
printFineCode :: SymTab -> LambdaTree -> String
printFineCode st lt = printCode TL_PlainC st (map (elemUnaryMinus.shrinkLN) lt)
-- noch unschön: Verbesserungen &&, || (False && * -> False; 1 || * = True)


-- | Indexumbenennung
-- Zeitindizes t1, t2, ...
-- Placementindizes p1, p2, ...
-- Additionals a1, a2, ...
renameSTInds :: SymTab -> [(a, [QualEntry])] -> SymTab
renameSTInds st qvs = let qv              = maxQV qvs
                          iqes            = zip [1..] qv
                          splitQEs        = [(filter (sndIs Sched) iqes),
                                             (filter (sndIs Place) iqes),
                                             (filter (sndIs Additional) iqes)]
                          sndIs qe iqe    = qe == snd iqe 
                          ijqes           = concat $ map (zip [1..])  $ splitQEs
                          m st (i,(j,qe)) = modifyLNSymbols (renIndex j i qe) st
                      in foldl m st ijqes

renIndex :: Int -> Int -> QualEntry -> LNSymbol -> LNSymbol
renIndex i j qe lns@(Symbol{sym_name=n, sym_type=LNS_Index}) | n == i 
                     = lns{sym_lexem=(qeString qe):(show j)}
renIndex _ _ _  lns  = lns

qeString :: QualEntry -> Char
qeString Sched      = 't'
qeString Place      = 'p'
qeString Additional = 'a' 

-- QualVects haben alle den gleichen Präfix, daher einfach den längsten heraussuchen.
maxQV :: [(a, [b])] -> [b]
maxQV tstmtqvs = let qvs     = map snd tstmtqvs
                     maxsize = maximum $ (map length qvs)
                 in maybe [] id $ find (\qv -> length qv == maxsize) qvs
