module Main

where

import List
import ScannerParser.ScannerParser
import ScannerParser.ParseTree
import ScannerParser.SymbolTable
import ScannerParser.PrintCode
import ScannerParser.FineCode
--import System.Directory (getDirectoryContents)
import Data.Char
import Data.Bits
import System 
import LoopoDS

import HXML.HXML -- http://www.flightlab.com/~joe/hxml/

main :: IO ()
main = do         
        args <- getArgs
        if (length args)<2 then showhelp else convert args
        return ()
       where
        showhelp = do 
                    putStrLn "Usage: ScannerParser Command InFile [OutFile]"
                    putStrLn "Command: One of the following"
                    putStrLn "         printcode       to generate pretty C code"
                    putStrLn "         printcodeexact  to generate C code from the original parse tree"
                    putStrLn "         xmlconvert      to convert the parse tree information into XML"
                    putStrLn "         targout         to generate pretty C code with using QualVect information"
                    putStrLn "InFile:  Name of the ScannerParser file to parse" 
                    putStrLn "Outfile: Name of the file to save the output to. If not given, screen is used."
                    return ()
        convert args = do
                    let command = map toLower (args!!0)
                        ptinfile  = if command=="targout"
                                     then ((args!!1)++".parse")
                                     else args!!1
                    (st, pt, flns, ivs) <- openParseFile (ptinfile)
                    let orginfile = args!!1
                        outfile = args!!2
                        loopoDS = unsafeReadLoopoDS orginfile
                        qualVects = loopoQUAL loopoDS
                        newSymTab = renameSTInds st qualVects
                        output  = if command=="printcode"
                                   then printFineCode TL_PlainC st pt
                                   else if command=="printcodeexact"
                                    then printCode st pt
                                    else if command=="xmlconvert"
                                     then convertFile (st, pt, flns, ivs) 
                                     else if command=="targout"
                                      then printFineCode newSymTab pt
                                      else error "Invalid command."
                    if (length args)==3 
                     then writeFile outfile output
                     else putStrLn  output
                    return ()                
