module ScannerParser.ScannerParser
    (
      parseFile, openParseFile, convertFile, parseAndSaveFile,
      InfoVector
    )
where

import Data.List
import ScannerParser.ParseTree
import ScannerParser.SymbolTable
import ScannerParser.LambdaExpression
import ScannerParser.FineCode
import System.Directory (getDirectoryContents)
import Data.Bits
import LoopoDS.IntegralID

import HXML.HXML -- http://www.flightlab.com/~joe/hxml/

-- | Infovector                   
type InfoVector = (Int,       -- Node ID of the InfoVect
                   Int,       -- Number of module
                   [String])  -- List of tags

-- |Argument list:
-- 
-- @ InfoVect ID occ type name state stmt op tstmt step up low cond indices children elsechild arguments next "label" # @
type MatrixLine = (Maybe InfoVector,Int,Int,Int,Int,Int,Int,Int,
                   Int,Int,Int, Int, Int, Int,
                   Int, Int, Int, Int, String)  
                   
instance Show (Maybe InfoVector,Int,Int,Int,Int,Int,Int,Int,
                   Int,Int,Int, Int, Int, Int,
                   Int, Int, Int, Int, String) where
 show (iv,i1,i2,i3,i4,i5,i6,i7, i8,i9,i10, i11, i12, i13, i14, i15, i16, i17, s) 
    = "(" ++ (show iv) ++ ", " 
          ++ (show i1) ++ ", "
          ++ (show i2) ++ ", "
          ++ (show i3) ++ ", "
          ++ (show i4) ++ ", "
          ++ (show i5) ++ ", "
          ++ (show i6) ++ ", "
          ++ (show i7) ++ ", "
          ++ (show i8) ++ ", "
          ++ (show i9) ++ ", "
          ++ (show i10) ++ ", "
          ++ (show i11) ++ ", "
          ++ (show i12) ++ ", "
          ++ (show i13) ++ ", "
          ++ (show i14) ++ ", "
          ++ (show i15) ++ ", "
          ++ (show i16) ++ ", "
          ++ (show i17) ++ ", "
          ++ (show s) ++ ")"

-- |teilt einen String am Vorkommen eines Substrings auf.
splitString :: String -> String -> (String, String)
splitString pattern text = (take (p_begin) text, 
                            drop (p_begin + (length pattern)) text) where 
                              Just p_begin = findPosition pattern text

findPosition :: String -> String -> Maybe Int
findPosition pattern text = 
   let fp [] _ = Nothing
       fp rest pos = if isPrefixOf pattern rest
                     then Just pos
                     else fp (tail rest) (pos+1)
   in fp text 0 
   
      
-- |reads <project>.out and gives back the syntax tree.
parseFile :: String -> (SymTab, LambdaTree, [LambdaNode], [InfoVector])
parseFile project_out = (getSymTabFromFile symtab,
                         pt, 
                         lns,
                         fstivs ++ sndivs) 
                        where
                         (symtab, triml) = splitString "MainProgram:\n\n" project_out
                         (mp_p_o, f_p_o) = splitString  "\n\nFunctions:\n" triml 
                         (pt,  fstivs) = getParseTreeFromMatrix mp_p_o
                         (lns, sndivs) = getFunctionsFromFile f_p_o

ifsrcoutmppo = "InfoVect: 2 -1\n0 1 0 1 2 2 -1 -1 -1 -1 -1 48 18 4 1 30 0 0 0 1 1 1 0 0 1 0 0 0 \"\" #\n47 5 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n46 4 2 0 0 0 280 0 0 0 0 0 0 1 1 0 0 \"\" #\n32 2 2 5 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n45 3 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n31 1 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n1 1 0 0 35 8 2 0 0 1 276 0 0 0 0 0 0 1 1 0 0 \"\" #\n33 6 2 1 4 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n34 7 2 2 3 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n2 2 48 -1 0 -1 -1 44 17 1 2 2 1 0 0 0 0 0 0 1 1 0 0 1 \"\" #\n51 -1 2 1 4 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n35 8 2 0 0 1 276 0 0 0 0 0 0 1 1 0 0 \"\" #\n33 6 2 1 4 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n34 7 2 2 3 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n0 1 0 1 2 2 48 44 0 3 -1 37 16 1 1 2 2 0 0 0 0 0 0 1 1 0 0 0 \"\" #\n36 15 2 1 4 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n43 14 2 0 0 2 279 0 0 0 0 0 0 1 1 0 0 \"\" #\n41 12 2 1 1 2 0 0 0 0 0 0 1 0 0 0 0 \"\" #\n40 11 2 0 0 2 280 0 0 0 0 0 0 1 1 0 0 \"\" #\n38 9 2 1 4 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n39 10 2 1 3 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n42 13 2 1 3 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #"

-- | teilt einen String mit Function-Liste in Strings mit je einer Function auf.
funcStrings :: String -> [String]
funcStrings s = if (isPrefixOf "================\n" s)
                then x : (funcStrings ys)
                else []
                where 
                 (_,xs) = splitString  "================\n" s
                 (x,ys) = splitString  "\n\n" xs

-- |opens <project>.out with the given file name and returns syntax tree
openParseFile :: String -> IO (SymTab, LambdaTree, [LambdaNode], [InfoVector])
openParseFile fn = do
                    project_out <- readFile fn   
                    {- putStrLn (if (findPosition  "InfoVect: " project_out /= Nothing) 
                               then  "*** Warning: InfoVect information will be ignored"
                               else "") -}
                    let parsedfile = parseFile project_out
                    return parsedfile
                    
parseAndPrintFile :: String -> IO ()
parseAndPrintFile fn = do
                        putStrLn ("- - - File " ++ fn)
                        (st, pt, flns, ivs) <- openParseFile fn
                        --putStrLn (show st) 
                        putStrLn (printFineCode st pt)
                        --putStrLn $ printCode flns
                        --putStrLn "Converting to and from XML..."
                        --let xml = (showXML (ScannerParser.SymbolTable.toXML st))
                        --    Just xmlst = ScannerParser.SymbolTable.fromXML (parseXML xml)                            
                        --putStrLn ("Checking equality: " ++ show (xmlst == st))
                        --putStrLn (show ivs)
                        putStrLn ""
                        return ()

-- | Converts a given ScannerParser2 file into an XML file.
convertFile :: (SymTab, LambdaTree, [LambdaNode], [InfoVector]) -> String
convertFile (st, pt, flns, _) = showXML (Tree (ELNode "ScannerParser" []) [
                             Tree (ELNode "MainProgram" []) [parseTreeToXML pt], 
                             Tree (ELNode "FuncDecls"   []) [parseTreeToXML flns],
                             symTabToXML st
                            ])
   

-- | ffnet eine C-Datei und speichert sie in eine XML-Datei   
--
-- InfoVector werden ignoriert!
parseAndSaveFile :: String -> String -> IO ()
parseAndSaveFile rfn wfn = do
                            (st, pt, flns, ivs) <- openParseFile rfn
                            let xml = convertFile (st, pt, flns, ivs)
                            writeFile wfn xml
                            return ()

test :: IO ()
test = do 
        allfiles <- getDirectoryContents "."
        let files = filter (inList "out") allfiles where
                        inList s xs = case findPosition s xs of
                                       Nothing -> False
                                       Just _  -> True
        putStrLn (unlines (zipWith (++) (map ((++ " ").show) [0..]) files)) -- Dateiliste ausgeben
        putStr "Dateinummer (a fuer alle): "
        fn <- getLine
        if (head fn) == 'a' then
          mapM parseAndPrintFile (files)
          else
          mapM parseAndPrintFile ([files !! read(fn)]) 
        return ()
          
-- |nimmt die gesamte Matrix aus <project>.out und gibt den Syntaxbaum zurck
-- Vorgehen: Aufsplitten nach Zeilen
--           Aufsplitten nach Leerzeichen
--           Aussortieren der zu kurzen Listen (InfoVect-Zeilen)
--           bersetzen in Typ MatrixLine (String->Int Konvertierung)
--           Erstellen des Syntaxbaums (oberste Ebene ist immer ein Block)
getParseTreeFromMatrix :: String -> (LambdaTree, [InfoVector])
getParseTreeFromMatrix s = if null rest then (lns, ivs) 
                           else error "Malformatted parsefile! Unreachable LambdaNodes."
 where 
  (rest, lns, ivs) = makeTreeFromList $ map translateMatrixLine $ findInfoVector 
                  $ map (words) $ lines s 
                  
-- | Gibt die Liste der Funktionen zurck
getFunctionsFromFile :: String -> ([LambdaNode], [InfoVector])
getFunctionsFromFile s = (lns, ivs) where
                         (_, lns, ivs) = makeTreeFromList $ map func2ML $ 
                                         map lines $ funcStrings s

-- | bersetzt Liste von Funktionen zu MatrixLines; Hilfsfunktion fr getFunctionsFromFile
-- 
-- Zum Verstndnis siehe project.out FctDecls
func2ML fs = (Nothing, gV "ID", gV "occ_ID", gV "typ", gV "name", gV "state",
              gV "stmnt", gV "op", gV "Tstmt", gV "step", gV "up",
              gV "low", gV "cond", gV "indices", gV "children", gV "elsechild", 
              gV "arguments", gV "next", "")
              where
              wfs             = map words $ map (drop 3) fs
              gV              = getVal wfs
              getVal []     s = error ("Not found: " ++ s)
              getVal (x:xs) s = if s==head x 
                                then read (x!!1)
                                else getVal xs s 
             
-- |sucht nach InfoVektor-Zeilen, um sie zu einer Zeile zusammenzufgen
findInfoVector :: [[String]] -> [[String]]             
findInfoVector (x:xs) | (length x == 3)
                      = (concat ((tail x):(take l xs))):(findInfoVector (drop l xs)) -- hier den neuen Code rein!
                        where
                         l = getInfoVectLength (x:xs)
findInfoVector (x:xs) = x:(findInfoVector xs)
findInfoVector []     = []

-- | Findet die Zeilenanzahl eines InfoVects in der Matrix heraus. 
-- Einleitungszeile ist Zeile 0!
getInfoVectLength :: [[String]] -> Int
getInfoVectLength (x:xs) = case (read (x!!1)) of -- Typ des InfoVects herausfinden
                           2 -> normLoopLength 
                           0 -> 1 -- Scanpars
                           1 -> 1 -- Targout
                           3 -> 1 -- Transif
                           _ -> error "Unknown InfoVect type - no length information available"
                           where
                           normLoopLength = case read ((head xs)!!3) of
                                            1 -> 1 -- keine eingebundenen Knoten -> nur eine Zeile
                                            0 -> getNormLoopLength 2 (tail xs) -- eingebundene Knoten, dynamische Lnge
                                            
-- | Sucht das Ende eines NormLoops-InfoVects, der Knoten enthlt und deshalb beliebig lang sein kann.
-- 
-- Erkennungszeichen: Nchste Zeile lnger als 19 Zeichen, die nicht ein anderer InfoVector ist.
getNormLoopLength :: Int -> [[String]] -> Int
getNormLoopLength i (y:ys) | (length y) == 3 = getNormLoopLength (i+l+1) (drop l ys) -- Inline-InfoVect
                                                where l = getInfoVectLength (y:ys)                                               
getNormLoopLength i (y:ys) | (length y) > 19 = i    -- Ende gefunden
getNormLoopLength i (y:ys) | otherwise       = getNormLoopLength (i+1) ys -- weitersuchen
getNormLoopLength i _      = error "Did not find end of NormLoops-InfoVect!"
                        
-- |bersetzt die Stringliste in ein MatrixLine-Paar
translateMatrixLine :: [String] -> MatrixLine
{- translateMatrixLine lns | (length lns>20) -- InfoVect umwandeln
                = translateMatrixLine ((unwords (take infoVectLength lns)) :(drop infoVectLength lns))
                  where infoVectLength = length lns-19
translateMatrixLine lns | (length lns == 19) -- Zeile ohne InfoVect
                = translateMatrixLine ("":lns) -}
translateMatrixLine lns|length lns>=19 
                = (infovect, read(lnid), read(occ), read(lntype), read(name), read(state),
                   read(stmt), read(op), read(tstmt), read(step), read(up),
                   read(low), read(cond), read(indices), read(children),
                   read(elsechild), read(arguments), read(next), read(label)) 
                  where
                    infoVectLength = length lns-19
                    infoVectList = take infoVectLength lns
                    infovect = case infoVectList of 
                               []                -> Nothing
                               (modnr:"-1":tags) -> Just (read lnid, read modnr, tags)
                               _                 -> Nothing -- error ("Wrong InfoVect pattern: " 
                                                    --       ++ (show infoVectList))
                                                    -- das falsche Pattern ist leider mglich bei
                                                    -- dynamischen NormLoops-InfoVects. 
                    [lnid,occ,lntype,name,state,stmt,op,tstmt,step,up,
                       low,cond,indices,children,elsechild,arguments,next,
                       label,"#"] = drop infoVectLength lns

translateMatrixLine xs = error ("Wrong matrix format:" ++ (show xs))


-- |berfhrt die Liste in eine Baum
-- Rckgabewert ([ML], [LN]): [ML] sind die briggebebliebenen Matrixlines nach 
-- Erstellen von LN. Die Rckgabe ist ntig, um weiter rekursiv auf der Matrix
-- abzusteigen.
-- LambdaTree ist eine Liste von Syntaxbumen. Die Liste ist ntig fr Schleifenkrper,
-- arguments, indices etc.
-- An dieser Stelle wird der InfoVector als zweites Tupelelement nach oben zurckgegeben;
-- erstes Tupelelement ist die entsprechende nodeid.
makeTreeFromList :: [MatrixLine] -> ([MatrixLine], LambdaTree, [InfoVector])
makeTreeFromList ((infovect, lnid,occ,lntype,name,state,stmt,op,tstmt,step,up,
                       low,cond,indices,children,elsechild,arguments,next,
                       label):xs) 
              = (next_xs, 
                 (case lntype of
                  1    -> Assign {nodeid=lnid, occurence=occ, name=name, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt),
                           indices=indices_tree,
                           firstExpr=head children_tree} 
                  4    -> ForLoop {nodeid=lnid, occurence=occ, name=name, 
                           state=getLNStateFromInt state, 
                           loopStride=listToMaybe stride_tree, 
                           loopUpperBound=listToMaybe up_tree,
                           loopLowerBound=listToMaybe low_tree, 
                           loopBody=children_tree}
                  6    -> If {nodeid=lnid, occurence=occ, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt),
                           condition=listToMaybe cond_tree,
                           thenbranch=children_tree, 
                           elsebranch=elsechild_tree}
                  3    -> Skip {nodeid=lnid, occurence=occ, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt)}
                  5    -> WhileLoop {nodeid=lnid, occurence=occ, name=name, 
                           state=getLNStateFromInt state, 
                           loopStride=listToMaybe stride_tree, 
                           loopLowerBound=listToMaybe low_tree, 
                           condition=listToMaybe cond_tree,
                           loopBody=children_tree} 
                  7    -> FuncDecl {nodeid=lnid, occurence=occ, name=name, 
                           state=getLNStateFromInt state, 
                           funcbody=children_tree, 
                           arguments=arguments_tree}
                  8    -> ProcCall {nodeid=lnid, occurence=occ, name=name, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt),
                           arguments=arguments_tree}
                  2    -> case (getLNStateFromInt state) of
                           LNS_OK       -> 
                             case op of
                             40 -> head children_tree -- ehemals Bracket; wird bersprungen, hat immer ein Kind
                             _  -> Expression {nodeid=lnid, occurence=occ, 
                                   operator=getLNOpFromInt op, 
                                   statement=(integerToID stmt),
                                   targetStatement=(integerToID tstmt),
                                   condition=listToMaybe cond_tree,        
                                   firstOperand=listToMaybe children_tree,
                                   secondOperand=listToMaybe elsechild_tree}
                           LNS_Read -> 
                             ExprRead {nodeid=lnid, occurence=occ, name=name, 
                             statement=(integerToID stmt),
                             targetStatement=(integerToID tstmt),
                             indices=indices_tree}               -- state FunctionCall
                           LNS_FunctionCall -> 
                             ExprFuncCall {nodeid=lnid, occurence=occ, name=name, 
                             statement=(integerToID stmt),
                             targetStatement=(integerToID tstmt),
                             arguments=arguments_tree}
                           LNS_Proto -> -- Funktionsargumente bei Deklarationen - evtl. anderer Typ
                             ExprIndBlobConst {nodeid=lnid, occurence=occ, name=name, 
                             state=getLNStateFromInt state, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt)}                     
                           _           -> 
                             if (state<=5 && state >= 3) then -- nur Const (3), Index (4), Blob (5) erlaubt
                             ExprIndBlobConst {nodeid=lnid, occurence=occ, name=name, 
                             state=getLNStateFromInt state, 
                           statement=(integerToID stmt),
                           targetStatement=(integerToID tstmt)}
                             else error ("Expression state not known: "++(show state))
                  9    -> Label {nodeid=lnid, occurence=occ, lexem=label}
                  10   -> GoTo {nodeid=lnid, occurence=occ, lexem=label})
                 :next_tree,
                 (case infovect of 
                  Nothing -> infvects
                  Just iv -> iv:infvects)
                 )
                 where
                  (stride_xs, stride_tree, iv1)   = if step==1 then 
                                                     makeTreeFromList xs
                                                   else (xs, [], [])
                  (up_xs    , up_tree    , iv2)   = if up==1 then 
                                                     makeTreeFromList stride_xs
                                                   else (stride_xs, [], [])
                  (low_xs   , low_tree   , iv3)   = if low==1 then 
                                                      makeTreeFromList up_xs
                                                   else (up_xs, [], [])
                  (cond_xs  , cond_tree  , iv4)   = if cond==1 then 
                                                      makeTreeFromList low_xs
                                                    else (low_xs, [], [])
                  (indices_xs, indices_tree, iv5) = if indices==1 then 
                                                     makeTreeFromList cond_xs
                                                   else (cond_xs, [], [])
                  (children_xs, children_tree, iv6) = if children==1 then 
                                                     makeTreeFromList indices_xs
                                                   else (indices_xs, [], [])
                  (elsechild_xs, elsechild_tree, iv7) = if elsechild==1 then 
                                                      makeTreeFromList children_xs
                                                   else (children_xs, [], [])
                  (arguments_xs, arguments_tree, iv8) = if arguments==1 then 
                                                      makeTreeFromList elsechild_xs
                                                   else (elsechild_xs, [], [])
                  (next_xs  , next_tree, iv9)     = if next==1 then 
                                                      makeTreeFromList arguments_xs
                                                   else (arguments_xs, [], [])
                  infvects = iv1 ++ iv2 ++ iv3 ++ iv4 ++ iv5 ++ iv6 ++ iv7 ++ iv8 ++ iv9
                  listToMaybe xs                  = case xs of 
                                                      []    -> Nothing
                                                      [x]   -> Just x
                                                      _     -> error "Cannot convert list to Maybe"                 
makeTreeFromList _ = error ("Wrong matrix format!")


-- | Wandelt die Konstanten der C-Klassen in Haskell-Daten um
getLNStateFromInt :: Int -> LambdaNodeState
getLNStateFromInt i = case i of
                         0  -> LNS_OK
                         2  -> LNS_Write
                         1  -> LNS_Read
                         3  -> LNS_Constant
                         4  -> LNS_Index
                         5  -> LNS_Blob
                         6  -> LNS_Array
                         7  -> LNS_Key
                         8  -> LNS_Function
                         9  -> LNS_FunctionCall
                         10 -> LNS_Proto
                         20 -> LNS_ReadAndWrite
                         50 -> LNS_Various             -- Functions
                         30 -> LNS_Do 
                         31 -> LNS_DoAll 
                         32 -> LNS_DoAcross 
                         40 -> LNS_TilDo 
                         41 -> LNS_TilPar 
                         42 -> LNS_TilAss
                         97 -> LNS_Statement                         
                         _  -> error ("State not known: " ++ (show i))
                         
-- | Wandelt die Konstanten der C-Klassen in Haskell-Daten um
getLNOpFromInt :: Int -> LambdaNodeOp
getLNOpFromInt i = case i of 
                     279  -> LNOp_Plus
                     280  -> LNOp_Minus
                     319  -> LNOp_Minus --unary Minus
                     {- 40   -> LNOp_Bracket --} 
                     281 -> LNOp_Times 
                     282 -> LNOp_Div 
                     303 -> LNOp_Mod 
                     310 -> LNOp_Not 
                     309 -> LNOp_And 
                     308 -> LNOp_Or
                     276 -> LNOp_GE 
                     275 -> LNOp_GEQ 
                     316 -> LNOp_NEQ 
                     273 -> LNOp_EQ 
                     274 -> LNOp_EQ --EQEQ
                     272 -> LNOp_LEQ 
                     271 -> LNOp_LE 
                     293 -> LNOp_IIf 
                     -- LNOp_Parallel 
                     0   -> LNOp_None -- nicht sicher
                     _   -> error ("Operator not known: " ++ show i)
                     
testMatrix = "InfoVect: 2 -1\n0 2 2 1 2 2 -1 -1 -1 -1 -1 59 29 4 1 30 0 0 0 1 1 1 0 0 1 0 0 0 \"\" #\n58 5 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n34 4 2 0 0 0 280 0 0 0 0 0 0 1 1 0 0 \"\" #\n32 2 2 1 5 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n33 3 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n31 1 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n52 23 4 2 30 0 0 0 1 1 1 0 0 1 0 0 1 \"\" #\n51 14 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n42 13 2 -1 9 0 0 0 0 0 0 0 0 0 0 1 0 \"\" #\n38 9 2 0 0 0 280 0 0 0 0 0 0 1 1 0 1 \"\" #\n36 7 2 1 5 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n37 8 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n41 12 2 0 0 0 279 0 0 0 0 0 0 1 1 0 0 \"\" #\n39 10 2 1 4 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n40 11 2 2 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n0 2 2 1 2 2 -1 -1 -1 -1 -1 59 35 6 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n44 22 1 1 2 1 0 0 0 0 0 0 1 1 0 0 0 \"\" #\n43 21 2 1 4 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n50 20 2 0 0 1 279 0 0 0 0 0 0 1 1 0 0 \"\" #\n46 16 2 1 1 1 0 0 0 0 0 0 1 0 0 0 0 \"\" #\n45 15 2 1 4 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n49 19 2 2 1 1 0 0 0 0 0 0 1 0 0 0 0 \"\" #\n47 17 2 1 4 1 0 0 0 0 0 0 0 0 0 0 1 \"\" #\n48 18 2 2 4 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n53 28 1 3 2 2 0 0 0 0 0 0 0 1 0 0 0 \"\" #\n57 27 2 0 0 2 279 0 0 0 0 0 0 1 1 0 0 \"\" #\n54 24 2 3 1 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n56 26 2 1 1 2 0 0 0 0 0 0 1 0 0 0 0 \"\" #\n55 25 2 1 4 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n"
testMatrix2 = "InfoVect: 2 -1\n0 1 1 1 2 1 -1 -1 -1 31 2 1 1 2 1 0 0 0 0 0 0 0 1 0 0 1 \"\" #\n32 1 2 0 3 1 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n4 1 1 1 2 1 -1 -1 -1 40 -1 1 3 2 3 0 0 0 0 0 0 0 1 0 0 1 \"\" #\n33 3 2 1 1 3 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n3 1 1 1 2 1 -1 -1 -1 55 -1 1 4 2 4 0 0 0 0 0 0 0 1 0 0 1 \"\" #\n42 -1 2 0 0 4 280 0 0 0 0 0 0 1 1 0 0 \"\" #\n34 4 2 42 3 4 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n41 -1 2 3 1 4 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n5 1 1 1 2 1 -1 -1 -1 59 -1 1 6 2 5 0 0 0 0 0 0 0 1 0 0 1 \"\" #\n58 -1 2 -2 9 5 0 0 0 0 0 0 0 0 0 1 0 \"\" #\n56 -1 2 6 1 5 0 0 0 0 0 0 0 0 0 0 1 \"\" #\n57 -1 2 4 1 5 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n0 1 1 0 70 -1 2 0 0 0 309 0 0 0 0 0 0 1 1 0 0 \"\" #\n66 -1 2 0 0 0 309 0 0 0 0 0 0 1 1 0 0 \"\" #\n63 -1 2 0 0 0 272 0 0 0 0 0 0 1 1 0 0 \"\" #\n43 -1 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n62 -1 2 1 4 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n65 -1 2 0 0 0 272 0 0 0 0 0 0 1 1 0 0 \"\" #\n64 -1 2 1 4 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n61 -1 2 4 1 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n69 -1 2 0 0 0 303 0 0 0 0 0 0 1 1 0 0 \"\" #\n68 -1 2 0 0 0 280 0 0 0 0 0 0 1 1 0 0 \"\" #\n67 -1 2 1 4 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n43 -1 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n35 5 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n2 1 -1 -1 -1 39 9 4 1 30 0 0 0 1 1 1 0 0 1 0 0 0 \"\" #\n35 5 2 1 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n60 -1 2 1 5 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n43 -1 2 0 3 0 0 0 0 0 0 0 0 0 0 0 0 \"\" #\nInfoVect: 2 -1\n0 1 1 1 2 1 39 1 5 37 8 1 2 2 2 0 0 0 0 0 0 1 1 0 0 0 \"\" #\n45 -1 2 0 0 2 279 0 0 0 0 0 0 1 1 0 0 \"\" #\n44 -1 2 1 4 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n41 -1 2 3 1 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #\n38 6 2 0 3 2 0 0 0 0 0 0 0 0 0 0 0 \"\" #"
-- test = getParseTreeFromMatrix testMatrix


-- getSymTabFromFile :: String -> SymTab
getSymTabFromFile s = mapSymTab makeSymTabFrame $
                      mapSymTab (map makeFrameSymbols) $
                      mapSymTab splitFrameSymbols $ 
                      splitSymTabFrames 
                      $ drop 2 $ lines s

-- | Teilt die Stringliste der Symbole in ihre Frames ein.
splitSymTabFrames :: [String] -> [(Int, [String])]
splitSymTabFrames (x:y:xs) | (x == "================") && 
                                (isPrefixOf "// frame " y)
                            = (read(drop 9 y), takeWhile isFrameEnd xs)
                              :(splitSymTabFrames xs)
                              where 
                               isFrameEnd s = not (isPrefixOf "// frame " s 
                                                   || s == "SymbolTable END")
splitSymTabFrames (x:xs)    = splitSymTabFrames xs
splitSymTabFrames []        = []

-- | Teilt die Liste der Symbole eines Frames in einzelne Symbole auf
-- 
-- [[[String]]]: Symbol,Zeilen
splitFrameSymbols :: [String] -> [[String]]
splitFrameSymbols []     = []
splitFrameSymbols (x:xs) | x == "================" && not (null xs)
                              = takeWhile (/= "================") xs
                                :(splitFrameSymbols xs)
                         | otherwise
                              = splitFrameSymbols xs
                                                

-- | Wandelt die Strings in Symbole mit Bool um, der bestimmt, ob 
--  das folgende Symbol ein Replacement ist.
makeFrameSymbols :: [String] -> (LNSymbol, Bool)
makeFrameSymbols sym = (Symbol {sym_lexem=gV "lexem",
                                sym_token=read (gV "token"),
                                sym_type=getLNStateFromInt (read (gV "type")),
                                sym_name=read (gV "number"),
                                sym_indices=read (gV "indices"), 
                                sym_level=read (gV "level"),
                                sym_intent=getLNSIntentFromInt (read (gV "intent")), 
                                sym_basictype=getLNSBasicTypeFromInt (read  (gV "basictype")), 
                                sym_identify=read (gV "identify"),
                                sym_status=getLNSStatusFromInt (read (gV "status"))}, 
                        toEnum (read (gV "next")))
                        where
                            wfs             = map words $ map (drop 3) sym
                            gV              = getVal wfs
                            getVal []     s = "0" -- standardmig 0 annehmen; gefhrlich!
                            getVal (([]):xs) s= getVal xs s 
                            getVal (x:xs) s = if s==head x 
                                              then (x!!1)
                                              else getVal xs s 
                                              
-- | Wandelt Symbol/Bool Tupel in geschachtelte Liste vom Typ SymTabFrame um
makeSymTabFrame :: [(LNSymbol, Bool)] -> SymTabFrame
makeSymTabFrame [] = []
makeSymTabFrame xs = let nexts      = takeWhile (snd) xs
                         (repl, _)  = unzip ((xs !! (length nexts)):nexts)
                         other      = drop (length repl) xs
                     in repl:(makeSymTabFrame other)

getLNSIntentFromInt :: Int -> Maybe LNSIntent
getLNSIntentFromInt 299 = Just LNSI_In
getLNSIntentFromInt 300 = Just LNSI_Out
getLNSIntentFromInt 301 = Just LNSI_InOut
getLNSIntentFromInt 0   = Nothing

getLNSBasicTypeFromInt :: Int -> Maybe LNSBasicType
getLNSBasicTypeFromInt 19 = Just LNST_Unknown
getLNSBasicTypeFromInt 20 = Just LNST_Void
getLNSBasicTypeFromInt 21 = Just LNST_Int
getLNSBasicTypeFromInt 22 = Just LNST_Bool
getLNSBasicTypeFromInt 23 = Just LNST_Float
getLNSBasicTypeFromInt 24 = Just LNST_Template
getLNSBasicTypeFromInt 25 = Just LNST_Processor
getLNSBasicTypeFromInt 26 = Just LNST_UnsignedInt
getLNSBasicTypeFromInt 0  = Nothing

getLNSStatusFromInt :: Int -> LNSStatus
getLNSStatusFromInt i = LNSS {sym_valid     =(i .&.   1 > 0), 
                              sym_pure      =(i .&.   2 > 0), 
                              sym_no_warn   =(i .&.   4 > 0),
                              sym_dynamic   =(i .&.   8 > 0), 
                              sym_local     =(i .&.  16 > 0), 
                              sym_pseudo    =(i .&.  32 > 0), 
                              sym_normLoops =(i .&.  64 > 0), 
                              sym_assign_idx=(i .&. 128 > 0)}
