-- |
-- The module generates the polyhedral subdomains of an array that
-- are read or written in a program. Alternativeley, the corresponding 
-- loop nest can be returned.
--
-- Author: Martin Griebl
-- $Id: TaskData.hs 3726 2007-07-20 13:36:22Z classen $
--
module TaskData (taskData, taskDataLoop)
where

import LoopoDS
import LMath
import List
import Maybe(fromJust)
import CodeGen
import Data.Map(Map, (!))
import qualified Data.Map as Map
import NewCommCode hiding(main)
import Debug.Trace(trace)

cmdLine ::  String -> String -> [[Integer]] -> [Integer] -> [Bool] -> CommandLine
cmdLine tmpdir projectname widths numbers selecteds = 
    let commandLine :: [String]
    	commandLine = [	tmpdir, projectname, show widths, show numbers, 
                        show selecteds, "../loopo/targout/targout", "", 
			"--useCommunication=True",
			"--cyclicMapping=True", "--useMPIWrapper=False", 
			"--unpackAfterAlltoall=True", "0", ""]
    in trace ("TaskData.cmdLine: must be modified when the new getopt-parser is active") $
       parseCommandLine commandLine

cloogData :: String -> String -> [[Integer]] -> [Integer] -> [Bool] -> CloogData
cloogData tmpdir projectname widths numbers selecteds = 
    unsafeGetCloogStmtInfos (cmdLine tmpdir projectname widths numbers selecteds)

commCodeDS :: String -> String -> [[Integer]] -> [Integer] -> [Bool] -> CommCodeDS
commCodeDS tmpdir projectname widths numbers selecteds = 
    unsafeCreateCommCodeDS (cmdLine tmpdir projectname widths numbers selecteds)


{--
let cd = cloogData "/tmp/loopo9880" "polyprod-imper.src.out"
let cc = commCodeDS "/tmp/loopo9880" "polyprod-imper.src.out"
cloogDataCompStmts cd   -- liefert Liste aller Berechnungsstmts, 
			-- bestehend aus einer Liste von InEQs + TrStmtNr
			-- (incl. Tiling)
			-- Hoffnung: ich bekomme die origKoordinaten daraus (*)
ccTISPCs cc    -- liefert SrcStmtID (die Rume sind ohne Tiling)
ccAccesses cc  -- liefern letztlich (invers) die Accesses je SrcStmdId
	       -- und die SrcAccessFunktion
-- diese SrcAccessFunktion nach den origKoordinaten angewendet liefert
-- die geschriebenen/gelesenen Daten


(*)  angeblich wird der TISPC wegprojeziert; es bleiben Tile und ISPC =>
  projectToVars ((cloogStmtIS((cloogDataCompStmts cd)!!0))!!0) Nothing  [2,3]
liefert die Quellkoordinaten des                     ^^^^-ten Stckchens
des Zielindexraums: 
0: 0 <= i <= n && j <= n && j >= n-i
1: j >= 0 && 0 <= i <= n-1 && j <= n-i-1

let theIspc0 = projectToVars ((cloogStmtIS((cloogDataCompStmts cd)!!0))!!0) Nothing  [2,3]
let theIspc1 = projectToVars ((cloogStmtIS((cloogDataCompStmts cd)!!1))!!0) Nothing  [2,3]

let theAccFct = accIndexFct((ccAccesses cc)!!1)

image theAccFct (Data.Maybe.fromJust theIspc0)
 liefert was der erste Teil beschreibt: 
 n <= x <= 2*n
image theAccFct (Data.Maybe.fromJust theIspc1)
 liefert was der zweite Teil beschreibt: 
 0 <= x <= n-1

Nach diesem Plausibilittstest gehen wir nun ohne die Projektion 
auf die Quellindizes vor, um parametrisiert in den Tile-Nummern 
die analogen Rume zu bekommen.

Also konvertiere die Tile-Dimensionen zu Konstanten, ehe man auf die
Quellkoorinaten projeziert:

let theParameterizedIspc0 = projectToVars (moveVarsToParams ((cloogStmtIS((cloogDataCompStmts cd)!!0))!!0) 0 2 0) Nothing [0,1]

let theColExtendedAccFct =  insertParams theAccFct 0 (zeroMatrix (nRows theAccFct) 2)
let theExtendedAccFct = affineFunction((nVars theColExtendedAccFct, nParams theColExtendedAccFct),  getMatrix(deleteAffineRows theColExtendedAccFct 1 2)) Inhomogeneous

---falsch:affineFunction((nVars theColExtendedAccFct, nParams theColExtendedAccFct), insertZeroRows(getMatrix theColExtendedAccFct) 1 2) Homogeneous

image theExtendedAccFct (Data.Maybe.fromJust theParameterizedIspc0)

 n, 400pt  <= x <= 400pt+399, 2n, 2n-20tt  und viele Bedingungen an die Params

analog:

let theParameterizedIspc1 = projectToVars (moveVarsToParams ((cloogStmtIS((cloogDataCompStmts cd)!!1))!!0) 0 2 0) Nothing [0,1]
image theExtendedAccFct (Data.Maybe.fromJust theParameterizedIspc1)
 0, 400pt, 20tt <= x <= n-1, 400pt+399   und viele Bedingungen an die Params


24.11.06:

Bemerkungen:
 * wenn ich "image theExtendedAccFct theParameterizedIspc" nehme,
   kann ich nur polytopische Subbereiche der Arrays bergeben,
   wogegen das Framework affine Funktionen (accFct) ber 
   polytopischen Indexrumen -- und damit lattices -- erlaubt.
 * wenn ich die polytopischen Indexrume aufzhle und die 
   accfunction explizit angebe, dann werden dieselben Arrayzellen
   mehrfach aufgezhlt (da sie im Lauf der Berechnung immer wieder
   berschrieben wurden). Das ist ein sehr hoher 
   Mehrfach-versendeaufwand!

 => Ich erlaube entweder polytopische Arraybereiche und verschicken
   damit kompat, oder ich entscheide mich gleich fr die komplett  
   dynamische Variante mit den mitgeschikcten Arrayindizes -- 
   das ist im Beispiel sogar krzer als die lattice-Variante.

Frage: ist das korrekt, die Mehrfach-berschreibungen nur lokal zu
haben ?

 - Wenn jemand einen alten Wert bruchte, gibt's von der Quelle zum
   jemand eine Abhngigkeit (z.B. true), und wenn sie Quellen-lokal 
   wieder berschrieben wird, dann auch eine Antiabhngigkeit vom 
   jemand zum berschreiber.

   Dies ist jedoch nicht mglich, wenn wir FCO haben und 
   der jemand und die Quelle bzw. der berschreiber nicht am selben
   Prozessor sind.

--}

---
-- Reads (unsafe) the <code>project</code> in the <code>tmpdir</code>
-- and returns, depending on the <code>theAccType</code>, all 
-- array cells that are correpondingly accessed in the program, or,
-- more precisely, in a tile of the program. For that purpose,
-- the tile coordinates are converted to be the first 
-- parameter dimensions
-- The Integer result is the number of Tile dimensions assumed. (cf HACK)
-- It is needed for the genLoops
--
-- taksData is the top function that collects all required info from
-- the different sources, adapts the dimensionalities, and delegates 
-- the computation of the accessed data in each ComputeStatement to 
-- taskDataPerTISPC
taskData :: String -> String -> [[Integer]] -> [Integer] -> [Bool] -> AccessType -> (Integer,[(ArrayID, IneqSystem)])
taskData tmpdir project widths numbers selecteds theAccType = (fromIntegral tileDims,
    concatMap (taskDataPerTISPC tileDims (ccTISPCs cc) trimmedAccs theAccType) 
	      (cloogDataCompStmts cd))
	where cc = commCodeDS tmpdir project widths numbers selecteds
	      cd = cloogData  tmpdir project widths numbers selecteds
	      lds = unsafeReadLoopoDS (tmpdir ++ "/" ++ project)
	      accs = loopoAccesses lds    
                   -- (ccAccesses cc) seems to have only such from accpairs!
                   -- (but in this case we need all accesses, because we
                   --  are not working with dependence driven communication)
              trimmedAccs = map (trimAccess (loopoSLP lds)) accs
	      tileDims = length $ tiOptTiledDims (ccTilingOptions cc)
	      -- MG HACK: this assumes that all first dimensions are
	      -- for tiling; it disallows a merge of tiled and non-tiled
	      -- dimensions!!! 24 Nov 2006

-- one TStatement can read several arrays, so <code>taskDataPerTISPC</code>
--   finds all accesses and delegates the computation of the TaskData to
--   <code>taskDataPerACC</code>
-- The "TISPC" (according to the function name an obvious input for this method) 
--   is hidden/embedded in the <code>cloogStatement</code>. 
-- Note that this space is neither in the format of ispc, nor of tispc. 
-- Instead, its representation is promised to consist of
--   the tile dimensions and the *original* (source) dimensions; the
--   target dimensions are promised to be projected away. This guarantee
--   is provided by the representation of cloogStmtIS (I just rely on that).
--
-- Thus, taskDataPerTISPC works in the following steps:
--   a) find all accesses in the <code>cloogStatement</code> of the desired type
--      <code>theAccType</code>; they are called <code>relevantAccesses</code>
--   b) in the index space, in which the access occurs, the tile
--      dimensions become parameters, so that the remaining variables enumerate
--      all (source, see above) index points for a given tile; this domain is 
--      called <code>parameterizedIspc</code>
--   c) the computation of the array cells accessed by 
--      <code>relevantAccesses</code> within <code>parameterizedIspc</code>
--      is delegated to <code>taskDataPerACC</code>
taskDataPerTISPC :: Int -> [TStatement] -> [Access] -> AccessType -> CloogStatement -> [(ArrayID, IneqSystem)]
taskDataPerTISPC tileDims tStmts accesses theAccType cloogStatement = 
--    trace ("relevantAccesses = " ++ show relevantAccesses ++ ", TSmtId=" ++ show tsID ++ ", cTispcParts = " ++ show  cTispcParts) $
    map (taskDataPerACC parameterizedIspc) relevantAccesses
	where relevantAccesses = 
		  filter (\a -> accType a == theAccType && 
			  accStmtID a == findSrcStmtID tsID)        accesses
	      tsID :: TStatementID
	      tsID = siTrStmtNr (cloogStmtInfo cloogStatement)
	      srcStmtID :: TStatement -> StatementID
	      srcStmtID (_,s,_) = s
	      findSrcStmtID :: TStatementID -> StatementID
	      findSrcStmtID tsid = 
		  srcStmtID $ Maybe.fromJust ( find (\(t,s,_) -> t==tsid) tStmts )
	      -- next, the target index spaces must be computed and the
	      -- tile dimensions in the tispc must be 
	      -- transformed into parameters
	      cTispcParts = cloogStmtIS cloogStatement
	      cTispc = if not(length cTispcParts == 1)
		       then error ("assert multiple (" ++ 
                                   show (length cTispcParts) ++ ") ISPCS")
		       else head cTispcParts
	      extendedCTispc = moveVarsToParams cTispc  0 tileDims 0 
	      -- MG HACK: the first 0 requires that the tile loops 
	      -- are the outermost indices 24. Nov 2006

	      parameterizedIspc = -- trace ("param. ISPC before projection: \n" ++ show extendedCTispc) $
                                  fromJust $
				  projectToVars extendedCTispc  Nothing [0..(nV-1)]
	      nV = nVars extendedCTispc

---
-- taskDataPerACC computes the array cells accessed by 
--   <code>acc</code> within <code>parameterizedIspc</code>.
-- 
-- Assumption: the number of variables is the number of loops surrounding
-- the access in the original original (source) program; 
-- the meanings of the variables are the source indices.
-- This is guaranteed by the caller (see above).
--
-- As <code>parameterizedIspc</code> has the tile dimensions as
-- additional parameters, these must be added to the access
-- function before computing the image. 
taskDataPerACC :: IneqSystem -> Access -> (ArrayID, IneqSystem)
taskDataPerACC parameterizedIspc acc =  
    if (nVars theAccFct == nVars parameterizedIspc) then
       (accArrayID acc, accessedElements)
    else
       error("dimensionalities of source access " ++ show (nVars theAccFct)
             ++ " and index set in taskDataPerACC "  
             ++ show (nVars parameterizedIspc) ++ " do not match")
    where accessedElements = -- trace ("AccessedElemets \n" ++
                             --       show extendedAccFct ++
                             --       "Non-extd. AccessedElemets \n" ++
                             --       show theAccFct ++
                             --       "Parameterisierte ISPC \n" ++
                             --       show parameterizedIspc ++ "\n" ) $
                             image extendedAccFct parameterizedIspc
	  theAccFct = accIndexFct acc

          -- Since, apart from the tile dimensions, the format and
          -- meaning of the colums of acc and parameterizedIspc is
          -- the same, the number of dimensions for the tiles must be the 
          -- difference in the parameter dimensionalities (added 
          -- to the parameterizedIspc by the caller):
          tileDims = nParams parameterizedIspc - nParams theAccFct

          -- the access function must be extended to the target
	  -- dimensionality; this is not problematic as the
	  -- access function does only refer to source index columns
          -- Thus: a zero matrix of appropriate size is what we need.
	  colExtendedAccFct =  insertParams theAccFct 0 
			       (zeroMatrix (nRows theAccFct) tileDims)
	  extendedAccFct = affineFunction((nVars colExtendedAccFct, 
					   nParams colExtendedAccFct), 
					  m) Inhomogeneous
	  m = getMatrix(deleteAffineRows colExtendedAccFct 
			(nRows theAccFct - nrToDrop) nrToDrop )
	  nrToDrop = nParams theAccFct + 1

taskDataLoop :: String -> String -> [[Integer]] -> [Integer] -> [Bool] -> AccessType -> [String]
taskDataLoop tmpdir project widths numbers selecteds theAccType = 
    map (\(a,i)-> loopNestPretty (names tileDim i) i ++ show a) (snd td)
	where td = taskData tmpdir project widths numbers selecteds theAccType
	      tileDim = fst td

names :: Integer -> IneqSystem -> Map Integer String
names tileDims is = Map.fromList (vars ++ tiles ++ params)
    where
    vars = [(i, 'i' : show i) |  i <- [0 .. nrv-1]]
    tiles = [(i+nrv, "tiledim" ++ show i) |  i <- [0 .. tileDims-1]]
    params = [(i+nrv+tileDims, 'n' : show i) |  i <- [0 .. nrp -1]]
    nrv = nVars is
    nrp = nParams is
    

demoWidths=[[20,0],[0,400]]
demoNumbers=[0,1]
demoSelecteds=[True,True]

-- examples:
-- let tdl3 = taskDataLoop "/tmp/loopo3455" "matmult-imper.src.out" [[20,0,0],[0,200,0],[0,0,400]] [0,1,2] [True,True,True] Write
-- still fails.

-- let tdl = taskDataLoop  "/tmp/loopo8675" "polyprod-imper.src.out" demoWidths demoNumbers demoSelecteds Read
-- is ok.

-- main is only for testing!!!
main = do
       let tdl = taskDataLoop "/tmp/matmult-imper-soak"  "matmult-imper.src.out" [[400,0],[0,300]] [0,1] [True,True] Read
       putStrLn (concat tdl)


