-- program to generate an allocation which has the 
-- Forward Communication Only property, if possible

import List
import Ratio
import Vector
import Loopo
import IOExts

----------------------------------------------------------------------
-- how to get input data from the ispc file
----------------------------------------------------------------------

load :: (FilePath -> IO (LPResult a)) -> FilePath -> IO a
load f file =
    do res <- f file
       case res of
           ParseOk x    -> return x
	   ParseFailed f -> error ("Fehler: " ++ show f)


readslpfile :: String -> IO ([(Integer,[Integer])])
readslpfile datei = 
     do resslps <- load readSLP (datei ++ ".slp")
	return resslps

readaccfile :: String -> IO ([(Integer, AccessType, Integer, Integer, AffineFunction)])
readaccfile datei = 
     do resaccs <- load readACC (datei ++ ".acc")
	return resaccs

domainOfStmt :: Integer -> [(Integer,IneqSystem)] -> IneqSystem
domainOfStmt stmntid  ispcs =  
    case ( find (\(s,d)-> stmntid==s) ispcs ) of
         Nothing    -> error("statement not found")
         Just (s,d) -> d




----------------------------------------------------------------------
-- hardcoded input
----------------------------------------------------------------------

--datei = "/home/ststud/sommer/jetztArbeit/sor/sor-1dim.src.out"
-- datei = "simplegauss"
-- datei = "exs/while.src.out"
-- datei = "exs/accbased/example10_2.src.out"
--datei = "exs/feaut/example10_2.src.out"
--datei = "exs/gauss-2.src.out"
--datei = "exs/gaussimple.src.out"
--thedatei = "exs/gauss-half.src.out"
--datei = "exs/feautgasusssimple/gaussimple.src.out"


----------------------------------------------------------------------
-- the farkas method in our case
----------------------------------------------------------------------


-- now the pieces of the farkas equation system:

row4const :: Integer -> Matrix
row4const c = rowMatrix ([0 | i <-[1..(c-1)]] ++ [1])

farkas :: Matrix -> Integer -> Integer -> IneqSystem -> Integer -> Matrix
farkas lhs fullWidth nrZeroBlocksBefore ispc nrZeroBlocksAfter = 
-- the ispc must be given in all indices!
    let cols = nVars ispc + nParams ispc + 1
        rows = fullWidth        --nRows lhs 
        res =  multiconc1 [
            lhs,
            (zeroMatrix rows nrZeroBlocksBefore),
            (smul (mtranspose 
                      -- we append the row for the constant before, because we
                      -- are used to call the respective lambda "lambda_0"
                      ( join (row4const cols) (getMatrix ispc) )
                  ) (-1)),
            (zeroMatrix rows nrZeroBlocksAfter)
          ]
    in -- trace ("Farkas liefert: " ++ show res) 
       res


formGroups :: [Integer] -> [a] -> [[a]]
formGroups [] _     = []
formGroups (w:ws) l = grp : formGroups ws rest
		      where (grp,rest) = splitAt (fromIntegral w) l

countZeroGroups :: [Integer] -> [Rational] -> [Integer]
countZeroGroups widths row = map fst $
			     filter (\(_,vs) -> all (==0) vs) $
			     zip [1..] $
			     formGroups widths row 


filterNonConstantSolution :: [Integer] -> ([Rational],[Rational]) -> Bool
-- one list of Rationals is one vector (farkas or nonFarkas)
filterNonConstantSolution poss (sols,_) = not $ or
            $ map ( all (0==) )
            $ map init
            $ formGroups poss sols


{--
neue einfachere Versino folgt
widthsToGroups :: AllocMethod -> [(Dependence,Integer)] -> [Integer]
widthsToGroups _ [] = []
widthsToGroups method ((d,w):xs) = if method==General &&
                               depTypeToAccType Dest (depType d) == Read then
                                  w:(w:(widthsToGroups method xs))
			     else w:(widthsToGroups method xs)
--}

widthsToGroups :: AllocMethod -> [(Dependence,Integer)] -> [Integer]
widthsToGroups General [] = []
widthsToGroups General ((d,w):xs) = 
                          if depTypeToAccType Dest (depType d) == Read then
                                  w:(w:(widthsToGroups General xs))
			  else w:(widthsToGroups General xs)
widthsToGroups _ l = map snd l


superDepAndOffsetList :: AllocMethod -> [Statement] -> [Dependence] -> (Integer,[(Dependence,(Integer,Integer))])
superDepAndOffsetList method stmts deps = 
    let simpleAndExtWidths = 
              map   (\d-> let width = 1+nRows (destination d)
                          in
                              if method == General && 
                                 depTypeToAccType Dest (depType d) == Read then
                                   (width, 2*width)
			      else (width, width))
		    deps
	(widths, extWidths) = unzip simpleAndExtWidths
        offsets = superPrefixSum  extWidths
    in 
        (last offsets, zip deps (zip (init offsets) widths))

superAccAndOffsetList :: Integer -> [Statement] -> [Access] -> (Integer,[(Access,(Integer,Integer))])
superAccAndOffsetList initOffs stmts accs = 
    let widths = map   (\a-> (1+nRows
				   (snd (errorFindStmt stmts (stmtId a)))))
				  accs
        offsets = map (initOffs+) (superPrefixSum widths)
    in 
        (last offsets, zip accs (zip (init offsets) widths))





data AllocMethod = DepDriven | OwnerDriven | General | Small  deriving (Show,Eq,Read)

---

main :: IO()
main = mainx "exs/feautgasusssimple/gaussimple.src.out" "DepDriven" -- "General"


mainx :: String -> String -> IO()
mainx datei methodstr =
    let 
--      datei = "exs/gauss-half.src.out"
--      datei = "exs/feautgasusssimple/gaussimple.src.out"
--        datei = "exs/gauss-2.src.out"
--        datei = "exs/move-read-acc.src.out"
--        method = General
        method :: AllocMethod
        method = read methodstr
    in
    do 
       stmts' <- load readISPC (datei ++ ".ispc")  -- :: [Statement]
       slps <- readslpfile datei
       deps <- load readDependences (datei ++ ".dep") -- :: [Dependence]
       acps <- load readACP  (datei ++ ".acp") -- :: [(Integer, Integer, Integer)]
       rawaccesses <- readaccfile datei
       let stmts = reducedIspcsStmts ((map (\ (w,(x,y))->((w,y),x))) (joinZip' slps stmts'))
--	   accs = acpsToAccs acps stmts deps rawaccesses
           accs = raccsToAccs rawaccesses stmts
	   stmtPosList = stmtPoss stmts
	   accPosList = accPoss accs stmts
           arrayPosList = arrayPoss accs (poss4Stmts stmts)
           arrayWidthList = snd (unzip (arraysWithDims accs))
	   stmtWidthList = stmtWidths stmts
	   accWidthList = accWidths accs 
	   tmpafffct = (hTrafo.head) deps
	   fullWidth = nVars tmpafffct + nParams tmpafffct +1
	   glhswidth = glhsWidth method stmts accs
	   (depRhsWidth, theDepAndOffsetList) = superDepAndOffsetList method stmts deps
	   (rhsWidth, theAccAndOffsetList) 
              | method == General =
                  superAccAndOffsetList depRhsWidth stmts (filter (\a->accessType a==Write) accs)
	      | method == OwnerDriven = superAccAndOffsetList 0 stmts accs
	      | otherwise =  (depRhsWidth,[])
           eqs1 = multijoin1 
		 [ farkas
		     anydep
                     fullWidth
		     offs 
--		     (snd stmt')
		     (destination dep)
		     (rhsWidth-offs-(nRows (destination dep))-1+1)       
                                       -- +1 fuer die unbenoetigte 1
	 	   | (dep,(offs,_)) <- theDepAndOffsetList, 
		     let stmt' = errorFindStmt stmts' (destinationNumber dep)
                         anydep = if method == DepDriven then
		                        depDrivenToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth 
                                 else   
                                        anyDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
		   ]
           eqs2 = multijoin1 
		 [ farkas 
		     otherdep
                     fullWidth
		     (offs+width)
		     (destination dep)
--		     (snd stmt')
		     (rhsWidth-offs-width-(nRows (destination dep))-1+1)   
                                       -- +1 fuer die unbenoetigte 1
	 	   | (dep,(offs,width)) <- theDepAndOffsetList, 
		     let stmt' = errorFindStmt stmts' (destinationNumber dep)
			 acctype = depTypeToAccType Dest (depType dep) 
			 otherdep = readDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
-- MG stimmt der Umgang mit non-true-deps?
                     ,
                     (acctype == Read)
		   ]
           eqs3 = if (length theAccAndOffsetList)>0 then
                      multijoin1 [ farkas 
  		        constraintMatrix
                        fullWidth
		        offs
		        (snd stmt')
		        (rhsWidth-offs-(nRows (snd stmt'))-1+1) 
                                       -- +1 fuer die unbenoetigte 1
	 	        | (acc,(offs,width)) <- theAccAndOffsetList, -- accessType acc == Write, 
		          let stmt' = errorFindStmt stmts' (stmtId acc)
			      constraintMatrix = writeAccOrOwnerDrivenToMatrix method acc fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
		        ]
                  else zeroMatrix 0 (nColumns eqs1)

	   eqs = case method of
                      General     -> join (join eqs1 eqs2) eqs3
--                      General     -> eqs1
                      Small       -> eqs1
		      DepDriven   -> eqs1 -- no eqs2 since no realloc
                                          -- no eqs3 since implicitly given (OCR)
                      OwnerDriven -> eqs3 -- error ("OwnerDriven not yet integrated.")
           nrNonFarkasVars = nColumns eqs-1-rhsWidth     
           eqsAsIeqs = ineqSystem((nColumns eqs-1,0), join eqs (-eqs))
	   nonNegFarkas = ineqSystem((nColumns eqs-1,0), multiconc1 [
               zeroMatrix rhsWidth nrNonFarkasVars,
	       unitMatrix rhsWidth rhsWidth,
	       zeroMatrix rhsWidth 1 ])
           allIneqs = join eqsAsIeqs nonNegFarkas
--	   lrv = dualOfIS allIneqs
           lrv = dualOfSystem (eqSystem((nColumns eqs-1,0), eqs)) nonNegFarkas
	   nonFf = map mk 
--                           (filter (\v->not (all (0==) 
--                                       (take (fromInteger(snd(head accPosList)-1)) 
--				       -- MG hack: incorrect for different A-dims,
--				       --    e.g., move-read-acc.src
--                                             (vectorEntries v))))
                                  (lrvRays lrv)
--                                              )
	   mk ray = splitAt (fromIntegral nrNonFarkasVars) $
		    vectorEntries ray

           rhsWidthList = widthsToGroups method
                         (zip (fst (unzip theDepAndOffsetList))
			      (snd (unzip (snd (unzip theDepAndOffsetList))))) 
                         ++  snd (unzip (snd (unzip theAccAndOffsetList)))
           lhsWidthList 
              | method == DepDriven   = stmtWidthList
              | method == General     = (stmtWidthList ++ accWidthList)
              | method == OwnerDriven = (stmtWidthList ++ arrayWidthList)
	      | otherwise = (stmtWidthList ++ accWidthList)

           sortedGroups = sortBy (\(r1,_,c1)(r2,_,c2) ->
			          compare (length c2) (length c1))
	     [ (formGroups lhsWidthList nonFarkas, farkas, countZeroGroups rhsWidthList farkas)  
	                        | (nonFarkas,farkas) <- nonFf
                                ,  if method == OwnerDriven then True 
                                   else  filterNonConstantSolution lhsWidthList
                                                             (nonFarkas,farkas)
                                ]
	   
       putStrLn ((show  deps) ++ "\n\n\n")
--       putStrLn (show  (combineDepAcp deps acps))
--       putStrLn (show  accs)
--       putStrLn (show(reducedIspcs (map (\(x,y)->(y,x)) (joinZip slps stmts'))))
--       putStrLn("The reduced statments are: " ++ show stmts)
--       putStrLn (show(oldStmtPoss (map (\(x,y)->(y,x)) (joinZip slps stmts))))
       putStrLn ("Accesses:" ++ show accs)
       putStrLn ("\nAccess-Pairs: "  ++ show acps)
       putStrLn ("glhsWidth = " ++ show glhswidth)
       putStrLn ("StmtWidths = " ++ (show stmtWidthList))
       putStrLn ("AccWidths = " ++ (show accWidthList))
       putStrLn ("ArrayWidths = " ++ (show arrayWidthList))
       putStrLn ("Positions: " ++ (show (map snd (stmtPoss stmts))) ++ " und " ++ (show (map snd (accPoss accs stmts))))
       putStrLn ("Positions: " ++ (show stmtPosList) ++ " und " ++ (show accPosList))
       putStrLn ("ArrayPositions: " ++ (show arrayPosList))
       putStrLn ("Position of the last stmt: " ++ (show (stmtPos stmtPosList (last stmts))))
       putStrLn ("fullWidth: " ++ (show fullWidth))
       putStrLn ("depAndOffestList: " ++ (show (fst(unzip(snd(unzip theDepAndOffsetList))))))
       putStrLn ("accAndOffestList: " ++ (show (fst(unzip(snd(unzip theAccAndOffsetList))))))
       putStrLn ("rhsWidth/depRhsWidth: " ++ (show rhsWidth) ++ "/" ++ (show depRhsWidth))
       putStrLn ("eqs3: " ++ (intShowMat eqs3 True))
       putStrLn ("eqs2: " ++ (intShowMat eqs2 True))
       putStrLn ("eqs1: " ++ (intShowMat eqs1 True))
       putStrLn ("Eqs: \n" ++ intShowMat eqs True)
----       putStrLn "\n\n Und nun alle Ineqs: \n"
----       putStrLn (show allIneqs)
---
--       putStrLn ( "rhsWidth = " ++ show rhsWidth ++ "\nFarkas-equalities\n" ++ show eqs )
       putStrLn ( "\nDual\nLines:\n" ++ show (lrvLines lrv) ++ "\nVertices:\n" ++ show (lrvVertices lrv) )
       putStrLn ( "\nrhsWidthList\n" ++ show rhsWidthList )
       putStrLn ( "\nlhsWidthList\n" ++ show lhsWidthList )
       putStrLn ( "\nSorted Groups\n" ++ unlines (map (\ (n,f,c) -> 
                                                 concat (concat (map (lines.(flip (++) "    ").show.vector) n))  ++ "\n" ++
                                                 show (vector f)  ++ 
                                                 show c
                                              ) (take 20 sortedGroups)) )
--       putStrLn ( "\nSorted Groups\n" ++ concat (map (\ (n,f,c) -> 
--                                                 show (map vector n)  ++ 
--                                                 show (vector f)  ++ 
--                                                 show c
--                                              ) sortedGroups) )
--       putStrLn ( "\nSorted Groups\n" ++ unlines (map (show.(\ (n,f,c)->(unlines(map (show.vector) n), vector f,c))) sortedGroups) )
       putStrLn ("")

filterdims :: (IneqSystem, [Integer]) -> IneqSystem
filterdims (ispc, surr) = ineqSystem((vars,pars), redmatrix)
         where 	vars = sum surr
		pars = nParams ispc
		redmatrix = multiconc1 
			[subMatrixVer (getMatrix ispc) i i |
				i <- [0..(nColumns ispc)-1], 
					(i>=genericLength surr || 
                                         surr!!(fromInteger i)>0)]

filterrowdims :: (IneqSystem, [Integer]) -> IneqSystem
filterrowdims (ispc, surr) = ineqSystem((vars,pars), redmatrix)
         where 	vars = nVars ispc
		pars = nParams ispc
		redmatrix = multijoin1 
			[subMatrixHor (getMatrix ispc) i i |
				i <- [0..(nRows ispc)-1], 
					(i>=genericLength surr || 
                                         surr!!(fromInteger i)>0)]

reducedIspcs :: [(IneqSystem, [Integer])] -> [IneqSystem]
reducedIspcs ispcSlpZip = map  filterdims  ispcSlpZip

reducedIspcsStmts :: [(Statement, [Integer])] -> [Statement]
reducedIspcsStmts ispcSlpZip = map (\ ((s,i),u) -> (s, filterdims(i,u)) )  ispcSlpZip

stmtWidths :: [Statement] -> [Integer]
stmtWidths ss = map (nColumns.snd) ss

accWidths :: [Access] -> [Integer]
accWidths acs = map (nColumns.domain) acs

glhsWidth :: AllocMethod -> [Statement] -> [Access] -> Integer
glhsWidth method stmts accs 
       | method == DepDriven  =  sum (stmtWidths stmts)
       | method == OwnerDriven = sum (stmtWidths stmts) + sum dims
       | otherwise =  sum (stmtWidths stmts) + sum (accWidths accs)
  where (arrays,dims) = unzip (arraysWithDims accs)


-- takes all accesses, computes the dimensions of every array access, and
-- returns a list of the pairs of array id and its dimensionality.
-- Every array is olisted only once.
arraysWithDims :: [Access] -> [(Integer,Integer)]  
arraysWithDims accs = nub $ map  -- von jedem Array einmal die Zeilenzahl
       (\a->(arrayID a, nRows(getMatrixIH(indexFct a))))  -- nein, ohne Parameter und 1
--     (\a->(arrayID a, nRows(indexFct a)))  -- ja, incl. Parameter und 1
       accs

-- computes a list, as long as the number of arrays in the program; every
-- list element holds the array id and the position (starting at offs) where
-- the coefficients for this array are located:
arrayPoss :: [Access] -> Integer -> [(Integer,Integer)]
arrayPoss accs offs = zip arrays (map (offs +) (superPrefixSum dims))
    where (arrays,dims) = unzip (arraysWithDims accs)  


stmtPoss ::  [Statement] -> [(Integer,Integer)]
stmtPoss s  = zip (map fst s) (superPrefixSum (stmtWidths s))


accPoss :: [Access] -> [Statement] -> [(Integer,Integer)]
accPoss a s = zip (map accessId a) ((map ( shift + )) (superPrefixSum (accWidths a)))
	      where shift = poss4Stmts s

-- how many positions are used by the statements (== column number of the 
-- next entry, since we start at 0)
poss4Stmts :: [Statement] -> Integer
poss4Stmts ss = snd(last (stmtPoss ss)) + (nColumns.snd) (last ss)


errorFind ::  (a -> Bool) -> [a] -> a
errorFind f l = case find f l of
		Nothing -> error "Error: Element not found.\n"
		Just x  -> x


accPos :: [(Integer,Integer)] -> Access -> Integer
accPos l a = snd(errorFind (\ (i,p)-> (accessId a)==i ) l)

stmtPos :: [(Integer,Integer)] -> Statement -> Integer
stmtPos l s = snd(errorFind (\ (i,p)-> (fst s)==i ) l)

accIdPos :: [(Integer,Integer)] -> Integer -> Integer
accIdPos l accid = snd(errorFind (\ (i,p)-> accid==i ) l)

stmtIdPos :: [(Integer,Integer)] -> Integer -> Integer
stmtIdPos l stmtid = snd(errorFind (\ (i,p)-> stmtid==i ) l)

getPosIn ::  [(Integer,Integer)] -> (a->Integer) -> a -> Integer
getPosIn xs a2i a =  snd(errorFind (\ (i,_)-> (a2i a)==i ) xs)


stmtMatrix :: Statement -> Integer -> [(Integer,[Integer])] -> Matrix
-- statement volldimensional bergeben, damit man was zum rausschneiden hat.
stmtMatrix (id, ispc) fullWidth slps = 
    getMatrix (filterrowdims ((ineqSystem((nVars ispc,nParams ispc), 
					     unitMatrix fullWidth fullWidth)),
			   (errorFindSlps slps id)))


accMatrix :: Access -> Integer -> [Statement] -> [(Integer,[Integer])] -> Matrix
-- stmts voll bergeben weils stmtMatrix so braucht
accMatrix acc fullWidth stmts' slps = getMatrixIH (indexFct acc) 


accIdMatrix :: Access -> Integer -> [Statement] -> [(Integer,[Integer])] -> Matrix
-- stmts voll bergeben weils stmtMatrix so braucht
accIdMatrix acc fullWidth stmts' slps = 
    stmtMatrix (errorFindStmt stmts' (stmtId acc)) fullWidth slps

accHMatrix :: Access -> AffineFunction -> [Statement] -> [(Integer,[Integer])] -> Matrix
-- stmts' oder stmts ist egal
accHMatrix acc htrafo stmts' slps = 
    getMatrix (filterrowdims (htrafoAsIes, slpVect))
    where
          nv = nVars htrafo
	  np = nParams htrafo
          fullWidth = nv + np +1
	  stmt' = errorFindStmt stmts' (stmtId acc)
	  slpVect = errorFindSlps slps (fst stmt')
	  htrafoAsIes = ineqSystem((nv,np), getMatrix htrafo)



errorFindSlps :: [(Integer, [Integer])] -> Integer -> [Integer]
errorFindSlps slps id = snd (errorFind (\ (key,_)-> key==id ) slps)

errorFindStmt :: [Statement] -> Integer -> Statement
errorFindStmt stmts id = errorFind (\ (key,_)-> key==id ) stmts


depDrivenToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
depDrivenToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let stmt = errorFindStmt stmts (destinationNumber dep)
	stmt' = errorFindStmt stmts' (destinationNumber dep)
	stmtMat = stmtMatrix stmt' fullWidth slps
	stmtpos = stmtPos stmtposs stmt
	srcacc = srcAcc dep acps accs
        srcstmtid = stmtId srcacc
	srcstmt = errorFindStmt stmts srcstmtid
	srcMat  = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	srcpos = stmtPos stmtposs srcstmt
	prezeros = min srcpos stmtpos
	(firstMat, midzeros, sndMat, postzeros) =
               if srcpos > stmtpos then
                  (stmtMat, 
                   srcpos - stmtpos - ((nColumns.snd) stmt),
                   srcMat,
                   glhswidth - srcpos - ((nColumns.snd) srcstmt))
               else if srcpos < stmtpos then
                  (srcMat,
                   stmtpos - srcpos - ((nColumns.snd) srcstmt),
                   stmtMat,
                   glhswidth - stmtpos - ((nColumns.snd) stmt))
               else -- source and destination are the same stmt
                  (srcMat+stmtMat,
                   0,
                   zeroMatrix 0 fullWidth,
                   glhswidth - stmtpos - ((nColumns.snd) stmt))
    in  
        multiconc1 [
		   zeroMatrix fullWidth prezeros,
		   mtranspose firstMat,
		   zeroMatrix fullWidth midzeros,
		   mtranspose sndMat,
		   zeroMatrix fullWidth postzeros
		   ] 


anyDepToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
anyDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let stmt = errorFindStmt stmts (destinationNumber dep)
	stmt' = errorFindStmt stmts' (destinationNumber dep)
	stmtMat = stmtMatrix stmt' fullWidth slps
	stmtpos = stmtPos stmtposs stmt
	srcacc = srcAcc dep acps accs
	accMat  = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	accpos = accPos accposs srcacc
	midzeros = accpos - stmtpos - ((nColumns.snd) stmt)
	postzeros = glhswidth - accpos - ((nColumns.domain) srcacc)
    in  -- trace ("stmtpos: " ++ show stmtpos ++ ", midzeros: " ++ show midzeros ++ ", postzeros: " ++ show postzeros ++ "\nSrc-Access: " ++ show srcacc ++ "\naccMatrix: " ++ show accMat ++ "\nstmtMat: " ++ show stmtMat)
        (multiconc1 [
		   zeroMatrix fullWidth stmtpos,
		   mtranspose stmtMat,
		   zeroMatrix fullWidth midzeros,
		   mtranspose accMat,
		   zeroMatrix fullWidth postzeros
		   ] )

-- now: if access (=current access) writes:

writeAccOrOwnerDrivenToMatrix :: AllocMethod -> Access -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix

writeAccOrOwnerDrivenToMatrix method acc fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth = 
    let stmt = errorFindStmt stmts (stmtId acc)
	stmt' = errorFindStmt stmts' (stmtId acc)
	stmtMat = -(stmtMatrix stmt' fullWidth slps)
	stmtpos = stmtPos stmtposs stmt
	accMat  
          | method == OwnerDriven = 
            accMatrix  acc fullWidth stmts' slps  
          | otherwise = 
            accIdMatrix  acc fullWidth stmts' slps    -- unitMatrix fullWidth fullWidth
	accpos = accPos accposs acc
        arraypos = getPosIn (arrayPoss accs (poss4Stmts stmts)) id (arrayID acc)
	midzeros 
           | method == OwnerDriven = arraypos - stmtpos - ((nColumns.snd) stmt)
           | otherwise = accpos - stmtpos - ((nColumns.snd) stmt)
	postzeros  
           | method == OwnerDriven = glhswidth - arraypos - (nRows accMat)
           | otherwise = glhswidth - accpos - ((nColumns.domain) acc)
        sign 
           | method == OwnerDriven = 
                   if accessType acc == Write then 1 else -1
	   | otherwise = 1
        gesamt =  trace ("Access " ++ show acc ++ " liefert Positionen stmtpos=" ++ show stmtpos ++ ", arrayPos=" ++ show arraypos ++ " und stmtMat " ++ show stmtMat ++ " und accMat " ++ show accMat ++ "; midzeros = " ++ show midzeros ++ ".\n") 
                  multiconc1 [
		   zeroMatrix fullWidth stmtpos,
		   mtranspose (smul stmtMat sign),
		   zeroMatrix fullWidth midzeros,
		   mtranspose (smul accMat sign),
		   zeroMatrix fullWidth postzeros
		   ]
    in trace ("WriteAccessMatrixLhs = " ++ show gesamt ++ ".\n\n") 
       gesamt



-- now: if dependence destination (=current access) reads:
readDepToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
readDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let srcacc = srcAcc dep acps accs
	srcmat = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	srcpos = accPos accposs srcacc
	dstacc = dstAcc dep acps accs
	dstmat  = accIdMatrix  dstacc fullWidth stmts' slps    -- unitMatrix fullWidth fullWidth
	dstpos = accPos accposs dstacc
        matrix = if dstpos == srcpos then 
                    multiconc1 [
	              	         zeroMatrix fullWidth srcpos,
		                 mtranspose (srcmat+dstmat),
		                 zeroMatrix fullWidth (glhswidth - srcpos - ((nColumns.domain) srcacc))
		                 ]
--  		    zeroMatrix fullWidth glhswidth
--  		    zeroMatrix 0 glhswidth
                 else 
                    let (midzeros, postzeros, matrix') =
                          if dstpos > srcpos then
	                      (
	                      dstpos - srcpos - ((nColumns.domain) srcacc),
	                      glhswidth - dstpos - ((nColumns.domain) dstacc),
	                      multiconc1 [
		                  zeroMatrix fullWidth srcpos,
		                  mtranspose srcmat,
		                  zeroMatrix fullWidth midzeros,
		                  mtranspose dstmat,
		                  zeroMatrix fullWidth postzeros
		                  ]
	                      )
	                  else (  --  dstpos < srcpos 
	                      srcpos - dstpos - ((nColumns.domain) dstacc),
	              	      glhswidth - srcpos - ((nColumns.domain) srcacc),
	              	      multiconc1 [
	              	         zeroMatrix fullWidth dstpos,
		                 mtranspose dstmat,
		                 zeroMatrix fullWidth midzeros,
		                 mtranspose srcmat,
		                 zeroMatrix fullWidth postzeros
		                 ]
	                      )
                    in  matrix'
    in --trace ("Lhs-Matrix for src " ++ show (accessId srcacc) ++ 
       --       " / dst "  ++ show (accessId dstacc) ++ 
       --       ": "  ++ show matrix
       --      ) 
--             (if depNumber dep /= 5 then matrix
--  		else     zeroMatrix fullWidth glhswidth ) -- ja
--	        else -matrix)                             -- nein
                  matrix                                  -- nein


sndOfThree :: (a,b,c) -> b
sndOfThree (_,y,_) = y

thrdOfThree :: (a,b,c) -> c
thrdOfThree (_,_,z) = z

srcAcc :: Dependence -> [(Integer, Integer, Integer)] -> [Access] -> Access 
srcAcc dep acps accs = errorFindAcc  accs
		            (sndOfThree (errorFindAcp  acps (depNumber dep)))
			   
dstAcc :: Dependence -> [(Integer, Integer, Integer)] -> [Access] -> Access 
dstAcc dep acps accs = errorFindAcc  accs
		            (thrdOfThree (errorFindAcp  acps (depNumber dep)))
			   

errorFindAcp :: [(Integer, Integer, Integer)] -> Integer -> (Integer, Integer, Integer)
errorFindAcp acps depId = errorFind (\ (d,_,_)->d==depId) acps

errorFindAcc :: [Access] -> Integer -> Access 
errorFindAcc accs key = errorFind (\ a->(accessId a)==key) accs


-- returns the prefix sum, but with 0 as the first and the sum over all as last 
-- list element. Example: superPrefixSum [2,5,10] = [0,2,7,17]
superPrefixSum :: [Integer] -> [Integer]
superPrefixSum xs = snd res ++ [fst res]
       where res = mapAccumL prefixSumAccuFct 0 xs
prefixSumAccuFct :: Integer -> Integer -> (Integer,Integer)
prefixSumAccuFct accu val = ( accu +  val,  accu )


joinZip' :: (Show a, Eq a) => [(a,b)] -> [(a,c)] -> [(a,(b,c))]
joinZip' [] _ = []
joinZip' ((key,dat):xs) l = 
      case ( find (\(s,d)-> key==s) l ) of
         Nothing    -> trace ("\nWarning: key " ++ show key ++
		              " in joinZip not found. Element dropped.")
		             (joinZip' xs l) 
         Just (s,d) -> (key,(dat,d)) : (joinZip' xs l) 


joinZip :: (Show a, Eq a) => [(a,b)] -> [(a,c)] -> [(b,c)]
joinZip xs ys = map snd (joinZip' xs ys)

combineDepAcp :: [Dependence] -> [(Integer,Integer,Integer)] -> [(Dependence,(Integer,Integer))]
combineDepAcp deps acps = joinZip  (zip (map depNumber deps) deps) 
			           (map (\ (a,b,c)->(a,(b,c))) acps)


findDep :: Integer -> [Dependence] -> Dependence
findDep id deps = 
    case ( find (\d -> depNumber d==id ) deps ) of
         Nothing    -> error("dependence " ++ show id ++ " not found")
         Just d     -> d

findIspc :: Integer -> [Statement] -> IneqSystem
findIspc key  stmts = 
    case ( find (\(s,_) -> s==key ) stmts ) of
         Nothing    -> error("statement " ++ show key ++ " findIspc not found")
         Just (s,i) -> i

findIndexFctAndArrID :: Integer -> [RawAccess] -> (AffineFunction,Integer)
findIndexFctAndArrID key raccs =
    case ( find (\(_,_,n,_,_) -> n==key ) raccs ) of
         Nothing    -> trace ("\nWarning: access " ++ show key ++ " not found - do not use cffada") (affineFunction((0,0),zeroMatrix 0 1) Inhomogeneous,0)
-- MG hack. Reason: 
-- with cffada the acc numbers (LaNode) do not match with the acp (Occ) numbers
         Just (a,_,_,_,f) -> (f,a)


data DepPart = Source | Dest deriving (Show,Eq)


depTypeToAccType :: DepPart -> DependenceType -> AccessType
depTypeToAccType dp dt = depTypeToAccType' dp 
			 (toEnum (if asInt > 6 then asInt-6 else asInt))
    where asInt = fromEnum dt


depTypeToAccType' Source UniformTrue = Write
depTypeToAccType' Source UniformFlow = Write
depTypeToAccType' Source UniformOutput = Write

depTypeToAccType' Source UniformAnti = Read
depTypeToAccType' Source UniformInput = Read

depTypeToAccType' Dest UniformTrue = Read
depTypeToAccType' Dest UniformFlow = Read
depTypeToAccType' Dest UniformInput = Read

depTypeToAccType' Dest UniformOutput = Write
depTypeToAccType' Dest UniformAnti = Write



acpsToAccs :: [(Integer,Integer,Integer)] -> [Statement] -> [Dependence] -> [RawAccess] -> [Access]
acpsToAccs acps stmts deps raccs = acpsToAccs' acps stmts deps raccs [] 


acpsToAccs' :: [(Integer,Integer,Integer)] -> [Statement] -> [Dependence] -> [RawAccess] -> [Integer] -> [Access]
acpsToAccs' [] _ _ _ _ = []
acpsToAccs' ((dep, src, dest):xs) stmts deps raccs knowns =
    (if (elem src knowns) then [] else [accOf(src, dep, Source)]) ++
    (if (elem dest knowns) then [] else [accOf(dest, dep, Dest)]) ++
    acpsToAccs' xs stmts deps raccs (src : (dest : knowns )) 
    where
	accOf(accid,depid,dp) = 
	    let dep = findDep depid deps
		stmtid = if dp==Source then sourceNumber dep
			 else destinationNumber dep
                (idxfct,arrId) = findIndexFctAndArrID accid raccs
		in Access {
		  accessId = accid,
		  indexFct = idxfct,
		  stmtId = stmtid,
		  domain = findIspc stmtid stmts,
		  accessType = depTypeToAccType dp (depType dep),
                  arrayID = arrId
		  }

raccsToAccs :: [RawAccess] -> [Statement] -> [Access]
raccsToAccs raccs stmts = map raccToAcc raccs
  where raccToAcc (arrayId,accType,laNode,stmtNr,idxFct) =
              Access{
                     accessId = laNode,
                     indexFct = idxFct,
                     stmtId = stmtNr,
                     domain = findIspc stmtNr stmts,
                     accessType = accType,
                     arrayID = arrayId }
                 
