-- program to generate an allocation which has the 
-- Forward Communication Only property, if possible

import List
import Ratio
import Vector
import Loopo
import IOExts

import DepAcc
import TrDep
import ReduceDims

----------------------------------------------------------------------
-- how to get input data from the ispc file
----------------------------------------------------------------------

load :: (FilePath -> IO (LPResult a)) -> FilePath -> IO a
load f file =
    do res <- f file
       case res of
           ParseOk x    -> return x
	   ParseFailed f -> error ("Fehler: " ++ show f)


readslpfile :: String -> IO ([(Integer,[Integer])])
readslpfile datei = 
     do resslps <- load readSLP (datei ++ ".slp")
	return resslps

readaccfile :: String -> IO ([(Integer, AccessType, Integer, Integer, AffineFunction)])
readaccfile datei = 
     do resaccs <- load readACC (datei ++ ".acc")
	return resaccs

domainOfStmt :: Integer -> [(Integer,IneqSystem)] -> IneqSystem
domainOfStmt stmntid  ispcs =  
    case ( find (\(s,d)-> stmntid==s) ispcs ) of
         Nothing    -> error("statement not found")
         Just (s,d) -> d




----------------------------------------------------------------------
-- hardcoded input
----------------------------------------------------------------------

--datei = "/home/ststud/sommer/jetztArbeit/sor/sor-1dim.src.out"
-- datei = "simplegauss"
-- datei = "exs/while.src.out"
-- datei = "exs/accbased/example10_2.src.out"
--datei = "exs/feaut/example10_2.src.out"
--datei = "exs/gauss-2.src.out"
--datei = "exs/gaussimple.src.out"
--thedatei = "exs/gauss-half.src.out"
--datei = "exs/feautgasusssimple/gaussimple.src.out"


----------------------------------------------------------------------
-- the farkas method in our case
----------------------------------------------------------------------


-- now the pieces of the farkas equation system:

row4const :: Integer -> Matrix
row4const c = rowMatrix ([0 | i <-[1..(c-1)]] ++ [1])

farkas :: Matrix -> Integer -> Integer -> IneqSystem -> Integer -> Matrix
farkas lhs fullWidth nrZeroBlocksBefore ispc nrZeroBlocksAfter = 
-- the ispc must be given in all indices!
    let cols = nVars ispc + nParams ispc + 1
        rows = fullWidth        --nRows lhs 
        res =  multiconc1 [
            lhs,
            (zeroMatrix rows nrZeroBlocksBefore),
            (smul (mtranspose 
                      -- we append the row for the constant before, because we
                      -- are used to call the respective lambda "lambda_0"
                      ( join (row4const cols) (getMatrix ispc) )
                  ) (-1)),
            (zeroMatrix rows nrZeroBlocksAfter)
          ]
    in -- trace ("Farkas liefert: " ++ show res) 
       res


formGroups :: [Integer] -> [a] -> [[a]]
formGroups [] _     = []
formGroups (w:ws) l = grp : formGroups ws rest
		      where (grp,rest) = splitAt (fromIntegral w) l

countZeroGroups :: [Integer] -> [Rational] -> [Integer]
countZeroGroups widths row = map fst $
			     filter (\(_,vs) -> all (==0) vs) $
			     zip [1..] $
			     formGroups widths row 


filterNonConstantSolution :: [Integer] -> ([Rational],[Rational]) -> Bool
-- one list of Rationals is one vector (farkas or nonFarkas)
filterNonConstantSolution poss (sols,_) = not $ or
            $ map ( all (0==) )
            $ map init
            $ formGroups poss sols


widthsToGroups :: AllocMethod -> [(Dependence,Integer)] -> [Integer]
widthsToGroups _ [] = []
widthsToGroups method ((d,w):xs) = if method==General &&
                               depTypeToAccType Dest (depType d) == Read then
                                  w:(w:(widthsToGroups method xs))
			     else w:(widthsToGroups method xs)


superDepAndOffsetList :: AllocMethod -> [Statement] -> [Dependence] -> (Integer,[(Dependence,(Integer,Integer))])
superDepAndOffsetList method stmts deps = 
    let simpleAndExtWidths = 
              map   (\d-> let width = 1+nRows (destination d)
                          in
                              if method == General && 
                                 depTypeToAccType Dest (depType d) == Read then
                                   (width, 2*width)
			      else (width, width))
		    deps
	(widths, extWidths) = unzip simpleAndExtWidths
        offsets = superPrefixSum  extWidths
    in 
        (last offsets, zip deps (zip (init offsets) widths))

superAccAndOffsetList :: Integer -> [Statement] -> [Access] -> (Integer,[(Access,(Integer,Integer))])
superAccAndOffsetList initOffs stmts accs = 
    let widths = map   (\a-> (1+nRows
				   (snd (errorFindStmt stmts (stmtId a)))))
				  accs
        offsets = map (initOffs+) (superPrefixSum widths)
    in 
        (last offsets, zip accs (zip (init offsets) widths))





-- type Statement = (Integer,IneqSystem)

data AllocMethod = DepDriven | OwnerDriven | General | Small  deriving (Show,Eq,Read)

---

main :: IO()
-- main = mainx "exs/feautgasusssimple/gaussimple.src.out" "General"
main = mainx "exs/feautgasusssimple/gaussimple.src.out" "DepDriven"


mainx :: String -> String -> IO()
mainx datei methodstr =
    let 
--      datei = "exs/gauss-half.src.out"
--      datei = "exs/feautgasusssimple/gaussimple.src.out"
--        datei = "exs/gauss-2.src.out"
--        datei = "exs/move-read-acc.src.out"
--        method = General
        method = read methodstr
    in
    do 
       stmts' <- load readISPC (datei ++ ".ispc")  -- :: [Statement]
       slps <- readslpfile datei
       deps <- load readDependences (datei ++ ".dep") -- :: [Dependence]
       acps <- load readACP  (datei ++ ".acp") -- :: [(Integer, Integer, Integer)]
       rawaccesses <- readaccfile datei
       let stmts = reducedIspcsStmts ((map (\ (w,(x,y))->((w,y),x))) (joinZip' slps stmts'))
	   accs = acpsToAccs slps acps stmts (map toTrDependence deps) rawaccesses  -- 19.2.03  slps dazu, damit sind die Accesses nur noch in umgebenden Schleifen indiziert -- keine Ahnung ob das passt -- MG
	   stmtPosList = stmtPoss stmts
	   accPosList = accPoss accs stmts
	   stmtWidthList = stmtWidths stmts
	   accWidthList = accWidths accs 
	   tmpafffct = (hTrafo.head) deps
	   fullWidth = nVars tmpafffct + nParams tmpafffct +1
	   glhswidth = glhsWidth method stmts accs
	   (depRhsWidth, theDepAndOffsetList) = superDepAndOffsetList method stmts deps
	   (rhsWidth, theAccAndOffsetList) = 
               if (method==General) then
                  superAccAndOffsetList depRhsWidth stmts (filter (\a->accessType a==Write) accs)
               else (depRhsWidth,[])
           eqs1 = multijoin1 
		 [ farkas
		     anydep
                     fullWidth
		     offs 
--		     (snd stmt')
		     (destination dep)
		     (rhsWidth-offs-(nRows (destination dep))-1+1)       
                                       -- +1 fuer die unbenoetigte 1
	 	   | (dep,(offs,_)) <- theDepAndOffsetList, 
		     let stmt' = errorFindStmt stmts' (destinationNumber dep)
                         anydep = if method == DepDriven then
		                        depDrivenToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth 
                                 else   
                                        anyDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
		   ]
           eqs2 = multijoin1 
		 [ farkas 
		     otherdep
                     fullWidth
		     (offs+width)
		     (destination dep)
--		     (snd stmt')
		     (rhsWidth-offs-width-(nRows (destination dep))-1+1)   
                                       -- +1 fuer die unbenoetigte 1
	 	   | (dep,(offs,width)) <- theDepAndOffsetList, 
		     let stmt' = errorFindStmt stmts' (destinationNumber dep)
			 acctype = depTypeToAccType Dest (depType dep) 
			 otherdep = readDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
-- MG stimmt der Umgang mit non-true-deps?
                     ,
                     (acctype == Read)
		   ]
           eqs3 = if (length theAccAndOffsetList)>0 then
                      multijoin1 [ farkas 
  		        writeconstraint
                        fullWidth
		        offs
		        (snd stmt')
		        (rhsWidth-offs-(nRows (snd stmt'))-1+1) 
                                       -- +1 fuer die unbenoetigte 1
	 	        | (acc,(offs,width)) <- theAccAndOffsetList, accessType acc == Write, 
		          let stmt' = errorFindStmt stmts' (stmtId acc)
			      writeconstraint = writeAccToMatrix acc fullWidth stmts stmts' slps acps accs stmtPosList accPosList glhswidth
		        ]
                  else zeroMatrix 0 (nColumns eqs1)

	   eqs = case method of
                      General     -> join (join eqs1 eqs2) eqs3
--                      General     -> eqs1
                      Small       -> eqs1
		      DepDriven   -> eqs1 -- no eqs2 since no realloc
                                          -- no eqs3 since implicitly given (OCR)
                      OwnerDriven -> error ("OwnerDriven not yet integrated.")
           nrNonFarkasVars = nColumns eqs-1-rhsWidth     
           eqsAsIeqs = ineqSystem((nColumns eqs-1,0), join eqs (-eqs))
	   nonNegFarkas = ineqSystem((nColumns eqs-1,0), multiconc1 [
               zeroMatrix rhsWidth nrNonFarkasVars,
	       unitMatrix rhsWidth rhsWidth,
	       zeroMatrix rhsWidth 1 ])
           allIneqs = join eqsAsIeqs nonNegFarkas
--	   lrv = dualOfIS allIneqs
           lrv = dualOfSystem (eqSystem((nColumns eqs-1,0), eqs)) nonNegFarkas
	   nonFf = map mk 
--                           (filter (\v->not (all (0==) 
--                                       (take (fromInteger(snd(head accPosList)-1)) 
--				       -- MG hack: incorrect for different A-dims,
--				       --    e.g., move-read-acc.src
--                                             (vectorEntries v))))
                                  (lrvRays lrv)
--                                              )
	   mk ray = splitAt (fromIntegral nrNonFarkasVars) $
		    vectorEntries ray

           rhsWidthList = widthsToGroups method
                         (zip (fst (unzip theDepAndOffsetList))
			      (snd (unzip (snd (unzip theDepAndOffsetList))))) ++
		       snd (unzip (snd (unzip theAccAndOffsetList)))
           lhsWidthList = if method==DepDriven then stmtWidthList
                          else (stmtWidthList ++ accWidthList)

           sortedGroups = sortBy (\(r1,_,c1)(r2,_,c2) ->
			          compare (length c2) (length c1))
	     [ (formGroups lhsWidthList nonFarkas, farkas, countZeroGroups rhsWidthList farkas)  
	                        | (nonFarkas,farkas) <- nonFf
                                ,  filterNonConstantSolution lhsWidthList
                                     (nonFarkas,farkas)
                                ]
	   
       putStrLn ((show  deps) ++ "\n\n\n")
--       putStrLn (show  (combineDepAcp deps acps))
--       putStrLn (show  accs)
--       putStrLn (show(reducedIspcs (map (\(x,y)->(y,x)) (joinZip slps stmts'))))
--       putStrLn("The reduced statments are: " ++ show stmts)
--       putStrLn (show(oldStmtPoss (map (\(x,y)->(y,x)) (joinZip slps stmts))))
       putStrLn ("Accesses:" ++ show accs)
       putStrLn ("\nAccess-Pairs: "  ++ show acps)
       putStrLn ("glhsWidth = " ++ show glhswidth)
       putStrLn ("StmtWidths = " ++ (show stmtWidthList))
       putStrLn ("AccWidths = " ++ (show accWidthList))
--       putStrLn ("Positions: " ++ (show (map snd (stmtPoss stmts))) ++ " und " ++ (show (map snd (accPoss accs stmts))))
       putStrLn ("Positions: " ++ (show stmtPosList) ++ " und " ++ (show accPosList))
       putStrLn ("Position of the last stmt: " ++ (show (stmtPos stmtPosList (last stmts))))
       putStrLn ("fullWidth: " ++ (show fullWidth))
       putStrLn ("depAndOffestList: " ++ (show (fst(unzip(snd(unzip theDepAndOffsetList))))))
       putStrLn ("accAndOffestList: " ++ (show (fst(unzip(snd(unzip theAccAndOffsetList))))))
       putStrLn ("rhsWidth/depRhsWidth: " ++ (show rhsWidth) ++ "/" ++ (show depRhsWidth))
--       putStrLn ("eqs3: " ++ (show eqs3))
--       putStrLn ("eqs2: " ++ (show eqs2))
       putStrLn ("Eqs: \n" ++ show eqs)
----       putStrLn "\n\n Und nun alle Ineqs: \n"
----       putStrLn (show allIneqs)
---
--       putStrLn ( "rhsWidth = " ++ show rhsWidth ++ "\nFarkas-equalities\n" ++ show eqs )
       putStrLn ( "\nDual\nLines:\n" ++ show (lrvLines lrv) ++ "\nVertices:\n" ++ show (lrvVertices lrv) )
       putStrLn ( "\nrhsWidthList\n" ++ show rhsWidthList )
       putStrLn ( "\nlhsWidthList\n" ++ show lhsWidthList )
       putStrLn ( "\nSorted Groups\n" ++ unlines (map (\ (n,f,c) -> 
                                                 concat (concat (map (lines.(flip (++) "    ").show.vector) n))  ++ "\n" ++
                                                 show (vector f)  ++ 
                                                 show c
                                              ) (take 20 sortedGroups)) )
--       putStrLn ( "\nSorted Groups\n" ++ concat (map (\ (n,f,c) -> 
--                                                 show (map vector n)  ++ 
--                                                 show (vector f)  ++ 
--                                                 show c
--                                              ) sortedGroups) )
--       putStrLn ( "\nSorted Groups\n" ++ unlines (map (show.(\ (n,f,c)->(unlines(map (show.vector) n), vector f,c))) sortedGroups) )
       putStrLn ("")


stmtWidths :: [Statement] -> [Integer]
stmtWidths ss = map (nColumns.snd) ss

accWidths :: [Access] -> [Integer]
accWidths acs = map (nColumns.domain) acs

glhsWidth :: AllocMethod -> [Statement] -> [Access] -> Integer
glhsWidth method stmts accs = 
       if method == DepDriven then sum (stmtWidths stmts)
       else sum (stmtWidths stmts) + sum (accWidths accs)


stmtPoss ::  [Statement] -> [(Integer,Integer)]
stmtPoss s  = zip (map fst s) (superPrefixSum (stmtWidths s))


accPoss :: [Access] -> [Statement] -> [(Integer,Integer)]
accPoss a s = zip (map accessId a) ((map (\ x-> shift + x)) (superPrefixSum (accWidths a)))
	      where shift = snd(last (stmtPoss s)) + (nColumns.snd) (last s)

accPos :: [(Integer,Integer)] -> Access -> Integer
accPos l a = snd(errorFind (\ (i,p)-> (accessId a)==i ) l)

stmtPos :: [(Integer,Integer)] -> Statement -> Integer
stmtPos l s = snd(errorFind (\ (i,p)-> (fst s)==i ) l)

accIdPos :: [(Integer,Integer)] -> Integer -> Integer
accIdPos l accid = snd(errorFind (\ (i,p)-> accid==i ) l)

stmtIdPos :: [(Integer,Integer)] -> Integer -> Integer
stmtIdPos l stmtid = snd(errorFind (\ (i,p)-> stmtid==i ) l)


stmtMatrix :: Statement -> Integer -> [(Integer,[Integer])] -> Matrix
-- statement volldimensional bergeben, damit man was zum rausschneiden hat.
stmtMatrix (id, ispc) fullWidth slps = 
    getMatrix (filterrowdims ((ineqSystem((nVars ispc,nParams ispc), 
					     unitMatrix fullWidth fullWidth)),
			   (errorFindSlps slps id)))


accIdMatrix :: Access -> Integer -> [Statement] -> [(Integer,[Integer])] -> Matrix
-- stmts voll bergeben weils stmtMatrix so braucht
accIdMatrix acc fullWidth stmts' slps = 
    stmtMatrix (errorFindStmt stmts' (stmtId acc)) fullWidth slps

accHMatrix :: Access -> AffineFunction -> [Statement] -> [(Integer,[Integer])] -> Matrix
-- stmts' oder stmts ist egal
accHMatrix acc htrafo stmts' slps = 
    getMatrix (filterrowdims (htrafoAsIes, slpVect))
    where
          nv = nVars htrafo
	  np = nParams htrafo
          fullWidth = nv + np +1
	  stmt' = errorFindStmt stmts' (stmtId acc)
	  slpVect = errorFindSlps slps (fst stmt')
	  htrafoAsIes = ineqSystem((nv,np), getMatrix htrafo)



errorFindSlps :: [(Integer, [Integer])] -> Integer -> [Integer]
errorFindSlps slps id = snd (errorFind (\ (key,_)-> key==id ) slps)

errorFindStmt :: [Statement] -> Integer -> Statement
errorFindStmt stmts id = errorFind (\ (key,_)-> key==id ) stmts


depDrivenToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
depDrivenToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let stmt = errorFindStmt stmts (destinationNumber dep)
	stmt' = errorFindStmt stmts' (destinationNumber dep)
	stmtMat = stmtMatrix stmt' fullWidth slps
	stmtpos = stmtPos stmtposs stmt
	srcacc = srcAcc (depNumber dep) acps accs
        srcstmtid = stmtId srcacc
	srcstmt = errorFindStmt stmts srcstmtid
	srcMat  = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	srcpos = stmtPos stmtposs srcstmt
	prezeros = min srcpos stmtpos
	(firstMat, midzeros, sndMat, postzeros) =
               if srcpos > stmtpos then
                  (stmtMat, 
                   srcpos - stmtpos - ((nColumns.snd) stmt),
                   srcMat,
                   glhswidth - srcpos - ((nColumns.snd) srcstmt))
               else if srcpos < stmtpos then
                  (srcMat,
                   stmtpos - srcpos - ((nColumns.snd) srcstmt),
                   stmtMat,
                   glhswidth - stmtpos - ((nColumns.snd) stmt))
               else -- source and destination are the same stmt
                  (srcMat+stmtMat,
                   0,
                   zeroMatrix 0 fullWidth,
                   glhswidth - stmtpos - ((nColumns.snd) stmt))
    in  
        multiconc1 [
		   zeroMatrix fullWidth prezeros,
		   mtranspose firstMat,
		   zeroMatrix fullWidth midzeros,
		   mtranspose sndMat,
		   zeroMatrix fullWidth postzeros
		   ] 


anyDepToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
anyDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let stmt = errorFindStmt stmts (destinationNumber dep)
	stmt' = errorFindStmt stmts' (destinationNumber dep)
	stmtMat = stmtMatrix stmt' fullWidth slps
	stmtpos = stmtPos stmtposs stmt
	srcacc = srcAcc (depNumber dep) acps accs
	accMat  = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	accpos = accPos accposs srcacc
	midzeros = accpos - stmtpos - ((nColumns.snd) stmt)
	postzeros = glhswidth - accpos - ((nColumns.domain) srcacc)
    in  -- trace ("stmtpos: " ++ show stmtpos ++ ", midzeros: " ++ show midzeros ++ ", postzeros: " ++ show postzeros ++ "\nSrc-Access: " ++ show srcacc ++ "\naccMatrix: " ++ show accMat ++ "\nstmtMat: " ++ show stmtMat)
        (multiconc1 [
		   zeroMatrix fullWidth stmtpos,
		   mtranspose stmtMat,
		   zeroMatrix fullWidth midzeros,
		   mtranspose accMat,
		   zeroMatrix fullWidth postzeros
		   ] )

-- now: if access (=current access) writes:

writeAccToMatrix :: Access -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix

writeAccToMatrix acc fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth = 
    let stmt = errorFindStmt stmts (stmtId acc)
	stmt' = errorFindStmt stmts' (stmtId acc)
	stmtMat = -(stmtMatrix stmt' fullWidth slps)
	stmtpos = stmtPos stmtposs stmt
	accMat  = accIdMatrix  acc fullWidth stmts' slps    -- unitMatrix fullWidth fullWidth
	accpos = accPos accposs acc
	midzeros = accpos - stmtpos - ((nColumns.snd) stmt)
	postzeros = glhswidth - accpos - ((nColumns.domain) acc)
        gesamt =  multiconc1 [
		   zeroMatrix fullWidth stmtpos,
		   mtranspose stmtMat,
		   zeroMatrix fullWidth midzeros,
		   mtranspose accMat,
		   zeroMatrix fullWidth postzeros
		   ]
    in gesamt


-- now: if dependence destination (=current access) reads:
readDepToMatrix :: Dependence -> Integer -> [Statement] -> [Statement] -> [(Integer,[Integer])] -> [(Integer, Integer, Integer)] -> [Access] -> [(Integer, Integer)] -> [(Integer, Integer)] -> Integer -> Matrix
readDepToMatrix dep fullWidth stmts stmts' slps acps accs stmtposs accposs glhswidth =
    let srcacc = srcAcc (depNumber dep) acps accs
	srcmat = -(accHMatrix srcacc (hTrafo dep) stmts' slps)
	srcpos = accPos accposs srcacc
	dstacc = dstAcc (depNumber dep) acps accs
	dstmat  = accIdMatrix  dstacc fullWidth stmts' slps    -- unitMatrix fullWidth fullWidth
	dstpos = accPos accposs dstacc
        matrix = if dstpos == srcpos then 
                    multiconc1 [
	              	         zeroMatrix fullWidth srcpos,
		                 mtranspose (srcmat+dstmat),
		                 zeroMatrix fullWidth (glhswidth - srcpos - ((nColumns.domain) srcacc))
		                 ]
--  		    zeroMatrix fullWidth glhswidth
--  		    zeroMatrix 0 glhswidth
                 else 
                    let (midzeros, postzeros, matrix') =
                          if dstpos > srcpos then
	                      (
	                      dstpos - srcpos - ((nColumns.domain) srcacc),
	                      glhswidth - dstpos - ((nColumns.domain) dstacc),
	                      multiconc1 [
		                  zeroMatrix fullWidth srcpos,
		                  mtranspose srcmat,
		                  zeroMatrix fullWidth midzeros,
		                  mtranspose dstmat,
		                  zeroMatrix fullWidth postzeros
		                  ]
	                      )
	                  else (  --  dstpos < srcpos 
	                      srcpos - dstpos - ((nColumns.domain) dstacc),
	              	      glhswidth - srcpos - ((nColumns.domain) srcacc),
	              	      multiconc1 [
	              	         zeroMatrix fullWidth dstpos,
		                 mtranspose dstmat,
		                 zeroMatrix fullWidth midzeros,
		                 mtranspose srcmat,
		                 zeroMatrix fullWidth postzeros
		                 ]
	                      )
                    in  matrix'
    in --trace ("Lhs-Matrix for src " ++ show (accessId srcacc) ++ 
       --       " / dst "  ++ show (accessId dstacc) ++ 
       --       ": "  ++ show matrix
       --      ) 
--             (if depNumber dep /= 5 then matrix
--  		else     zeroMatrix fullWidth glhswidth ) -- ja
--	        else -matrix)                             -- nein
                  matrix                                  -- nein

