module Renum where

import Control.Exception ( assert, evaluate )
import Data.List
import Data.Maybe
import Data.Map ( Map )
import qualified Data.Map as Map
import System.Environment
import System.IO

import LMath
import LMath.Algo.EhrhartArith
-- import LMath.Algo.LGen
import LMath.Algo.OmegaTest

import LoopoDS

import Debug.Trace


--main :: IO ()
--main = getArgs >>= mainA

mainA :: [String] -> IO ()
mainA args = undefined


localise :: RealFrac a
  => [String]       -- variable names
  -> [String]       -- param names
  -> GIneqSystem a  -- context in the parameters
  -> [(ZDomain, [GAffineFunction a])]  -- index spaces and access functions
  -> Integer        -- index of t dimension
  -> Maybe ZDomain  -- supply another C
  -> (EPartition a
--     ,ZDomain a
     , EPartition a
--     , [(ZPolyhedron a, GAffineFunction a)]
--     , [(ZDomain a, [EValue a])]
--     , [[(ZDomain a, [EValue a])]]
     , LGenParams a
     , String
     )
--  String -- LGenParams
localise varNames paramNames (ctx :: GIneqSystem a) ispcsAndAccFns tdim mbC =
--    lgen lgenOptionsStd (lgpContext lgp) (lgpStatements lgp)
--    last $ runLGen lgp
--    lgp
--    last $ runLGenWithStatements lgp stmtFn
    (delta
--    ,c
    , rho
--    ,df_list
--    , accCombs $ snd $ head ispcsAndAccFns
--    , map splitByRho ispcsAndAccFns
    , lgp
    , runLGenWithStatements lgo lgfRealFrac lgp lgoo numOrdDTreeContext (\str -> error "no conditions should be necessary") 0 -- (tD+1)
    )
    where
    df_list = [moveTtoParams tdim (a,f)
              | (dom,f) <- accSet, a <- dom
              ]
    c       = case mbC of
                  Nothing    -> computeC df_list
                  Just userC -> userC
    (i,o,r) = computeIOR tdim c mbW0
    rho     = computeRho c
    delta   = computeDelta (aDim+tdim) rho
    mbW0    = Nothing

    aDim = targetDimIH $ snd $ head accSet
    innerDims = zpTargetDim (head $ fst $ head accSet) - tdim - 1

    accSet :: [(ZDomain, GAffineFunction a)]
    accSet = [(ispc,accFn) | (ispc,accFns) <- ispcsAndAccFns, accFn <- accFns]

    nP = nParams ctx
    nIspcDims = zpTargetDim $ head $ fst $ head ispcsAndAccFns

    ctx' = insertZeroParams ctx 0 (tdim+1)
    (lgp, lgoo', tab) = computeLGenParams varNames paramNames tdim ctx splitByRho delta ispcsAndAccFns (i,o) rr
    tD = let (is:_,sc):_ = lgpStatements lgp in nVars sc - nVars is
--    lgoo = lgoo' { lgooStmtGen = Just stmtFn, lgooVarIFs=[2] }
    lgoo = lgoo' { lgooStmtGen = stmtFn
                 , lgooLoopGen = loopFn -- lgooVarParallelOff=[(2,("self","np"))]
                 }
    lgo = lgenOptionsStd { lgoSimplifyIFs=True
                         , lgoSimplifyOuterLoops=True
                         }

    getSchedPos :: Integer -> SchedPos
    getSchedPos stmtID = case lookup (fromIntegral stmtID) tab of
                           Just (_,sp,_) -> sp

    -- r has aDim target variables (the enumerated array indices)
    -- in dependence of (tdim+1) parameters (outer loops and t loop)
    -- and the structural parameters.
    -- neg/pos have 0 vars and aDim+(tdim+1)+#structural  parameters
    rr :: Either ZDomain (ZDomain,ZDomain)
    rr = case signumDelta delta of
            Nothing -> Left r
            Just (neg,pos) ->
                Right (intersect neg, intersect pos)
        where
        r' = map (zpMoveParamsToTargetDims 0 (tdim+1) aDim) r
        intersect :: [GIneqSystem a] -> ZDomain
        intersect syss =
             map (zpMoveTargetDimsToParams aDim (tdim+1) 0) $
             zdIntersection r' $
             map (\sys -> zpFromIneqs $
                          moveParamsToVars sys 0 (aDim+tdim+1) 0
                 ) syss

    loopFn :: Integer -> String -> [[GVector a]] -> [[GVector a]]
           -> String -> String -> [Integer] -> Maybe String
    loopFn n var [[]] [[]] _ _ _ = Just "if (1)"
    loopFn n var ls us lb ub stmtIDs
        | n == 2 && schedpos == [Reorg MoveBackward] =
            Just $ concat ["int iters = (", ub, "-(", lb, "))/np+1, ub="
                          , ub, ";\n", "for (", var, "=", lb
                          , "+self, count=0; count<iters; count++,"
                          , var, "+=np)"
                          ]
        | n == 2     =  Just $ loopGenSelfNP "self" "np" n var ls us lb ub
        | otherwise  =  Nothing
        where
        schedpos = map getSchedPos stmtIDs

    --  exprs  do not contain expressions for _local_dummy as
    --  _local_dummy is only introduced by the scatter functions
    stmtFn :: Integer -> [GVector a] -> Maybe String
    stmtFn stmtID exprs''
      | all (==1) (map denominator $ concat exprs') =
        Just $
        case schedPos of
           Sync _  -> "__syncthreads();"
           MoveIn  -> concat ["// move in "
                             ,show (last exprs), " to ", show evs, "\n"
                             ,"L[", lOff, "] = A[", aOffs, "];"
                             ]
           MoveOut -> concat ["// move out ", show evs, " to "
                             ,show (last exprs), "\n"
                             ,"A[", aOffs, "] = L[", lOff, "];"
                             ]
           Reorg t -> 
               concat ["// reorg ", show t, " ", show evs, " by "
                      , show dev, "\n"
                      ,"int x=", lOff, ";\n"
                      ,"float val;\n"
                      ,"if (p <= ub) val = L[x];\n"
                      ,"__syncthreads();\n"
                      ,"if (p <= ub) L[x+", deltaE, "]=val;"
                      ]
           S sid   -> let accFns = snd (ispcsAndAccFns !! (sid-1)) in
                      -- "/* " ++ show (zip (map newAccFn accFns) (map (evInsertZeroParams (tdim+2) innerDims) evs)) ++ " */\n" ++
                      concat ["S", show sid, "("
                             ,concat (intersperse "," $
                                      zipWith showSubFn evs accFns
                                     )
                             ,");"
                             ]
 --show evs ++ "(" ++ show (zipWith subFn evs accFns) ++ ")"
      | otherwise = error "stmtFn: fractional case not handled"
      where
      Just (evs, schedPos, dev) = lookup (fromIntegral stmtID) tab
      subFn :: EValue a -> GAffineFunction a -> EValue a
      subFn ev accFn = foldl (.) id (zipWith evSubstParam [0..] naccFns) nev
          where
          naccFns = map (\v -> insertEntries v 0 (genericReplicate aDim 0)) $
                    rowVectors $ getMatrixIH (newAccFn accFn)
          nev = evInsertZeroParams (tdim+2) innerDims ev
      showSubFn ev accFn = evalueToCExpression computeParamNamesNEV $
                           subFn ev accFn
      exprs' = map (map toRational . vectorEntries) exprs''
      exprs = map (map numerator) exprs'

      -- assumes that the dummy dimension has coefficient 0
      exprsWOdummies = map (\v -> let (a,_:b) = genericSplitAt (tdim+1) v in a ++ b) $
                       genericTake nIspcDims exprs

      resubFn = fmap fromIntegral $
                affineFunction ((nV,nP), matrix exprsWOdummies) Inhomogeneous
           where
           nV = genericLength (head exprsWOdummies) - nP - 1

      newAccFn accFn =
          assert (all (==1) $ map (denominator.toRational) $ entries accFn) $
          composeAffine accFn' resubFn
          where
          accFn' = fmap (numerator.toRational) accFn

      lOff = evalueToCExpression moveParamNamesEV $ head evs
      aOffs = concat $ intersperse "][" [evalueToCExpression moveParamNamesA $ evFromAffine (vector e) | e <- aExprs]
      aExprs = reverse $ genericTake aDim $ reverse exprs
      deltaE = evalueToCExpression moveParamNamesEV dev

    computeParamNamesNEV = error "should never be used" : varNames ++ paramNames
    moveParamNamesEV = genericTake aDim vars2 ++ vars1 ++ paramNames
      where
      (vars1, vars2) = genericSplitAt (tdim+1) varNamesWOdummy
    moveParamNamesA = varNamesWithDummy ++ paramNames

    varNamesWOdummy = vars1 ++ vars2
        where
        (vars1, _:vars2) = genericSplitAt (tdim+1) varNamesWithDummy
    varNamesWithDummy = lgooVarNames lgoo

    splitByRho :: (ZDomain,[GAffineFunction a]) -> [(ZDomain, [EValue a])]
    splitByRho (ispc,accs) =
--        trace "splitByRho" $
        sbr ispc $ accCombs accs

    sbr :: ZDomain -> [(ZDomain, [EValue a])] -> [(ZDomain, [EValue a])]
    sbr []   _  = []
    sbr ispc [] = []
    sbr ispc ((sys,evs):rest)
        | not (null inters)  =  (inters,evs) : cont
        | otherwise          =  sbr ispc rest
        where
        inters = ispc `zdIntersection` sys
        cont   = sbr (ispc `zdDifference` sys) rest


    accTable :: Map [a] [(ZDomain, [EValue a])]
    accTable = Map.fromList $
               map (\afs -> (concatMap entries afs, compAccCombs afs)) $
               ([ident] : [negIdent] : [accFns | (_,accFns) <- ispcsAndAccFns])
        where
        ident    = insertZeroVars (afIdentity aDim nP) 0 (tdim+1)
        negIdent = afMirror ident
                

    accCombs :: [GAffineFunction a] -> [(ZDomain, [EValue a])]
    accCombs accs = case Map.lookup (concatMap entries accs) accTable of
                        Just res -> {- trace ("FOUND:" ++ show accs) -} res
                        Nothing  -> trace ("WARNING: accCombs" ++ show accs) $
                                    compAccCombs accs
   

    compAccCombs :: [GAffineFunction a] -> [(ZDomain, [EValue a])]
    compAccCombs accs =
        ac [([zpFromIneqs $ ineqSystem ((nV,nP), zeroMatrix 0 (nV+nP+1))], [])]
           accs
        where
        nV = nVars   $ head $ accs
        nP = nParams $ head $ accs

    ac :: [(ZDomain, [EValue a])] -> [GAffineFunction a] -> [(ZDomain, [EValue a])]
    ac combs [] = combs
    ac combs (af:afs) =
        ac combs2 afs
        where
        combs2 = [ (dom2, evs ++ [ev])
                 | (sys,ev) <- rhoParts
                 , let pre = inverseImage (withT af) (moveParamsToVars sys 0 (targetDimIH af+tdim+1) 0)
                 , (dom,evs) <- combs
                 , let dom2 = dom `zdIntersection` [zpFromIneqs pre]
                 , not (null dom2)
                 ]

    withT af = affineFunction ((nV,nP), matrix (upper ++ tvec ++ lower))
                              Inhomogeneous
        where
        upper, lower, tvec :: [[a]]
        (upper,lower) = genericSplitAt (targetDimIH af + tdim) $
                        rows' $ getMatrixIH af
        tvec = rows' $ conc (unitMatrix (tdim+1) (tdim+1))
                            (zeroMatrix (tdim+1) (nV+nP+1-(tdim+1)))
        nV = nVars af
        nP = nParams af

    rhoParts' = epPartsDisjoint rho
    rhoParts = rhoParts' ++ [(s,0) | s <- sys]
        where
        sys = disjointUnionDifference Nothing [univ] $ map fst rhoParts'
        univ = ineqSystem ((0,nP), zeroMatrix 0 (nP+1))
        -- fails if  rhoParts'==[]
        nP = nParams $ fst $ head rhoParts'
--                                                    t  p  n
bsp1ia = [([zpFromIneqs (ineqSystem ((2,1), matrix [[ 0, 1, 0, 0%1]
                                                   ,[ 0,-1, 1, 0]
                                                   ,[ 1, 0, 0, 0]
                                                   ,[-1, 0, 1, 0]
                                                   ]
                                  )
                      )
        ],
                 [affineFunction ((2,1), matrix [[ 1, 1, 0, 1%1]])
                                 Inhomogeneous
                 ]
        )
       ]
bsp1ctx = ineqSystem ((0,1), matrix [[1,-2%1]])
bsp1 = localise ["t","p"] ["n"] bsp1ctx bsp1ia 0 Nothing

bsp2ia   = [([bsp2stmt1], [affineFunction ((2,2), matrix [[0, 1, 0, 0, -1]])
                                          Inhomogeneous
                          ,affineFunction ((2,2), matrix [[0, 1, 0, 0, 0]])
                                          Inhomogeneous
                          ,affineFunction ((2,2), matrix [[0, 1, 0, 0, 1]])
                                          Inhomogeneous
                          ]
            )
           ]
--                                          k  i  m  n
bsp2stmt1src = ineqSystem ((2,2), matrix [[ 1, 0, 0, 0,-1]
                                         ,[-1, 0, 1, 0, 0]
                                         ,[ 0, 1, 0, 0,-1]
                                         ,[ 0,-1, 0, 1,-1]
                                         ]
                          )
bsp2stmt1 = zpFromIneqs bsp2stmt1src
bsp2ctx = ineqSystem ((0,2), matrix [[ 1, 0,-4], [ 0, 1,-4]])

bsp2 = localise ["k", "i"] ["m", "n"] bsp2ctx bsp2ia 1 Nothing

bsp3stmt1 = zpolyhedron bsp2stmt1src $
            affineFunction ((2,2), matrix [[ 2, 1, 0, 0,-3] -- t=i+2k-3
                                          ,[ 1, 0, 0, 0,-1] -- p=k-1
                                          ]
                           ) Inhomogeneous
-- i = t-2k+3 = t-2(p+1)+3 = t-2p+1
bsp3ia = [([bsp3stmt1], [affineFunction ((2,2), matrix [[1,-2,0,0,0%1]]) Inhomogeneous
                        ,affineFunction ((2,2), matrix [[1,-2,0,0,1]]) Inhomogeneous
                        ,affineFunction ((2,2), matrix [[1,-2,0,0,2]]) Inhomogeneous
                        ]
          )]
--                                                      x  t  m  n
bsp3alt2C = [zpFromIneqs $ ineqSystem ((1,3), matrix [[ 1, 0, 0, 0, 0]
                                                     ,[ 1,-1, 2, 0, 2]
                                                     ,[-1, 0, 0, 1, 0]
                                                     ,[-1, 1, 0, 0, 2]
                                                     ]
                                      )
            ]
bsp3ctx = bsp2ctx
bsp3 = localise ["t", "p"] ["m", "n"] bsp3ctx bsp3ia 0 Nothing
-- bsp3alt  = localise ["t", "p"] ["m", "n"] bsp3ctx bsp3ia 0 (Just bsp3altC)
bsp3alt2 = localise ["t", "p"] ["m", "n"] bsp3ctx bsp3ia 0 (Just bsp3alt2C)

--                                          p  t  n
bsp1df = [(zpFromIneqs $
           ineqSystem ((1,2),     matrix [[ 1, 0, 0, 0%1]
                                         ,[-1, 0, 1, 0]
                                         ,[ 0, 1, 0, 0]
                                         ,[ 0,-1, 1, 0]
                                         ]
                      )
          ,affineFunction ((1,2), matrix [[ 1, 1, 0, 1%1]]) Inhomogeneous
          )
         ]

flatIA :: [(ZDomain, [GAffineFunction a])] -> [(ZPolyhedron, GAffineFunction a)]
flatIA ia = [(d,af) | (dom,afs) <- ia, d <- dom, af <- afs]

-- Move dimension t and all outer dimensions 
-- to the parameters (in leading position).
-- Dimension t has index  tdim  in the parameters, then.
moveTtoParams :: RealFrac a
  => Integer
  -> (ZPolyhedron, GAffineFunction a)
  -> (ZPolyhedron, GAffineFunction a)
moveTtoParams tdim (dom,f) =
    (zpMoveTargetDimsToParams 0 (tdim+1) 0 dom, moveDims f)
    where
    moveDims f =
        affineFunction ((nV-tdim-1,nP+tdim+1), getMatrix $ moveVarsToParams ae 0 (tdim+1) 0) Inhomogeneous
        where
        ae = affineExp ((nV,nP), getMatrixIH f)
        nV = nVars f
        nP = nParams f

computeC :: RealFrac a
  => [(ZPolyhedron, GAffineFunction a)]   -- (D_j,f_j) list
  -> ZDomain
computeC df_list = foldl zdUnion [] [[zpImage f d] | (d,f) <- df_list]

computeIOR ::
     Integer
  -> ZDomain               -- C
  -> Maybe ZDomain         -- W_0 
  -> (ZDomain, ZDomain, ZDomain)
computeIOR tdim' c mbW0 =
    (zdDifference c (maybe c_m1 (zdUnion c_m1) mbW0)
    ,zdDifference c c_1
    ,zdIntersection c c_1
    )
    where
    c_1  = map (zpSubstParam tdim' $ t   1 ) c
    c_m1 = map (zpSubstParam tdim' $ t (-1)) c

    t x = vector (replicate tdim 0 ++ 1 : replicate (nP-1-tdim) 0 ++ [x])
    nP = zpParams $ head c
    tdim = fromIntegral tdim'

computeRho :: RealFrac a
  => ZDomain     -- C
  -> EPartition a
computeRho c = zdLexicographicIndex True Nothing c

computeDelta :: RealFrac a
  => Integer           -- index of t dimension (note: rho has array dims)
  -> EPartition a      -- rho
  -> EPartition a
computeDelta tDim rho =
    rho1 - rho
    where
    rho1 = epSubstParamSum tDim 1 rho

signumDelta :: Real a
  => EPartition a                 -- Delta
  -> Maybe ([GIneqSystem a]       -- where Delta is negative
           ,[GIneqSystem a])      -- where Delta is positive
signumDelta delta
    | all (isJust.snd) xs =
        Just ( simplifyDomain [] $ oT $ map mkNeg xs
             , simplifyDomain [] $ oT $ map mkPos xs
             )
    | otherwise = Nothing
    where
    xs = [(cond,evAffine (nParams cond) ev) | (cond,ev) <- epPartitions delta]

    mkPos (cond,Just v) = mk cond (sub1 v)
    mkNeg (cond,Just v) = mk cond (sub1 $ negate v)

    sub1 v = v + vector (replicate (vectorLength v-1) 0 ++ [-1])

    mk cond v = cond `join` ineqSystem ((nV,nP),m)
        where
        m = rowVectorsToMatrix1 [v]
        nV = nVars cond
        nP = nParams cond

oT :: Real a => [GIneqSystem a] -> [GIneqSystem a]
oT syss = [sys | sys <- syss, omegaTest (Just sys) Nothing]

data DirectionOfMovement = MoveBackward
                         | MoveBackwardCheck
                         | MoveForward
                         | MoveForwardCheck
                         deriving (Eq, Show)

data SchedPos = S Int
              | MoveIn
              | MoveOut
              | Reorg DirectionOfMovement
              | Sync SchedPos
              deriving (Eq, Show)

computeLGenParams :: RealFrac a
 => [String]           -- variable names
 -> [String]           -- parameter names
 -> Integer
 -> GIneqSystem a
 -> ( (ZDomain, [GAffineFunction a]) -> [(ZDomain, [EValue a])] )
 -> EPartition a   -- Delta
 -> [(ZDomain, [GAffineFunction a])]        -- index spaces of statements with t as td at idx tdim and access functions
 -> (ZDomain, ZDomain)  -- I, O with t as first param
 -> Either ZDomain (ZDomain,ZDomain)  -- R  (bw,fw)
 -> (LGenParams a, LGenOutputOptions a, [(Int, ([EValue a], SchedPos, EValue a))])
computeLGenParams varNames paramNames tdim (ctx :: GIneqSystem a) splitByRho delta ispcsAndAccFns (i,o) rr =
--    trace ("scatterDims=" ++ show (scatterDims,splitList)) $
    (LGP { lgpContext = ctx,
           lgpStatements = stmts
         }
    ,realFracLGOO iterNames varNamesWithDummy paramNames
--    , [(stmtID,(evs,schedPos,dev)) | ((_,evs,schedPos,dev),stmtID) <- zip splitList [1..]]
    , zip [1..] [(evs,schedPos,dev) | (ispcs,evs,schedPos,dev) <- splitList, zp <- ispcs]
    )
    where
    stmts = [([zpPreimage zp]
             ,equalitiesToInequalities $ 
              afRelation True $
              insDim sid $ zpLattice zp
             )
            | (ispc,_,sid,_) <- splitList, zp <- ispc
            ]
    l = fromIntegral $ length ispcsAndAccFns
    
    moveDims zd = map (zpMoveParamsToTargetDims 0 (tdim+1) 0) zd

--    iterNames = ["iter"++show i | i <- [1..]]

    (varNamesWithDummy,iterNames) = (a ++ "_local_dummy" : b ++ fill, [])
        where
        (a,b) = splitAt (fromIntegral tdim+1) varNames
        fill = genericTake (scatterDims - genericLength varNames - 1)
                           ["_local_dummy" ++ show i | i <- [0..]]

    insDim sid af =
        affineFunction ((nV,nP), multijoin1 (upper ++ vPart ++ lower ++ fill))
                       Inhomogeneous
        where
        (upper, lower) = splitAt (fromIntegral tdim + 1) $ rows $ getMatrixIH af
        vPart = rows $ getMatrixIH $ constantAffineFunction nV nP (matrix [[schedPos sid]])
        fill = rows $ zeroMatrix (scatterDims - (targetDimIH af + 1)) (nV+nP+1)
        nV = nVars af
        nP = nParams af

    schedPos sid =
        case sid of
              S _     -> 2
              MoveIn  -> 0
              MoveOut -> 4
              Reorg MoveBackward      -> 6
              Reorg MoveBackwardCheck -> 8
              Reorg MoveForward       -> 10
              Reorg MoveForwardCheck  -> 12
              Sync sid'               -> schedPos sid' + 1 

    scatterDims = 1 + maximum [zpTargetDim d | (dom,_,_,_) <- splitList, d <- dom]


    splitList =
        concat
        [[(dom,evs,sid,dev)
         , (makeDimsUniv [tdim+1 .. zpTargetDim (head dom)-1] dom,[],Sync sid,0)
         ]
        | ((ispc,accFns),sid,dev) <- zip3 ispcsAndAccFns (map S [1..]) (repeat 0) ++ ior
        , (dom,evs) <- splitByRho (ispc,accFns)
        , not (null dom)
        ]
    movedI = moveDims i
    ior = -- trace (" ===>  (nV,movedI) = " ++ show (nV,movedI)) $
          [((movedI,[idFn]),MoveIn,0), ((moveDims o,[idFn]),MoveOut,0)] ++
          case rr of
             Left r          -> rdefs r MoveBackwardCheck r MoveForwardCheck
             Right (rBw,rFw) -> rdefs rBw MoveBackward rFw MoveForward
    rdefs rBw movedirBw rFw movedirFw =
       splitByDelta ((moveDims rBw,[idFn]),Reorg movedirBw) ++
       map mirrorDims (splitByDelta ((moveDims rFw,[negIdFn]),Reorg movedirFw))



    idFn    = affineFunction ((nV,nP), zeroMatrix aDim (nV-aDim) `conc` unitMatrix aDim aDim `conc` zeroMatrix aDim (nP+1)) Inhomogeneous
    negIdFn = affineFunction ((nV,nP), zeroMatrix aDim (nV-aDim) `conc` negate (unitMatrix aDim aDim) `conc` zeroMatrix aDim (nP+1)) Inhomogeneous

    nV = zpTargetDim $ head movedI
    nP = zpParams    $ head movedI

    aDim = targetDimIH $ head $ concatMap snd ispcsAndAccFns

    -- mirror in the array-enumerating dimensions
    mirrorDims ((zd,b),c,d) = ((map (zpMirrorDims [nV-aDim .. nV-1]) zd,b),c,d)


    splitByDelta :: ((ZDomain,b),c) -> [((ZDomain,b),c, EValue a)]
    splitByDelta ((dom,b),c) =
        [((d,b),c,ev) | (d,ev) <- spd dom parts]
        where
        parts = epPartsDisjoint delta

    spd :: ZDomain -> [(GIneqSystem a,EValue a)] -> [(ZDomain, EValue a)]
    spd []  _  = []
    spd dom [] = [(dom, 0)]
    spd dom ((sys',ev):rest)
        | null inters  =  spd dom rest
        | otherwise    =  (inters, ev) : cont
        where
        sys = moveParamsToVars (moveParamsToVars sys' 0 aDim 0) 0 (tdim+1) 0
        zd = [zpFromIneqs sys]
        inters = zdIntersection dom zd
        cont   = spd (zdDifference dom zd) rest

makeDimsUniv :: [Integer] -> ZDomain -> ZDomain
makeDimsUniv dims zd = foldl (flip makeDimUniv) zd dims

makeDimUniv :: Integer -> ZDomain -> ZDomain
makeDimUniv dim zd = map mdu zd
    where
    mdu zp = zpolyhedron sys2 af2
        where
        sys = zpPreimage zp
        af  = zpLattice  zp
        sys2 = insertZeroVars sys nV 1
        af2  = affineFunction ((nV+1,nP),
                               rowVectorsToMatrix1 (take (tD-1) (rowVectors (getMatrixIH $ insertZeroVars af nV 1)) ++
                                                    [vector (genericReplicate nV 0 ++ [1] ++ genericReplicate (nP+1) 0)]
                                                   )
                              ) Inhomogeneous

        tD = targetDimIH af
        nV = nVars sys
        nP = nParams sys



main = -- mapM_ print (x 0 7)
       -- print $ zpUnionDifference [e] [f,e]
       -- print $ zpUnionDifference [e,f] []
       -- print $ localise ["t", "p"] ["n"] bsp1ctx bsp1ia 0
       let (delta,rho,lgp,cd) = bsp3alt2 in
         hPutStrLn stderr ("delta = " ++ show delta) >>
         putStrLn ("lgp = " ++ show lgp) >>
         putStrLn cd

{-
       let c = computeC $ map (moveTtoParams 1) $ flatIA bsp2ia
           (i,o,r) = computeIOR 1 c Nothing
           rho = computeRho c
           delta = computeDelta (1+1) rho
       in print (c,(i,o,r),rho,delta)
-}

xprint :: Show a => [a] -> IO ()
xprint xs =
    do evaluate $ length $ concat zs
       mapM_ putStrLn zs
    where
    zs = map show xs

--                                    p t0 t1  n
tispc0 = ineqSystem ((1,3), matrix [[ 0, 1, 0, 0, 0%1]
                                   ,[ 0,-1, 0, 1, 0]
                                   ,[ 1, 0, 0, 0, 0]
                                   ,[-1, 0, 0, 1, 0]
                                   ]
                    )
tispc1 = ineqSystem ((1,3), matrix [[ 0, 0, 1, 0, 0%1]
                                   ,[ 0, 0,-1, 1, 0]
                                   ,[ 1, 0, 0, 0, 0]
                                   ,[-1, 0, 0, 1, 0]
                                   ]
                    )
{-

-}

{-
-- i = t+p+1
-- A[2t+2p+2] (original: A[2i])
-- A[t+p+1]   (original: A[i])
idx1_0 = affineFunction ((1,3), matrix [[2, 2, 0, 0, 2]]) Inhomogeneous
idx2_0 = affineFunction ((1,3), matrix [[1, 1, 0, 0, 1]]) Inhomogeneous
idx1_1 = affineFunction ((1,3), matrix [[2, 0, 2, 0, 2]]) Inhomogeneous
idx2_1 = affineFunction ((1,3), matrix [[1, 0, 1, 0, 1]]) Inhomogeneous

z1_0 = zpolyhedron tispc0 idx1_0
z2_0 = zpolyhedron tispc0 idx2_0
z1_1 = zpolyhedron tispc1 idx1_1
z2_1 = zpolyhedron tispc1 idx2_1
-}

----


--                                     p  t  n
tispc_0 = ineqSystem ((1,2), matrix [[ 0, 1, 0, 0]
                                    ,[ 0,-1, 1, 0]
                                    ,[ 1, 0, 0, 0]
                                    ,[-1, 0, 1, 0%1]
                                    ]
                     )
tispc_0v = moveParamsToVars tispc_0 0 1 0

tispc_1 = ineqSystem ((1,2), matrix [[ 0, 1, 0, 0+1]
                                    ,[ 0,-1, 1, 0-1]
                                    ,[ 1, 0, 0, 0]
                                    ,[-1, 0, 1, 0%1]
                                   ]
                     )
{-
tispc_2 = ineqSystem ((1,2), matrix [[ 0, 1, 0, 0+2]
                                    ,[ 0,-1, 1, 0-2]
                                    ,[ 1, 0, 0, 0]
                                    ,[-1, 0, 1, 0%1]
                                   ]
                     )
-}

--                                 t  n
ctx = ineqSystem ((0,2), matrix [[ 1, 0, 0]
                                ,[ 0, 1,-1]
                                ,[-1, 1, 0%1]
                                ]
                 )
--                                  n
ctx_v = ineqSystem ((0,1), matrix [[1,-1]])

idx2_0 = affineFunction ((1,2), matrix [[2, 2, 0, 2]]) Inhomogeneous
idx1_0 = affineFunction ((1,2), matrix [[1, 1, 0, 1]]) Inhomogeneous
idx2_1 = affineFunction ((1,2), matrix [[2, 2, 0, 2+2]]) Inhomogeneous
idx1_1 = affineFunction ((1,2), matrix [[1, 1, 0, 1+1]]) Inhomogeneous
--idx2_2 = affineFunction ((1,2), matrix [[2, 2, 0, 2+4]]) Inhomogeneous
--idx1_2 = affineFunction ((1,2), matrix [[1, 1, 0, 1+2]]) Inhomogeneous

z1_0 = zpolyhedron tispc_0 idx1_0
z2_0 = zpolyhedron tispc_0 idx2_0
z1_1 = zpolyhedron tispc_1 idx1_1
z2_1 = zpolyhedron tispc_1 idx2_1
--z1_2 = zpolyhedron tispc_2 idx1_2
--z2_2 = zpolyhedron tispc_2 idx2_2


v = zdTaggedDisjointUnion $ zip [z1_0,z2_0,z1_1,z2_1] $ map (\x -> [x]) [0..]
x t n = [(zpEnumerate [t,n] zp, tag) | (zp,tag) <- v]
{-
q = zpTDU $ zip [z1_0,z2_0,z1_1,z2_1] $ map (\x -> [x]) [0..]
v = concatMap zpTaggedDisjointUnion {-bfDU-} q
w = [[enumerateZPolyhedron [0,7] zp | (zp,tags) <- l] | l<-q]
k = [(enumerateZPolyhedron [0,7] zp,tags) | (zp,tags) <- zpTaggedDisjointUnion (q!!6)]
-}

{-
images = map zpRealImage [z1_0,z2_0,z1_1,z2_1]

aa, bb :: ZPolyhedron Rational
aa = zpolyhedron (ineqSystem ((1,1),matrix[[-1,0,100],[0,-1,99],[0,1,1],[1,0,0]])) (affineFunction ((1,1),matrix [[1,1,2]]) Inhomogeneous)
bb = zpolyhedron (ineqSystem ((1,1),matrix[[-1,0,100],[0,-1,100],[0,1,0],[1,0,0]])) (affineFunction ((1,1),matrix [[1,1,1]]) Inhomogeneous)



v1 = zpTaggedDisjointUnion $ zip [z1_0,z1_1] $ map (\x -> [x]) [0..]
x1 t n = [(zpEnumerate [t,n] zp, tag) | (zp,tag) <- v1]

v2 = [(zp,[1]) | (zp,[1]) <- v1] ++
     (zpTaggedDisjointUnion $ zip [z1_0] $ map (\x -> [x]) [0..])

-- x < x' nicht so guenstig, da der Fall x' = min{x} (und damit
-- leerer Polyeder) speziell behandelt wird, d.h. Spezialfall fuer
-- erstes Arrayelement. Besser: x <= x' und nachher 1 vom
-- Ehrhart-Polynom abziehen.
--                                       x  x' t  n
iLeConstr = ineqSystem ((1,3), matrix [[-1, 1, 0, 0,0%1]])
ictx = insertZeroParams ctx 0 1 `join`
       moveVarsToParams (fromJust (convexUnionAll $ map zpRealImage [z1_0])) 0 1 0

inames = ["x", "t", "n"]

v'' = [(f zp,tags) | (zp,tags) <- v2]
    where
    f zp = zpAddTargetConstraints (iLeConstr) $ zpInsertZeroParams zp 0 1

v' = [(disjCount fzp, tags) | (fzp,tags) <- v'']
    where
    -- simplifizieren ueberfluessig, wenn zpCount... mit (Just ictx) ?
    -- nein, zpCount... mit (Just ictx) ist notwendig (sonst
    -- ueberfluessige Fallunterscheidungen), aber nicht ausreichend
    -- fuer einfachste Bedingungen -> simplifizieren notwendig
    disjCount zp = [(simpl,ehp) | (cond,ehp) <- epPartitions $ zpCountIntegralPointsDisj (Just ictx) zp, let simpl = simplifyIneqs (Just $ ictx) cond, omegaTest (Just simpl) Nothing]

s = [(evaluesWithCondsToCExpression inames $ epartition $ map simpl evcs, tags) | (evcs, tags) <- v']
    where
    simpl (cond,ev) = (simplifyIneqs (Just ictx) cond, ev)
s0 = concat (intersperse "+" ['(':expr++")" | (expr,tags) <- s, 0 `elem` tags]) ++ "-1"
-- s1 = concat $ intersperse "+" ['(':expr++")" | (expr,tags) <- s, 1 `elem` tags]


lgp = LGP { lgpContext = ctx_v,
            lgpParamNames = ["n"],
            lgpVarNames = ["t", "dummy", "p"],
            lgpStatements =
                [[(tispc_0v, idSched 1 0)]] ++
                [[(fromJust (projectToVars tispc_0v Nothing [0]), affineFunction ((1,1), matrix [[1,0,0], [0,0,2],[0,0,0]]) Inhomogeneous)]] ++
                [[zp2s d 0 zp] | (zp,tags) <- v1 | d <- [3..]]
                   
          }

zp2s :: Integer -> Integer -> ZPolyhedron Rational -> (IneqSystem,GAffineFunction Integer)
zp2s d toff zp = (moveParamsToVars sys 0 1 0
                 ,idSched d toff
                 )
        where
        sys :: IneqSystem
        sys = zpPreimage zp

idSched :: Integer -> Integer -> GAffineFunction Integer
idSched d toff = 
    affineFunction ((2,1), matrix [[1,0,0,toff]
                                  ,[0,0,0,d]
                                  ,[0,1,0,0]
                                  ]
                   ) Inhomogeneous


insertConstVar :: Num a => Int -> a -> GIneqSystem a -> GIneqSystem a
insertConstVar pos val sys =
    insertZeroVars sys pos 1 `join`
    ineqSystem ((nV+1,nP), zeroMatrix 2 pos `conc` matrix [[1],[-1]] `conc`
                           zeroMatrix 2 (nV-pos+nP+1)
               )
    where
    nV = nVars sys
    nP = nParams sys

c = codeWithStatements (lgpVarNames lgp) (lgpParamNames lgp) stmtGen (lgen lgenOptionsStd (lgpContext lgp) (lgpStatements lgp) !! 3) 

stmtGen :: Integer -> [String] -> String
stmtGen 1 [t,p] = "int i = (" ++ t ++ ")+(" ++ p ++ ")+1;\n" ++
                    "int x = i;\n" ++     -- acces function
                    "curr[s0(x,t,n)] = curr[s0(x,t,n)] + 1;\n"
stmtGen 2 [_]   = "int *tmp = curr; curr = prev; prev = tmp;\n"
stmtGen d [t,i] = "int x = " ++ show (entry af 0 0) ++ "*(" ++ i ++ ")+" ++
                  show (entry af 0 1) ++ "*(" ++ t ++ ")+" ++
                  show (entry af 0 2) ++ "*n" ++ "+(" ++ show (entry af 0 3) ++ ");\n" ++
                  left ++ " = " ++ right ++ ";\n"
    where
    left
       | 1 `elem` tags  =  "curr[s0(x,t+1,n)]"
       | otherwise      =  "A[x]"

    right
       | 0 `elem` tags  =  "prev[s0(x,t,n)]"
       | otherwise      =  "A[x]"

    (zp, tags) = v1 !! fromIntegral (d-3)
    af' = zpLattice zp
    af = fmap numerator af'

 



e = zpolyhedron (ineqSystem ((1,0),matrix[[1,0],[-1,10]])) (affineFunction ((1,0),matrix [[2,0]]) Inhomogeneous)
f = zpolyhedron (ineqSystem ((1,0),matrix[[1,0],[-1,10]])) (affineFunction ((1,0),matrix [[2,1]]) Inhomogeneous)

ep = zpolyhedron (ineqSystem ((1,1),matrix[[1,0,0],[-1,1,0]])) (affineFunction ((1,1),matrix [[2,0,0]]) Inhomogeneous)
fp = zpolyhedron (ineqSystem ((1,1),matrix[[1,0,0],[-1,1,0]])) (affineFunction ((1,1),matrix [[2,0,1]]) Inhomogeneous)

ee = zpUnionDifference [e] []
-}


-- dimensions inside t dimension
xxx_innerdims = 3
--                                                      t  i  j  k  n  
xxx_df = [moveTtoParams 0 (a,f) | (a,f) <-
           [(zpFromIneqs $ ineqSystem ((xxx_innerdims+1,1),
                                              matrix [[ 1, 0, 0, 0, 0, 0 ]
                                                     ,[-1, 0, 0, 0, 1, 0 ]
                                                     ,[-1, 1, 0, 0, 0, 0 ]
                                                     ,[ 2,-1, 0, 0, 0, 0 ]
                                                     ,[-1, 0, 1, 0, 0, 0 ]
                                                     ,[ 2, 0,-1, 0, 0, 0 ]
                                                     ,[-1, 0, 0, 1, 0, 0 ]
                                                     ,[ 2, 0, 0,-1, 0, 0%1 ]]
                                      )
            ,affineFunction ((xxx_innerdims+1,1), matrix [[ 0, 1, 0, 0, 0, 0 ]
                                                         ,[ 0, 0, 1, 0, 0, 0%1 ]
                                                         ,[ 0, 0, 0, 1, 0, 0%1 ]
                                                    ]
                            ) Inhomogeneous
            )
           ]
         ]

xxx_rho :: EPartition Rational
xxx_rho = computeRho $ computeC xxx_df
{-
  xxx_rho = [...,
(Inequality system in 0 variables and 5 parameters
Matrix with 3 rows and 6 columns
 0%1  0%1  1%1  (-1)%1  0%1  (-1)%1
 0%1  1%1  0%1  (-1)%1  0%1  (-1)%1
 1%1  0%1  0%1  (-1)%1  0%1  (-1)%1
,"(((1)*t+2)*t+1)*i+((((-1))*t+(-2))*t+(-1))*t+((1)*t+1)*j+(((-1))*t+(-1))*t+(1)*k+((-1))*t+0")
    , ...]
-}
xxx_delta = computeDelta (xxx_innerdims+0) xxx_rho
{- FEHLERHAFT ?! (passt nicht zu rho(t+1,(i,j,k))-rho(t,(i,j,k)) ?)
  xxx_delta = [...,
(Inequality system in 0 variables and 5 parameters
Matrix with 3 rows and 6 columns
 0%1  1%1  0%1  (-1)%1  0%1  (-2)%1
 1%1  0%1  0%1  (-1)%1  0%1  (-2)%1
 0%1  0%1  1%1  (-1)%1  0%1  (-2)%1
,"(((1)+(((1)*t+2)*t+1)+((2)*t+2))*i+(((1)*t+1)*j+(((-1))*t+(-1))*t+(1)*k+((-1))*t+0)+(((-1))*t+(-1))+(((((-1))*t+(-3))*t+(-3))*t+(-1))+((((-2))*t+(-4))*t+(-2)))+(((((-1))*t+(-2))*t+(-1))*i+(((1)*t+2)*t+1)*t+(((-1))*t+(-1))*j+((1)*t+1)*t+((-1))*k+(1)*t+0)")
             , ... ]
-}
